#ifdef INCLUDE_FLANN
	// see http://people.cs.ubc.ca/~mariusm/index.php/FLANN/FLANN
	#include "flann/flann.hpp"
#endif


#include "LinAlgMatrixSparseReal.h"
#include "LinAlgMatrixReal.h"
#include "LinAlgVectorReal.h"


#ifndef __DIFFUSION_MAP_H
#define __DIFFUSION_MAP_H

class DiffusionMap
{	
	public:
		enum DiffusedSearchType
		{
			SEARCH_EXHAUSTIVE,
			SEARCH_USE_FLANN
		};
		
		enum EigenDecompositionLibrary
		{
			EIGS_MATLAB,
			EIGS_SVDLIBC
		};
	
		enum DistToSimNormType
		{
			D2S_PRESCRIBE_REGIONS,
			D2S_PER_POINT,
			D2S_GLOBAL_MIN,
			D2S_GLOBAL_MAX,
			D2S_GROUPED_PER_POINT
		};
		
	
	// Preparing the map
		DiffusionMap(int N, int nEigs, 
					 const std::vector<int> & numPointsInGroup = std::vector<int>(),
					 DiffusedSearchType srch = SEARCH_EXHAUSTIVE, 
					 EigenDecompositionLibrary decomp = EIGS_MATLAB,
					 DistToSimNormType normType = D2S_PER_POINT);
	
		~DiffusionMap();
		int NEntries();
		int NEigs();
		void FreeData();
		void SetCorrs(DiffusionMap * otherMap, bool makeCopy = false);
		void SetCorrs(LinAlgMatrixSparseReal * corrMtx, bool makeCopy = false);
		void AddCorrValue(int i, int j, double val);
		void ClearAllCorrValues();
		void Embed(bool addIdentities=true, bool symmetrize=true, 
				   bool rowNormalize=true, bool normalizeEigenvectors=true, 
				   double * timeInSec = NULL, const char * prefix = NULL);
		void UpdateEmbedding(bool addIdentities=true, bool symmetrize=true, 
							 bool rowNormalize=true, bool normalizeEigenvectors=true,
							 double * timeInSec = NULL);
		
	// k-means clustering
		void ClusterKMeans(int nClusters, std::vector<int> & clusterIDs,
						   std::vector<int> * clusterCentroids = NULL);
	
	// Map parameters (can be changed interactively)
		double DiffTime();
		double DiffDistPercentile();
		void SetTime(double t);
		void SetMaxDistancePercentile(double percentile);
		
		virtual void SetNormalizationMethod(DistToSimNormType normType);
		virtual DistToSimNormType GetNormalizationMethod();
		virtual double NormalizeSimilaritiesBasedOnRegion(std::vector<int> & regionIDs,
												  bool useMin = true);
		void SetSearchType(DiffusedSearchType srchTp);
		void SetEigsLib(EigenDecompositionLibrary eigsLib);
	
		void ClearNNCache(bool clearNNEmbedding);
	
	// Get raw data (just in case)
		void Eigs(LinAlgMatrixReal ** eigVecs, LinAlgVectorReal ** eigVals);
	
	// Queries: distances
		double EmbeddedDistanceExact2(int i, int j);
		double EmbeddedDistance2(int i, int j);	
		void ClosestPnts(int pntID, int ** closestIDs, double ** closestDists, int * nFoundPnts);
	
	// Queries: similarities
		double EmbeddedSimilarityExact(int i, int j);
		double EmbeddedSimilarity(int i, int j);
		double GetSimSigma2();
		
		void MostSimilarPoints(int pntID, int ** closestIDs, 
							   double ** closestSims, int * nFoundPnts);
	
		void MostSimilarPointsInGroup(int pntID, int groupID, int ** closestIDs, 
									  double ** closestSims, int * nFoundPnts);

	// reading / writing the data
		void WriteBinaryData(const char * dataDir, const char * filePrefix);
		bool ReadBinaryData(const char * dataDir, const char * filePrefix);
	
	protected:	
		void FillInputOutputPtrs();
#ifdef INCLUDE_FLANN
		flann::Index<flann::L2<double> > * m_pEmbeddedSpaceANN;
		flann::Matrix<double> * m_pANNDataset;
		flann::Matrix<int> * m_pANNOutIndices;
		flann::Matrix<double> * m_pANNOutDistances;
		flann::Matrix<double> * m_pANNQuery;
		void BuildFLANNIndex();
	
	// per group (e.g. per model) flann
		flann::Index<flann::L2<double> > ** m_pEmbeddedSpaceANNPerGroup;
		flann::Matrix<double> ** m_pANNDatasetPerGroup;
//		flann::Matrix<int> ** m_pANNOutIndicesPerGroup;
//		flann::Matrix<double> ** m_pANNOutDistancesPerGroup;
//		flann::Matrix<double> ** m_pANNQueryPerGroup;
		void BuildFLANNForGroup(int groupID);
		void FillInputOutputPtrsFLANN();
#endif
		void FillRowOfEmbeddedDistances(int rowID);
		void FillRowOfEmbeddedDistancesWithGlobalNorms(int rowID);
		void FillRowOfEmbeddedDistancesWithPerGroupNorms(int rowID);
		void FillRowOfEmbeddedDistancesWithPerGroupNorms(int rowID, int groupID);
		
		int m_iN;
		int m_iNEigs;
		double m_fDiffTime;
		double m_fMaxDistPercentile;
		DistToSimNormType m_NormalizationType;
		
		bool m_bFreeCorrs;
		LinAlgMatrixSparseReal * m_pCorrsProcessedForDiffusion;
		LinAlgMatrixSparseReal * m_pCorrVals;
		LinAlgVectorReal * m_pCorrVectorRowSum;
		
	// eigen decomposition
		EigenDecompositionLibrary m_DecompositionLib;
		LinAlgMatrixReal * m_pEigenVectors;
		LinAlgVectorReal * m_pEigenValues;

	// NN library data
		DiffusedSearchType m_SearchType;	
		double * m_aQueryFeatures;
		int * m_aClosestPntIDs;
		double * m_aClosestDistances;
		double * m_aClosestSimilarities;
	
	// NN caching 
		std::map<int, double> ** m_aComputedCorrsArray;
		double * m_afComputerNormDistArray;	
		bool * m_abFilledComputedCoorsRow;
		double m_fCurrentNormDist;
		virtual void FillGlobalNorms();
		double m_fGlobalNormMax;
		double m_fGlobalNormMin;
		
	// grouping (e.g. points per model)	
		int m_iNGroups;
		int * m_aPointToGroup;
		int * m_aGroupToOffset;
		int * m_aGroupToNumPoints;
};

#endif





