
/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "AppBenchmark.h"
#include "CollectionGroundTruth.h"
#include "DistanceEuclidean.h"
#include "CollectionIdentityPointsets.h"
#include "CollectionDiffuseDynamicGraph.h"
#include "TimeProfiler.h"

AppBenchmark::PerMapStats::PerMapStats()
{
	m_pModelFrom = NULL;
	m_pModelTo = NULL;
	m_pPntsFrom = NULL;
	m_bFilled = false;
}

AppBenchmark::AppBenchmark(AnalysisPipeline & pipeline, const VKString & moduleName)
: PipelineStepApplication(pipeline, moduleName)
{
	bool exists;	
	int nDynIters = m_Pipeline.m_Params.GetIntValue(m_sModuleName, 
													"INPUT_TestIterations", exists);
	if (!exists)
		nDynIters = 0;
	m_bIterated = nDynIters==0;
	m_bWritingImages = false;
	m_Benchmark = BENCHMARK_NONE;
	m_CorrEvaluation = EVAL_CORRS_NONE;
	m_DrawType = BENCHDRAW_ERROR;
	//m_DrawType = BENCHDRAW_NONE;
	m_DrawableLinkMain = -1;
	ClearDrawableErrors();
	
	ParamParser & par = m_Pipeline.m_Params;
	VKString benchmark = par.GetStrValue(m_sModuleName, "INPUT_BenchmarkType");
		
	if (benchmark=="BENCHMARK_CORRS")
	{
		std::cout<<" Correspondences"<<std::endl;
		// INPUT: eval method, map pairs, eval pnts, diffed corrs, 
		//			ground truth map, output dir, 
		// OUTPUT: stats, date, copy settings file
		m_Benchmark=BENCHMARK_CORRS;
		
		VKString evalType = par.GetStrValue(m_sModuleName, "INPUT_EvalType");
		
		if (evalType == "EVAL_CORRS_VAL_AS_PENALTY")
			m_CorrEvaluation = EVAL_CORRS_VAL_AS_PENALTY;
		else
			assert(false);
		
		VKString groundTruthMap = par.GetStrValue(m_sModuleName, "INPUT_GroundTruthMap");
		VKString groundTruthValue = par.GetStrValue(m_sModuleName, "INPUT_GroundTruthCorrValue");
		VKString testMap = par.GetStrValue(m_sModuleName, "INPUT_CorrsToTest");
		VKString testCorrVal = par.GetStrValue(m_sModuleName, "INPUT_CorrValToTest");
		
		if (groundTruthMap=="default")
			groundTruthMap = CollectionGroundTruth::DefaultGroundTruthMapName();
		if (groundTruthValue=="default")
			groundTruthValue = CollectionGroundTruth::DefaultGroundTruthCorrValue();
		
		m_pGroundTruthCollection = m_Pipeline.GetDB()->GetCorrCollection(groundTruthMap);
		assert(m_pGroundTruthCollection!=NULL);
		m_pGroundTruthValue = m_pGroundTruthCollection->GetCorrValueByName(groundTruthValue);
		assert(m_pGroundTruthValue!=NULL);
		
		m_pCorrCollection = m_Pipeline.GetDB()->GetCorrCollection(testMap);
		assert(m_pCorrCollection!=NULL);
		m_pCorrValue = m_pCorrCollection->GetCorrValueByName(testCorrVal);
		assert(m_pCorrValue!=NULL);
		
		m_bWriteErrors = par.GetStrValue(m_sModuleName, 
										 "INPUT_WriteErrors")=="true";
		m_bTakeBest = par.GetStrValue(m_sModuleName, 
									  "INPUT_TakeBestCorr")=="true";
		m_bWriteFuzzyForPair = par.GetStrValue(m_sModuleName, 
											   "INPUT_WriteFuzzyPair")=="true";
		m_bWriteCoarseForPair = par.GetStrValue(m_sModuleName, 
												"INPUT_WriteCoarsePair")=="true";
		m_bWriteFineForPair = par.GetStrValue(m_sModuleName, 
											  "INPUT_WriteFinePair")=="true";
		m_sDistanceName = par.GetStrValue(m_sModuleName, 
										  "INPUT_DistanceMetric");
	}
}

VKString AppBenchmark::GetAppType()
{
	return "AppBenchmark";
}

void AppBenchmark::Execute()
{
	std::cout<<"Benchmark: "<<std::flush;
	srand(0);
	bool ok;
	ParamParser & par = m_Pipeline.m_Params;
	VKString benchmark = par.GetStrValue(m_sModuleName, "INPUT_BenchmarkType");
	
	if (m_Benchmark==BENCHMARK_CORRS)
	{
		std::cout<<" Correspondences"<<std::endl;
		// INPUT: eval method, map pairs, eval pnts, diffed corrs, 
		//			ground truth map, output dir, 
		// OUTPUT: stats, date, copy settings file		
		int nModels = m_Pipeline.GetDB()->NumModels();
		
		std::vector<int> modelPairs;
		VKString whichMaps = par.GetStrValue(m_sModuleName, "INPUT_ModelPairs");		

		int MAX_ITERATIONS = 10000;
		if (whichMaps=="InputQueryFile")
		{
			std::map<VKString, int > tripletToModel;
			ModelDatabase * db = m_Pipeline.GetDB();
			for (int i=0; i<nModels; i++)
			{
				VKStringList triplet = db->GetModelByID(i)->m_sBIMQueryTriplet;
				assert (triplet.count()==3);
				VKString str = triplet[0] + "_" + triplet[1] + "_" + triplet[2];
				tripletToModel[str] = i;
			}
			VKString queryFile = par.GetStrValue(m_sModuleName, "INPUT_QueryFile");
			std::ifstream textStream(queryFile.c_str());
			assert(textStream.is_open());
			VKString t1, t2, t3, t4, t5, t6;
			while(!textStream.eof())
			{
				t1.readToken(textStream);
				if (textStream.eof())
					break;
				t2.readToken(textStream);
				t3.readToken(textStream);
				t4.readToken(textStream);
				t5.readToken(textStream);
				t6.readToken(textStream);
				
				VKString str1 = t1 + "_" + t2 + "_" + t3;
				VKString str2 = t4 + "_" + t5 + "_" + t6;
				if (tripletToModel.find(str1)==tripletToModel.end())
					std::cout<<"[WARNING] Cannot find "<<str1.c_str()<<std::endl;
				else if (tripletToModel.find(str2)==tripletToModel.end())
					std::cout<<"[WARNING] Cannot find "<<str2.c_str()<<std::endl;
				else
				{
					modelPairs.push_back(tripletToModel[str1]);
					modelPairs.push_back(tripletToModel[str2]);
				}
			}
		}
		else if (whichMaps=="AllPairs")
		{
			for (int i=0; i<nModels; i++)
			for (int j=0; j<nModels; j++)
			{
				modelPairs.push_back(i);
				modelPairs.push_back(j);
			}
		}
		else if (whichMaps=="OneToAll")
		{
			VKStringList oneFrom = par.GetStrValues(m_sModuleName, "INPUT_ModelsFrom");
			std::vector<int> selectedFrom;
			std::map<int, bool> modelsIncluded;
			for (int i=0; i<oneFrom.count(); i++)
			{
				if (oneFrom[i]=="Random")
				{
					assert(i+1 < oneFrom.count());
					int nFrom = oneFrom[i+1].toInt(&ok);
					assert(ok);
					
					int nAdded = 0;
					while (nAdded < nFrom)
					{
						int modelID = rand() % nModels;
						if (modelID<0)
							modelID *= -1;
						if (modelsIncluded.find(modelID)==modelsIncluded.end())
						{
							nAdded++;
							modelsIncluded[modelID] = true;
							selectedFrom.push_back(modelID);
						}
						MAX_ITERATIONS--;
						if (MAX_ITERATIONS < 0)
						{
							std::cout<<"[WARNING] Could not add "<<nFrom<<" models"<<std::endl;
							break;
						}
					}
					i++;
				}
				else
				{
					int modelID = oneFrom[i].toInt(&ok);
					if (!ok)
					{
						Model3D * model = m_Pipeline.GetDB()->GetModelByName(oneFrom[i]);
						assert(model!=NULL);
						modelID = model->ModelID();
					}
					assert(m_Pipeline.GetDB()->GetModelByID(modelID)!=NULL);
					selectedFrom.push_back(modelID);
				}
			}
			for (int i=0; i<(int)selectedFrom.size(); i++)
			for (int j=0; j<nModels; j++)
			{
				modelPairs.push_back(selectedFrom[i]);
				modelPairs.push_back(j);
			}
		}
		else if (whichMaps=="SomePairs")
		{
			std::map<int, bool> occupiedPairs;
			VKStringList pairsList = par.GetStrValues(m_sModuleName, "INPUT_ModelsPairList");
			assert(pairsList.count()%2 == 0);
			for (int i=0; i<pairsList.count(); i+=2)
			{
				assert(i+1 < pairsList.count());
				if (pairsList[i]=="Random")
				{
					int nFrom = pairsList[i+1].toInt(&ok);
					assert(ok);
					int nAdded = 0;
					while (nAdded < nFrom)
					{
						int model1ID = rand() % nModels;
						int model2ID = rand() % nModels;
						if (model1ID<0)
							model1ID *= -1;
						if (model2ID<0)
							model2ID *= -1;
						if (model1ID == model2ID)
							continue;
						int pairID = model2ID * nModels + model1ID;
						if (occupiedPairs.find(pairID)==occupiedPairs.end())
						{
							nAdded++;
							occupiedPairs[pairID] = true;
							modelPairs.push_back(model1ID);
							modelPairs.push_back(model2ID);
						}
						MAX_ITERATIONS--;
						if (MAX_ITERATIONS < 0)
						{
							std::cout<<"[WARNING] Could not add "<<nFrom<<" models"<<std::endl;
							break;
						}
					}
				}
				else
				{
					bool ok1, ok2;
					int model1ID = pairsList[i].toInt(&ok1);
					int model2ID = pairsList[i+1].toInt(&ok2);
					if (!ok1 || !ok2)
					{
						Model3D * model1 = m_Pipeline.GetDB()->GetModelByName(pairsList[i]);
						Model3D * model2 = m_Pipeline.GetDB()->GetModelByName(pairsList[i+1]);
						assert(model1!=NULL && model2!=NULL);
						model1ID = model1->ModelID();
						model2ID = model2->ModelID();
					}
					assert(m_Pipeline.GetDB()->GetModelByID(model1ID)!=NULL);
					assert(m_Pipeline.GetDB()->GetModelByID(model2ID)!=NULL);
					modelPairs.push_back(model1ID);
					modelPairs.push_back(model2ID);
				}
			}
		}
		
		std::cout<<"Testing "<<modelPairs.size()/2<<" pairs:"<<std::endl;
		assert(modelPairs.size() > 0);
		assert(modelPairs.size()%2==0);
		for (int i=0; i<(int)modelPairs.size(); i+=2)
		{
			if (modelPairs[i] != modelPairs[i+1])
				AddPairStats(modelPairs[i], modelPairs[i+1]);
		}
		TimeProfiler timer;
		for (int i=0; i<(int)m_EvalCorrStats.size(); i++)
		{
			timer.WriteProgress("WritePairs", i, (int)m_EvalCorrStats.size());
			if (m_bWriteErrors)
				FillPerMapStats(m_EvalCorrStats[i]);
			if (m_bWriteFuzzyForPair)
				WriteFuzzyMapForPair(m_EvalCorrStats[i]);
			if (m_bWriteCoarseForPair)
				WriteCoarseMapForPair(m_EvalCorrStats[i]);
			if (m_bWriteFineForPair)
				WriteFineMapForPair(m_EvalCorrStats[i]);
		}
		std::cout<<std::endl;
		WriteErrorsToConcatFile();
	}
	else
		assert(false);
}

void AppBenchmark::AddPairStats(int fromModel, int toModel)
{
	assert(fromModel < m_Pipeline.GetDB()->NumModels());
	assert(toModel < m_Pipeline.GetDB()->NumModels());
	bool evalPnsExist;
	ParamParser & par = m_Pipeline.m_Params;
	
	int currID = (int)m_EvalCorrStats.size();
	m_EvalCorrStats.push_back(PerMapStats());
	m_EvalCorrStats[currID].m_pModelFrom = m_Pipeline.GetDB()->GetModelByID(fromModel);
	m_EvalCorrStats[currID].m_pModelTo = m_Pipeline.GetDB()->GetModelByID(toModel);

	VKString evalPntsName = par.GetStrValue(m_sModuleName, "INPUT_EvalOnPnts", evalPnsExist);
	PointSet * pntsFrom = m_EvalCorrStats[currID].m_pModelFrom->GetPointSet(evalPntsName);
	if (!evalPnsExist || pntsFrom==NULL)
		pntsFrom = m_EvalCorrStats[currID].m_pModelFrom->PointCloud();
	assert(pntsFrom!=NULL);
	m_EvalCorrStats[currID].m_pPntsFrom = pntsFrom;
	
	Model3D * mFrom = m_EvalCorrStats[currID].m_pModelFrom;
	Model3D * mTo = m_EvalCorrStats[currID].m_pModelTo;
	m_SrcModelToPairs[mFrom][mTo] = currID;
}

VKString AppBenchmark::GetMeshStr(VKStringList & t)
{
	bool ok;	
	assert(t.count()==3);
	if (t[0]=="SCAPE")
	{
		if (t[2].toInt(&ok) < 10)
		{
			assert(ok);
			return VKString("mesh00")+t[2];
		}
		else
		{
			assert(ok);
			return VKString("mesh0")+t[2];
		}
	}
	else if (t[0]=="TOSCA")
		return t[1]+t[2];
	else if (t[0]=="SHREC")
	{
		int c = t[1].toInt(&ok);
		assert(ok);
		
		int id=t[2].toInt(&ok);
		assert(ok);
		return VKString::number((c-1) * 20 + id);
	}
	assert(false);
	return "";
}

VKString AppBenchmark::GetMapStr(PerMapStats & perMapStats)
{
	VKString filename = m_Pipeline.GetDB()->GetAppDir(m_sAppName);
	VKStringList t1 = perMapStats.m_pModelFrom->m_sBIMQueryTriplet;
	VKStringList t2 = perMapStats.m_pModelTo->m_sBIMQueryTriplet;
	
	VKString str1 = GetMeshStr(t1);
	VKString str2 = GetMeshStr(t2);
	return filename + str1+"_to_" + str2;
}

void AppBenchmark::WriteFuzzyMapForPair(PerMapStats & perMapStats)
{
	VKString filename = GetMapStr(perMapStats)+"_fuzzy.bin";
	std::cout<<"Writing Fuzzy Map: "<<filename.c_str()<<std::endl;
	m_Pipeline.GetDB()->CreateDirsOnPathToFile(filename);
	FILE * fuzzyFile = fopen(filename.c_str(), "wb");
	assert(fuzzyFile!=NULL);
	
	int nModels = 1;
	fwrite(&nModels, sizeof(int), 1, fuzzyFile);
	int mID=0;
	int mID2=1;
	int nPoints=perMapStats.m_pPntsFrom->NumPoints();
	fwrite(&mID, sizeof(int), 1, fuzzyFile);
	fwrite(&nPoints, sizeof(int), 1, fuzzyFile);

	PointSet * pntsTo = perMapStats.m_pModelTo->PointCloud();
	std::map<int, double> * perPointMap = new std::map<int, double>[nPoints];
	m_pCorrCollection->GetCorrsForPairWithVal(perMapStats.m_pModelFrom,
											  perMapStats.m_pModelTo,
											  perPointMap);
	for (int i=0; i<perMapStats.m_pPntsFrom->NumPoints(); i++)
	{
		ModelPoint pFrom = perMapStats.m_pPntsFrom->GetPoint(i);		
		int vID = pFrom.NearestVertex();
		fwrite(&vID, sizeof(int), 1, fuzzyFile);
		int nCorrs = (int)perPointMap[i].size();
		fwrite(&nCorrs, sizeof(int), 1, fuzzyFile);
		
		for (std::map<int, double>::iterator iter = perPointMap[i].begin();
			 iter != perPointMap[i].end(); iter++)
		{
			int vID2 = pntsTo->GetPoint(iter->first).NearestVertex();
			double val = iter->second;
			fwrite(&mID2, sizeof(int), 1, fuzzyFile);
			fwrite(&vID2, sizeof(int), 1, fuzzyFile);
			fwrite(&val, sizeof(double), 1, fuzzyFile);
		}		
	}
	fclose(fuzzyFile);
	delete [] perPointMap;
}

void AppBenchmark::WriteCoarseMapForPair(PerMapStats & perMapStats)
{
	VKString filename = GetMapStr(perMapStats)+"_fuzzybest.cor";
	m_Pipeline.GetDB()->CreateDirsOnPathToFile(filename);
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	for (int i=0; i<perMapStats.m_pPntsFrom->NumPoints(); i++)
	{
		ModelPoint pFrom = perMapStats.m_pPntsFrom->GetPoint(i);
		ModelPoint bestCorr = m_pCorrCollection->GetBestCorr(pFrom, perMapStats.m_pModelTo, 
															 "auto", NULL);
		int v1ID = pFrom.NearestVertex();
		int v2ID = 0;
		if (!bestCorr.Invalid())
			v2ID = bestCorr.NearestVertex();
		textStream<<v1ID<<"\n"<<v2ID<<"\n";
	}
	textStream.close();
	return;
}

void AppBenchmark::WriteFineMapForPair(PerMapStats & perMapStats)
{
//	VKString filename = GetMapStr(perMapStats)+"_fuzzybest.map";
//	m_Pipeline.GetDB()->CreateDirsOnPathToFile(filename);
//	std::ofstream textStream(filename.c_str());
//	assert(textStream.is_open());
//	PointSet * allVertices
//	for (int i=0; i<perMapStats.m_pModelFrom->GetMesh()->NVertices(); i++)
//	{
//		ModelPoint pFrom = perMapStats.m_pPntsFrom->GetPoint(i);
//		ModelPoint bestCorr = m_pCorrCollection->GetBestCorr(pFrom, perMapStats.m_pModelTo, 
//															 "auto", NULL);
//		int v1ID = pFrom.NearestVertex();
//		int v2ID = 0;
//		if (!bestCorr.Invalid())
//			v2ID = bestCorr.NearestVertex();
//		textStream<<v1ID<<"\n"<<v2ID<<"\n";
//	}
//	textStream.close();
//	return;
}

void AppBenchmark::FillPerMapStats(PerMapStats & perMapStats)
{
	if (perMapStats.m_bFilled)
		return;
	
	assert(m_Benchmark==BENCHMARK_CORRS);
	assert(m_CorrEvaluation==EVAL_CORRS_VAL_AS_PENALTY);
	assert(m_pGroundTruthCollection!=NULL && m_pGroundTruthValue!=NULL);
	assert(m_pCorrCollection!=NULL && m_pCorrValue!=NULL);
	assert(perMapStats.m_pModelFrom!=NULL);
	assert(perMapStats.m_pModelTo!=NULL);
	assert(perMapStats.m_pPntsFrom!=NULL);
	
//	std::cout<<"Mapping Pair: "<<perMapStats.m_pModelFrom->ModelName().c_str();
//	std::cout<<" -> "<<perMapStats.m_pModelTo->ModelName().c_str()<<std::endl;
	
	perMapStats.m_bFilled = true;

	std::map<Model3D*, int> & allToModels = m_SrcModelToPairs[perMapStats.m_pModelFrom];
	std::vector<Model3D*> toModels;
	if (m_bTakeBest)
	{
		for (std::map<Model3D*, int>::iterator iter = allToModels.begin();
			 iter != allToModels.end(); iter++)
			toModels.push_back(iter->first);
	}
	for (int i=0; i<perMapStats.m_pPntsFrom->NumPoints(); i++)
	{
		// get fuzzy correspondences
		ModelPoint pFrom = perMapStats.m_pPntsFrom->GetPoint(i);
		std::vector<ModelPoint> corrs;
		std::vector<double> errors;
		GetPerPointErrors(pFrom, corrs, errors);

		// initialize values for all 'to' models
		for (std::map<Model3D*, int>::iterator iter = allToModels.begin(); 
			 iter != allToModels.end(); iter++)
		{
			int id = iter->second;
			assert(id >= 0 && id < (int)m_EvalCorrStats.size());
			if (m_EvalCorrStats[id].m_EvalCorrValues.size() != i)
			{
				std::cout<<"[ERROR] m_EvalCorrValues: id="<<id<<" i=="<<i;
				std::cout<<" size="<<m_EvalCorrStats[id].m_EvalCorrValues.size()<<std::endl;
			}
			assert((int)m_EvalCorrStats[id].m_EvalCorrValues.size() == i);
			m_EvalCorrStats[id].m_EvalCorrValues.push_back(-1);
			m_EvalCorrStats[id].m_bFilled = true;		
		}
		
		// fill error values
		for (int j=0; j<(int)corrs.size(); j++)
		{
			Model3D * toModel = (Model3D*)corrs[j].GetModel();
			std::map<Model3D*, int>::iterator iter2 = allToModels.find(toModel);
			if (iter2 != allToModels.end())
			{
				assert(iter2->second >= 0 && iter2->second < (int)m_EvalCorrStats.size());
				assert((int)m_EvalCorrStats[iter2->second].m_EvalCorrValues.size() == i+1);
				
				std::vector<double> & corrVals = m_EvalCorrStats[iter2->second].m_EvalCorrValues;
				assert(errors[j] != -1);
				if (corrVals[i]==-1)
					corrVals[i] = errors[j];
				else
					corrVals[i] += errors[j];
			}
		}

		// set to maxErr
		for (std::map<Model3D*, int>::iterator iter = allToModels.begin(); 
			 iter != allToModels.end(); iter++)
		{
			int id = iter->second;
			assert(id >= 0 && id < (int)m_EvalCorrStats.size());
			assert((int)m_EvalCorrStats[id].m_EvalCorrValues.size() == i+1);
			if (m_EvalCorrStats[id].m_EvalCorrValues[i] > 1. 
				|| m_EvalCorrStats[id].m_EvalCorrValues[i]==-1)
				m_EvalCorrStats[id].m_EvalCorrValues[i] = 1;
		}
	}
}


void AppBenchmark::GetPerPointErrors(const ModelPoint & pFrom,
									 std::vector<ModelPoint> & corrs,
									 std::vector<double> & errors)
{
	bool takeBestOnly = true;
	assert(m_Benchmark==BENCHMARK_CORRS);
	assert(m_CorrEvaluation==EVAL_CORRS_VAL_AS_PENALTY);
	assert(m_pGroundTruthCollection!=NULL && m_pGroundTruthValue!=NULL);
	assert(m_pCorrCollection!=NULL && m_pCorrValue!=NULL);

	if (takeBestOnly)
	{
		for (int i=0; i<m_Pipeline.GetDB()->NumModels(); i++)
		{
			Model3D * model = m_Pipeline.GetDB()->GetModelByID(i);
			if (pFrom.GetModel() != model)
			{
				ModelPoint bestCorr = m_pCorrCollection->GetBestCorr(pFrom, model, "none", NULL);
				corrs.push_back(bestCorr);
			}
		}
	}
	else
		m_pCorrCollection->GetCorrs(pFrom, &corrs);
	
	// get ground truth correspondences
	std::map<Model3D*, ModelPoint> groundTruth;
	std::vector<ModelPoint> trueCorrs;
	m_pGroundTruthCollection->GetCorrs(pFrom, &trueCorrs);
	for (int j=0; j<(int)trueCorrs.size(); j++)
	{
		Model3D * modelTo = (Model3D*)trueCorrs[j].GetModel();
		assert(groundTruth.find(modelTo) == groundTruth.end());
		groundTruth[modelTo] = trueCorrs[j];
		assert(m_pGroundTruthValue->Value(pFrom, trueCorrs[j])==1.);
	}
	
	// find normalization for models
	std::map<Model3D*, double> modelToTotalCorrValue;
	double totalCorrVal = 0;
	for (int j=0; j<(int)corrs.size(); j++)
	{
		Model3D * toModel = (Model3D*)corrs[j].GetModel();
		std::map<Model3D*, double>::iterator iter = modelToTotalCorrValue.find(toModel);
		double val;
		if (takeBestOnly)
			val = 1.;
		else
			val = m_pCorrValue->Value(pFrom, corrs[j]);
		
		if (iter == modelToTotalCorrValue.end())
			modelToTotalCorrValue[toModel] = val;
		else
			modelToTotalCorrValue[toModel] += val;
	}
		
	// fill errors
	for (int j=0; j<(int)corrs.size(); j++)
	{
		Model3D * toModel = (Model3D*)corrs[j].GetModel();
		
		// TODO: replace with geodesic
		ModelDistance * dist = toModel->DefaultEuclideanDistance();
		assert(dist!=NULL);
//		assert(m_pDistance!=NULL);
//		ModelDistance * dist = toModel->GetModelDistance(m_sDistanceName);
//		assert (dist!=NULL);
		
		std::map<Model3D*, ModelPoint>::iterator iter = groundTruth.find(toModel);
		assert(iter!=groundTruth.end());
		if (iter==groundTruth.end() || corrs[j].Invalid())
			errors.push_back(-1);
		else
		{
			const ModelPoint & trueCorr = iter->second;
			double distToTrue = dist->Distance(corrs[j], trueCorr);
//			if (distToTrue > 1.)
//				distToTrue = 1.;
			double valNorm = modelToTotalCorrValue[toModel];
			double val;
			if (takeBestOnly)
				val = 1.;
			else
				val = m_pCorrValue->Value(pFrom, corrs[j]) / valNorm;
			errors.push_back(val * distToTrue);		
		}
	}
	
}

bool AppBenchmark::CaptureMouseButtonEvent(DatabaseVisWindow * window, int x, int y, 
										   VisWindow::MouseButton mouseButtonID, 
										   VisWindow::MouseButtonState newState)
{
	return false;
}

bool AppBenchmark::CaptureMouseMotionEvent(DatabaseVisWindow * window, 
										   int x, int y, int dx, int dy)
{
	return false;
}

bool AppBenchmark::CaptureKeyPressed(DatabaseVisWindow * window, 
									 const char key, int mouseX, int mouseY)
{
	bool keyCaptured = true;
	switch (key)
	{
		case '`':
			std::cout<<"DB Benchmark App:"<<std::endl;
			std::cout<<"\te\tErrors for selected point"<<std::endl;
			std::cout<<"\tE\tErrors for all points"<<std::endl;
			std::cout<<"\tc\tClear errors"<<std::endl;
			std::cout<<"\t0-9\tDraw error w.r.t to a different model"<<std::endl;
			std::cout<<"\t!\tWrite Errors to a directory"<<std::endl;
			keyCaptured = false;
			break;
		case '!':
			WritePairsTable(window);
			break;
		case 'c':
			ClearDrawableErrors();
			break;
		case 'e':
		{
			ClearDrawableErrors();
			Model3D * mainModel = m_Pipeline.GetDB()->GetModelInSubwindow(0);
			assert(mainModel!=NULL);
			Model3DRenderingParams * mainRend = mainModel->RenderingParams();
			ModelPoint selected = mainRend->GetSelectedPoint();
			FillDrawAtPnt(window, selected, true);
			break;
		}
		case 'E':
		{
			ClearDrawableErrors();
			if (m_DrawableLinkMain<0)
				m_DrawableLinkMain = 1;
			Model3D * mainModel = m_Pipeline.GetDB()->GetModelInSubwindow(0);
			assert(mainModel!=NULL);
			FillDrawMeshErr(window, mainModel, m_DrawableLinkMain);
			break;
		}
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
		{
				// link error
			int newlink = (int)(key - '0');
			if (newlink < (int)m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID.size())
				m_DrawableLinkMain = newlink;
			std::cout<<"Link to : "<<m_DrawableLinkMain<<std::endl;
			Model3D * mainModel = m_Pipeline.GetDB()->GetModelInSubwindow(0);
			assert(mainModel!=NULL);
			FillDrawMeshErr(window, mainModel, m_DrawableLinkMain);
			break;
		}
		case 'd':
			DynamicGraphIterations(window);
			break;
        default:
            keyCaptured = false;
            break;
    }
	
    if (keyCaptured)
        window->GetRenderer()->UpdateScreen();
    return keyCaptured;
}

void AppBenchmark::Draw(DatabaseVisWindow * window)
{
	if (!m_bIterated)
	{
		m_bIterated = true;
		DynamicGraphIterations(window);
	}
	// nothing. might draw plots / show stats in the future
	//std::cout<<"drawing: "<<m_DrawType<<std::endl;
	Model3D * model0 = m_Pipeline.GetDB()->GetModelInSubwindow(0);
	assert(model0!=NULL);
	if (m_DrawType==BENCHDRAW_ERROR)
	{
		if (m_DrawableLinkMain > 0)
		{
			if (!m_bWritingImages)
				FillDrawMeshErr(window, model0, m_DrawableLinkMain);

			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();

			//glRasterPos2d(pnt[0], pnt[1]);
			
			double cx0, cy0, scale0, dx0, dy0;
			double cx1, cy1, scale1, dx1, dy1;
			window->GetWindowParams(0, &cx0, &cy0, &scale0, &dx0, &dy0);
			window->GetWindowParams(m_DrawableLinkMain, &cx1, &cy1, &scale1, &dx1, &dy1);
			
			glLoadIdentity();
			R3Point p1(cx0 / window->Width(), cy0 / window->Height(), 0);
			R3Point p2(cx1 / window->Width(), cy1 / window->Height(), 0);
			
			R3Ray ray(p1, p2);
			p1 = ray.Point(.5);
			p2 = ray.Point(.85);
			glDisable(GL_LIGHTING);
			glLineWidth(2.);
			glColor3d(.3, .9, 0);
			glBegin(GL_LINES);
			glVertex3d(p1.X(), p1.Y(), p1.Z());
			glVertex3d(p2.X(), p2.Y(), p2.Z());
			glEnd();
			
			glMatrixMode(GL_PROJECTION);
			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);
			glPopMatrix();
		}
		else if (!model0->RenderingParams()->GetSelectedPoint().Invalid() 
				 && !m_bWritingImages)
			FillDrawAtPnt(window, model0->RenderingParams()->GetSelectedPoint(), true);
		
		int nWindows = (int)m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID.size();
		for (int subWindowID = 0; subWindowID < nWindows; subWindowID++)
		{
			//if (m_bWritingImages && subWindowID!=0 && subWindowID!=2)
			//	continue;
			
			int modelID = m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID[subWindowID];
			if (modelID!=-1)
			{
				Model3D * model = m_Pipeline.GetDB()->GetModelByID(modelID);
				assert(model!=NULL);
				std::map<Model3D*, ModelPerVertexValues* >::iterator iter;
				iter = m_DrawableError.find(model);
				if (iter != m_DrawableError.end())
				{
					iter->first->RenderingParams()->LoadMyCamera(subWindowID);
					iter->second->Draw(window);
					iter->first->RenderingParams()->RestoreCameraAfterLoad(subWindowID);
				}
			}
		}    
		
	}
}

void AppBenchmark::DynamicGraphIterations(DatabaseVisWindow * window)
{
	std::cout<<"Dynami graph iteration"<<std::endl;
	std::vector<int> seconds; 
	std::vector<double> errors;
	
	int nDynIters = m_Pipeline.m_Params.GetIntValue(m_sModuleName, "INPUT_TestIterations");
	// write current results
	double err = WritePairsTable(window);
	
	CollectionDiffuseDynamicGraph * diffusedGraph = NULL;
	if (m_pCorrCollection==NULL 
		|| m_pCorrCollection->GetCollectionType()!="CollectionDiffuseDynamicGraph")
	{
		std::cout<<"[WARNING] Benchmarking non-dynamic graph"<<std::endl;
		return;
	}
	diffusedGraph = (CollectionDiffuseDynamicGraph*)m_pCorrCollection;

	bool converged = false;
	std::vector<double> scores;	
	seconds.push_back(0);
	errors.push_back(err);
	TimeProfiler iterationsProfile;	
	for (int i=0; i<nDynIters && !converged; i++)
	{	
		scores.clear();
		
		// iteration
		iterationsProfile.startedProcess("Iterations");
		converged = diffusedGraph->Step3FullIteration(1, &scores);
		iterationsProfile.finishedProcess("Iterations", false);
		int secs = iterationsProfile.getTotalProcessLengthInSecondsInt("Iterations");
		seconds.push_back(secs);
		
		// re-fill benchmark		
		ClearDrawableErrors();
		TimeProfiler timer;
		std::cout<<"Calculating error for: "<<m_EvalCorrStats.size()<<std::endl;
		for (int j=0; j<(int)m_EvalCorrStats.size(); j++)
		{
			m_EvalCorrStats[j].m_EvalCorrValues.clear();
			m_EvalCorrStats[j].m_bFilled = false;			
		}
		for (int j=0; j<(int)m_EvalCorrStats.size(); j++)
		{
			timer.WriteProgress("TotalProgress", j, (int)m_EvalCorrStats.size());
			FillPerMapStats(m_EvalCorrStats[j]);
		}
		std::cout<<std::endl;
		
		// write new iteration results
		err = WritePairsTable(window);
		errors.push_back(err);
	}
	
	std::cout<<"End Benchmark Dynamic Iterations"<<std::endl;
	std::cout<<"Scores=[";
	for (int i=0; i<(int)scores.size(); i++)
		std::cout<<scores[i]<<" "<<std::flush;
	std::cout<<"]"<<std::endl;
	std::cout<<"Errors=[";
	for (int i=0; i<(int)errors.size(); i++)
		std::cout<<errors[i]<<" "<<std::flush;
	std::cout<<"]"<<std::endl;
	std::cout<<"Times=[";
	for (int i=0; i<(int)seconds.size(); i++)
		std::cout<<seconds[i]<<" "<<std::flush;
	std::cout<<"]"<<std::endl;
}

void AppBenchmark::ClearDrawableErrors()
{
	m_pMainModel = NULL;
	m_pModelTo = NULL;
	m_DrawableForPnt = ModelPoint();
	m_DrawableError.clear();
}

void AppBenchmark::FillDrawAtPnt(DatabaseVisWindow * window, 
								 const ModelPoint & selPnt, bool sort)
{
	if (!m_bWriteErrors)
		return;
	if (selPnt==m_DrawableForPnt)
		return;
	if (selPnt.Invalid())
		return;
	
	m_fMaxErr = 0;
	m_fMaxErrRender = m_Pipeline.m_Params.GetDoubleValue(m_sModuleName, "INPUT_MaxErrRender");
	
	std::cout<<"Draw error on meshes due to selected point"<<std::endl;
	ClearDrawableErrors();
	m_DrawableForPnt = selPnt;
	m_DrawableLinkMain = -1;
	
	ParamParser & par = m_Pipeline.m_Params;
	bool evalPnsExist = true;
	VKString evalPntsName = par.GetStrValue(m_sModuleName, "INPUT_EvalOnPnts", evalPnsExist);
	if (!evalPnsExist)
		evalPntsName = "auto";
	
	std::vector<Sortable> sortedModels;
	for (int i=0; i<m_Pipeline.GetDB()->NumModels(); i++)
	{
		Model3D * model = m_Pipeline.GetDB()->GetModelByID(i);
		std::map<Model3D*, ModelPerVertexValues*>::iterator iter;
		iter = m_CachedPerVertexVals.find(model);
		if (iter == m_CachedPerVertexVals.end())
		{
			m_CachedPerVertexVals[model] = new ModelPerVertexValues(model, false, false); 
			iter = m_CachedPerVertexVals.find(model);
		}
		assert(iter != m_CachedPerVertexVals.end());
		
		iter->second->ClearValues();
		iter->second->SetValues(evalPntsName, 0);
		m_DrawableError[model] = iter->second;
		sortedModels.push_back(Sortable(0, model, i));
	}
	
	std::vector<ModelPoint> pnts;
	std::vector<double> errs;
	
	GetPerPointErrors(selPnt, pnts, errs);
	assert(pnts.size()==errs.size());
	for (int i=0; i<(int)errs.size(); i++)
	{
		//std::cout<<"pnt["<<i<<"].err="<<errs[i]<<std::endl;
		double errVal = 0;
		if (errs[i] < 0)
			errVal = m_fMaxErr;
		else
			errVal=errs[i];
			
		Model3D * model = (Model3D*)pnts[i].GetModel();
		m_fMaxErr = vkMax(m_fMaxErr, errs[i]);
		m_DrawableError[model]->SetVertexValue(pnts[i], errVal);
		sortedModels[model->ModelID()].value += errVal;
		//std::cout<<"Err="<<errVal<<" / "<<m_fMaxErrRender<<std::endl;
	}
	
	for (int i=0; i<m_Pipeline.GetDB()->NumModels(); i++)
	{
		Model3D * model = m_Pipeline.GetDB()->GetModelByID(i);
		m_DrawableError[model]->Interpolate(GetColorInterpolator(model));
		m_DrawableError[model]->InitColorsValueRange(0, m_fMaxErrRender, false, window);
		//m_DrawableError[model]->InitColors(window);
	}
	
	std::sort(sortedModels.begin(), sortedModels.end());
	window->m_NewModelOrder.clear();
	int srcID = selPnt.GetModel()->ModelID();
	window->m_iCurrModelTraversalID = 0;
	for (int i=(int)sortedModels.size()-1; i--; i>=0)
	{
		int modelID = sortedModels[i].id;
		if (modelID != srcID)
			window->m_NewModelOrder.push_back(modelID);
	}
	window->m_NewModelOrder.push_back(srcID);
	window->FillMostSimilarModels(window->m_iCurrModelTraversalID, 
								  window->m_NewModelOrder);
}

CollectionIdentityPointsets * AppBenchmark::GetColorInterpolator(Model3D * model)
{

	std::map<Model3D*, CollectionIdentityPointsets *>::iterator iter;
	iter = m_ColorInterpolators.find(model);
	if (iter != m_ColorInterpolators.end())
		return iter->second;
		
	CollectionIdentityPointsets * retVal = NULL;
	VKString myInterp = "tempBenchmarkToColorInterp";

	bool evalPnsExist;
	ParamParser & par = m_Pipeline.m_Params;
	VKString evalPntsName = par.GetStrValue(m_sModuleName, "INPUT_EvalOnPnts", evalPnsExist);
	assert(evalPnsExist);
	
	retVal = new CollectionIdentityPointsets(myInterp, m_Pipeline.GetDB(), -1,
											 model->ModelName(), "AllVertices",
											 evalPntsName, 1, true);
	m_ColorInterpolators[model] = retVal;
	return retVal;
}

double AppBenchmark::WritePairsTable(DatabaseVisWindow * window)
{
	m_DrawableLinkMain = 2;
	m_bWritingImages = true;
	
	std::vector<int> flags;
	for (int i=0; i<m_Pipeline.GetDB()->NumModels(); i++)
	{
		Model3DRenderingParams * mRend = m_Pipeline.GetDB()->GetModelByID(i)->RenderingParams();
		flags.push_back(mRend->m_iRenderingFlags);
		mRend->SetFlag(Model3DRenderingParams::RENDER_FACES, false);
	}

	// for every pair that was compared
	std::vector<Sortable> sortedVals;
	VKStringList filenames;
	VKString dirName;
	std::vector<int> copyIDs = m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID;
	std::vector<int> & ids = m_Pipeline.GetDB()->GetRenderParams()->m_mWindowToModelID;
	for (int i=0; i<(int)ids.size(); i++)
		ids[i] = -1;
	assert(m_DrawableLinkMain < (int)ids.size());
	int old0 = ids[0];
	int oldID = ids[m_DrawableLinkMain];
	double totalErr=0;
	VKString prefix = VKString("Benchmark/")+window->GetTime()+"/";
	for (int i=0; i<(int)m_EvalCorrStats.size(); i++)
	{
		// render per model error
		ids[0] = m_EvalCorrStats[i].m_pModelFrom->ModelID();
		ids[m_DrawableLinkMain] = m_EvalCorrStats[i].m_pModelTo->ModelID();
		Model3DRenderingParams * mRend = m_EvalCorrStats[i].m_pModelTo->RenderingParams();
		mRend->SetFlag(Model3DRenderingParams::RENDER_FACES, true);
		FillDrawMeshErr(window, m_EvalCorrStats[i].m_pModelFrom, m_DrawableLinkMain);
		double cumVal = 0;
		for (int j=0; j<(int)m_EvalCorrStats[i].m_EvalCorrValues.size(); j++)
			cumVal += m_EvalCorrStats[i].m_EvalCorrValues[j];

		cumVal /= m_EvalCorrStats[i].m_EvalCorrValues.size();
		totalErr += cumVal;
		
		sortedVals.push_back(Sortable(cumVal, NULL, i));
		
		// write to image
		window->WriteImages(prefix, DatabaseVisWindow::WRITE_SCREENSHOT, 
							i, false, false, 
							&dirName, &filenames);
		mRend->SetFlag(Model3DRenderingParams::RENDER_FACES, false);
	}
	
	totalErr /= m_EvalCorrStats.size();
	
	std::sort(sortedVals.begin(), sortedVals.end());
	
	VKString htmlFile = (dirName+"/index.html");
	bool couldGen = m_Pipeline.GetDB()->CreateDirsOnPathToFile(htmlFile);
	assert(couldGen);
	std::ofstream textStream(htmlFile.c_str());
	assert(textStream.is_open());
	textStream<<"<html><head><title>"<<prefix.c_str();
	textStream<<"</title></head><body>";
	textStream<<"<h3>Total Error: "<<totalErr<<"</h3>";
	textStream<<"\n<table border=1>";
	for (int i=(int)sortedVals.size()-1; i>=0; i--)
	{
		int id = sortedVals[i].id;
		textStream<<"\t<tr>\n";
		textStream<<"\t\t<td>"<<sortedVals[i].value;
		textStream<<"<br>"<<m_EvalCorrStats[id].m_pModelFrom->ModelName().c_str();
		textStream<<" -> "<<m_EvalCorrStats[id].m_pModelTo->ModelName().c_str();
		textStream<<"</td>\n";
		textStream<<"\t\t<td><img src=\""<<(filenames[id]+".png").c_str()<<"\"></td>\n";
		textStream<<"\t</tr>\n";
	}
	
	textStream<<"</table></body></html>\n";
	ids = copyIDs;
	for (int i=0; i<m_Pipeline.GetDB()->NumModels(); i++)
	{
		Model3DRenderingParams * mRend = m_Pipeline.GetDB()->GetModelByID(i)->RenderingParams();
		mRend->m_iRenderingFlags = flags[i];
	}

	m_bWritingImages = false;
	ClearDrawableErrors();
	return totalErr;
}

void AppBenchmark::FillDrawMeshErr(DatabaseVisWindow * window, 
								   Model3D * fromModel, int drawableLink)
{
	Model3D * toModel = m_Pipeline.GetDB()->GetModelInSubwindow(drawableLink);
	assert(toModel!=NULL);

	if (m_pMainModel==fromModel && m_pModelTo==toModel)
		return;
	
	m_fMaxErr = m_Pipeline.m_Params.GetDoubleValue(m_sModuleName, "INPUT_MaxErr");
	m_fMaxErrRender = m_Pipeline.m_Params.GetDoubleValue(m_sModuleName, "INPUT_MaxErrRender");

	ClearDrawableErrors();
	std::map<Model3D*, ModelPerVertexValues*>::iterator iter;
	iter = m_CachedPerVertexVals.find(fromModel);
	if (iter == m_CachedPerVertexVals.end())
	{
		m_CachedPerVertexVals[fromModel] = new ModelPerVertexValues(fromModel, false, false); 
		iter = m_CachedPerVertexVals.find(fromModel);
	}
	assert(iter != m_CachedPerVertexVals.end());
	
	iter->second->ClearValues();
	
	m_pMainModel = fromModel;
	m_pModelTo = toModel;
	std::map<Model3D *, std::map<Model3D*, int> >:: iterator iter1;
	iter1 = m_SrcModelToPairs.find(fromModel);
	if (iter1 != m_SrcModelToPairs.end())
	{
		std::map<Model3D*, int>::iterator iter2 = iter1->second.find(toModel);
		
		if (iter2 != iter1->second.end())
		{

			int id = iter2->second;
			
			// fill values
			PerMapStats & info = m_EvalCorrStats[id];
			assert((int)info.m_EvalCorrValues.size()==info.m_pPntsFrom->NumPoints());
			for (int i=0; i<info.m_pPntsFrom->NumPoints(); i++)
			{
				const ModelPoint & p = info.m_pPntsFrom->GetPoint(i);
				double err = info.m_EvalCorrValues[i];
				iter->second->SetVertexValue(p, err);
			}
			
			// prepare values for rendering
			iter->second->Interpolate(GetColorInterpolator(fromModel));	
			//iter->second->InitColors(window);
			bool couldInit = iter->second->InitColorsValueRange(0, m_fMaxErrRender, false, window);
			assert(couldInit);
			m_DrawableError[fromModel] = iter ->second;
		}
		else
			std::cout<<"[WARNING] No bench. stats for "<<toModel->ModelName().c_str()<<std::endl;
	}
	else
		std::cout<<"[WARNING] No bench. stats for "<<fromModel->ModelName().c_str()<<std::endl;
}

void AppBenchmark::WriteErrorsToConcatFile()
{
	int N_SAMPLES = 200;
	double minErr=0;
	double maxErr=.25;
	double derr = (maxErr - minErr) / N_SAMPLES;
	
	std::vector<double> errors;

	for (int i=0; i<(int)m_EvalCorrStats.size(); i++)
	{
		for (int e=0; e<(int)m_EvalCorrStats[i].m_EvalCorrValues.size(); e++)
		{
			double val = m_EvalCorrStats[i].m_EvalCorrValues[e];
			if (val < 0)
				val = maxErr+1.;
			errors.push_back(val);
		}
	}
	
	VKString allErrFile2 = m_Pipeline.GetDB()->GetAppDir(m_sAppName) + "errors_raw_unsorted.txt";
	m_Pipeline.GetDB()->CreateDirsOnPathToFile(allErrFile2);
	std::ofstream textStreamRaw2(allErrFile2.c_str());
	assert(textStreamRaw2.is_open());
	textStreamRaw2<<"E=[ ";
	for (int i=0; i<(int)errors.size(); i++)
		textStreamRaw2<<errors[i]<<" ";
	textStreamRaw2<<"];\n";
	textStreamRaw2.close();
	
	std::sort(errors.begin(), errors.end());
	
	std::vector<double> X;
	std::vector<double> Y;
	for (double err = minErr; err<=maxErr; err+=derr)
	{
		int counter = 0;
		
		while (counter < (int)errors.size()
			   && errors[counter] < err)
			counter++;
		double percentile = (double)counter / (double)errors.size();
		X.push_back(err);
		Y.push_back(percentile);
	}

	VKString errFile = m_Pipeline.GetDB()->GetAppDir(m_sAppName) + "errors.m";
	std::ofstream textStream(errFile.c_str());
	assert(textStream.is_open());

	textStream<<"X=[ ";
	for (int i=0; i<(int)X.size(); i++)
		textStream<<X[i]<<" ";
	textStream<<"];\n";

	textStream<<"Y=[ ";
	for (int i=0; i<(int)Y.size(); i++)
		textStream<<Y[i]<<" ";
	textStream<<"];\n";
	textStream.close();
	
	VKString allErrFile = m_Pipeline.GetDB()->GetAppDir(m_sAppName) + "errors_raw_sorted.txt";
	std::ofstream textStreamRaw(allErrFile.c_str());
	assert(textStreamRaw.is_open());
	
	textStreamRaw<<"E=[ ";
	for (int i=0; i<(int)errors.size(); i++)
		textStreamRaw<<errors[i]<<" ";
	textStreamRaw<<"];\n";
	textStreamRaw.close();
}

void AppBenchmark::WriteResultsChild(std::ofstream & textStream)
{
	textStream<<"Benchmark "<<m_Benchmark<<" "<<m_CorrEvaluation<<"\n";
	textStream<<"NStats "<<m_EvalCorrStats.size()<<"\n";
	for (int i=0; i<(int)m_EvalCorrStats.size(); i++)
	{
		PerMapStats & info = m_EvalCorrStats[i];
		
		int modelFrom = info.m_pModelFrom->ModelID();
		int modelTo = info.m_pModelTo->ModelID();
		std::vector<double> & vals = info.m_EvalCorrValues;
		textStream<<modelFrom<<" "<<modelTo<<" "<<info.m_bFilled<<"\n";
		textStream<<vals.size()<<"\n";
		for (int j=0; j<(int)vals.size(); j++)
			textStream<<vals[j]<<" ";		
	}
	textStream<<"BenchmarkAlign\n";
}

bool AppBenchmark::ReadResultsChild(std::ifstream & textStream)
{
	VKString tempStr;
	tempStr.readToken(textStream);
	assert(tempStr=="Benchmark");
	int benchmark, correval;
	textStream>>benchmark>>correval;
	m_Benchmark = (BenchmarkType)benchmark;
	m_CorrEvaluation = (CorrsEvaluationType)correval;
	int nstats;
	tempStr.readToken(textStream);
	assert(tempStr=="NStats");
	textStream>>nstats;
	
	for (int i=0; i<nstats; i++)
	{		
		int modelFrom, modelTo;
		textStream>>modelFrom>>modelTo;
		AddPairStats(modelFrom, modelTo);
	
		PerMapStats & info = m_EvalCorrStats[i];
		
		textStream>>info.m_bFilled;
		int nVals;
		textStream>>nVals;
		double val;
		for (int j=0; j<nVals; j++)
		{
			textStream>>val;
			info.m_EvalCorrValues.push_back(val);
		}		
	}
	tempStr.readToken(textStream);
	assert(tempStr=="BenchmarkAlign");

	return true;
}

