/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "AppLabeling.h"
#include "ModelValueLabel.h"
#include "Sortable.h"

AppLabeling::AppLabeling(AnalysisPipeline & pipeline, 
						 const VKString & moduleName)
: PipelineStepApplication(pipeline, moduleName)
{
}

int AppLabeling::NumLabelParams()
{
	VKStringList collections = m_Pipeline.m_Params.GetStrValues(m_sModuleName, "INPUT_Collection");
	VKStringList collectionCorrs = m_Pipeline.m_Params.GetStrValues(m_sModuleName, 
																	"INPUT_CollectionCorrValue");
	VKStringList labelNames = m_Pipeline.m_Params.GetStrValues(m_sModuleName, "MEMORY_CreateLabels");
	assert(collections.count()==collectionCorrs.count() && labelNames.count()==collections.count());
	return collections.count();
}

void AppLabeling::FillPtrsFromParams(int i)
{
	VKStringList corrs = m_Pipeline.m_Params.GetStrValues(m_sModuleName, "INPUT_Collection");
	VKStringList corrValues = m_Pipeline.m_Params.GetStrValues(m_sModuleName, "INPUT_CollectionCorrValue");
	VKStringList labelsList = m_Pipeline.m_Params.GetStrValues(m_sModuleName, "INPUT_LabelsModelVal");
	FillPtrs(corrs[i], -1, corrValues[i], labelsList);
}

void AppLabeling::FillPtrsForDraw(DatabaseVisWindow * window)
{
	if (m_Pipeline.m_Params.GetStrValue(m_sModuleName, "INPUT_UseUserSelected")!="true")
	{
		FillPtrsFromParams(0);
		return;
	}
	
	int corrID;
	VKString corrCollection = window->CollectionToRnder(&corrID);
	if (m_Pipeline.GetMainDB()->GetCorrCollection(corrCollection)==NULL)	
	{
		corrCollection = m_Pipeline.m_Params.GetStrValue(m_sModuleName, "INPUT_Collection");
		corrID = -1;
	}
	
	VKString corrCollectionValue = window->CorrValueToRender();
	if (corrCollectionValue=="")
		corrCollectionValue = m_Pipeline.m_Params.GetStrValue(m_sModuleName, "INPUT_CollectionCorrValue");
	
	VKStringList labelsList;
	VKString modelValueToRender = window->ModelValueToRender(0);
	if (modelValueToRender=="")
		labelsList = m_Pipeline.m_Params.GetStrValues(m_sModuleName, "INPUT_LabelsModelVal");
	else
		labelsList.push_back(modelValueToRender);
	
	FillPtrs(corrCollection, corrID, corrCollectionValue, labelsList);
}

void AppLabeling::FillPtrs(const VKString & corrCollection, int corrID, const VKString & corrValue, 
						   VKStringList & labelsList)
{
	m_pCorrVals = NULL;
	m_pCorrs = NULL;
	m_sModelLabelsValName = labelsList;

	if (corrID==-1)
		m_pCorrs = m_Pipeline.GetMainDB()->GetCorrCollection(corrCollection);
	else
	{
		CorrCollection * coll = m_Pipeline.GetMainDB()->GetCorrCollection(corrCollection);
		if (coll==NULL)
			return;
		CollectionPairwise * collPariwise = coll->SafeCastToPairwise();
		if (collPariwise==NULL)
			return;
		m_pCorrs = collPariwise->GetModelCorr(corrID);
	}
	
	if (m_pCorrs!=NULL)
		m_pCorrVals = m_pCorrs->GetCorrValueByName(corrValue);

}

VKString AppLabeling::GetAppType()
{
	return "AppLabeling"; 
}

void AppLabeling::Execute()
{
	if (m_sCreatedModelValues.count()==0)
		m_sCreatedModelValues = m_Pipeline.m_Params.GetStrValues(m_sModuleName, "MEMORY_CreateLabels");
	
	int NParams = NumLabelParams();
	for (int paramID=0; paramID<NParams; paramID++)
	{
		VKString currLabelName = m_sCreatedModelValues[paramID];
		FillPtrsFromParams(paramID);
		VKString models = m_Pipeline.m_Params.GetStrValue(m_sModuleName, "INPUT_ModelsToLabel");	
		
		if (models=="all")
		{
			TimeProfiler timer;
			std::cout<<"Labeling Models"<<std::endl;
			int nModels=m_Pipeline.GetMainDB()->NumModels();
			for (int i=0; i<nModels; i++)
			{
				LabelModel(i, currLabelName);
				timer.WriteProgress("Labeling", paramID * nModels + i, NParams * nModels);
			}
			std::cout<<std::endl;
		}
		else
			assert(false);
	}
}

void AppLabeling::LabelModel(int modelID, const VKString & modelLabelsName)
{
	Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(modelID);

	ModelValueLabel * labelsModelValue = new ModelValueLabel(model, modelLabelsName);
	model->AddModelValue(modelLabelsName, labelsModelValue);
	
	int currSegID = 0;
	std::map<VKString, int> labelSegment;	
	PointSet * pntsToLabel = model->PointCloud();
	for (int j=0; j<pntsToLabel->NumPoints(); j++)
	{
		ModelPoint pnt = pntsToLabel->GetPoint(j);
		
		VKStringList labels;
		std::vector<double> labelScores;
		FillLabelsAtAPoint(pnt, labels, labelScores);
		int segID = (labels.count()==0) ? -1 : GetSegID(labels[0]);
		labelsModelValue->AssignSegmentID(pnt, segID);
	}
	AssignGlobalLabels(labelsModelValue);
	//labelsModelValue->AssignFacesFromPointSet("auto", "auto");
	labelsModelValue->AssignFaceLabels();
}

int AppLabeling::GetSegID(const VKString & label)
{
	std::map<VKString, int>::iterator iter = m_GlobalLabelToSegID.find(label);
	
	if (iter==m_GlobalLabelToSegID.end())
	{
		int newSegID = (int)m_GlobalLabelToSegID.size();
		m_GlobalLabelToSegID[label] = newSegID;
		return newSegID;
	}
	else
		return iter->second;
}

void AppLabeling::AssignGlobalLabels(ModelValueLabel * labelValue)
{
	for (std::map<VKString, int>::iterator iter = m_GlobalLabelToSegID.begin();
		 iter != m_GlobalLabelToSegID.end(); iter++)
		labelValue->AssignLabel(iter->second, iter->first);			
}

void AppLabeling::Draw(DatabaseVisWindow * window)
{
	// draw labels for the clicked point (with probabilities)
	ModelPoint pnt = window->SelectedPoint();
	bool cached = (m_CachedPnt==pnt) && !window->PendingCorrValueRenderingUpdate();
	if (!cached && !pnt.Invalid())
	{
		FillPtrsForDraw(window);
		m_CachedPnt = pnt;
		FillLabelsAtAPoint(pnt, m_CachedLabels, m_CachedLabelValues);
	}

	if (pnt.Invalid())
	{
		m_CachedLabels.clear();
		m_CachedLabelValues.clear();
		m_pCorrs = NULL;
		m_pCorrVals = NULL;
	}
	else 
	{
		for (int i=0; i<m_CachedLabels.count(); i++)
		{
			window->GetRenderer()->RenderText(R2Point(-.05, .95 - i * .025), 
											  m_CachedLabels[i]+" ("+m_CachedLabelValues[i]+")", 
											  1, 0, 0, VisWindowRenderer::TEXT_SIZE_SMALL);
		}
	}
}

void AppLabeling::FillLabelsAtAPoint(const ModelPoint & pnt, 
									 VKStringList & labels, std::vector<double> & labelValues)
{
	labels.clear();
	labelValues.clear();
	
	if (m_pCorrs==NULL)
		return;

	std::vector<ModelPoint> pnts;
	m_pCorrs->GetCorrs(pnt, &pnts);
	std::map<VKString, double> mapLabels;
	for (int i=0; i<(int)pnts.size(); i++)
	{
		if (pnts[i].GetModel()==pnt.GetModel())
			continue;
		double val = (m_pCorrVals!=NULL) ? (m_pCorrVals->Value(pnt, pnts[i])) : (1.);
		for (int j=0; j<m_sModelLabelsValName.count(); j++)
		{
			Model3D * model = ((Model3D *)pnts[i].GetModel());
			ModelValue * modelVal = model->GetModelValue(m_sModelLabelsValName[j]);
			if (modelVal!=NULL && modelVal->GetModelValueType()=="ModelValueLabel")
			{
				ModelValueLabel * labelsModelVal = (ModelValueLabel*)modelVal;
				VKString currLabel = labelsModelVal->GetLabel(pnts[i]);

				
				std::map<VKString, double>::iterator iter = mapLabels.find(currLabel);
				VKString modelName = ((Model3D*)pnts[i].GetModel())->ModelName();
				if (iter==mapLabels.end())
					mapLabels[currLabel] = val;
				else
					iter->second += val;
			}
		}
	}
	
	VKStringList unsortedNames;
	std::vector<Sortable> sortedVals;	
	for (std::map<VKString, double>::iterator iter = mapLabels.begin(); 
		 iter != mapLabels.end(); iter++)
	{
		sortedVals.push_back(Sortable(iter->second, NULL, unsortedNames.count()));
		unsortedNames.push_back(iter->first);
	}
	
	std::sort(sortedVals.begin(), sortedVals.end());
	
	//double threshRatio = 1E-10;
	double threshRatio = 0;
	int nEnd = (int)sortedVals.size()-1;
	for (int i=nEnd; i>=0; i--)
	//for (int i=0; i<nEnd; i++)
	{
		if (sortedVals[nEnd].value * threshRatio < sortedVals[i].value)
		{
			labels.push_back(unsortedNames[sortedVals[i].id]);
			labelValues.push_back(sortedVals[i].value);		
		}
	}
	if (labels.count()==0)
	{
		labels.push_back("UnknownLabel");
		labelValues.push_back(1);
	}
	//assert(labels.count()>0);
}


void AppLabeling::WriteResultsChild(std::ofstream & textStream)
{
	// write co-segmentation / co-labeling results (for the DB)
	textStream<<"AutoLabelingVals "<<m_sCreatedModelValues.count()<<"\n";
	for (int i=0; i<m_sCreatedModelValues.count(); i++)
		textStream<<m_sCreatedModelValues[i].c_str()<<" ";

	textStream<<"\n";
	
	for (int mID = 0; mID < m_Pipeline.GetMainDB()->NumModels(); mID++)
	{
		Model3D * model = m_Pipeline.GetMainDB()->GetModelByID(mID);
		for (int i=0; i<m_sCreatedModelValues.count(); i++)
		{
			ModelValue * modelVal = model->GetModelValue(m_sCreatedModelValues[i]);
			if (modelVal!=NULL)
				modelVal->WriteModelValue(m_Pipeline.GetMainDB());
		}
	}
	
	
	// write benchmark results 
}

bool AppLabeling::ReadResultsChild(std::ifstream & textStream)
{
	// read co-segmentation / co-labeling results (for the DB)
	VKString text;
	text.readToken(textStream);
	assert(text=="AutoLabelingVals");
	int nVals = 0;
	textStream>>nVals;
	
	for (int i=0; i<nVals; i++)
	{
		text.readToken(textStream);
		
		int loaded=0;
		for (int j=0; j<m_Pipeline.GetMainDB()->NumModels(); j++)
			loaded += (int)ModelValue::ReadModelValue(m_Pipeline.GetMainDB()->GetModelByID(j)->ModelName(), 
													  text, m_Pipeline.GetMainDB());
		assert(loaded>0);	// at least one model has this new value
	}
		
	
	// reads benchmark results 
	return true;
}





