/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TimeProfiler.h"
#include "ModelDatabase.h"
#include "PointDescriptor.h"
#include "DatabaseVisWindow.h"


#include "CollectionAggregatePairwise.h"

CollectionAggregatePairwise::CollectionAggregatePairwise(const VKString & collectionName, 
														 ModelDatabase * db)
: CollectionDiffuse(collectionName, db)
{
	m_aPerPointCounter = NULL;
	m_aModelToInClusterGroupID = NULL;
	m_pDiffMap = NULL;
#ifdef INCLUDE_LS_FUZZY
	m_pLSFuzzyCorrs = NULL;
#endif
	m_AggMode = AGG_UNDEFINED;
	
	m_iClusterID = -1;
	m_iNumClusters = 0;
	m_GlobalIDToClusterID = NULL;
	m_GlobalIDToInClusterID = NULL;
	m_InClusterToGlobalID = NULL;
	m_InClusterToGlobalIDSize = NULL;

}

CollectionAggregatePairwise::~CollectionAggregatePairwise()
{
	if (m_pDiffMap!=NULL)
		delete m_pDiffMap;
#ifdef INCLUDE_LS_FUZZY
	if (m_pLSFuzzyCorrs!=NULL)
		delete m_pLSFuzzyCorrs;
#endif
}

DiffusionMap * CollectionAggregatePairwise::GetDiffusionMap()
{
	return m_pDiffMap;
}

void CollectionAggregatePairwise::InitDiffMapAggregation(const CorrMtxParams & mtp,
														 const DiffusionParams & diffP,
														 int nEigs,
														 int * globalIDToClusterID,
														 int * globalIDToInClusterID,
														 int ** inClusterToGlobalID,
														 int * inClusterToGlobalIDSize,
														 int numClusters,
														 int clusterID)

{
	m_iClusterID = clusterID;
	m_iNumClusters = numClusters;
	assert(m_iClusterID < m_iNumClusters && m_iClusterID>=0);
	m_GlobalIDToClusterID = globalIDToClusterID;
	m_GlobalIDToInClusterID = globalIDToInClusterID;
	m_InClusterToGlobalID = inClusterToGlobalID;
	m_InClusterToGlobalIDSize = inClusterToGlobalIDSize;

	m_AggMode = AGG_DIFFUSION_MAP;
	
	std::vector<int> grouping;
	FillModelGrouping(grouping);
		
	m_pDiffMap = new DiffusionMap(inClusterToGlobalIDSize[clusterID], 
								  nEigs, grouping, 
								  diffP.m_ANNSearchType,
								  diffP.m_EigsLib,
								  diffP.m_Dist2SimNorm);

	m_pDiffMap->SetTime(diffP.m_fDiffTime);
	m_pDiffMap->SetMaxDistancePercentile(diffP.m_fDistPercentile);
}

void CollectionAggregatePairwise::InitLinSysMapAggregation(const CorrMtxParams & mtp,
														   int * globalIDToClusterID,
														   int * globalIDToInClusterID,
														   int ** inClusterToGlobalID,
														   int * inClusterToGlobalIDSize,
														   int numClusters,
														   int clusterID)
{
	m_iClusterID = clusterID;
	m_iNumClusters = numClusters;
	assert(m_iClusterID < m_iNumClusters && m_iClusterID>=0);
	m_GlobalIDToClusterID = globalIDToClusterID;
	m_GlobalIDToInClusterID = globalIDToInClusterID;
	m_InClusterToGlobalID = inClusterToGlobalID;
	m_InClusterToGlobalIDSize = inClusterToGlobalIDSize;

	m_AggMode = AGG_LINEAR_SYSTEM;
	std::vector<int> grouping;
	FillModelGrouping(grouping);
	
#ifdef INCLUDE_LS_FUZZY	
	m_pLSFuzzyCorrs = new LSFuzzy(LSFuzzy::TP_MATLAB_SCRIPT, grouping);
//	m_pLSFuzzyCorrs->SetMatlabScript("diffuseWithRenorm");
	m_pLSFuzzyCorrs->SetMatlabScript("completeWithSVT");	
#endif
}

void CollectionAggregatePairwise::FillModelGrouping(std::vector<int> & grouping)
{
	if (m_aModelToInClusterGroupID==NULL)
		m_aModelToInClusterGroupID = new int[m_pDB->NumModels()];
	
	std::vector<Sortable> modelOrder;
	for (int i=0; i<m_pDB->NumModels(); i++)	
	{
		PointSet * pnts = m_pDB->GetModelByID(i)->PointCloud();
		const ModelPoint & p = pnts->GetPoint(0);
		int globalID = m_pDB->GlobalID(p);
		int clusterID = m_GlobalIDToClusterID[globalID];
		if (clusterID==m_iClusterID)
			modelOrder.push_back(Sortable(m_GlobalIDToInClusterID[globalID], NULL, i));
		m_aModelToInClusterGroupID[i] = -1;
	}
	
	std::sort(modelOrder.begin(), modelOrder.end());
	for (int i=0; i<(int)modelOrder.size(); i++)
	{
		int inCluster0 = (int)modelOrder[i].value;
		int mID = modelOrder[i].id;
		PointSet * pnts = m_pDB->GetModelByID(mID)->PointCloud();
		const ModelPoint & p = pnts->GetPoint(0);
		int globalID = m_pDB->GlobalID(p);
		assert(inCluster0==m_GlobalIDToInClusterID[globalID]);
		grouping.push_back(pnts->NumPoints());
		m_aModelToInClusterGroupID[mID] = i;
		if (i<(int)modelOrder.size()-1)
		{
			int nextInCluster0 = (int)modelOrder[i+1].value;
			if (nextInCluster0 != inCluster0 + pnts->NumPoints())
			{
				std::cout<<"[ERROR] Model Order: "<<inCluster0<<" / "<<nextInCluster0<<std::endl;
				assert(nextInCluster0 == inCluster0 + pnts->NumPoints());
			}
		}
	}
	
//	for (int i=0; i<m_pDB->NumModels(); i++)
//	{
//		PointSet * pnts = m_pDB->GetModelByID(i)->PointCloud();
//		const ModelPoint & p = pnts->GetPoint(0);
//		int clusterID = m_GlobalIDToClusterID[m_pDB->GlobalID(p)];
//		if (clusterID==m_iClusterID)
//		{
//			//m_aModelToInClusterGroupID[i] = (int)grouping.size();
//			grouping.push_back(pnts->NumPoints());
//		}
//		else
//			m_aModelToInClusterGroupID[i] = -1;
//
//	}
}

void CollectionAggregatePairwise::ClearPerPointCounter()
{
	if (m_aPerPointCounter==NULL)
		m_aPerPointCounter = new int[m_pDB->NGlobalPnts()];
	for (int i=0; i<m_pDB->NGlobalPnts(); i++)
		m_aPerPointCounter[i] = 0;
}
void CollectionAggregatePairwise::AddEntry(int p1, int p2, double val)
{
	if (m_aPerPointCounter!=NULL)
	{
		m_aPerPointCounter[p1]++;
		m_aPerPointCounter[p2]++;
	}
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert (m_pDiffMap!=NULL);
		m_pDiffMap->AddCorrValue(InClusterID(p1), InClusterID(p2), val);
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		m_pLSFuzzyCorrs->AddValue(InClusterID(p1), InClusterID(p2), val);
	}
#endif
	else
		assert(false);
}

double CollectionAggregatePairwise::GetAggValue(int p1, int p2)
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert (m_pDiffMap!=NULL);
//		// DELETEME: just for testing
//		int * closestIDs;
//		double * closestSims;
//		int nCorrs;
//		int p2ModelID =  m_pDB->GetPoint(p2).GetModel()->ModelID();
//		m_pDiffMap->MostSimilarPointsInGroup(InClusterID(p1), 
//											 m_aModelToInClusterGroupID[p2ModelID],
//											 &closestIDs, &closestSims, &nCorrs);
//		for (int i=0; i<nCorrs; i++)
//			if (m_pDB->GlobalID(GetPoint(closestIDs[i]))==p2)
//				return closestSims[i];
//		assert(false);
//		return -1;
//		// just for testing
		
		return m_pDiffMap->EmbeddedSimilarity(InClusterID(p1), InClusterID(p2));
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		return m_pLSFuzzyCorrs->FuzzyCorr(InClusterID(p1), InClusterID(p2));
	}
#endif
	else
		assert(false);
}

double CollectionAggregatePairwise::DiffusedCorrValue(const ModelPoint & pntFrom, 
													const ModelPoint & pntTo)
{
	return GetAggValue(m_pDB->GlobalID(pntFrom),
					   m_pDB->GlobalID(pntTo));
}

void CollectionAggregatePairwise::InteractiveChangeDiffusionTime(double dt)
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		double newT = m_pDiffMap->DiffTime() + dt;
		if (newT < 0)
			newT = 0;
		std::cout<<"Interactive: Diff. Time["<<m_iClusterID<<"] = "<<newT<<std::endl;
		m_pDiffMap->SetTime(newT);
	}
	else
		std::cout<<"[WARNING] Not using diff maps"<<std::endl;
	
}

void CollectionAggregatePairwise::InteractiveChangeDiffusionSigma(double ds)
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		double diffP = m_pDiffMap->DiffDistPercentile() + ds;
		if (diffP <= 0)
			diffP = 10E-5;
		
		std::cout<<"Interactive: Diff. % ["<<m_iClusterID<<"] = "<<diffP<<std::endl;
		m_pDiffMap->SetMaxDistancePercentile(diffP);
	}
	else
		std::cout<<"[WARNING] Not using diff maps"<<std::endl;	
}


void CollectionAggregatePairwise::ClearPairwiseEntries()
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		m_pDiffMap->ClearAllCorrValues();
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		m_pLSFuzzyCorrs->ClearAll();
	}
#endif
	else
		assert(false);
}

void CollectionAggregatePairwise::Embed(const VKString & tempDir,
										double * timeInSec)
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		bool addId  = true;
		bool symm  = true;
		bool rownorm  = true;
		bool normevecs = true;
		m_pDiffMap->Embed(addId, symm, rownorm, normevecs, timeInSec, tempDir.c_str());
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		m_pLSFuzzyCorrs->Embed(tempDir, 
							   timeInSec);
	}
#endif
	else
		assert(false);
}

VKString CollectionAggregatePairwise::GetCollectionType()
{
	return "CollectionAggregatePairwise";
}

CollectionAggregatePairwise::AggregationMode CollectionAggregatePairwise::GetAggMode() const
{
	return m_AggMode;
}

void CollectionAggregatePairwise::FillValsForPair(int modelFromID, int modelToID,
												  std::map<int, double> * saveVals,
												  int nRowsInSaveVals)
{
	assert(saveVals!=NULL);
	Model3D * modelFrom = m_pDB->GetModelByID(modelFromID);
	PointSet * inPnts = modelFrom->PointCloud();
	int nPoints = inPnts->NumPoints();
	assert(nPoints==nRowsInSaveVals);

	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_aModelToInClusterGroupID!=NULL);
		
		int * closestIDs = NULL;
		double * closestSims = NULL;
		int nCorrs=0;
		int modelToGroupID = m_aModelToInClusterGroupID[modelToID];
		assert(modelToGroupID>=0);
		assert(m_pDiffMap!=NULL);
		for (int i=0; i<nPoints; i++)
		{
			double norm = 0;
			saveVals[i].clear();
			int inClusterID = InClusterID(inPnts->GetPoint(i));
			m_pDiffMap->MostSimilarPointsInGroup(inClusterID, modelToGroupID, &closestIDs, 
												 &closestSims, &nCorrs);

			for (int j=0; j<nCorrs; j++)
			{
				ModelPoint p = GetPoint(closestIDs[j]);
				
//				std::cout<<"p="<<i<<" fromMDL="<<modelFromID<<" toMDL="<<modelToID<<std::endl;
//				std::cout<<"toModelGroup="<<modelToGroupID<<" closestID="<<closestIDs[j]<<std::endl;
//				std::cout<<"Retrieved globalID: "<<m_InClusterToGlobalID[m_iClusterID][inClusterID]<<std::endl;

				if (p.GetModel()->ModelID()!=modelToID)
				{
					std::cout<<"[ERROR] Incorrect model: "<<p.GetModel()->ModelID();
					std::cout<<" need: "<<modelToID<<std::endl;
					assert(p.GetModel()->ModelID()==modelToID);
				}
				norm += closestSims[j];
				saveVals[i][p.PntCloudID()] = closestSims[j];
			}

			if (norm != 0)
			{
				std::map<int, double>::iterator iter;
				for (iter = saveVals[i].begin(); 
					 iter != saveVals[i].end(); iter++)
					iter->second /= norm;
			}
			else
				assert(saveVals[i].size()==0);
		}
	}
	else
	{		
		// there is a better way to do below...
		for (int i=0; i<nPoints; i++)
		{
			double norm = 0;
			
			std::vector<ModelPoint> corrs;
			std::vector<double> vals;
			
			GetCorrs(inPnts->GetPoint(i), &corrs, &vals);
			
			for (int j=0; j<(int)corrs.size(); j++)
			{
				
				if (corrs[j].GetModel()->ModelID() != modelToID)
					continue;
				
				saveVals[i][corrs[j].PntCloudID()] = vals[j];
				norm += vals[j];			
			}
			
			if (norm != 0)
			{
				std::map<int, double>::iterator iter;
				for (iter = saveVals[i].begin(); 
					 iter != saveVals[i].end(); iter++)
					iter->second /= norm;
			}
			else
				assert(saveVals[i].size()==0);
		}
	}
}

double CollectionAggregatePairwise::IntegratedFuzzy(int model1, int model2,
													double ** savedVals)
{
	assert(savedVals!=NULL);
	int m1 = vkMin(model1, model2);
	int m2 = vkMax(model1, model2);
	if (savedVals[m1]!=NULL)
		return savedVals[m1][m2];
	
	Model3D * modelFrom = m_pDB->GetModelByID(m1);
	assert(modelFrom!=NULL);
	PointSet * pntsFrom = modelFrom->PointCloud();

	assert(savedVals[m1]==NULL);
	savedVals[m1] = new double[m_pDB->NumModels()];
	for (int i=0; i<m_pDB->NumModels(); i++)
		savedVals[m1][i] = 0;
	
	for (int i=0; i<pntsFrom->NumPoints(); i++)
	{
		ModelPoint pFrom = pntsFrom->GetPoint(i);
		std::vector<ModelPoint> corrs;
		std::vector<double> vals;
		GetCorrs(pFrom, &corrs, &vals);
		for (int c=0; c<(int)corrs.size(); c++)
		{
			int fm2 = corrs[c].GetModel()->ModelID();
			savedVals[m1][fm2] += vals[c];
		}
	}
	return savedVals[m1][m2];
}

void CollectionAggregatePairwise::GetCorrs(const ModelPoint & pnt, 
										   std::vector<ModelPoint> * corrPnts,
										   std::vector<double> * vals)
{
	int * closestIDs = NULL;
	double * closestSims = NULL;
	int nCorrs=0;
	int inClusterID = InClusterID(pnt);
	
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		m_pDiffMap->MostSimilarPoints(inClusterID, &closestIDs, 
									  &closestSims, &nCorrs);
		
//		// DELETEME JUST FOR TESTING
//		nCorrs = 0;
//		for (int i=0; i<m_pDB->NumModels(); i++)
//		{
//			int toGroupID = m_aModelToInClusterGroupID[i];
//			if (toGroupID==-1)
//				continue;
//			m_pDiffMap->MostSimilarPointsInGroup(inClusterID, toGroupID, &closestIDs, 
//												 &closestSims, &nCorrs);
//			for (int j=0; j<nCorrs; j++)
//			{
//				const ModelPoint p = GetPoint(closestIDs[j]);
//				corrPnts->push_back(p);
//				if (vals!=NULL)
//					vals->push_back(closestSims[j]);
//			}		
//		}
//		return;
//		// JUST FOR TESTING
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		m_pLSFuzzyCorrs->FuzzyCorrs(inClusterID, &closestIDs, 
									&closestSims, &nCorrs);
	}
#endif
	else
		assert(false);
	
	assert(closestIDs!=NULL && closestSims!=NULL);
	for (int i=0; i<nCorrs; i++)
	{
		const ModelPoint p = GetPoint(closestIDs[i]);
		corrPnts->push_back(p);
		if (vals!=NULL)
			vals->push_back(closestSims[i]);
	}
}

void CollectionAggregatePairwise::GetCorrs(const ModelPoint & pnt, 
										   std::vector<ModelPoint> * corrPnts)
{
	GetCorrs(pnt, corrPnts, NULL);
}

ModelPoint CollectionAggregatePairwise::GetBestCorr(const ModelPoint & pnt, 
												Model3D * toModel, 
												const VKString & modelValue, 
												double * value)
{
	int locID1 = InClusterID(pnt);
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		int bestPnt = -1;
		assert(m_pDiffMap!=NULL);

		int bestLoc2= -1;
		PointSet * pnts = toModel->PointCloud();
		double minDist = FLT_MAX;
		for (int i=0; i<pnts->NumPoints(); i++)
		{
			int locID2 = InClusterID(pnts->GetPoint(i));
			double dist = m_pDiffMap->EmbeddedDistanceExact2(locID1, locID2);
			if (dist < minDist)
			{
				minDist = dist;
				bestPnt = i;
				bestLoc2 = locID2;
			}
		}
		
		assert(bestLoc2>=0);
		if (value!=NULL)
			*value = m_pDiffMap->EmbeddedSimilarityExact(locID1, bestLoc2);
		assert(bestPnt>=0);
		return pnts->GetPoint(bestPnt);
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		int locID2 = InClusterID(toModel->PointCloud()->GetPoint(0));
		int g2 = m_pLSFuzzyCorrs->GetGroup(locID2);
		int bestID = m_pLSFuzzyCorrs->BestCorr(locID1, g2, value);
		return m_pDB->GetPoint(bestID);
	}
#endif
	assert(false);
	return ModelPoint();
}

int CollectionAggregatePairwise::InClusterID(int globalID)
{
	if (!(m_GlobalIDToClusterID!=NULL && globalID >= 0))
	{
		std::cout<<"[ERROR] Invalid global: "<<globalID;
		if (m_GlobalIDToClusterID==NULL)
			std::cout<<". INVALID ARRAY"<<std::endl;
		else
			std::cout<<". Array is ok"<<std::endl;
		assert(m_GlobalIDToClusterID!=NULL && globalID >= 0);
	}
	int cID = m_GlobalIDToClusterID[globalID];
	if (cID!=m_iClusterID)
	{
		std::cout<<"[ERROR] Getting ID for wrong cluster: "<<cID;
		std::cout<<" should be: "<<m_iClusterID<<std::endl;
		assert(cID==m_iClusterID);
	}
	assert(m_GlobalIDToInClusterID!=NULL);
	return m_GlobalIDToInClusterID[globalID];
}

int CollectionAggregatePairwise::InClusterID(const ModelPoint & pnt)
{
	int gp = m_pDB->GlobalID(pnt);	
	return InClusterID(gp);
}

ModelPoint CollectionAggregatePairwise::GetPoint(int inClusterID)
{
	int globalID = m_InClusterToGlobalID[m_iClusterID][inClusterID];
	assert(globalID < m_pDB->NGlobalPnts());
	return m_pDB->GetPoint(globalID);
}


void CollectionAggregatePairwise::Write(const VKString & directory, const VKString & prefix)
{
	VKString mainFile = directory + prefix + "_agg.txt";
	std::ofstream textStream(mainFile.c_str());
	assert(textStream.is_open());
	textStream<<"AggType "<<m_AggMode<<"\n";
	//textStream<<"NEntries "<<m_AggMode<<"\n";
	//textStream<<"NEigs "<<m_AggMode<<"\n";
	
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		m_pDiffMap->WriteBinaryData(directory.c_str(), prefix.c_str());
		//	// save eigenvectors / eigenvalues
		//	if (m_aDiffusionMaps.size()==1)
		//	{
		//		if (m_bKeepDrawableLog)
		//			AddEigenvaluesToGraphLog();
		//		
		//		VKString mainDir = m_pDB->GetCollectionDir(m_sCollectionName);
		//		
		//		LinAlgMatrixReal * evecs=NULL;
		//		LinAlgVectorReal * evals=NULL;
		//		m_aDiffusionMaps[0]->Eigs(&evecs, &evals);
		//		assert(evecs!=NULL && evals!=NULL);
		//		int iterID = 0;
		//		VKString evalsStr;
		//		VKString evecsStr;
		//		VKString iterIDStr = "none";
		//		while (true)
		//		{
		//			iterIDStr = VKString::number(iterID);
		//			while (iterIDStr.length() < 4)
		//				iterIDStr = VKString("0") + iterIDStr;
		//			evalsStr = mainDir + "/Iter_" + iterIDStr + "_evals.bin";
		//			evecsStr = mainDir + "/Iter_" + iterIDStr + "_evecs.bin";
		//			if (!m_pDB->FileExists(evalsStr))
		//				break;
		//			iterID++;
		//		}
		//		m_pDB->CreateDirsOnPathToFile(evalsStr);
		//		evecs->WriteMatrixBin(evecsStr.c_str());
		//		evals->WriteVectorBin(evalsStr.c_str());
		//	}
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		m_pLSFuzzyCorrs->Write(directory.c_str(), prefix.c_str());
	}
#endif
	else
		assert(false);
	textStream.close();
}

void CollectionAggregatePairwise::Read(const VKString & directory, const VKString & prefix)
{
	VKString tempStr;
	VKString mainFile = directory + prefix + "_agg.txt";
	std::ifstream textStream(mainFile.c_str());
	assert(textStream.is_open());
	tempStr.readToken(textStream);
	assert(tempStr=="AggType");
	int aggType;
	textStream>>aggType;
	m_AggMode = (AggregationMode)aggType;
	
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		bool couldRead = m_pDiffMap->ReadBinaryData(directory.c_str(), prefix.c_str());
		assert(couldRead);
	}
#ifdef INCLUDE_LS_FUZZY	
	else if (m_AggMode==AGG_LINEAR_SYSTEM)
	{
		assert(m_pLSFuzzyCorrs!=NULL);
		m_pLSFuzzyCorrs->Read(directory.c_str(), prefix.c_str());
	}
#endif
	else
		assert(false);
}


void CollectionAggregatePairwise::DrawAllEmbeddedPoints(DatabaseVisWindow * window,
														int basedOnSingleModel)
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		LinAlgMatrixReal * evecs=NULL;
		LinAlgVectorReal * evals=NULL;
		
		m_pDiffMap->Eigs(&evecs, &evals);
		
		double t = m_pDiffMap->DiffTime();
		
		double minX, maxX, minY, maxY;
		GetEmbeddedBounds(minX, minY, maxX, maxY, basedOnSingleModel);
		
		glDisable(GL_LIGHTING);
		//	glDisable(GL_DEPTH_TEST);
		glPointSize(1);
		
		glBegin(GL_POINTS);
		glColor3d(0, 1., 0);
		// render all points
		for (int i=0; i<m_pDiffMap->NEntries(); i++)
		{
			double x = pow((*evals)(1), t) * (*evecs)(i, 1);
			double y = pow((*evals)(2), t) * (*evecs)(i, 2);
			
			x = (x - minX) / (maxX-minX);
			y = (y - minY) / (maxY-minY);
			
			glVertex3d(x,y, -.2);
		}
		
		glEnd();
		
		// render selected
		ModelPoint selPnt = window->SelectedPoint();
		if (!selPnt.Invalid())
		{
			int selID = m_pDB->GlobalID(selPnt);
			int inClusterID = -1;
			if (m_GlobalIDToClusterID[selID]!=-1)
			{
				int localID = m_GlobalIDToInClusterID[selID];
				assert(localID!=-1);
				
				double x = pow((*evals)(1), t) * (*evecs)(localID, 1);
				double y = pow((*evals)(2), t) * (*evecs)(localID, 2);
				
				x = .05 + .9 * (x - minX) / (maxX-minX);
				y = .05 + .9 * (y - minY) / (maxY-minY);
				
				glPointSize(4);		
				glBegin(GL_POINTS);
				glColor3d(0, 0, 1);	
				glVertex3d(x, y, -.1);
				glEnd();
			}
		}	
	}
}

void CollectionAggregatePairwise::SelectEmbeddedPoint(DatabaseVisWindow * window,
													  double inx, double iny,
													  int basedOnSingleModel)
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		LinAlgMatrixReal * evecs = NULL;
		LinAlgVectorReal * evals = NULL;
		m_pDiffMap->Eigs(&evecs, &evals);
		
		if (evecs==NULL || evals==NULL 
			|| evals->Rows() < 3 || evecs->Cols() < 3)
		{
			std::cout<<"[WARNING] Diffusion map not defined"<<std::endl;
			return;
		}
		
		double minX, minY, maxX, maxY;
		GetEmbeddedBounds(minX, minY, maxX, maxY, basedOnSingleModel);
		double t = m_pDiffMap->DiffTime();
		double bestdist = FLT_MAX;
		int bestID = -1;
		int nPoints = m_pDB->NGlobalPnts();
		for (int i=0; i<nPoints; i++)
		{
			int inClusterID = -1;
			if (m_GlobalIDToClusterID[i]==-1
				|| m_GlobalIDToClusterID[i]!=m_iClusterID)
				continue;
			
			int localID = m_GlobalIDToInClusterID[i];
			assert(localID!=-1);
			
			double x = pow((*evals)(1), t) * (*evecs)(localID, 1);
			double y = pow((*evals)(2), t) * (*evecs)(localID, 2);
			
			x = .05 + .9 * (x - minX) / (maxX-minX);
			y = .05 + .9 * (y - minY) / (maxY-minY);
			
			double dist = pow(x-inx, 2.) + pow(y-iny, 2.);
			if (dist < bestdist)
			{
				bestdist = dist;
				bestID = i;
			}
		}
		
		if (bestID==-1)
		{
			std::cout<<"[WARNING] bestID=-1"<<std::endl;
			return;
		}
		ModelPoint bestPnt = m_pDB->GetPoint(bestID);
		if (bestPnt.Invalid())
		{
			std::cout<<"[WARNING] BestPnt=INVALID"<<std::endl;
			return;
		}
		int modelID = ((Model3D*)bestPnt.GetModel())->ModelID();
		
		int wndID = window->SelectedWindowID();
		if (wndID<0)
			m_pDB->GetRenderParams()->m_mWindowToModelID[0] = modelID;
		else
			m_pDB->GetRenderParams()->m_mWindowToModelID[wndID] = modelID;
		
		((Model3D*)bestPnt.GetModel())->RenderingParams()->SelectPoint(bestPnt.PntCloudID());
	}
}

void CollectionAggregatePairwise::GetEmbeddedBounds(double &minX, double &minY, 
													double &maxX, double &maxY,
													int basedOnSingleModel)
{
	if (m_AggMode==AGG_DIFFUSION_MAP)
	{
		assert(m_pDiffMap!=NULL);
		minX = FLT_MAX;
		minY = FLT_MAX;
		maxX = -FLT_MAX;
		maxY = -FLT_MAX;
		
		LinAlgMatrixReal * evecs=NULL;
		LinAlgVectorReal * evals=NULL;
		
		m_pDiffMap->Eigs(&evecs, &evals);
		
		double t = m_pDiffMap->DiffTime();
		// find min/max
		//for (int i=0; i<nPoints; i++)
		for (int i=0; i<m_pDiffMap->NEntries(); i++)
		{
			int globalID = m_InClusterToGlobalID[m_iClusterID][i];
			int modelID = m_pDB->GetPoint(globalID).GetModel()->ModelID();
			
			if (basedOnSingleModel == -1 || basedOnSingleModel==modelID)
			{
				double x = pow((*evals)(1), t) * (*evecs)(i, 1);
				double y = pow((*evals)(2), t) * (*evecs)(i, 2);
				minX = vkMin(minX, x);
				minY = vkMin(minY, y);
				maxX = vkMax(maxX, x);
				maxY = vkMax(maxY, y);
			}
		}	
	}
}

