/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TimeProfiler.h"
#include "ModelDatabase.h"
#include "PointDescriptor.h"


#include "CollectionDiffuse.h"

CollectionDiffuse::CollectionDiffuse(const VKString & collectionName, ModelDatabase * db)
: CorrCollection(collectionName, db, -1)
{
}

CollectionDiffuse::CollectionDiffuse(const VKString & collectionName, ModelDatabase * db, 
									 const CorrMtxParams & corrMtxParams,
									 const DiffusionParams & diffusionParams)
: CorrCollection(collectionName, db, -1)
{
	m_DiffusionParams = diffusionParams;
	m_CorrMtxParams = corrMtxParams;
}

void CollectionDiffuse::CallAddEntryForAll()
{
	if (m_CorrMtxParams.m_Type==CORR_MTX_PAIRWISE_CORRS)
		CallAddEntryForAllPairwiseCollections();
	else if (m_CorrMtxParams.m_Type==CORR_MTX_POINT_DESCRIPTORS)
		CallAddEntryForAllPointDescriptors();
	else if (m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_SCRIPT)
		CallAddEntryForAllExternalScript();
	else if(m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_FILE)
		CallAddEntryForAllExternalFile();
	else 
		assert(false);
}

void CollectionDiffuse::CallAddEntryForAllPairwiseCollections()
{
	assert(m_CorrMtxParams.m_Type==CORR_MTX_PAIRWISE_CORRS);
	int nModels = m_pDB->NumModels();
	int nGlobalPnts= m_pDB->NGlobalPnts();
	
	std::cout<<"Similarity Matrix: Fill "<<nGlobalPnts<<" x "<<nGlobalPnts<<std::endl;
	TimeProfiler timer;		
	// fill matrix (iterate over all correspondences)
	CorrCollection * collection = m_pDB->GetCorrCollection(m_CorrMtxParams.m_sInputCollection);
	assert(collection!=NULL);
	CollectionPairwise * singleSrcColl = collection->SafeCastToPairwise();
	assert(singleSrcColl!=NULL);
	for (int i=0; i<singleSrcColl->NumModelCorrs(); i++)
	{
		timer.WriteProgress("FillMtx", i, singleSrcColl->NumModelCorrs());
		
		CollectionSinglePair * corr = singleSrcColl->GetModelCorr(i);
		assert(corr!=NULL);
		
		int fromID = corr->ModelFrom()->ModelID();
		int toID = corr->ModelTo()->ModelID();
		
		Model3D * model = corr->ModelFrom();		
		PointSet * pntCld = model->PointCloud();
		assert(pntCld!=NULL);
		
		CorrValue * globalVal = corr->GetCorrValueByName(m_CorrMtxParams.m_sGlobalCorrValue);
		double globalScale = 1;
		if (globalVal!=NULL)
		{
			double integratedVal = globalVal->IntegratedValue(model->ModelName());
			assert(integratedVal<=1. + 1E-10 && integratedVal>=0);
			if (integratedVal < m_CorrMtxParams.m_fMinOverlap)
				globalScale = 0;
			else if (integratedVal > m_CorrMtxParams.m_fMaxOverlap)
				globalScale = 1.;
			else
				globalScale = integratedVal / m_CorrMtxParams.m_fMaxOverlap;
		}
		
		if (globalScale <= m_CorrMtxParams.m_fValueThreshold)	// enforce sparsity
			continue;
		
		CorrValue * localVal = corr->GetCorrValueByName(m_CorrMtxParams.m_sLocalCorrValue);
		
		for (int p1ID=0; p1ID < pntCld->NumPoints(); p1ID++)
		{
			ModelPoint p1 = pntCld->GetPoint(p1ID);
//			int mtxID1 = MatrixPosition(p1, m_aModelPntToGlobalID);
			int mtxID1 = m_pDB->GlobalID(p1);
			std::vector<ModelPoint> corrPnts;
			corr->GetCorrs(p1, &corrPnts);
			
			for (int p2ID=0; p2ID<(int)corrPnts.size(); p2ID++)
			{
				ModelPoint p2 = corrPnts[p2ID];
				int toModelID = p2.GetModel()->ModelID();
//				int mtxID2 = MatrixPosition(p2, m_aModelPntToGlobalID);
				int mtxID2 = m_pDB->GlobalID(p2);
				
				double atPntVal = 1.;
				if (localVal!=NULL)
					atPntVal = localVal->Value(p1, p2);
				
						// enforce sparsity
				if (globalScale * atPntVal > m_CorrMtxParams.m_fValueThreshold)
					AddEntry(mtxID1, mtxID2, globalScale * atPntVal);
			}
		}
	}
	std::cout<<std::endl;	
}

void CollectionDiffuse::CallAddEntryForAllPointDescriptors()
{
	assert(m_CorrMtxParams.m_Type==CORR_MTX_POINT_DESCRIPTORS);
	
	int nModels = m_pDB->NumModels();

	VKString descrStr = m_CorrMtxParams.m_sPointDescriptor;
	for (int m1=0; m1<nModels; m1++)
	{
		Model3D * model1 = m_pDB->GetModelByID(m1);
		assert(model1!=NULL);
		PointSet * pnts1 = model1->PointCloud();
		PointDescriptorsContainer * container1 = pnts1->GetSetDescriptor(descrStr);
		assert(container1!=NULL);		
		
		for (int p1=0; p1<pnts1->NumPoints(); p1++)
		{
			ModelPoint pnt1 = pnts1->GetPoint(p1);
			PointDescriptor * descr1 = container1->GetDescriptor(p1);
			assert(descr1!=NULL);
			//int globID1 = MatrixPosition(pnt1, m_aModelPntToGlobalID);
			int globID1 = m_pDB->GlobalID(pnt1);
			
			for (int m2=0; m2<nModels; m2++)
			{
				Model3D * model2 = m_pDB->GetModelByID(m2);
				assert(model2!=NULL);
				PointSet * pnts2 = model2->PointCloud();
				PointDescriptorsContainer * container2 = pnts2->GetSetDescriptor(descrStr);
				assert(container2!=NULL);
				
				for (int p2=0; p2<pnts2->NumPoints(); p2++)
				{
					ModelPoint pnt2 = pnts2->GetPoint(p2);
					//int globID2 = MatrixPosition(pnt2, m_aModelPntToGlobalID);
					int globID2 = m_pDB->GlobalID(pnt2);
					PointDescriptor * descr2 = container2->GetDescriptor(p2);
					assert(descr2!=NULL);
					double val = descr1->MeasureSimilarity(*descr2);
					AddEntry(globID1, globID2, val);
				}
			}	
		}
	}
}

void CollectionDiffuse::CallAddEntryForAllExternalScript()
{
	assert (m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_SCRIPT);
	assert(false);
}

void CollectionDiffuse::CallAddEntryForAllExternalFile()
{
	assert (m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_FILE);
	
	VKString fileDir = m_CorrMtxParams.m_sExternalScriptName;
	VKString simFileName = m_CorrMtxParams.m_sGlobalCorrValue;
	VKString dbFileName = m_CorrMtxParams.m_sLocalCorrValue;
	std::cout<<"Reading similarities from external file"<<std::endl;
	std::cout<<"\tDir="<<fileDir.c_str()<<" Sim="<<simFileName.c_str();
	std::cout<<" DB="<<dbFileName.c_str()<<std::endl;
	
	std::ifstream textStreamSim((fileDir+simFileName).c_str());
	assert(textStreamSim.is_open());
	std::ifstream textStreamDB((fileDir+dbFileName).c_str());
	assert(textStreamDB.is_open());
	
	std::map<int, int> inputModelIDToMyID;
	VKString dbLineStr;
	int lineID=0;
	while (!textStreamDB.eof())
	{
		dbLineStr.readLine(textStreamDB);
		VKStringList dbLineList = dbLineStr.split("/");
		if (dbLineList.count()>1)
		{
			VKString modelName = dbLineList[dbLineList.count()-1].replace(".cld", "");
			Model3D * model = m_pDB->GetModelByName(modelName);
			assert(model!=NULL);
			inputModelIDToMyID[lineID] = model->ModelID();
			lineID++;
		}
	}	
	
	VKString simLineStr;
	while (!textStreamSim.eof())
	{
		simLineStr.readLine(textStreamSim);
		VKStringList simLineList = simLineStr.split(" ");
		if (simLineList.count()>1)
		{
			assert(simLineList.count()==5);
			bool ok; 
			
			int m1ID = simLineList[0].toInt(&ok);		assert(ok);
			int p1ID = simLineList[1].toInt(&ok);		assert(ok);
			int m2ID = simLineList[2].toInt(&ok);		assert(ok);
			int p2ID = simLineList[3].toInt(&ok);		assert(ok);
			double val = simLineList[4].toDouble(&ok);	assert(ok);
			// switch to 0-based
			p1ID--;
			p2ID--;
			
			assert(inputModelIDToMyID.find(m1ID)!=inputModelIDToMyID.end());
			assert(inputModelIDToMyID.find(m2ID)!=inputModelIDToMyID.end());
			
			Model3D * model1 = m_pDB->GetModelByID(inputModelIDToMyID[m1ID]);
			assert(model1!=NULL);
			Model3D * model2 = m_pDB->GetModelByID(inputModelIDToMyID[m2ID]);
			assert(model2!=NULL);
			ModelPoint pnt1 = model1->PointCloud()->GetPoint(p1ID);
			ModelPoint pnt2 = model2->PointCloud()->GetPoint(p2ID);
			
			AddEntry(m_pDB->GlobalID(pnt1), m_pDB->GlobalID(pnt2), val);
		}
	}
}

//int CollectionDiffuse::NPoints(ModelDatabase * db)
//{
//	int NModels = db->NumModels();
//	int nPoints = -1;
//	
//	for (int i=0; i<NModels; i++)
//	{
//		int pointsOnModel = db->GetModelByID(i)->PointCloud()->NumPoints();
//		if (nPoints==-1)
//			nPoints = pointsOnModel;
//		else
//			assert(nPoints==pointsOnModel);
//	}
//	assert(nPoints!=-1);
//	return nPoints;
//}

DiffusionMap::DistToSimNormType CollectionDiffuse::StrToDist2Sim(const VKString & str)
{
	if (str=="D2S_PRESCRIBE_REGIONS")
		return DiffusionMap::D2S_PRESCRIBE_REGIONS;
	else if (str=="D2S_PER_POINT")
		return DiffusionMap::D2S_PER_POINT;
	else if (str=="D2S_GLOBAL_MIN")
		return DiffusionMap::D2S_GLOBAL_MIN;
	else if (str=="D2S_GLOBAL_MAX")
		return DiffusionMap::D2S_GLOBAL_MAX;
	
	assert(false);
	return DiffusionMap::D2S_PRESCRIBE_REGIONS;	
}

DiffusionMap::DiffusedSearchType CollectionDiffuse::StrToDiffSearch(const VKString & str)
{
	if (str=="SEARCH_EXHAUSTIVE")
		return DiffusionMap::SEARCH_EXHAUSTIVE;
	else if (str=="SEARCH_USE_FLANN")
		return DiffusionMap::SEARCH_USE_FLANN;
	
	assert(false);
	return DiffusionMap::SEARCH_EXHAUSTIVE;	
}

DiffusionMap::EigenDecompositionLibrary CollectionDiffuse::StrToEigsLib(const VKString & str)
{
	if (str=="EIGS_MATLAB")
		return DiffusionMap::EIGS_MATLAB;
	else if (str=="EIGS_SVDLIBC")
		return DiffusionMap::EIGS_SVDLIBC;
	
	assert(false);
	return DiffusionMap::EIGS_MATLAB;
}

void CollectionDiffuse::ReadMtxParams(std::ifstream &textStream)
{
	VKString text;
	
	text.readToken(textStream);
	assert(text=="CorrMtxType");
	int type;
	textStream>>type;
	m_CorrMtxParams.m_Type = (CorrMtxType)type;
	
	if (m_CorrMtxParams.m_Type==CORR_MTX_PAIRWISE_CORRS)
	{
		m_CorrMtxParams.m_sInputCollection.readToken(textStream);
		m_CorrMtxParams.m_sLocalCorrValue.readToken(textStream);
		m_CorrMtxParams.m_sGlobalCorrValue.readToken(textStream);
		text.readToken(textStream);
		assert(text=="SurfFrac");
		textStream>>m_CorrMtxParams.m_fMinOverlap>>m_CorrMtxParams.m_fMaxOverlap;
	}
	else if (m_CorrMtxParams.m_Type==CORR_MTX_PAIRWISE_CORRS)
	{
		m_CorrMtxParams.m_sPointDescriptor.readToken(textStream);
	}
	else if (m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_SCRIPT)
	{
		m_CorrMtxParams.m_sExternalScriptName.readToken(textStream);
	}
	else if (m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_FILE)
	{
		m_CorrMtxParams.m_sExternalScriptName.readToken(textStream);
		m_CorrMtxParams.m_sLocalCorrValue.readToken(textStream);
		m_CorrMtxParams.m_sGlobalCorrValue.readToken(textStream);
	}
	else
		assert(false);
	text.readToken(textStream);
	assert(text=="MinValue");
	textStream>>m_CorrMtxParams.m_fValueThreshold;
}


void CollectionDiffuse::WriteMtxParams(std::ofstream &textStream)
{
	textStream<<"CorrMtxType "<<m_CorrMtxParams.m_Type<<"\n";
	if (m_CorrMtxParams.m_Type==CORR_MTX_PAIRWISE_CORRS)
	{		
		textStream<<m_CorrMtxParams.m_sInputCollection.c_str();
		textStream<<" "<<m_CorrMtxParams.m_sLocalCorrValue.c_str();
		textStream<<" "<<m_CorrMtxParams.m_sGlobalCorrValue.c_str()<<"\n";
		textStream<<"SurfFrac "<<m_CorrMtxParams.m_fMinOverlap<<" ";
		textStream<<m_CorrMtxParams.m_fMaxOverlap<<"\n";
	}
	else if (m_CorrMtxParams.m_Type==CORR_MTX_POINT_DESCRIPTORS)
	{
		textStream<<m_CorrMtxParams.m_sPointDescriptor.c_str()<<"\n";
	}
	else if (m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_SCRIPT)
	{
		textStream<<m_CorrMtxParams.m_sExternalScriptName.c_str()<<"\n";
	}
	else if (m_CorrMtxParams.m_Type==CORR_MTX_EXTERNAL_FILE)
	{
		textStream<<m_CorrMtxParams.m_sExternalScriptName.c_str();
		textStream<<" "<<m_CorrMtxParams.m_sLocalCorrValue.c_str();
		textStream<<" "<<m_CorrMtxParams.m_sGlobalCorrValue.c_str()<<"\n";		
	}
	else 
		assert(false);
	
	textStream<<"MinValue "<<m_CorrMtxParams.m_fValueThreshold<<"\n";	
	
}

void CollectionDiffuse::ReadDiffParams(std::ifstream &textStream)
{
	assert(false);	//todo
}

void CollectionDiffuse::WriteDiffParams(std::ofstream &textStream)
{
	assert(false);	//todo
}


