/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CollectionDiffuseDynamicGraph_H
#define __CollectionDiffuseDynamicGraph_H

#include "CollectionAggregatePairwise.h"
#include "DynamicGraphElements.h"
#include "CollectionDiffuse.h"
#include "DiffusionMap.h"
#include "TimeProfiler.h"

class AnalysisPipeline;
class DiffusionMap;

struct NodeConsistencyData
{
	int nodeID;
	std::vector<Sortable> edgeConsistency;
};

/**
 * Main (dynamic graph-based) alogrithm for finding diffused correspondences:
 * iteratively update links (inter-surface maps) to create the best mapping 
 * between models.
 */ 
class CollectionDiffuseDynamicGraph : public CollectionDiffuse
{
	public:
		friend class DynamicGraphLogParams;
		friend class GraphEdgeMap;
		friend class GraphNodeModel;
	
		enum DynamicGraphState
		{
			STATE_UNINITALIZED,
			STATE_INITIALIZED_GRAPH,
			STATE_EXPLORED_INITIAL_LINKS,
			STATE_DIFFUSED_INITIAL_LINKS,
			STATE_ITER_CONSISTENCY_OPTIM,
			STATE_ITER_DIFFUSION,
			STATE_CONVERGED
		};
		CollectionDiffuseDynamicGraph(const VKString & collectionName, 
									  AnalysisPipeline * pipeline, 
									  const VKString & moduleName);
		
	// execute the whole pipeline
		virtual void CreateOptimalDiffusedMap();
	// steps of the pipeline (added just for the code readibility)
		virtual void Step0FillNodesAndEdges();
		virtual void Step1FillEdgeValuesWithShapeDescriptors();
		virtual void Step2ConstructInitialSimilarityGraph();
		virtual void EnsureConnectivity(std::vector<int> & nodeToCC,
										std::vector<std::vector<int> > & ccToNodes,
										std::vector<Sortable> & remainingEdges,
										int kEdge, std::vector<GraphEdgeMap*> & addEdges);
		virtual bool Step3FullIteration(int nIterations, 
										std::vector<double> * saveScores = NULL);
	
		virtual void DiffuseModifiedLinks();
		//virtual int OptimizeConsistency();		// version 1
		virtual int OptimizeAlignmentGraph();	// version 2
		
		virtual void UpdateHistoryWithEdges(std::vector<int> & includedEdges,
											std::vector<int> &iterDataEdgeID,
											std::vector<int> &iterDataOldLinks,
											std::vector<int> &iterDataNewLinks,
											std::vector<std::vector<double> > &iterDataAllScores,
											std::vector<std::vector<double> > &iterDataAllCorrVals);

		virtual void FindRandomShortcutEdges(std::vector<int> & existingEdges,
											std::vector<int> & edgesToInclude,
											int &nApproxNeeded,
											std::vector<Sortable> & nodeOrder);

		virtual int FindGoodShortcutEdges(std::vector<int> & existingEdges,
										  std::vector<int> & edgesToInclude,
										  int &nApproxNeeded);
		virtual void TreatIsolatedNodes(std::vector<int> & existingEdges,
										std::vector<int> & edgesToInclude,
										std::vector<int> & isolatedNodes);
	
		virtual void WriteEdgeUpdates(std::vector<int> &iterDataEdgeID,
									  std::vector<int> &iterDataOldLinks,
									  std::vector<int> &iterDataNewLinks,
									  std::vector<std::vector<double> > &iterDataAllScores,
									  std::vector<std::vector<double> > &iterDataAllCorrVals);
		
	
		//virtual void OptimizeConsistencyForCluster(int clusterID);
		//virtual void FindEdgesFromWeakNodes(std::vector<int> & edgeIDs);
		virtual void FindExistingEdges(std::vector<int> & existingEdges);
		virtual void PruneWeakEdges(NodeConsistencyData * nodeData,
									double globalThreshold,
									std::vector<int> &iterDataEdgeID,
									std::vector<int> &iterDataOldLinks,
									std::vector<int> &iterDataNewLinks,
									std::vector<std::vector<double> > &iterDataAllScores,
									std::vector<std::vector<double> > &iterDataAllCorrVals,
									int * saveNExcludedPerNode,
									int * saveNExcludedGlobal,
									std::vector<Sortable> & sortedPrunnedNodes);
	
		virtual double OptimizeAllEdges(int * saveNIncluded,
									 int * saveNModified,
									 bool optimize,
									 NodeConsistencyData * nodeData,
									 std::vector<int> &iterDataEdgeID,
									 std::vector<int> &iterDataOldLinks,
									 std::vector<int> &iterDataNewLinks,
									 std::vector<std::vector<double> > &iterDataAllScores,
									 std::vector<std::vector<double> > &iterDataAllCorrVals);
		virtual int ImproveWeakEdgeConsistency(int edgeID, double threshold,
											   bool checkAllLinks, bool addAsModified,
							    std::vector<int> &iterDataEdgeID,
								std::vector<int> &iterDataOldLinks,
							    std::vector<int> &iterDataNewLinks,
								std::vector<std::vector<double> > &iterDataAllScores,
								std::vector<std::vector<double> > &iterDataAllCorrVals);
	
		virtual void CreateAndExploreDynamicLink(int edgeID);
	
		virtual double CalculateConsistencyScore();
	
	// add individual pairwise maps to the collection
		virtual void ExploreEdge(GraphEdgeMap * edge);
		virtual void ExploreEdges(std::vector<int> & edgeIDs);
		virtual void ExploreEdges(std::vector<GraphEdgeMap*> &edges);
		virtual void AddSinglePairCorrs(GraphEdgeMap * edge, 
										const VKString & corrModuleName,
										const VKString & corrValueName,
										bool pickOptimal = true);
	
		virtual double DiffusedCorrValue(const ModelPoint & pntFrom, 
										 const ModelPoint & pntTo);
		virtual void AddEntry(int p1, int p2, double val);
	
		virtual void GetCorrs(const ModelPoint & pnt, std::vector<ModelPoint> * corrPnts);
		virtual ModelPoint GetBestCorr(const ModelPoint & pnt, 
									   Model3D * toModel, 
									   const VKString & modelValue, 
									   double * value);

		virtual void GetCorrsForPairWithVal(Model3D * fromModel,
											Model3D * toModel, 
											std::map<int, double> * savePerPointVals);

		virtual VKString GetCollectionType();
		virtual bool ReadCollectionChild(std::ifstream & textStream);
		virtual void WriteCollectionChild(std::ofstream & textStream);

		static int EdgeID(int node1ID, int node2ID, int nNodes);
		
	
	// interaction
		void InteractiveChangeDiffusionTime(double dt);
		void InteractiveChangeDiffusionSigma(double ds);
	
		TimeProfiler m_Profiler;
		virtual void WriteLog(const VKString & msg);
		virtual void WriteStatsToLog();
		virtual void DrawLog(DatabaseVisWindow * window);
		virtual void DrawAllEmbeddedPoints(DatabaseVisWindow * window);
		virtual void SelectEmbeddedPoint(DatabaseVisWindow * window,
										 double x, double y);

//		virtual void GetEmbeddedBounds(double &minX, double &minY, 
//									   double &maxX, double &maxY);
		
		int m_iEmbeddBasedOnSingleModel;
	
		virtual void DrawAlignmentGraph(DatabaseVisWindow * window);
		virtual void SelectNodeInAlignmentGraph(DatabaseVisWindow * window,
												double x, double y, 
												bool justPressed);
		virtual void EmbedNodes(DatabaseVisWindow * window);
	
		virtual void DrawGraphsMenu(DatabaseVisWindow * window);
		virtual void SelectGraphInMenu(DatabaseVisWindow * window,
									   double x, double y);
		virtual void FillDiffusionParams(int clusterID, int * nEigs);
	
//		void GlobalIDToClusterID(int modelID, int pointID,
//								 DiffusionMap ** diffMap, int * inClusterID);
//		void GlobalIDToClusterID(int globalID,
//								 DiffusionMap ** diffMap, int * inClusterID);
//		int ClusteringIDToGlobalID(int clusterID, int inClusterID);
		int m_iSelectedSubgraph;
		
		virtual void ComputeConservation(double * perPntValues);
		virtual void AddConservationToModelValue(const VKString & modelValue, 
												 double * consData = NULL);
		virtual void ReadConservationAsFile(const VKString & inFile, 
											double * consData);
		virtual void SaveConservationAsFile(const VKString & outFile, 
											double * consData = NULL);

		bool m_bConservativeOnMemory;
	protected:

		DynamicGraphState m_State;
		bool m_bValidEmbeddingExists;	// deleteme
	
		bool m_bEmbeddedAlignmentGraph;
	
		std::vector<GraphEdgeMap * > m_ModifiedEdges;
	
		AnalysisPipeline * m_pPipeline;	
		
		void UpdateClusteringOfModels();
		//DiffusionMap* CreateDiffusionMap(int nPoints, int nEigs);
		CollectionAggregatePairwise* CreatePairwiseAggregation(int clusterID);
	
		int * m_GlobalIDToClusterID;
		int * m_GlobalIDToInClusterID;
		int ** m_InClusterToGlobalID;
		int * m_InClusterToGlobalIDSize;
		std::vector<std::vector<int> > m_mCCsToNodes;
//		std::vector<DiffusionMap *> m_aDiffusionMaps;
	
		int GetClusterID(const ModelPoint & pnt);
		CollectionAggregatePairwise * GetClusterAgg(const ModelPoint & pnt);
		CollectionAggregatePairwise * GetClusterAgg(Model3D * model);
		
		std::vector<CollectionAggregatePairwise *> m_aPairAgg;
	
		int m_iNumModels;
		GraphNodeModel * m_aAllNodes;
		GraphEdgeMap* m_aAllEdges;
		VKString m_sModuleName;	
	
		std::vector<double> m_Scores;
	
		void AddEigenvaluesToGraphLog();
		void FillEdgeBestLinks(std::vector<int> & inEdgeIDs, 
							   std::vector<int> & outLinkIDs);
	
		void WriteFinalCorrespondences();
	
		bool m_bKeepDrawableLog;
		bool m_bKeepTrackOfScore;
		bool m_bWriteFinalCorrs;
	
		DynamicGraphLogParams m_DrawableLog;
		
		int m_iNumMapsComputedParallel;
		int m_iNumMapsComputedSequential;
		std::map<int, bool> m_MapsComputedSequential;
		double m_fActualAggTime;
		double m_fDiffusionTime;
		int m_iNumDiffusions;
		double m_fEdgeExplorationTime;
		double m_fEdgeOptimizationTime;
		int m_iNumOptimizations;
};


#endif


