/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TimeProfiler.h"

#include "CollectionDiffuseFullMtx.h"

#include "CollectionSinglePair.h"
#include "CollectionPairwise.h"
#include "ModelDatabase.h"

CollectionDiffuseFullMtx::CollectionDiffuseFullMtx(const VKString & collectionName, 
												   ModelDatabase * db)
: CollectionDiffuse(collectionName, db)
{
	InitCollection();
}

CollectionDiffuseFullMtx::CollectionDiffuseFullMtx(const VKString & collectionName, 
												   ModelDatabase * db, 
//												   const VKString & inputCollection,
//												   const VKString & localCorrValue,
//												   const VKString & globalCorrValue,
//												   double minOverlap, double maxOverlap,
//												   double valueThreshold, 
												   CollectionDiffuse::CorrMtxParams & corrMtx,
												   bool takeEigs, int depth,
												   double embeddedSpaceDistRatio,
												   EigenDecompositionLibrary eigsLib, 
												   DiffusedSearchType searchType)
: CollectionDiffuse(collectionName, db, corrMtx, CollectionDiffuse::DiffusionParams())
{
	InitCollection();

	m_EigsLib = eigsLib;
	m_EmbeddedSpaceNNSearchType = searchType;
	
	m_fMaxEmbeddedDistanceRatio = embeddedSpaceDistRatio;

//	m_sInputCollectionName = inputCollection;
//	m_sLocalCorrValue = localCorrValue;
//	m_sGlobalCorrValue = globalCorrValue;
//
//	m_fMinOverlap = minOverlap;
//	m_fMaxOverlap = maxOverlap;
//	m_fValueThreshold = valueThreshold;
	
	m_bTakeEigs = takeEigs;
	m_iMaxDepth = depth;

}

CollectionDiffuseFullMtx::CollectionDiffuseFullMtx(const VKString & collectionName,
												   ModelDatabase * db, 
												   const VKString & inputDiffuseCollection,
												   bool takeEigs, int depth, double embeddedSpaceDistRatio,
												   EigenDecompositionLibrary eigsLib, 
												   DiffusedSearchType searchType)
: CollectionDiffuse(collectionName, db)
{
	InitCollection();
	
	m_EigsLib = eigsLib;
	m_EmbeddedSpaceNNSearchType = searchType;
	
	m_fMaxEmbeddedDistanceRatio = embeddedSpaceDistRatio;

	m_bCopiedVersion = true;
	m_sCopiedCorrFrom = inputDiffuseCollection;
	m_bTakeEigs = takeEigs;
	m_iMaxDepth = depth;
	
	CorrCollection * coll = m_pDB->GetCorrCollection(m_sCopiedCorrFrom);
	assert(coll!=NULL && coll->GetCollectionType() == "CollectionDiffuseFullMtx");
	CollectionDiffuseFullMtx * diffuseColl = (CollectionDiffuseFullMtx*)coll;

	m_CorrMtxParams = diffuseColl->m_CorrMtxParams;
	
//	m_fMinOverlap = diffuseColl->m_fMinOverlap;
//	m_fMaxOverlap = diffuseColl->m_fMaxOverlap;
//	m_fValueThreshold = diffuseColl->m_fValueThreshold;
//	
//	m_sLocalCorrValue = diffuseColl->m_sLocalCorrValue;
//	m_sGlobalCorrValue = diffuseColl->m_sGlobalCorrValue;
//	m_sInputCollectionName = diffuseColl->m_sInputCollectionName;
		// plain copy of matrices - should not be modified
	if (diffuseColl->m_pCorrMatrix==NULL && diffuseColl->m_pCorrMatrixSparse==NULL)
		diffuseColl->ExecuteRandomWalk();
	
	assert(diffuseColl->m_pCorrMatrix!=NULL || diffuseColl->m_pCorrMatrixSparse!=NULL);
	m_pCorrMatrixSparse = diffuseColl->m_pCorrMatrixSparse;
	m_pCorrMatrix = diffuseColl->m_pCorrMatrix;
	m_pCorrVectorRowSum = diffuseColl->m_pCorrVectorRowSum;
	
	if (takeEigs && diffuseColl->m_bTakeEigs)
	{
		m_pEigenValues = diffuseColl->m_pEigenValues;
		m_pEigenVectors = diffuseColl->m_pEigenVectors;
	}
}

void CollectionDiffuseFullMtx::InitCollection()
{
#ifdef INCLUDE_FLANN
	m_pEmbeddedSpaceANN = NULL;
	m_pANNDataset = NULL;
	m_pANNOutIndices = NULL;
	m_pANNOutDistances = NULL;
	m_pANNQuery = NULL;
#endif
	
	m_bCopiedVersion = false;
	m_pCorrVectorRowSum = NULL;
	m_abFilledComputedCoorsRow = NULL;
	m_iNEigs = -1;
	m_pEigenValues = NULL;
	m_pEigenVectors = NULL;

	m_fCurrentMinEmbeddedDistance = 1;
	
	m_iMaxDepth = 10;
	m_bTakeEigs = false;
	m_pInteractiveSteadyIterationMatrix = NULL;
	m_iIterationID = -1;
	m_bUseSparse = true;
//	m_sInputCollectionName = "LoadFromFile";
//	m_sLocalCorrValue = "LoadFromFile";
//	m_sGlobalCorrValue = "LoadFromFile";
//	m_fMinOverlap = -1;
//	m_fMaxOverlap = -1;
//	m_fValueThreshold = -1;
	m_pCorrMatrixSparse = NULL;
	m_pSteadyStateMatrixSparse = NULL;
	m_pCorrMatrix = NULL;
	m_pSteadyStateMatrix = NULL;
//	m_iNPoints = -1;
	m_iMtxFormat = LinAlgMatrixSparseReal::SUPPORT_ALL_DYNAMIC_CAST;
	m_aComputedCorrsArray = NULL;	
}

VKString CollectionDiffuseFullMtx::GetCollectionType()
{
	return "CollectionDiffuseFullMtx";
}

int CollectionDiffuseFullMtx::GetNEigs()
{
//	int NDims = GetNPoints() *  m_pDB->NumModels();
	int NDims = m_pDB->NGlobalPnts();
	
	if (m_iNEigs > NDims)
		m_iNEigs = NDims;
	
	if (m_iNEigs<=0)
		m_iNEigs = vkMin(NDims, 2 * NDims / m_pDB->NumModels());
	
	return m_iNEigs;
}


//int CollectionDiffuseFullMtx::GetNPoints()
//{
//	if (m_iNPoints < 0 )
//		m_iNPoints = NPoints(m_pDB);
//	
//	return m_iNPoints;
//}

void CollectionDiffuseFullMtx::ExecuteRandomWalk()
{
	bool save = (m_pCorrMatrix==NULL) || (m_bTakeEigs && m_pEigenVectors==NULL);
	if (m_bUseSparse)
		save = (m_pCorrMatrixSparse==NULL) || (m_bTakeEigs && m_pEigenVectors==NULL);
	
	if (m_pCorrMatrix==NULL && m_pCorrMatrixSparse==NULL)
		FillMatrix();
	
	if (m_bTakeEigs)
	{
		if (m_pEigenValues==NULL && m_pEigenVectors==NULL)
			FindEigenvalues();
	}
	else
	{
		if ((m_pSteadyStateMatrix==NULL && m_pSteadyStateMatrixSparse==NULL)
			|| m_iIterationID != m_iMaxDepth)
			FindSteadyState();
	}
	
	m_Timer.startedProcess("SavingCollection");
	if (save)
		WriteCollection();
	m_Timer.finishedProcess("SavingCollection", false);
}

void CollectionDiffuseFullMtx::AddEntry(int mtxID1, int mtxID2, double val)
{
//	int NPoints = GetNPoints();
//	int NModels = m_pDB->NumModels();
//	int NDims = NPoints * NModels;
	int NDims = m_pDB->NGlobalPnts();
	
	if (m_bUseSparse)
	{
		if (m_pCorrMatrixSparse==NULL)
			m_pCorrMatrixSparse = new LinAlgMatrixSparseReal(NDims, NDims, m_iMtxFormat);
		m_pCorrMatrixSparse->AddVal(mtxID1, mtxID2, val);
	}
	else
	{
		if (m_pCorrMatrix==NULL)
			m_pCorrMatrix = new LinAlgMatrixReal(NDims, NDims, 0);
		m_pCorrMatrix->AddVal(mtxID1, mtxID2, val);	
	}
}

void CollectionDiffuseFullMtx::FillMatrix()
{	
	int NDims = m_pDB->NGlobalPnts();
	
	TimeProfiler timer;	
	CallAddEntryForAll();
	
	assert((m_bUseSparse && m_pCorrMatrixSparse!=NULL) 
		   || (!m_bUseSparse && m_pCorrMatrix!=NULL));
	
	std::cout<<"Adding identities..."<<std::endl;
	bool identityToMax = true;
	
	double maxVal = 1.;
	if (identityToMax)
		maxVal = (m_bUseSparse ? m_pCorrMatrixSparse->GetMax() : m_pCorrMatrix->GetMax());
	
	// Adding identities
	for (int i=0; i<NDims; i++)
	{
		timer.WriteProgress("AddIdentity", i, NDims);
		if (m_bUseSparse)
		{
			if (m_pCorrMatrixSparse->GetVal(i, i)==0)
				m_pCorrMatrixSparse->AddVal(i, i, maxVal);
		}
		else
		{
			if (m_pCorrMatrix->GetVal(i, i)==0)
				m_pCorrMatrix->AddVal(i, i, maxVal);
		}
	}

	std::cout<<std::endl;
	
	std::cout<<"Row Normalization... "<<std::flush;
	bool rowNormalize = true;
	
	// Row normalization
	assert(m_pCorrVectorRowSum==NULL);	// do not leak memory
	m_pCorrVectorRowSum = new LinAlgVectorReal(m_pCorrMatrixSparse->Rows(), rowNormalize ? 0. : 1.);

	if (rowNormalize)
	{
		if (m_bUseSparse)
		{
			m_pCorrMatrixSparse->Symmetrize();
			m_pCorrMatrixSparse->RowNormalize(m_pCorrVectorRowSum);
		}
		else
		{
			m_pCorrMatrix->Symmetrize();
			m_pCorrMatrix->RowNormalize(m_pCorrVectorRowSum);
		}
	}
	std::cout<<"Done"<<std::endl;

	assert((m_pCorrMatrixSparse!=NULL || m_pCorrMatrix!=NULL) && m_pCorrVectorRowSum!=NULL);
}


void CollectionDiffuseFullMtx::FindEigenvalues()
{
	std::cout<<"Diffusion Distance: Eigenanalysis"<<std::endl;
	//int NDims = GetNPoints() * m_pDB->NumModels();
	int NDims = m_pDB->NGlobalPnts();
	
	if (m_pEigenValues==NULL)
		m_pEigenValues = new LinAlgVectorReal(GetNEigs());
	
	if (m_pEigenVectors==NULL)
		m_pEigenVectors = new LinAlgMatrixReal(NDims, GetNEigs());
	
	if (!m_bUseSparse)		// dense matrix eigendecomposition
	{
		if (m_pCorrMatrix==NULL)
			FillMatrix();
		assert(m_pCorrMatrix!=NULL);

		LinAlgMatrixReal * newMtx = new LinAlgMatrixReal(*m_pCorrMatrix);
		newMtx->EigendecompositionWithMATLAB(*m_pEigenValues, *m_pEigenVectors, GetNEigs(), "DiffuseMtx_");
		delete newMtx;		
	}
	else		// sparse matrix eigendecomposition
	{
		if (m_pCorrMatrixSparse==NULL)
			FillMatrix();
		assert(m_pCorrMatrixSparse!=NULL);
		
		LinAlgMatrixSparseReal * newMtx = new LinAlgMatrixSparseReal(*m_pCorrMatrixSparse);
		static const VKString pfx("DiffuseMtx_");
		if (m_EigsLib==EIGS_MATLAB)
			newMtx->EigendecompositionSymmetricMatlab(*m_pEigenValues, *m_pEigenVectors, 
													  GetNEigs(), pfx.c_str());
		else if (m_EigsLib==EIGS_SVDLIBC)
			newMtx->EigendecompositionSymmetricSVDLib(*m_pEigenValues, *m_pEigenVectors, GetNEigs());
		else
			assert(false);
		delete newMtx;	
	}
	
	// Normalization of eigenvectors (as Nader et al.)
	assert(m_pCorrVectorRowSum!=NULL);
	for (int i=0; i<m_pEigenVectors->Cols(); i++)
	{
		double norm = pow((*m_pCorrVectorRowSum)(i), -.5);
		double intVal = 0;
		for (int j=0; j<m_pEigenVectors->Rows(); j++)
		{
			(*m_pEigenVectors)(j,i) *= norm;
			intVal += pow((*m_pEigenVectors)(j,i), 2);
		}
	}	
}

void CollectionDiffuseFullMtx::FindSteadyState()
{
	if (m_bUseSparse)
	{
		if (m_pCorrMatrixSparse==NULL)
			FillMatrix();
		assert(m_pCorrMatrixSparse!=NULL);
		//assert(m_iMaxDepth==1 || m_iMaxDepth==0);
		m_iIterationID = m_iMaxDepth;
		
		if (m_pSteadyStateMatrixSparse==NULL)
			m_pSteadyStateMatrixSparse = new LinAlgMatrixSparseReal(*m_pCorrMatrixSparse);
		else 
			(*m_pSteadyStateMatrixSparse) = *m_pCorrMatrixSparse;
		m_pSteadyStateMatrixSparse->MultiplyEachRowColByScalar(*m_pCorrVectorRowSum, true);
		return;
	}
	
	assert(!m_bUseSparse);
	// update steady state matrix
	if (m_pCorrMatrix==NULL)
		FillMatrix();
	
	assert(m_pCorrMatrix!=NULL && m_pCorrVectorRowSum!=NULL);
	
	if (m_pSteadyStateMatrix==NULL)
	{
		m_pSteadyStateMatrix = new LinAlgMatrixReal(*m_pCorrMatrix);
		m_iIterationID = 1;
	}
	else if (m_iIterationID==-1 || m_iIterationID > m_iMaxDepth)
	{
		(*m_pSteadyStateMatrix) = *m_pCorrMatrix;
		m_iIterationID = 1;
	}
	else if (m_iIterationID == m_iMaxDepth)
		return;
	
	if (m_pInteractiveSteadyIterationMatrix==NULL)
		m_pInteractiveSteadyIterationMatrix = new LinAlgMatrixReal(*m_pCorrMatrix);
	
	
	// multiply powers of 2
	while (2 * m_iIterationID < m_iMaxDepth)
	{
		m_iIterationID *= 2;
		m_pSteadyStateMatrix->Multiply(*m_pSteadyStateMatrix, *m_pInteractiveSteadyIterationMatrix);
		(*m_pSteadyStateMatrix) = (*m_pInteractiveSteadyIterationMatrix);
		m_pSteadyStateMatrix->RowNormalize(NULL);
	}
	
	// just multiply
	while (m_iIterationID < m_iMaxDepth)
	{
		m_iIterationID++;
		m_pSteadyStateMatrix->Multiply(*m_pCorrMatrix, *m_pInteractiveSteadyIterationMatrix);
		(*m_pSteadyStateMatrix) = (*m_pInteractiveSteadyIterationMatrix);
		m_pSteadyStateMatrix->RowNormalize(NULL);
	}
}

void CollectionDiffuseFullMtx::UpdateCache()
{
	if (m_bTakeEigs)
	{
		if (m_pEigenValues==NULL && m_pEigenVectors==NULL)
			FindEigenvalues();
		
		int NDims = m_pDB->NGlobalPnts();
		// clear precomputed eigenvector stuff
		if (m_aComputedCorrsArray!=NULL)
		{
			for (int i=0; i<NDims; i++)
			{
				for (int j=0; j<m_pDB->NumModels(); j++)
					m_aComputedCorrsArray[i][j].clear();
				m_abFilledComputedCoorsRow[i] = false;
			}
		}
		
#ifdef INCLUDE_FLANN
		// clear ANN
		if (m_pEmbeddedSpaceANN!=NULL)
		{
			delete m_pEmbeddedSpaceANN;
			m_pEmbeddedSpaceANN = NULL;
		}
#endif
	}
	else
	{
		FindSteadyState();
	}
}

void CollectionDiffuseFullMtx::InteractivelyChangeDepth(int changeInDepth)
{
	m_iMaxDepth += changeInDepth;
	if (m_iMaxDepth <= 0)
		m_iMaxDepth = 0;
	
	std::cout<<"NewDepth="<<m_iMaxDepth<<std::endl;
	if (m_iMaxDepth==0)
		m_bTakeEigs = false;
	else
		m_bTakeEigs = true;
	UpdateCache();
}

void CollectionDiffuseFullMtx::InteractivelyChangeEmbeddedSpaceDistance(double change)
{
	m_fMaxEmbeddedDistanceRatio += change;
	if (m_fMaxEmbeddedDistanceRatio <= 0)
		m_fMaxEmbeddedDistanceRatio = 0;
//	if (m_fMaxEmbeddedDistanceRatio >= 1)
//		m_fMaxEmbeddedDistanceRatio = 1;
	
	std::cout<<"Embedded Sigma = "<<m_fMaxEmbeddedDistanceRatio<<std::endl;
	UpdateCache();
}

void CollectionDiffuseFullMtx::InteractivelyChangeSearchType()
{
	std::cout<<"Setting NN search in embedded space: ";
	switch (m_EmbeddedSpaceNNSearchType)
	{
		case SEARCH_EXHAUSTIVE:
			std::cout<<"FLANN"<<std::endl;
			m_EmbeddedSpaceNNSearchType = SEARCH_USE_FLANN;
			break;
		case SEARCH_USE_FLANN:
			std::cout<<"EXHAUSTIVE (brute force)"<<std::endl;
			m_EmbeddedSpaceNNSearchType = SEARCH_EXHAUSTIVE;
			break;
	}
	UpdateCache();
}

double CollectionDiffuseFullMtx::GetEmbeddedSpaceSigma2()
{
	return pow(.5 * GetEmbeddedSpaceMaxDistance(), 2.);
}

double CollectionDiffuseFullMtx::GetEmbeddedSpaceMaxDistance()
{
	assert(m_fCurrentMinEmbeddedDistance>0);
	return m_fCurrentMinEmbeddedDistance;
}

void CollectionDiffuseFullMtx::SetRegionNormalization(std::vector<ModelPoint> & pnts)
{
	m_bAutoRegionNormalize = false;
	double globalMin = FLT_MAX;
	for (int i=0; i<(int)pnts.size(); i++)
	{
		int mtxPosFrom = MatrixPosition(pnts[i]);
		if (m_aComputedCorrsArray==NULL || !m_abFilledComputedCoorsRow[mtxPosFrom])
			FindDiffusedValuesClosestFirst(pnts[i]);
		globalMin = vkMin(m_afComputerMinDistArray[mtxPosFrom], globalMin);
	}
	m_fCurrentMinEmbeddedDistance = globalMin;
}

double CollectionDiffuseFullMtx::DiffusedCorrValue(const ModelPoint & pntFrom, 
												   const ModelPoint & pntTo)
{
	m_Timer.startedProcess(VKString("DiffusedCorrValue"));
	int mtxPosFrom = MatrixPosition(pntFrom);
	int mtxPosTo = MatrixPosition(pntTo);
	int modeltoID = pntTo.GetModel()->ModelID();
	int pntToID = pntTo.PntCloudID();
	assert(modeltoID>=0 && modeltoID < m_pDB->NumModels());
	//assert(pntToID>=0 && pntToID < GetNPoints());
	//int NDims = GetNPoints() * m_pDB->NumModels();
	int NDims = m_pDB->NGlobalPnts();
	assert(mtxPosFrom<NDims && mtxPosTo<NDims && mtxPosFrom>=0 && mtxPosTo>=0);
	double val=-1;
	
	ExecuteRandomWalk();
	
	if (m_bTakeEigs)
	{
		assert(m_pEigenValues!=NULL && m_pEigenVectors!=NULL);

		// if row is cached
		if (m_aComputedCorrsArray==NULL || !m_abFilledComputedCoorsRow[mtxPosFrom])
			FindDiffusedValuesClosestFirst(pntFrom);
		assert(m_abFilledComputedCoorsRow[mtxPosFrom]);
		std::map<int, double>::iterator iter;
		iter = m_aComputedCorrsArray[mtxPosFrom][modeltoID].find(pntToID);
		if (m_bAutoRegionNormalize)
			m_fCurrentMinEmbeddedDistance = m_afComputerMinDistArray[mtxPosFrom];
		if (iter!=m_aComputedCorrsArray[mtxPosFrom][modeltoID].end())
			val = exp(-iter->second / GetEmbeddedSpaceSigma2());
		else
			val = 0;
	}
	else 
	{
		m_Timer.startedProcess(VKString("SteadyMtxGetVal"));		
		if (m_bUseSparse)
		{
			assert(m_pSteadyStateMatrixSparse!=NULL);
			val = m_pSteadyStateMatrixSparse->GetVal(mtxPosFrom, mtxPosTo);
		}
		else
		{
			assert(m_pSteadyStateMatrix!=NULL);
			val = m_pSteadyStateMatrix->GetVal(mtxPosFrom, mtxPosTo);
		}
		m_Timer.finishedProcess(VKString("SteadyMtxGetVal"), false);
	}
	
	m_Timer.finishedProcess(VKString("DiffusedCorrValue"), false);
//	static int counter = 0;
//	if ((counter++) % 10 == 0)
//	{
//		std::cout<<"Iterations: "<<counter<<std::endl;
//		std::cout<<m_Timer.profile().c_str()<<std::endl;
//	}
	return val;
}

void CollectionDiffuseFullMtx::FindDiffusedValuesClosestFirst(const ModelPoint & pntFrom)
{
//	std::cout<<"CollectionDiffuseFullMtx::FindDiffusedValuesClosestFirst - 1"<<std::endl;
//	int NDims = GetNPoints() * m_pDB->NumModels();
	int NDims = m_pDB->NGlobalPnts();
	// create cache if it doesn't exist
	if (m_aComputedCorrsArray==NULL)
	{
		m_abFilledComputedCoorsRow = new bool[NDims];
		m_aComputedCorrsArray = new std::map<int, double>*[NDims]; 
		m_afComputerMinDistArray = new double[NDims];
		for (int i=0; i<NDims; i++)
		{
			m_aComputedCorrsArray[i] = new std::map<int, double>[m_pDB->NumModels()];
			m_abFilledComputedCoorsRow[i] = false;
			m_afComputerMinDistArray[i] = 1.;
		}
	}
	
	int mtxPosFrom = MatrixPosition(pntFrom);
	assert(m_abFilledComputedCoorsRow[mtxPosFrom]==false);
	m_abFilledComputedCoorsRow[mtxPosFrom] = true;
	m_Timer.startedProcess(VKString("FillingRow"));
	
	int numNN = (int)ceil(m_fMaxEmbeddedDistanceRatio * NDims);
	if (numNN > NDims)
		numNN = NDims;
//	std::cout<<"CollectionDiffuseFullMtx::FindDiffusedValuesClosestFirst - 2"<<std::endl;
	// slow approach: examine every pair - calculate distances - sort
	if (m_EmbeddedSpaceNNSearchType==SEARCH_EXHAUSTIVE)
	{
		std::vector<Sortable> otherPnts;
		for (int i=0; i<NDims; i++)
		{
//			std::cout<<"\tPos2Pnt: "<<i<<std::endl;
			if (PositionToPoint(i).GetModel() == pntFrom.GetModel())
				continue;	// no self-correspondences for now
			
			double dist = 0;
			for (int e=0; e<m_pEigenValues->Rows(); e++)	// top N eigenvectors
			{
				double eigVal = pow((*m_pEigenValues)(e), (double)m_iMaxDepth);
				double e1 = (*m_pEigenVectors)(mtxPosFrom, e);
				double e2 = (*m_pEigenVectors)(i, e);
				dist += pow(eigVal * (e1-e2), 2);
			}
			
			dist = sqrt(dist);
			otherPnts.push_back(Sortable(dist, NULL, i));
		}
//		std::cout<<"CollectionDiffuseFullMtx::FindDiffusedValuesClosestFirst - 3"<<std::endl;
		std::sort(otherPnts.begin(), otherPnts.end());
		
		for (int i=0; i<vkMin((int)otherPnts.size(), numNN); i++)
		{
			int pntID = otherPnts[i].id;
			double corrDist = otherPnts[i].value;
//			std::cout<<"\tPos2Pnt: "<<i<<std::endl;
			ModelPoint currPntTo = PositionToPoint(pntID);
			int currModelID = currPntTo.GetModel()->ModelID();
			int currPntID = currPntTo.PntCloudID();
			assert(currModelID>=0 && currModelID < m_pDB->NumModels());
//			assert(currPntID>=0 && currPntID < GetNPoints());
			//std::cout<<"Corr["<<i<<"] Dist="<<corrDist<<std::endl;
			
			m_aComputedCorrsArray[mtxPosFrom][currModelID][currPntID] = corrDist*corrDist;
		}
		int pickID = vkMin((int)otherPnts.size()-1, numNN-1);
//		std::cout<<"CollectionDiffuseFullMtx::FindDiffusedValuesClosestFirst - 4"<<std::endl;		
		//int pickID = (numNN>=(int)otherPnts.size()) ? ((int)otherPnts.size()-1) : (numNN-1);
		m_afComputerMinDistArray[mtxPosFrom] = otherPnts[pickID].value;
	}
#ifdef INCLUDE_FLANN
	else if (m_EmbeddedSpaceNNSearchType==SEARCH_USE_FLANN)	// fast approach: use FLANN
	{
		int NEigs = m_pEigenValues->Rows();
		if (m_pEmbeddedSpaceANN==NULL)	// construct ANN index
		{
			std::cout<<"Constructing ANN index..."<<std::flush;
			if (m_pANNDataset == NULL)
				m_pANNDataset = new flann::Matrix<double>(new double[NDims*NEigs], 
														  m_pEigenVectors->Rows(), m_pEigenVectors->Cols());
			
			if (m_pANNOutIndices == NULL)
				m_pANNOutIndices = new flann::Matrix<int>(new int[NDims], 1, NDims);
			
			if (m_pANNOutDistances == NULL)
				m_pANNOutDistances = new flann::Matrix<double>(new double[NDims], 1, NDims);
			
			if (m_pANNQuery==NULL)
				m_pANNQuery = new flann::Matrix<double>(new double[NEigs], 1, NEigs);
			
			for (int i=0; i<NDims; i++)
			{
				(*m_pANNOutIndices)[0][i] = -1;
				(*m_pANNOutDistances)[0][i] = FLT_MAX;
			}
			
			for (int i=0; i<m_pEigenVectors->Rows(); i++)
			for (int j=0; j<m_pEigenVectors->Cols(); j++)
				(*m_pANNDataset)[i][j] = pow((*m_pEigenValues)(j), m_iMaxDepth) * (*m_pEigenVectors)(i,j);
			
			m_pEmbeddedSpaceANN = new flann::Index<flann::L2<double> >(*m_pANNDataset, 
																	   flann::KDTreeIndexParams(1));
			m_pEmbeddedSpaceANN->buildIndex();
			std::cout<<" - DONE"<<std::endl;
		}
		
		assert(m_pEmbeddedSpaceANN!=NULL && m_pANNQuery!=NULL && m_pANNDataset!=NULL 
			   && m_pANNOutIndices!=NULL && m_pANNOutDistances!=NULL);
		
		// fill m_pANNQuery
		for (int e=0; e<NEigs; e++)
			(*m_pANNQuery)[0][e] = (*m_pANNDataset)[mtxPosFrom][e];
		
		// find nearest neighbors (can set FLANN_CHECKS_UNLIMITED)
		// if approximate: set flann::KDTreeIndexParams(4..16), flann::SearchParams(numNN*10(?))
		m_pEmbeddedSpaceANN->knnSearch(*m_pANNQuery, *m_pANNOutIndices, *m_pANNOutDistances, 
									   numNN, flann::SearchParams(FLANN_CHECKS_UNLIMITED));
		
		// for each neighbor
		for (int i=0; i<numNN; i++)
		{
			int pntID = (*m_pANNOutIndices)[0][i];
			double corrDist = (*m_pANNOutDistances)[0][i];
			assert(pntID>=0 && pntID<NDims);
			
			ModelPoint currPntTo = PositionToPoint(pntID);
			int currModelID = currPntTo.GetModel()->ModelID();
			int currPntID = currPntTo.PntCloudID();
			assert(currModelID>=0 && currModelID < m_pDB->NumModels());
//			assert(currPntID>=0 && currPntID < GetNPoints());
			
			m_aComputedCorrsArray[mtxPosFrom][currModelID][currPntID] = corrDist;
		}
		m_afComputerMinDistArray[mtxPosFrom] = sqrt((*m_pANNOutDistances)[0][numNN-1]);
	}
#endif
	else
		assert(false);
	
	if (m_fMaxEmbeddedDistanceRatio>1.)
		m_afComputerMinDistArray[mtxPosFrom] *= m_fMaxEmbeddedDistanceRatio;
	
	m_Timer.finishedProcess(VKString("FillingRow"), false);
//	std::cout<<"CollectionDiffuseFullMtx::FindDiffusedValuesClosestFirst - DONE"<<std::endl;
}

void CollectionDiffuseFullMtx::GetCorrs(const ModelPoint & pnt,	
										std::vector<ModelPoint> * corrPnts)
{
	m_Timer.startedProcess("GetCorrs");
	ExecuteRandomWalk();
	int NDims = m_pDB->NGlobalPnts();
	
	int mtxPos = MatrixPosition(pnt);

	if (m_bTakeEigs)
	{
		if (m_aComputedCorrsArray==NULL || !m_abFilledComputedCoorsRow[mtxPos])
			FindDiffusedValuesClosestFirst(pnt);

		for (int i=0; i<m_pDB->NumModels(); i++)
		{
			PointSet * modelPnts = m_pDB->GetModelByID(i)->PointCloud();
			for (std::map<int, double>::iterator iter = m_aComputedCorrsArray[mtxPos][i].begin(); 
				 iter != m_aComputedCorrsArray[mtxPos][i].end(); iter++)
				corrPnts->push_back(modelPnts->GetPoint(iter->first));
		}
	}
	else
	{
		double thresh = 0;
		for (int i=0; i<NDims; i++)
		{
			ModelPoint pntTo = PositionToPoint(i);
			
			double corrVal = DiffusedCorrValue(pnt, pntTo);
			if (corrVal > thresh)
				corrPnts->push_back(pntTo);
		}
	}
	m_Timer.finishedProcess("GetCorrs", false);	
}


int CollectionDiffuseFullMtx::MatrixPosition(const ModelPoint & pnt)
{
	//return CollectionDiffuse::MatrixPosition(pnt, GetNPoints(), m_pDB->NumModels());
	return m_pDB->GlobalID(pnt);
}

ModelPoint CollectionDiffuseFullMtx::PositionToPoint(int position)
{
	//return CollectionDiffuse::PositionToPoint(position, m_pDB, GetNPoints());
	return m_pDB->GetPoint(position);
}

bool CollectionDiffuseFullMtx::ReadCollectionChild(std::ifstream & textStream)
{		
	std::cout<<"Reading Diffused: "<<std::flush;
	VKString text;
	text.readToken(textStream);
	assert(text=="DiffuseMtxSparse");
	textStream>>m_bUseSparse;
	
	ReadMtxParams(textStream);
	text.readToken(textStream);
	assert(text=="Depth");
	textStream>>m_iMaxDepth;
	text.readToken(textStream);
	assert(text=="TakeEigs");
	textStream>>m_bTakeEigs;
	text.readToken(textStream);
	assert(text=="CurrSteady");
	textStream>>m_iIterationID;
	text.readToken(textStream);
	assert(text=="MaxEmbeddedDistRatio");
	textStream>>m_fMaxEmbeddedDistanceRatio;
	text.readToken(textStream);
	assert(text=="CompFlags");
	int nnSearchType;
	int eigsLib;
	textStream>>nnSearchType>>eigsLib;
	m_EmbeddedSpaceNNSearchType = (DiffusedSearchType)nnSearchType;
	m_EigsLib = (EigenDecompositionLibrary)eigsLib;
	
//	int NDims = GetNPoints() * m_pDB->NumModels();
	int NDims = m_pDB->NGlobalPnts();
	VKString collectionDir = m_pDB->GetCollectionDir(m_sCollectionName);
		
	VKString eigenValuesVec = collectionDir + "evals.vector.bin";
	VKString eigenVectorMtx = collectionDir + "evecs.dense.matrix.bin";

	if (m_bUseSparse)
	{
		VKString similarityMtx = collectionDir + "corrs.sparse.matrix.bin";
		VKString steadyStateMtx = collectionDir + "steady.sparse.matrix.bin";

		if (m_pDB->FileExists(similarityMtx))
		{
			m_pCorrMatrixSparse = new LinAlgMatrixSparseReal(NDims, NDims, m_iMtxFormat);
			bool couldRead = m_pCorrMatrixSparse->ReadMatrixBin(similarityMtx.c_str());
			assert(couldRead);
			std::cout<<"Corrs "<<std::flush;		
		}

		if (m_pDB->FileExists(steadyStateMtx))
		{
			m_pSteadyStateMatrixSparse = new LinAlgMatrixSparseReal(NDims, NDims, m_iMtxFormat);
			bool couldRead = m_pSteadyStateMatrixSparse->ReadMatrixBin(steadyStateMtx.c_str());
			assert(couldRead);
			std::cout<<"Steady "<<std::flush;
		}
	}
	else 
	{
		VKString similarityMtx = collectionDir + "corrs.dense.matrix.bin";
		VKString steadyStateMtx = collectionDir + "steady.dense.matrix.bin";
		
		if (m_pDB->FileExists(similarityMtx))
		{
			m_pCorrMatrix = new LinAlgMatrixReal(NDims, NDims, 0);
			bool couldRead = m_pCorrMatrix->ReadMatrixBin(similarityMtx.c_str());
			assert(couldRead);
		}
		if (m_pDB->FileExists(steadyStateMtx))
		{
			m_pSteadyStateMatrix = new LinAlgMatrixReal(NDims, NDims, 0);
			bool couldRead = m_pSteadyStateMatrix->ReadMatrixBin(steadyStateMtx.c_str());
			assert(couldRead);
		}		
	}
	
	if (m_pCorrMatrix!=NULL || m_pCorrMatrixSparse!=NULL)
	{
		VKString similarityNormVec = collectionDir + "rownorms.vector.bin";
		m_pCorrVectorRowSum = new LinAlgVectorReal(NDims, 0);
		bool couldRead = m_pCorrVectorRowSum->ReadVectorBin(similarityNormVec.c_str());
		assert(couldRead);
		std::cout<<"RowNorm "<<std::flush;
	}

	if (m_pDB->FileExists(eigenValuesVec))
	{
		m_pEigenValues = new LinAlgVectorReal(GetNEigs(), 0);
		bool couldRead = m_pEigenValues->ReadVectorBin(eigenValuesVec.c_str());
		assert(couldRead);
		std::cout<<"Evals "<<std::flush;
	}	
	
	if (m_pDB->FileExists(eigenVectorMtx))
	{
		m_pEigenVectors = new LinAlgMatrixReal(NDims, GetNEigs(), 0);
		bool couldRead = m_pEigenVectors->ReadMatrixBin(eigenVectorMtx.c_str());
		assert(couldRead);
		std::cout<<"Evecs "<<std::flush;
	}
	std::cout<<" - DONE!"<<std::endl;
	return true;
}

void CollectionDiffuseFullMtx::WriteCollectionChild(std::ofstream & textStream)
{
	std::cout<<"\tWriting Diffused: "<<std::flush;
	
	textStream<<"DiffuseMtxSparse "<<m_bUseSparse<<"\n";
	WriteMtxParams(textStream);
//	textStream<<m_CorrMtxParams.m_sInputCollection.c_str();
//	textStream<<" "<<m_CorrMtxParams.m_sLocalCorrValue.c_str();
//	textStream<<" "<<m_CorrMtxParams.m_sGlobalCorrValue.c_str()<<"\n";
//	textStream<<"SurfFrac "<<m_CorrMtxParams.m_fMinOverlap<<" "<<m_CorrMtxParams.m_fMaxOverlap<<"\n";
//	textStream<<"MinValue "<<m_CorrMtxParams.m_fValueThreshold<<"\n";	
	
	textStream<<"Depth "<<m_iMaxDepth<<"\n";
	textStream<<"TakeEigs "<<m_bTakeEigs<<"\n";
	textStream<<"CurrSteady "<<m_iIterationID<<"\n";
	textStream<<"MaxEmbeddedDistRatio "<<m_fMaxEmbeddedDistanceRatio<<"\n";
	textStream<<"CompFlags "<<m_EmbeddedSpaceNNSearchType<<" "<<m_EigsLib<<"\n";
	
	VKString collectionDir = m_pDB->GetCollectionDir(m_sCollectionName);
	
	VKString eigenValuesVec = collectionDir + "evals.vector.bin";
	VKString eigenVectorMtx = collectionDir + "evecs.dense.matrix.bin";

	if (m_bUseSparse)
	{
		VKString similarityMtx = collectionDir + "corrs.sparse.matrix.bin";
		VKString steadyStateMtx = collectionDir + "steady.sparse.matrix.bin";

		if (m_pCorrMatrixSparse!=NULL)
		{
			m_pCorrMatrixSparse->WriteMatrixBin(similarityMtx.c_str());
			std::cout<<"Corrs "<<std::flush;
		}
		
		if (m_pSteadyStateMatrixSparse!=NULL)
		{
			m_pSteadyStateMatrixSparse->WriteMatrixBin(steadyStateMtx.c_str());
			std::cout<<"Steady "<<std::flush;
		}
	}
	else
	{
		VKString similarityMtx = collectionDir + "corrs.dense.matrix.bin";
		VKString steadyStateMtx = collectionDir + "steady.dense.matrix.bin";
		
		if (m_pCorrMatrix!=NULL)
			m_pCorrMatrix->WriteMatrixBin(similarityMtx.c_str());
		if (m_pSteadyStateMatrix!=NULL)
			m_pSteadyStateMatrix->WriteMatrixBin(steadyStateMtx.c_str());
	}
	
	if (m_pCorrMatrix!=NULL || m_pCorrMatrixSparse!=NULL)
	{
		VKString similarityNormVec = collectionDir + "rownorms.vector.bin";
		assert(m_pCorrVectorRowSum!=NULL);
		m_pCorrVectorRowSum->WriteVectorBin(similarityNormVec.c_str());
		
	}
	
	if (m_pEigenValues!=NULL)
		m_pEigenValues->WriteVectorBin(eigenValuesVec.c_str());
	if (m_pEigenVectors!=NULL)
	{
		m_pEigenVectors->WriteMatrixBin(eigenVectorMtx.c_str());
		std::cout<<"Evecs "<<std::flush;
	}
	std::cout<<std::endl;
}

CollectionDiffuseFullMtx::DiffusedSearchType 
CollectionDiffuseFullMtx::StrToDiffSearch(const VKString & str)
{	
	if (str=="SEARCH_EXHAUSTIVE")
		return SEARCH_EXHAUSTIVE;
	else if (str=="SEARCH_USE_FLANN")
		return SEARCH_USE_FLANN;
	
	assert(false);
	return SEARCH_EXHAUSTIVE;
}

CollectionDiffuseFullMtx::EigenDecompositionLibrary 
CollectionDiffuseFullMtx::StrToEigsLib(const VKString & str)
{
	if (str=="EIGS_MATLAB")
		return EIGS_MATLAB;
	else if (str=="EIGS_SVDLIBC")
		return EIGS_SVDLIBC;
	
	assert(false);
	return EIGS_MATLAB;
}


