/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#ifndef __CollectionDiffuseFullMtx_H
#define __CollectionDiffuseFullMtx_H

#include "CollectionDiffuse.h"
#include "LinAlgMatrixSparseReal.h"
#include "CorrCollection.h"
#include "TimeProfiler.h"
#include "Sortable.h"

#ifdef INCLUDE_FLANN
	// see http://people.cs.ubc.ca/~mariusm/index.php/FLANN/FLANN
	#include "flann/flann.hpp"
#endif

class CollectionDiffuseFullMtx : public CollectionDiffuse
{
	public:
		enum DiffusedSearchType
		{
			SEARCH_EXHAUSTIVE,
			SEARCH_USE_FLANN
		};
	
		enum EigenDecompositionLibrary
		{
			EIGS_MATLAB,
			EIGS_SVDLIBC
		};
	
		CollectionDiffuseFullMtx(const VKString & collectionName, 
								 ModelDatabase * db);
	
		CollectionDiffuseFullMtx(const VKString & collectionName,
								 ModelDatabase * db, const VKString & inputDiffuseCollection,
								 bool takeEigs, int depth, double embeddedSpaceDistRatio,
								 EigenDecompositionLibrary eigsLib, 
								 DiffusedSearchType searchType);
	
		CollectionDiffuseFullMtx(const VKString & collectionName, 
								 ModelDatabase * db, 
								 CollectionDiffuse::CorrMtxParams & corrMtx,
//								 const VKString & inputCollection,
//								 const VKString & localCorrValue,
//								 const VKString & globalCorrValue,
//								 double minOverlap, double maxOverlap,
//								 double valueThreshold, 
								 bool takeEigs, 
								 int depth, double embeddedSpaceDistRatio, 
								 EigenDecompositionLibrary eigsLib, 
								 DiffusedSearchType searchType);
		void InitCollection();
	
		virtual void GetCorrs(const ModelPoint & pnt,	
							  std::vector<ModelPoint> * corrPnts);			
		virtual void ExecuteRandomWalk();
	
		virtual VKString GetCollectionType();
	
		virtual double DiffusedCorrValue(const ModelPoint & pntFrom, 
										 const ModelPoint & pntTo);
	
		virtual bool ReadCollectionChild(std::ifstream & textStream);
		virtual void WriteCollectionChild(std::ofstream & textStream);
	
		virtual void UpdateCache();
		virtual void InteractivelyChangeDepth(int change);
		virtual void InteractivelyChangeEmbeddedSpaceDistance(double change);
		virtual double GetEmbeddedSpaceSigma2();
		virtual double GetEmbeddedSpaceMaxDistance();
		virtual void SetRegionNormalization(std::vector<ModelPoint> & pnts);
	
		virtual void InteractivelyChangeSearchType();
	
		static DiffusedSearchType StrToDiffSearch(const VKString & str);
		static EigenDecompositionLibrary StrToEigsLib(const VKString & str);
		virtual void AddEntry(int p1, int p2, double val);
			
	protected:
#ifdef INCLUDE_FLANN
		flann::Index<flann::L2<double> > * m_pEmbeddedSpaceANN;
		flann::Matrix<double> * m_pANNDataset;
		flann::Matrix<int> * m_pANNOutIndices;
		flann::Matrix<double> * m_pANNOutDistances;
		flann::Matrix<double> * m_pANNQuery;
#endif
	
		DiffusedSearchType m_EmbeddedSpaceNNSearchType;
		EigenDecompositionLibrary m_EigsLib;
	
		TimeProfiler m_Timer;
		virtual void FillMatrix();
		virtual void FindEigenvalues();
		virtual void FindSteadyState();
	
		virtual void FindDiffusedValuesClosestFirst(const ModelPoint & pntFrom);

		
		int MatrixPosition(const ModelPoint & pnt);
		ModelPoint PositionToPoint(int position);
	
		// constants
		int m_iNEigs;
		bool m_bUseSparse;
	
		// if corrs copied
		bool m_bCopiedVersion;
		VKString m_sCopiedCorrFrom;
	
		// if corrs filled with params
//		double m_fMinOverlap;
//		double m_fMaxOverlap;
//		double m_fValueThreshold;
//		VKString m_sLocalCorrValue;
//		VKString m_sGlobalCorrValue;
//		VKString m_sInputCollectionName;
	
		// adjustable parameters for convergence
		bool m_bTakeEigs;
		int m_iMaxDepth;
		double m_fCurrentMinEmbeddedDistance;
		double m_fMaxEmbeddedDistanceRatio;
	
		// caching (not saved)
		std::map<int, double> ** m_aComputedCorrsArray;
		double * m_afComputerMinDistArray;	
		bool * m_abFilledComputedCoorsRow;

		// saved parameters:
			// sparse: correspondence matrix, steady state matrix
		LinAlgMatrixSparseReal * m_pCorrMatrixSparse;
		LinAlgMatrixSparseReal * m_pSteadyStateMatrixSparse;

			// dense: correspondence, steady state
		LinAlgMatrixReal * m_pCorrMatrix;
		LinAlgMatrixReal * m_pSteadyStateMatrix;
		LinAlgMatrixReal * m_pInteractiveSteadyIterationMatrix;
			// number of iterations for the current steady state
		int m_iIterationID;
	
			// Eigenvectors / eigenvalues
		LinAlgVectorReal * m_pCorrVectorRowSum;	// used to normalize eigenvectors (see SFE paper)
		LinAlgVectorReal * m_pEigenValues;
		LinAlgMatrixReal * m_pEigenVectors;
	
			// points
//		int m_iNPoints;
//		virtual int GetNPoints();
	
		virtual int GetNEigs();
	
		int m_iMtxFormat;
};


#endif


