/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CollectionDiffuseHierarchical.h"
#include "ModelDatabase.h"
#include "TimeProfiler.h"

///////////////////////////////////////////////////
///////////////////////////////////////////////////
//////      								///////
//////		LevelOfDiffusionHierarchy		///////
//////      								///////
///////////////////////////////////////////////////
///////////////////////////////////////////////////

LevelOfDiffusionHierarchy::LevelOfDiffusionHierarchy(ModelDatabase * db, 
													 const CollectionDiffuse::DiffusionParams & diffParams, 
													 GroupingType gt, int levelID)
{
	m_pDB = db;
	m_GroupingType = gt;
	m_iLevelID = levelID;
	m_iNPoints = CollectionDiffuse::NPoints(m_pDB);
	m_iNModels = m_pDB->NumModels();
	m_DiffusionParams = diffParams;
}

void LevelOfDiffusionHierarchy::SetElementsToGroupMapping(std::vector<int> & elementToGroupMapping)
{
	// elements to group map
	m_mElementToGroupID = elementToGroupMapping;
	for (int i=0; i<(int)m_mElementToGroupID.size(); i++)
	{
		int groupID = m_mElementToGroupID[i];
		while (groupID <= (int)m_mGroupIDToElements.size())
			m_mGroupIDToElements.push_back(std::vector<int>());
		m_mGroupIDToElements[groupID].push_back(i);
	}
}

void LevelOfDiffusionHierarchy::SetElementsToGroupMapping(GroupingType elementsType)
{
	std::vector<int> elementToGroupID;
	if (elementsType==GROUP_PNTS && m_GroupingType==GROUP_MODELS)
	{
		for (int i=0; i<m_iNPoints * m_iNModels; i++)
			elementToGroupID.push_back(i % m_iNPoints);
	}
	else if (elementsType==GROUP_PNTS && m_GroupingType==GROUP_PNTS)
	{
		for (int i=0; i<m_iNPoints * m_iNModels; i++)
			elementToGroupID.push_back(i);
	}
	else
		assert(false);
	SetElementsToGroupMapping(elementToGroupID);
}

void LevelOfDiffusionHierarchy::SetClustering(std::vector<int> & groupClustering)
{
	int nGroups = (int)groupClustering.size();
	// Fill clusters to groupIDs map
	for (int i=0; i<nGroups; i++)
	{
		int clusterID = groupClustering[i];
		
		// for each cluster + in-matrix position find group
		while ((int)m_ClusterAndMatrixPositionToGroupID.size() <= clusterID)
			m_ClusterAndMatrixPositionToGroupID.push_back(std::vector<int>());
		m_ClusterAndMatrixPositionToGroupID[clusterID].push_back(i);
		
		// for each group - find cluster ID + in-matrix position
		DiffuseMapID groupsMapID;
		groupsMapID.m_iMatrixID = clusterID;
		groupsMapID.m_iEntryID = (int)m_ClusterAndMatrixPositionToGroupID[clusterID].size()-1;
		m_GroupClustering.push_back(groupsMapID);
	}	
	InitDiffusionMaps();
}

void LevelOfDiffusionHierarchy::InitDiffusionMaps()
{
	int nCacheGroups = 1;
	int nEigs = 256;
	int nGroups = (int)m_GroupClustering.size();
	int nElements = (int)m_mElementToGroupID.size();
	assert(nGroups!=0 && nElements!=0);
	
	// grouping type-specific vars: 
	//		how to group IDs in cache, number of e-vals, number of entries
	if (m_GroupingType==GROUP_MODELS)
	{
		nCacheGroups = 1;
		nEigs = vkMin(nEigs, m_iNModels);
		assert(nGroups==m_iNModels);
	}
	else if (m_GroupingType==GROUP_PNTS)
	{
		nCacheGroups = m_iNModels;		// group points for speedup... 
		// would only work if points come in per model chunks
		// set to 1 if ever do per part grouping
		nEigs = vkMin(nEigs, m_iNPoints * 2);
		assert(m_iNPoints*m_iNModels==nGroups);
		assert(nGroups==nElements);		// cannot break any further
	}
	else 
		assert(false);
	
	assert(m_ClusterAndMatrixPositionToGroupID.size()>0);
	// create diffusion maps based on cluster -> group mapping
	for (int i=0; i<(int)m_ClusterAndMatrixPositionToGroupID.size(); i++)
	{
		int nGroupsInCluster = (int)m_ClusterAndMatrixPositionToGroupID[i].size();
		DiffusionMap * diffMap = new DiffusionMap(nGroupsInCluster, nEigs, nCacheGroups,
												  m_DiffusionParams.m_ANNSearchType, 
												  m_DiffusionParams.m_EigsLib);
		diffMap->SetTime(m_DiffusionParams.m_fDiffTime);
		diffMap->SetMaxDistancePercentile(m_DiffusionParams.m_fDistPercentile);
		m_apDiffusionMaps.push_back(diffMap);
	}	
}

/**
 * From group clusters construct array of element clusters 
 *	(elements - groups at previous hierarchy levels)
 * e.g. for 2-level hierarchy: element = points, groups = models (at level 0=model-level)
 * 
 */
void LevelOfDiffusionHierarchy::ExpandGroupIntoElements(std::vector<int> & mGroupIDToClusterID, 
														std::vector<int> & mElementsIDToClusterID)
{
	for (int i=0; i<(int)m_mElementToGroupID.size(); i++)
	{
		int groupID = m_mElementToGroupID[i];
		assert(groupID >= 0 && groupID < (int)mGroupIDToClusterID.size());
		mElementsIDToClusterID.push_back(mGroupIDToClusterID[groupID]);
	}
}

/**
 * Add a value for a group
 */
void LevelOfDiffusionHierarchy::AddCorrValue(int groupID1, int groupID2, double value)
{
	assert(groupID1 < (int)m_GroupClustering.size());
	int matrixID1 = m_GroupClustering[groupID1].m_iMatrixID;
	int inMatrixID1 = m_GroupClustering[groupID1].m_iEntryID;

	assert(groupID2 < (int)m_GroupClustering.size());
	int matrixID2 = m_GroupClustering[groupID2].m_iMatrixID;
	int inMatrixID2 = m_GroupClustering[groupID2].m_iEntryID;
	
	if (matrixID1 != matrixID2)
		return;
	
	assert(matrixID1 < (int)m_apDiffusionMaps.size());
	DiffusionMap * diffMap = m_apDiffusionMaps[matrixID1];
	diffMap->AddCorrValue(inMatrixID1, inMatrixID2, value);
}

/**
 * Get a value for pair of groups (after spectral analysis)
 */
double LevelOfDiffusionHierarchy::EmbeddedSimilarity(int groupID1, int groupID2)
{
	assert(groupID1 < (int)m_GroupClustering.size());
	int matrixID1 = m_GroupClustering[groupID1].m_iMatrixID;
	int inMatrixID1 = m_GroupClustering[groupID1].m_iEntryID;
	
	assert(groupID2 < (int)m_GroupClustering.size());
	int matrixID2 = m_GroupClustering[groupID2].m_iMatrixID;
	int inMatrixID2 = m_GroupClustering[groupID2].m_iEntryID;
	
	if (matrixID1 != matrixID2)
		return 0;
	
	assert(matrixID1 < (int)m_apDiffusionMaps.size());
	DiffusionMap * diffMap = m_apDiffusionMaps[matrixID1];
	return diffMap->EmbeddedSimilarity(inMatrixID1, inMatrixID2);
}

/**
 * Eigenanalysis of all diffusion maps (all clusters)
 */
void LevelOfDiffusionHierarchy::EmbedAll()
{
	for (int i=0; i<(int)m_apDiffusionMaps.size(); i++)
		m_apDiffusionMaps[i]->Embed();
}

/**
 * Cluster all groups - note that clustering is done globally
 * that is for all groups - since they'll be broken into elements anyway.
 * e.g. to diffuse points - one need to cluster 'models' - but only 
 * cluster assignments at the point level is relevant.
 */
void LevelOfDiffusionHierarchy::Cluster(int nElementsInCluster, 
										std::vector<int> & mGroupIDToClusterID)
{
	int currClusterID = 0;
	for (int i=0; i<(int)m_apDiffusionMaps.size(); i++)
	{
		// find number of elements in groups
		int nElements = 0;
		int nEntries = m_apDiffusionMaps[i]->NEntries();
		for (int pID=0; pID < nEntries; pID++)
		{
			int groupID = m_ClusterAndMatrixPositionToGroupID[i][pID];
			nElements += (int)m_mGroupIDToElements[groupID].size();
		}
		
		// find number of desired clusters
		int nClusters = nElements / nElementsInCluster;
		
		// cluster groups in the diffusion map
		std::vector<int> pIDtoClusterID;
		if (nClusters<=1)	// pretty straightforward
			for (int i=0; i<nEntries; i++)
				pIDtoClusterID.push_back(0);
		else				// do actual spectral clustering
			m_apDiffusionMaps[i]->ClusterKMeans(nClusters, pIDtoClusterID);
		
		// update cluster IDs
		int localNClusters = 0;
		for (int pID=0; pID < nEntries; pID++)
		{
			int groupID = m_ClusterAndMatrixPositionToGroupID[i][pID];
			int localClusterID = pIDtoClusterID[pID];
			while (groupID >= (int)mGroupIDToClusterID.size())
				mGroupIDToClusterID.push_back(-1);
			localNClusters = vkMax(localNClusters, localClusterID+1);
			mGroupIDToClusterID[groupID] = currClusterID + localClusterID;
		}
		currClusterID += localNClusters;
	}
}

///////////////////////////////////////////////////
///////////////////////////////////////////////////
//////      								///////
//////		CollectionDiffuseHierarchical   ///////
//////      								///////
///////////////////////////////////////////////////
///////////////////////////////////////////////////
CollectionDiffuseHierarchical::CollectionDiffuseHierarchical(const VKString & collectionName, 
															 ModelDatabase * db)
: CollectionDiffuse(collectionName, db)
{
	m_iNModels = m_pDB->NumModels();
	m_iNPoints = CollectionDiffuse::NPoints(m_pDB);
}

CollectionDiffuseHierarchical::CollectionDiffuseHierarchical(const VKString & collectionName, 
															 ModelDatabase * db, 
															 const CorrMtxParams & corrMtxFill, 
															 const DiffusionParams & diffusionParams)
: CollectionDiffuse(collectionName, db, corrMtxFill, diffusionParams)
{
	m_iNModels = m_pDB->NumModels();
	m_iNPoints = CollectionDiffuse::NPoints(m_pDB);
}

void CollectionDiffuseHierarchical::GetCorrs(const ModelPoint & pnt,	
											 std::vector<ModelPoint> * corrPnts)

{
	int pntID = CollectionDiffuse::MatrixPosition(pnt, m_iNPoints, m_iNModels);
	LevelOfDiffusionHierarchy * level = m_aHierarchies[(int)m_aHierarchies.size()-1];
	int clusterID = level->m_GroupClustering[pntID].m_iMatrixID;
	int clusterPosID = level->m_GroupClustering[pntID].m_iEntryID; 
	
	int * ids = NULL;
	double * sims = NULL;
	int npnts = 0;
	
	level->m_apDiffusionMaps[clusterID]->MostSimilarPoints(clusterPosID, &ids, &sims, &npnts);	
	
	for (int i=0; i<npnts; i++)
	{
		int globalPntID = level->m_ClusterAndMatrixPositionToGroupID[clusterID][ids[i]];
		corrPnts->push_back(CollectionDiffuse::PositionToPoint(globalPntID, m_pDB, m_iNPoints));
	}
}

double CollectionDiffuseHierarchical::DiffusedCorrValue(const ModelPoint & pntFrom, 
														const ModelPoint & pntTo)
{
	if (m_bAutoRegionNormalize)
	{
		std::vector<ModelPoint> normPnts;
		normPnts.push_back(pntFrom);
		SetRegionNormalization(normPnts);
		m_bAutoRegionNormalize = true;
	}
	LevelOfDiffusionHierarchy * level = m_aHierarchies[(int)m_aHierarchies.size()-1];
	
	int pntID1 = CollectionDiffuse::MatrixPosition(pntFrom, m_iNPoints, m_iNModels);
	int pntID2 = CollectionDiffuse::MatrixPosition(pntTo, m_iNPoints, m_iNModels);
	
	return level->EmbeddedSimilarity(pntID1, pntID2);
}

void CollectionDiffuseHierarchical::SetRegionNormalization(std::vector<ModelPoint> & pnts)
{
	m_bAutoRegionNormalize = false;
	LevelOfDiffusionHierarchy * level = m_aHierarchies[(int)m_aHierarchies.size()-1];
	std::map<int, std::vector<int> > diffMapIDToNormRegion;
	for (int i=0; i<(int)pnts.size(); i++)
	{
		int globalPntID = CollectionDiffuse::MatrixPosition(pnts[i], m_iNPoints, m_iNModels);
		int clusterID = level->m_GroupClustering[globalPntID].m_iMatrixID;
		int clusterPosID = level->m_GroupClustering[clusterPosID].m_iEntryID; 
		diffMapIDToNormRegion[clusterID].push_back(clusterPosID);
	}
	
	std::map<int, std::vector<int> >::iterator iter;
	for (iter = diffMapIDToNormRegion.begin(); iter != diffMapIDToNormRegion.end(); iter++)
		level->m_apDiffusionMaps[iter->first]->NormalizeSimilaritiesBasedOnRegion(iter->second);
}

void CollectionDiffuseHierarchical::CreateHierarchies()
{
	m_aHierarchies.push_back(new LevelOfDiffusionHierarchy(m_pDB, m_DiffusionParams, 
														   LevelOfDiffusionHierarchy::GROUP_MODELS, 0));
	m_aHierarchies.push_back(new LevelOfDiffusionHierarchy(m_pDB, m_DiffusionParams, 
														   LevelOfDiffusionHierarchy::GROUP_PNTS, 1));
	m_aHierarchies[0]->SetElementsToGroupMapping(LevelOfDiffusionHierarchy::GROUP_PNTS);
	m_aHierarchies[1]->SetElementsToGroupMapping(LevelOfDiffusionHierarchy::GROUP_PNTS);
}

void CollectionDiffuseHierarchical::FillHierarchies()
{
	// propagation of values & clustering
	// calculate values:
	CallAddEntryForAll();
	
	// fill hierarchies
	for (int i=0; i<(int)m_aHierarchies.size(); i++)
		FillClustering(i);
}

void CollectionDiffuseHierarchical::FillClustering(int levelID)
{
	assert(levelID<(int)m_aHierarchies.size());
	LevelOfDiffusionHierarchy * thisLevel = m_aHierarchies[levelID];
	
	int nMaxElementsToDiffuse = 10000;
	// fill clusters
	std::vector<int> clustering;
	if (levelID==0)
	{
		for (int i=0; i<m_iNModels; i++)
			clustering.push_back(0);
	}
	else
	{
		LevelOfDiffusionHierarchy * prevLevel = m_aHierarchies[levelID-1];
		std::vector<int> groupClustering;
		prevLevel->Cluster(nMaxElementsToDiffuse, groupClustering);
		prevLevel->ExpandGroupIntoElements(groupClustering, clustering);
	}
	thisLevel->SetClustering(clustering);
	
	// fill values
	for (int pIDFrom=0; pIDFrom<m_iNPoints*m_iNModels; pIDFrom++)
	for (int mIDTo=0; mIDTo<m_iNModels; mIDTo++)
	for (int pTo=0; pTo<(int)m_PointToPointGroupedByModel[pIDFrom][mIDTo].size(); pTo++)
	{
		double value = m_PointToPointGroupedByModel[pIDFrom][mIDTo][pTo].val;
		if (thisLevel->m_GroupingType==LevelOfDiffusionHierarchy::GROUP_MODELS)
			thisLevel->AddCorrValue(pIDFrom / m_iNPoints, mIDTo, value);
		else if (thisLevel->m_GroupingType==LevelOfDiffusionHierarchy::GROUP_PNTS)
			thisLevel->AddCorrValue(pIDFrom, m_PointToPointGroupedByModel[pIDFrom][mIDTo][pTo].id, value);
		else
			assert(false);
	}
}

void CollectionDiffuseHierarchical::AddEntry(int mtxID1, int mtxID2, double val)
{
	int NDims = m_iNModels * m_iNPoints;
	if (m_PointToPointGroupedByModel==NULL)
	{
		m_PointToPointGroupedByModel = new std::vector<CorrValueID> *[NDims];
		for (int i=0; i<NDims; i++)
			m_PointToPointGroupedByModel[i] = new std::vector<CorrValueID>[m_iNModels];
	}
	
	int toModelID = CollectionDiffuse::PositionToPoint(mtxID2, m_pDB, m_iNPoints).GetModel()->ModelID();

	CorrValueID cv;
	cv.id=mtxID2;
	cv.val = val;
	m_PointToPointGroupedByModel[mtxID1][toModelID].push_back(cv);	
}

//void CollectionDiffuseHierarchical::FillValues()
//{	
//	assert(m_PointToPointGroupedByModel==NULL);
//	int NDims = m_iNModels * m_iNPoints;
//	m_PointToPointGroupedByModel = new std::vector<CorrValueID> *[NDims];
//	for (int i=0; i<NDims; i++)
//		m_PointToPointGroupedByModel[i] = new std::vector<CorrValueID>[m_iNModels];
//	
//	std::cout<<"Similarity Matrix: Fill "<<NDims<<" x "<<NDims<<std::endl;
//	TimeProfiler timer;		
//	// fill matrix (iterate over all correspondences)
//	CorrCollection * collection = m_pDB->GetCorrCollection(m_CorrMtxParams.m_sInputCollection);
//	assert(collection!=NULL);
//	CollectionPairwise * singleSrcColl = collection->SafeCastToPairwise();
//	assert(singleSrcColl!=NULL);
//	for (int i=0; i<singleSrcColl->NumModelCorrs(); i++)
//	{
//		timer.WriteProgress("FillMtx", i, singleSrcColl->NumModelCorrs());
//		
//		CollectionSinglePair * corr = singleSrcColl->GetModelCorr(i);
//		assert(corr!=NULL);
//		
//		int fromID = corr->ModelFrom()->ModelID();
//		int toID = corr->ModelTo()->ModelID();
//		
//		Model3D * model = corr->ModelFrom();		
//		PointSet * pntCld = model->PointCloud();
//		assert(pntCld!=NULL);
//		
//		CorrValue * globalVal = corr->GetCorrValueByName(m_CorrMtxParams.m_sGlobalCorrValue);
//		double globalScale = 1;
//		if (globalVal!=NULL)
//		{
//			double integratedVal = globalVal->IntegratedValue(model->ModelName());
//			assert(integratedVal<=1. + 1E-10 && integratedVal>=0);
//			if (integratedVal < m_CorrMtxParams.m_fMinOverlap)
//				globalScale = 0;
//			else if (integratedVal > m_CorrMtxParams.m_fMaxOverlap)
//				globalScale = 1.;
//			else
//				globalScale = integratedVal / m_CorrMtxParams.m_fMaxOverlap;
//		}
//		
//		if (globalScale <= m_CorrMtxParams.m_fValueThreshold)	// enforce sparsity
//			continue;
//		
//		CorrValue * localVal = corr->GetCorrValueByName(m_CorrMtxParams.m_sLocalCorrValue);
//		
//		ModelPoint p1 = pntCld->GetPoint(0);
//		for (int p1ID=0; p1ID < m_iNPoints; p1ID++)
//		{
//			ModelPoint p1 = pntCld->GetPoint(p1ID);
//			int mtxID1 = fromID * m_iNPoints + p1ID; 
//			std::vector<ModelPoint> corrPnts;
//			corr->GetCorrs(p1, &corrPnts);
//			
//			for (int p2ID=0; p2ID<(int)corrPnts.size(); p2ID++)
//			{
//				ModelPoint p2 = corrPnts[p2ID];
//				int toModelID = p2.GetModel()->ModelID();
//				int mtxID2 = CollectionDiffuse::MatrixPosition(p2, m_iNPoints, m_iNModels);
//				
//				double atPntVal = 1.;
//				if (localVal!=NULL)
//					atPntVal = localVal->Value(p1, p2);
//				
//				if (globalScale * atPntVal > m_CorrMtxParams.m_fValueThreshold)	// enforce sparsity
//				{
//					CorrValueID cv;
//					cv.id=mtxID2;
//					cv.val = globalScale*atPntVal;
//					m_PointToPointGroupedByModel[mtxID1][toModelID].push_back(cv);
//				}
//			}
//		}
//	}
//	std::cout<<std::endl;
//}


VKString CollectionDiffuseHierarchical::GetCollectionType()
{
	return "CollectionDiffuseHierarchical";
}

bool CollectionDiffuseHierarchical::ReadCollectionChild(std::ifstream & textStream)
{
	assert(false); // TODO
//	VKString textStr;
//	textStr.readToken(textStream);
//	assert(textStr=="NHierarchies");
//	m_aHierarchies
	
}

void CollectionDiffuseHierarchical::WriteCollectionChild(std::ofstream & textStream)
{
	assert(false); // TODO 
//	textStream<<"NHierarchies "<<m_aHierarchies.size()<<"\n";
//	VKString collectionDir = m_pDB->GetCollectionDir(m_sCollectionName);	
//	
}


