/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CollectionDiffuseInterpolator.h"
#include "ModelDatabase.h"

CollectionDiffuseInterpolator::CollectionDiffuseInterpolator
									(const VKString & collectionName, 
									 ModelDatabase * db)
: CollectionDiffuse(collectionName, db)
{
	m_iCacheSize = 200;
	m_pDiffuseCollection = NULL;
	m_pDiffuseCorrValue = NULL;
	
	m_sInputDiffCollection = "";
	m_sInputCorrValue = "";
	m_sInputPntSet = "";
	m_sOutputPntSet = "";

}

CollectionDiffuseInterpolator::CollectionDiffuseInterpolator
									(const VKString & collectionName, 
									 ModelDatabase * db,
									 const VKString & inputDiffuse, 
									 const VKString & inputDiffVal,
									 const VKString & inputPntSet,
									 const VKString & outputPntSet)
: CollectionDiffuse(collectionName, db)
{
	m_iCacheSize = 200;
	m_sInputDiffCollection = inputDiffuse;
	m_sInputCorrValue = inputDiffVal;
	m_sInputPntSet = inputPntSet;
	m_sOutputPntSet = outputPntSet;

	m_pDiffuseCollection = NULL;
	m_pDiffuseCorrValue = NULL;

}

void CollectionDiffuseInterpolator::FillPtrs()
{
	m_pDiffuseCollection = m_pDB->GetCorrCollection(m_sInputDiffCollection);
	if (m_pDiffuseCollection==NULL)
		std::cout<<"[ERROR] Cannot find collection: "<<m_sInputDiffCollection.c_str()<<std::endl;
	assert(m_pDiffuseCollection!=NULL);	
	m_pDiffuseCorrValue = m_pDiffuseCollection->GetCorrValueByName(m_sInputCorrValue);
//	if (m_pDiffuseCorrValue==NULL)
//	{
//		std::cout<<"[ERROR] Cannof find corr value: "<<m_sInputCorrValue.c_str();
//		std::cout<<" in collection "<<m_sInputDiffCollection.c_str()<<std::endl;
//	}
//	assert(m_pDiffuseCorrValue!=NULL);
}

ModelPoint CollectionDiffuseInterpolator::GetBestCorr(const ModelPoint & pnt, 
													  Model3D * toModel, 
													  const VKString & modelValue, 
													  double * value)
{
	FillPtrs();
	
	// for point (should be in 'output' point set) - get correspondence
	// for which diffusion is known
	CollectionIdentityPointsets * interp = GetInterpolator((Model3D*)pnt.GetModel());
	
	int * corrIDs;
	double * weights;
	int nCorrs;
	CollectionIdentityPointsets::PntSetType tp;
	interp->QuickCorrsWithWeights(pnt, &corrIDs, &weights, &nCorrs, &tp);	
	int bestWeight = -1;
	double bwVal = 0;
	for (int i=0; i<nCorrs; i++)
	{
		if (bestWeight==-1 || bwVal > weights[i])
		{
			bwVal = weights[i];
			bestWeight = i;
		}			
	}
	assert(bestWeight>=0);
	// we know corrs for this point:
	ModelPoint p = interp->m_PntsTo->GetPoint(corrIDs[bestWeight]);
	
	// map by diffused map
	return m_pDiffuseCollection->GetBestCorr(p, toModel, modelValue, value);
}

void CollectionDiffuseInterpolator::GetCorrs(const ModelPoint & pnt,	
											 std::vector<ModelPoint> * corrPnts)
{
	std::map<ModelPoint, double> & corrs = GetCachedCorrs(pnt);
	
	for (std::map<ModelPoint, double>::iterator iter = corrs.begin(); 
		 iter != corrs.end(); iter++)
		corrPnts->push_back(iter->first);
}

std::map<ModelPoint, double> & 
CollectionDiffuseInterpolator::GetCachedCorrs(const ModelPoint & pnt)
{
	std::map<ModelPoint, std::map<ModelPoint, double> >::iterator iter;
	iter = m_CachedCorrs.find(pnt);
	if (iter != m_CachedCorrs.end())
		return iter->second;

	m_AddedPntsInOrder.push_back(pnt);
	if (m_CachedCorrs.size() > m_iCacheSize)
	{
		iter = m_CachedCorrs.find(m_AddedPntsInOrder[0]);
		assert(iter != m_CachedCorrs.end());
		m_CachedCorrs.erase(iter);
		m_AddedPntsInOrder.erase(m_AddedPntsInOrder.begin());
	}

	FillPtrs();
	
	// for point (should be in 'output' point set) - get correspondence
	// for which diffusion is known
	CollectionIdentityPointsets * interp = GetInterpolator((Model3D*)pnt.GetModel());
	
	int * corrIDs;
	double * weights;
	int nCorrs;
	CollectionIdentityPointsets::PntSetType tp;
	interp->QuickCorrsWithWeights(pnt, &corrIDs, &weights, &nCorrs, &tp);
	std::map<ModelPoint, double> & currCorrs = m_CachedCorrs[pnt];
	
	for (int i=0; i<nCorrs; i++)
	{
		// we know corrs for this point:
		ModelPoint p = interp->m_PntsTo->GetPoint(corrIDs[i]);
		double w = weights[i];
		
		// map by diffused map
		std::vector<ModelPoint> diffCorrs;
		m_pDiffuseCollection->GetCorrs(p, &diffCorrs);
		
//		std::cout<<"Interp Corrs: "<<nCorrs<<" w="<<w<<std::endl;
//		std::cout<<"Diffuse Corrs: "<<diffCorrs.size()<<std::endl;
		// accumulate values
		for (int j=0; j<(int)diffCorrs.size(); j++)
		{
			double value = 0;
			
			if (m_pDiffuseCorrValue!=NULL)
				value = m_pDiffuseCorrValue->Value(p, diffCorrs[j]) * w;
			else	// corr value might not be defined for collection
			{
				CollectionPairwise * collPair = m_pDiffuseCollection->SafeCastToPairwise();
				assert (collPair!=NULL);
				int m1ID = pnt.GetModel()->ModelID();
				int m2ID = diffCorrs[j].GetModel()->ModelID();
				assert(collPair->NumModelCorrs(m1ID, m2ID)==1);
				int globalCorrID = collPair->GetModelCorrGlobalID(m1ID, m2ID, 0);
				CollectionSinglePair * singlePair = collPair->GetModelCorr(globalCorrID);
				assert(singlePair!=NULL);
				CorrValue * corrVal = singlePair->GetCorrValueByName(m_sInputCorrValue);
				assert(corrVal!=NULL);
				value = corrVal->Value(p, diffCorrs[j]) * w;
			}
			std::map<ModelPoint, double>::iterator iter = currCorrs.find(diffCorrs[j]);
			if (iter == currCorrs.end())
				currCorrs[diffCorrs[j]] = value;
			else
				iter->second += value;
		}
	}
	iter = m_CachedCorrs.find(pnt);
	assert (iter != m_CachedCorrs.end());

	return iter->second;
}

double CollectionDiffuseInterpolator::DiffusedCorrValue(const ModelPoint & pntFrom, 
														const ModelPoint & pntTo)
{
	std::map<ModelPoint, double> & corrs = GetCachedCorrs(pntFrom);
	std::map<ModelPoint, double>::iterator iter = corrs.find(pntTo);
	if (iter == corrs.end())
		return 0.;
	else
		return iter->second;
}

VKString CollectionDiffuseInterpolator::GetCollectionType()
{
	return "CollectionDiffuseInterpolator";
}

void CollectionDiffuseInterpolator::AddEntry(int p1, int p2, double val)
{
	assert(false);
}

CollectionIdentityPointsets * CollectionDiffuseInterpolator::GetInterpolator(Model3D * model)
{
	FillPtrs();
	std::map<Model3D*, CollectionIdentityPointsets *>::iterator iter;
	iter = m_apInterpolators.find(model);
	if (iter==m_apInterpolators.end())
	{
		VKString interpName = "DefaultDiffuseInterp";
		m_apInterpolators[model] = new CollectionIdentityPointsets(interpName, m_pDB, -1,
																   model->ModelName(), 
																   m_sOutputPntSet, 
																   m_sInputPntSet, 
																   1, true);
		iter =  m_apInterpolators.find(model);
	}
	assert(iter != m_apInterpolators.end());
	return iter->second;
}

bool CollectionDiffuseInterpolator::ReadCollectionChild(std::ifstream & textStream)
{
	m_sInputDiffCollection.readToken(textStream);
	m_sInputCorrValue.readToken(textStream);
	m_sInputPntSet.readToken(textStream);
	m_sOutputPntSet.readToken(textStream);
	
	FillPtrs();
	return true;
}

void CollectionDiffuseInterpolator::WriteCollectionChild(std::ofstream & textStream)
{
	textStream<<m_sInputDiffCollection.c_str()<<" "<<m_sInputCorrValue.c_str()<<" ";
	textStream<<m_sInputPntSet.c_str()<<" "<<m_sOutputPntSet.c_str()<<"\n";
}



