/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#include "CollectionGroundTruth.h"
#include "CorrValue.h"
#include "ModelDatabase.h"

CollectionGroundTruth::GTCorrInfo::GTCorrInfo()
{
	m_bAffineDefined = false;
}

CollectionGroundTruth::CollectionGroundTruth(const VKString & collectionName,
											 ModelDatabase * database)
: CorrCollection(collectionName, database, -1)
{
	assert(database!=NULL);
	m_iConstantCorrValue = -1;
	m_iRigidAlignCorrValue = -1;
	SetPreferedGTType(GTTYPE_FEAT_PNTS);
}

void CollectionGroundTruth::SetPreferedGTType(GTType tp)
{
	if (tp!=GTTYPE_FEAT_PNTS)
		std::cout<<"[ERROR] Gt type: "<<tp<<std::endl;
	assert(tp==GTTYPE_FEAT_PNTS);
	m_PreferedGTType = tp;
	
	if (m_iConstantCorrValue==-1)
	{
		m_iConstantCorrValue = (int)m_CorrValues.size();
		m_CorrValues.push_back(new CorrValueConstant(m_pDB));
	}
	
	m_NameToCorrValue[DefaultGroundTruthCorrValue()] = m_iConstantCorrValue;
}

CollectionGroundTruth::~CollectionGroundTruth()
{
	
}

void CollectionGroundTruth::RemoveModel(int modelID)
{
	Model3D * modelToBeRemoved = m_pDB->GetModelByID(modelID);
	assert(modelToBeRemoved!=NULL);
	std::map<Model3D *, GTCorrInfo>::iterator iter = m_ModelInfo.find(modelToBeRemoved);
	if (iter != m_ModelInfo.end())
		m_ModelInfo.erase(iter);
	
	for (std::map<VKString, std::vector<Model3D*> >::iterator iter = m_ModelsInClass.begin();
		 iter != m_ModelsInClass.end(); iter++)
	{
		std::vector<Model3D*> & models = iter->second;
		for (int i=0; i<(int)models.size(); i++)
		{
			if (models[i]==modelToBeRemoved)
			{
				models.erase(models.begin()+i);
				i--;
			}
		}
	}	
}

void CollectionGroundTruth::GetCorrs(const ModelPoint & pnt, 
									 std::vector<ModelPoint> * corrPnts)
{
	Model3D * modelFrom = (Model3D *)pnt.GetModel();
	assert(m_ModelInfo.find(modelFrom)!=m_ModelInfo.end());
	GTCorrInfo & info = m_ModelInfo[modelFrom];
	
	std::map<Model3D*, bool> foundCorr;

	for (int i=0; i<info.m_ModelInClass.count(); i++)
	{
		VKString className = info.m_ModelInClass[i];
		int pntID = GetPntID(info, i, pnt);
		if (pntID >= 0)
		{
			std::map<VKString, std::vector<Model3D*> >::iterator iter;
			iter = m_ModelsInClass.find(className);
			assert(iter!=m_ModelsInClass.end());
			for (int j=0; j<(int)iter->second.size(); j++)
			{
				Model3D * model = iter->second[j];
				if (model != modelFrom && foundCorr.find(model)==foundCorr.end())
				{
					std::map<Model3D *, GTCorrInfo>::iterator iter2;
					iter2 = m_ModelInfo.find(model);
					assert(iter2!=m_ModelInfo.end());

					for (int k = 0; k < iter2->second.m_ModelInClass.count(); k++)
					{
						if (iter2->second.m_ModelInClass[k]==className)
						{
							PointSet * toPnts = model->GetPointSet(info.m_FeaturePntSetName[k]);
							assert(toPnts!=NULL);
							const ModelPoint & goodCorr = toPnts->GetPoint(pntID);
							corrPnts->push_back(goodCorr);
							foundCorr[model] = true;
							break;
						}
					}
				}
			}
		}
	}
}

void CollectionGroundTruth::AddModelFeaturePntsGT(Model3D* model, 
												  const VKString & featPntsSetName,
												  const VKString & className)
{
	std::map<Model3D *, GTCorrInfo>::iterator iter;
	iter = m_ModelInfo.find(model);
	if (iter==m_ModelInfo.end())
	{
		m_ModelInfo[model] = GTCorrInfo();
		iter = m_ModelInfo.find(model);
		assert(iter!=m_ModelInfo.end());
	}
	m_ModelsInClass[className].push_back(model);

	iter->second.m_ModelInClass.push_back(className);
	iter->second.m_FeaturePntSetName.push_back(featPntsSetName);
	int addPntsTo = (int)iter->second.m_ModelPointToID.size();
	iter->second.m_ModelPointToID.push_back(std::map<ModelPoint, int>());
	
	if (featPntsSetName=="AllVertices")
		iter->second.m_PntsTypes.push_back(PNTS_VERTEXID);
	else if (featPntsSetName=="AllFaces")
		iter->second.m_PntsTypes.push_back(PNTS_FACEID);
	else if (featPntsSetName=="ConnectedPointCloud")
		iter->second.m_PntsTypes.push_back(PNTS_POINTCLDID);
	else
	{
		iter->second.m_PntsTypes.push_back(PNTS_GENERAL);
		std::map<ModelPoint, int> & pnt2ID = iter->second.m_ModelPointToID[addPntsTo];

		PointSet * pnts = model->GetPointSet(featPntsSetName);
		if (pnts==NULL)
			std::cout<<"[ERROR] Cannot find pnts: "<<featPntsSetName.c_str()<<std::endl;
		assert(pnts!=NULL);

		for (int i=0; i < pnts->NumPoints(); i++)
			pnt2ID[pnts->GetPoint(i)] = i;
	}
}

int CollectionGroundTruth::GetPntID(GTCorrInfo & info, int classID, const ModelPoint & pnt)
{
	if (info.m_PntsTypes[classID]==PNTS_VERTEXID)
	{
		bool atVtx;
		int vID = pnt.NearestVertex(&atVtx);
		if (atVtx)
			return vID;
		else
			return -1;
	}
	else if (info.m_PntsTypes[classID]==PNTS_FACEID)
	{
		return pnt.TriID();
	}
	else if (info.m_PntsTypes[classID]==PNTS_POINTCLDID)
	{
		return pnt.PntCloudID();
	}
	else if (info.m_PntsTypes[classID]==PNTS_GENERAL)
	{
		std::map<ModelPoint, int>::iterator iter = info.m_ModelPointToID[classID].find(pnt);
		if (iter==info.m_ModelPointToID[classID].end())
			return -1;
		else
			return iter->second;		
	}

	assert(false);
	return -1;
}

void CollectionGroundTruth::AddModelRigidToCanonicalGT(Model3D* model, R3Affine & gtXForm)
{
	assert(false);	// todo
}


VKString CollectionGroundTruth::GetCollectionType()
{
	return "CollectionGroundTruth";
}

VKString CollectionGroundTruth::DefaultGroundTruthMapName()
{
	return "GroundTruthCorrsName";
}

VKString CollectionGroundTruth::DefaultGroundTruthCorrValue()
{
	return "GroundTruthCorrsValueName";
}

bool CollectionGroundTruth::ReadCollectionChild(std::ifstream & textStream)
{	
	VKString tempStr;
	tempStr.readToken(textStream);		assert(tempStr=="GroundTruth");
	tempStr.readToken(textStream);		assert(tempStr=="Type");
	int prefTp;
	textStream>>prefTp;
	SetPreferedGTType((GTType)prefTp);
	
	tempStr.readToken(textStream);		assert(tempStr=="NModels");
	int nModels;
	textStream>>nModels;
	
	int unknownModels = 0;
	
	for (int i=0; i<nModels; i++)
	{
		tempStr.readToken(textStream);	assert(tempStr=="Model");
		int modelID;
		textStream>>modelID;
		Model3D * model = m_pDB->GetModelByID(modelID);
		tempStr.readToken(textStream);
		if (model==NULL || model->ModelName()!=tempStr)
			model = m_pDB->GetModelByName(tempStr);
				
		bool affineExists;
		tempStr.readToken(textStream);	assert(tempStr=="AffineExists");
		textStream>>affineExists;
		assert(!affineExists);

		tempStr.readToken(textStream);	assert(tempStr=="NClasses");
		int nClasses;
		textStream>>nClasses;

		if (model==NULL)
			unknownModels++;
		
		for (int c=0; c<nClasses; c++)
		{
			VKString className;
			VKString pntSetName;
			className.readToken(textStream);
			pntSetName.readToken(textStream);
			if (model!=NULL)
				AddModelFeaturePntsGT(model, pntSetName, className);
		}
	}
	if (unknownModels>0)
		std::cout<<"[WARNING] Ground Truth has "<<unknownModels<<" unknown models"<<std::endl;
	
	tempStr.readToken(textStream);		assert(tempStr=="GroundTruthAlign");
	return true;
}

void CollectionGroundTruth::WriteCollectionChild(std::ofstream & textStream)
{
	textStream<<"GroundTruth Type "<<m_PreferedGTType<<"\n";
	textStream<<"NModels "<<m_ModelInfo.size()<<"\n";
	
	for (std::map<Model3D *, GTCorrInfo>::iterator iter = m_ModelInfo.begin();
		 iter != m_ModelInfo.end(); iter++)
	{
		textStream<<"Model "<<iter->first->ModelID();
		textStream<<" "<<iter->first->ModelName().c_str()<<"\n";
		int nClasses = iter->second.m_ModelInClass.count();
		textStream<<"AffineExists "<<iter->second.m_bAffineDefined<<"\n";
		assert(!iter->second.m_bAffineDefined);
		textStream<<"NClasses "<<nClasses<<"\n";
		for (int i=0; i<nClasses; i++)
		{
			VKString className = iter->second.m_ModelInClass[i];
			VKString pntSetName = iter->second.m_FeaturePntSetName[i];
			textStream<<className.c_str()<<" "<<pntSetName.c_str()<<"\n";
		}
	}
	
	textStream<<"GroundTruthAlign"<<"\n";
}

