/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CollectionSinglePair.h"
#include "CollectionPairwise.h"
#include "ModelDatabase.h"
#include "TimeProfiler.h"

CollectionPairwise::CollectionPairwise(const VKString & collectionName, 
									   ModelDatabase * database, bool writeToSingleFile)
: CorrCollection(collectionName, database, -1)
{
	m_bWriteToSingleFile = writeToSingleFile;
}

CollectionPairwise::~CollectionPairwise()
{
}

CollectionPairwise * CollectionPairwise::SafeCastToPairwise()
{
	return this;
}

void CollectionPairwise::GetCorrs(const ModelPoint & pnt, 
								  std::vector<ModelPoint> * corrPnts)
{
	int modelFromID = pnt.GetModel()->ModelID();
	std::map<int, std::vector<int> >::iterator iter;
	iter = m_mModelFromIDToCorrIDs.find(modelFromID);
	assert(iter!=m_mModelFromIDToCorrIDs.end());
	for (int i=0; i<(int)iter->second.size(); i++)
		GetModelCorr(iter->second[i])->GetCorrs(pnt, corrPnts);
}

int CollectionPairwise::GetFromModelCorrGlobalID(int modelFromID, int corrID)
{
	std::map<int, std::vector<int> >::iterator iter;
	iter = m_mModelFromIDToCorrIDs.find(modelFromID);
	assert(iter!=m_mModelFromIDToCorrIDs.end());
	assert(corrID>=0 && corrID < iter->second.size());
	return iter->second[corrID];
}

int CollectionPairwise::GetToModelCorrGlobalID(int modelToID, int corrID)
{
	std::map<int, std::vector<int> >::iterator iter;
	iter = m_mModelToIDToCorrIDs.find(modelToID);
	assert(iter!=m_mModelToIDToCorrIDs.end());
	assert(corrID>=0 && corrID < iter->second.size());
	return iter->second[corrID];	
}

int CollectionPairwise::GetModelCorrGlobalID(int modelFromID, int modelToID, int corrID)
{
	std::map<int, std::map<int, std::vector<int> > >::iterator iter;
	iter = m_mModelPairToCorrIDs.find(modelFromID);
	assert(iter!=m_mModelPairToCorrIDs.end());
	std::map<int, std::vector<int> >::iterator iter2 = iter->second.find(modelToID);
	assert(iter2!=iter->second.end());
	assert(corrID>=0 && corrID < iter2->second.size());
	return iter2->second[corrID];
}

CollectionSinglePair * CollectionPairwise::GetModelCorr(int corrID)
{
	if (corrID<0 || corrID >= (int)m_ModelCorrs.size())
	{
		std::cout<<"[ERROR] Trying to get map: "<<corrID;
		std::cout<<" / "<<m_ModelCorrs.size()<<std::endl;
	}
	assert(corrID >= 0 && corrID < (int)m_ModelCorrs.size());
	return m_ModelCorrs[corrID];
}

int CollectionPairwise::NumCorrsFromModel(int modelFrom)
{
	return (int)m_mModelFromIDToCorrIDs[modelFrom].size();
}

int CollectionPairwise::NumCorrsToModel(int modelTo)
{
	return (int)m_mModelToIDToCorrIDs[modelTo].size();
}

int CollectionPairwise::NumModelCorrs(int modelFrom, int modelTo)
{
	return (int)m_mModelPairToCorrIDs[modelFrom][modelTo].size();
}

int CollectionPairwise::NumModelCorrs()
{
	return (int)m_ModelCorrs.size();
}

ModelPoint CollectionPairwise::GetBestCorr(const ModelPoint & pnt, 
										  Model3D * toModel, 
										  const VKString & modelValue, 
										  double * value)
{
	int m1 = pnt.GetModel()->ModelID();
	int m2 = toModel->ModelID();
	assert(NumModelCorrs(m1, m2)==1);

	int gID = GetModelCorrGlobalID(m1, m2, 0);
	CollectionSinglePair * singlePair = GetModelCorr(gID);
	assert(singlePair!=NULL);
	return singlePair->GetBestCorr(pnt, toModel, modelValue, value);
}


void CollectionPairwise::ReplaceModelCorr(int mapID, CollectionSinglePair * modelCorr)
{
	assert(mapID>=0 && mapID<(int)m_ModelCorrs.size());
	assert(modelCorr->ModelFrom()==m_ModelCorrs[mapID]->ModelFrom());
	assert(modelCorr->ModelTo()==m_ModelCorrs[mapID]->ModelTo());
	delete m_ModelCorrs[mapID];
	m_ModelCorrs[mapID] = modelCorr;
}

void CollectionPairwise::AddModelCorr(CollectionSinglePair * modelCorr)
{
	assert(modelCorr!=NULL);
	int corrID = (int)m_ModelCorrs.size();
	assert(modelCorr->ModelFrom()!=NULL);
	assert(modelCorr->ModelTo()!=NULL);
	m_ModelCorrs.push_back(modelCorr);
	int modelFromID = modelCorr->ModelFrom()->ModelID();
	int modelToID = modelCorr->ModelTo()->ModelID();
	m_mCorrIDToModelFromID.push_back(modelFromID);
	m_mCorrIDToModelToID.push_back(modelToID);
	
	m_mModelFromIDToCorrIDs[modelFromID].push_back(corrID);
	m_mModelToIDToCorrIDs[modelToID].push_back(corrID);
	m_mModelPairToCorrIDs[modelFromID][modelToID].push_back(corrID);
}

VKString CollectionPairwise::GetCollectionType()
{
	return "CollectionPairwise";
}

bool CollectionPairwise::ReadCollectionChild(std::ifstream & textStream)
{
	TimeProfiler timer;
	assert(textStream.is_open());
	VKString text;
	
	text.readToken(textStream);
	assert(text=="ConcatColl");
	textStream>>m_bWriteToSingleFile;
	
	int NCorrs;
	text.readToken(textStream);
	assert(text=="NCorrs");
	textStream>>NCorrs;

	for (int i=0; i<NCorrs; i++)
	{
		if (i%10==0)
			timer.WriteProgress("ReadingColleciton", i, NCorrs);		
		text.readToken(textStream);
		assert(text=="ModelPair");
		int modelFromID, modelToID;
		textStream>>modelFromID>>modelToID;
		bool couldRead = false;
		if (m_bWriteToSingleFile)
			couldRead = CorrCollection::ReadAndCreateCollectionSinglePairWithID(
															m_sCollectionName, 
													        i, m_pDB, textStream);
		else
			couldRead = CorrCollection::ReadAndCreateCollectionSinglePairWithID(
														    m_sCollectionName, 
															i, m_pDB);
		assert(couldRead);
	}
	std::cout<<std::endl;
	return true;
}

void CollectionPairwise::WriteCollectionChild(std::ofstream & textStream)
{
	TimeProfiler timer;
	assert(textStream.is_open());
	
	textStream<<"ConcatColl "<<m_bWriteToSingleFile<<"\n";
	textStream<<"NCorrs "<<NumModelCorrs()<<"\n";
	for (int i=0; i<NumModelCorrs(); i++)
	{
		if (i%10==0)
			timer.WriteProgress("WritingCollection", i, NumModelCorrs());
		textStream<<"ModelPair "<<m_mCorrIDToModelFromID[i];
		textStream<<" "<<m_mCorrIDToModelToID[i]<<"\n";
		if (m_bWriteToSingleFile)
			GetModelCorr(i)->WriteCollection(textStream);	
		else
			GetModelCorr(i)->WriteCollection();	
	}
	std::cout<<std::endl;
}

bool CollectionPairwise::ReadCorrValues(const VKString & corrValueName)
{
//	std::cout<<"Reading Corr Values: "<<corrValueName.c_str()<<std::endl;
//	std::cout<<"\tRet="<<(m_NameToCorrValue.find(corrValueName)==m_NameToCorrValue.end())<<std::endl;
//	if (m_NameToCorrValue.find(corrValueName)==m_NameToCorrValue.end())
//		return true;
	
	assert(m_bWriteToSingleFile);
	VKString streamFile = m_pDB->GetCorrValueFile(m_sCollectionName, corrValueName);
	if (!m_pDB->FileExists(streamFile))
		return false;
	
	std::ifstream textStream(streamFile.c_str());
	assert(textStream.is_open());
	for (int i=0; i<(int)m_ModelCorrs.size(); i++)
	{
		bool couldRead = CorrValue::ReadCorrValue(textStream, m_pDB, 
												  m_sCollectionName, i, corrValueName);
		assert(couldRead);
	}
	textStream.close();
	return true;
}

void CollectionPairwise::WriteCorrValues(const VKString & corrValueName)
{
	// for now assume it's only concatenated
	assert(m_bWriteToSingleFile);
	VKString streamFile = m_pDB->GetCorrValueFile(m_sCollectionName, corrValueName);
	m_pDB->CreateDirsOnPathToFile(streamFile);
	std::ofstream textStream(streamFile.c_str());
	assert(textStream.is_open());
	for (int i=0; i<(int)m_ModelCorrs.size(); i++)
	{
		CollectionSinglePair * corrs = m_ModelCorrs[i];
		CorrValue * corrValue = corrs->GetCorrValueByName(corrValueName);
		assert(corrValue!=NULL);
		corrValue->WriteCorrValue(textStream);
	}
	textStream.close();
}






