/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "DistanceEuclidean.h"
#include "PointSet.h"
#include "Model3D.h"
#include "VkFunctions.h"

DistanceEuclidean::DistanceEuclidean(Model3D * model)
: ModelDistance(model)
{
}

double DistanceEuclidean::Distance(const ModelPoint & p1, const ModelPoint & p2)
{
	return R3Distance(p1.GetPosition(), p2.GetPosition()) / GetStandardNorm();
}

ModelPoint DistanceEuclidean::GetClosestPoint(const R3Point & p1, 
											  const VKString & pointSet)
{
	assert(m_pModel!=NULL);
	PointSet * pnts = m_pModel->GetPointSet(pointSet);
	if (pnts==NULL)
		pnts = m_pModel->PointCloud();
	
	R3Kdtree<const PointInSet *> * kdtree = pnts->GetKdTree();
	const PointInSet * closestPnt = kdtree->FindClosest(p1, 0, FLT_MAX);
	assert(closestPnt!=NULL);
	return closestPnt->GetModelPoint();
}

ModelPoint DistanceEuclidean::GetClosestPoint(const ModelPoint & p1, 
											  const VKString & pointSet)
{
	Model3D * model = (Model3D*)p1.GetModel();
	assert(model==m_pModel);
	return GetClosestPoint(p1.GetPosition(), pointSet);
}

void DistanceEuclidean::GetNeighborhood(const ModelPoint & p1, double threshold, 
										std::vector<ModelPoint> * trgPnts, 
										const VKString & pointSet)
{

	//std::cout<<"Threshold = "<<threshold<<" * "<<GetStandardNorm()<<std::endl;
	threshold = threshold * GetStandardNorm();
	
	Model3D * model = (Model3D*)p1.GetModel();
	
	if (model==NULL || model!=m_pModel)
	{
		if (model!=NULL && m_pModel!=NULL)
		{
			std::cout<<"[ERROR] Different models: "<<model->ModelName().c_str();
			std::cout<<" and "<<m_pModel->ModelName().c_str()<<std::endl;
		}
		else
			std::cout<<"[ERROR] NullPtr: "<<model<<" "<<m_pModel<<std::endl;
		assert(model!=NULL && model==m_pModel);
	}
	PointSet * pnts = model->GetPointSet(pointSet);
	if (pnts==NULL)
		pnts = model->PointCloud();
	
	R3Kdtree<const PointInSet *> * kdtree = pnts->GetKdTree();
	RNArray<const PointInSet *> neighbors;
	kdtree->FindAll(p1.GetPosition(), 0, threshold, neighbors);

	for (int i=0; i<neighbors.NEntries(); i++)
		trgPnts->push_back(neighbors[i]->GetModelPoint());

}

double DistanceEuclidean::CalculateStandardNorm()
{
	int norm=0;
	double aveRad = 0;
	double maxDiam = 0;
	PointSet * pntCloud = m_pModel->PointCloud();
	for (int i=0; i<pntCloud->NumPoints(); i++)
	for (int j=i+1; j<pntCloud->NumPoints(); j++)
	{
		double dist = R3Distance(pntCloud->GetPoint(i).GetPosition(), 
								 pntCloud->GetPoint(j).GetPosition());
		norm++;
		aveRad += dist;
		maxDiam = vkMax(maxDiam, dist);
	}
	
//	if (m_pModel->ModelID()==0)
//		std::cout<<"Calculated Norm = "<<maxDiam<<" pairs: "<<norm<<" aveDist="<<aveRad<<std::endl;
	//return aveRad / norm;
	return maxDiam;
}

VKString DistanceEuclidean::DistanceType()
{
	return "DistanceEuclidean";
}
