/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "DynamicGraphElements.h"
#include "CollectionDiffuseDynamicGraph.h"
#include "ModelDatabase.h"
#include "Model3D.h"
#include "AnalysisPipeline.h"
#include "DatabaseVisWindow.h"

//////////////////	GraphNodeModel	///////////////
GraphNodeModel::GraphNodeModel()
{
	m_iClusterID = -1;
	m_aAllEdges = NULL;
	m_pModel = NULL;
	m_iIsolationCounter = 0;
	m_iEdgesTriedCounter = 0;
}

GraphNodeModel::GraphNodeModel(Model3D * model, GraphEdgeMap * allEdges)
{
	m_aAllEdges = allEdges;
	m_pModel = model;
	m_iIsolationCounter = 0;
	m_iEdgesTriedCounter = 0;
}

int GraphNodeModel::NodeID()
{
	return m_pModel->ModelID();
}

int GraphNodeModel::NEdges()
{
	//assert(m_pModel!=NULL);
	//return m_pModel->GetDB()->NumModels()-1;
	return (int)m_IncludedEdges.size();
}

void GraphNodeModel::IncludeEdge(GraphEdgeMap * edge)
{
	for (int i=0; i<(int)m_IncludedEdges.size(); i++)
	{
		if (m_IncludedEdges[i] == edge)
			return;
	}
	
	m_IncludedEdges.push_back(edge);
	edge->IncludeEdge();
}
void GraphNodeModel::ExcludeEdge(GraphEdgeMap * edge)
{
	for (int i=0; i<(int)m_IncludedEdges.size(); i++)
		if (m_IncludedEdges[i] == edge)
		{
			m_IncludedEdges.erase(m_IncludedEdges.begin() + i);
			edge->ExcludeEdge();
			break;
		}
}

GraphEdgeMap * GraphNodeModel::AdjacentEdge(int id)
{
	assert(id<(int)m_IncludedEdges.size() && id>=0);
	return m_IncludedEdges[id];
//	assert(m_pModel!=NULL);
//	assert(m_aAllEdges!=NULL);
//	int nModels = m_pModel->GetDB()->NumModels();
//	assert(id < nModels-1);
//	assert(m_pModel!=NULL);
//	int myID = m_pModel->ModelID();
//	if (id > myID)
//		id--;
//	
//	if (id==myID)
//		return NULL;
//	else
//		return &(m_aAllEdges[CollectionDiffuseDynamicGraph::EdgeID(myID, id, nModels)]);
}

Model3D * GraphNodeModel::NodeModel()
{
	return m_pModel;
}

void GraphNodeModel::WriteNode(FILE * binaryFile)
{
	int nodeID = NodeID();
	fwrite(&nodeID, sizeof(int), 1, binaryFile);
	fwrite(&m_iClusterID, sizeof(int), 1, binaryFile);
	fwrite(&m_iInClusterOffset, sizeof(int), 1, binaryFile);
}

void GraphNodeModel::ReadNode(FILE * binaryFile)
{
	int nodeID;
	fread(&nodeID, sizeof(int), 1, binaryFile);
	fread(&m_iClusterID, sizeof(int), 1, binaryFile);
	fread(&m_iInClusterOffset, sizeof(int), 1, binaryFile);
	if (NodeID()!=nodeID)
		std::cout<<"[ERROR] Node ids are not same "<<NodeID()<<"!="<<nodeID<<std::endl;
	assert(NodeID()==nodeID);
}


//////////////////	GraphEdgeMap	///////////////
GraphEdgeMap::GraphEdgeMap()
{
	Init();
}

GraphEdgeMap::GraphEdgeMap(GraphNodeModel * node1, GraphNodeModel * node2,
						   CollectionDiffuseDynamicGraph* graph)
{
	Init();
	m_pNode1 = node1;
	m_pNode2 = node2;
	m_pGraph = graph;
}

void GraphEdgeMap::Init()
{	
	m_aStoreCurrentDiffusedMap = NULL;
	m_bIncluded = false;
	m_pNode1 = NULL;
	m_pNode2 = NULL;
	m_pGraph = NULL;

	m_iDynamicAlignmentID = -1;
	m_CachedMaps = NULL;
	m_iMaxCacheSize = 3;
	m_iPreviousSelectedBest = -1;
	m_iSelectedBestID = -1;
	m_fConsistencyEdgeValue = -1;
	m_fIntegratedEdgeValue = -1;
	m_fShapeDescriptorEdgeValue = -1;
	m_iExploredCollections = -1;
	m_iNumConsistencyOptimizations = 0;
}

int GraphEdgeMap::EdgeID()
{
	assert(m_pNode1!=NULL && m_pNode2!=NULL);
	
	int nModels = m_pNode1->NodeModel()->GetDB()->NumModels();
	int node1 = m_pNode1->NodeID();
	int node2 = m_pNode2->NodeID();
	return CollectionDiffuseDynamicGraph::EdgeID(node1, node2, nModels); 
}

GraphNodeModel * GraphEdgeMap::Node(int id)
{
	if (id==0)
		return m_pNode1;
	else if (id==1)
		return m_pNode2;
	assert(false);
	return NULL;
}

void GraphEdgeMap::IncludeEdge()
{
	if (!m_bIncluded)
	{
		m_bIncluded = true;
		Node(0)->IncludeEdge(this);
		Node(1)->IncludeEdge(this);
	}
}

void GraphEdgeMap::ExcludeEdge()
{
	if (m_bIncluded)
	{
		m_bIncluded = false;
		Node(0)->ExcludeEdge(this);
		Node(1)->ExcludeEdge(this);
		ClearCache();
		ClearDiffLinkCache();
	}
}

bool GraphEdgeMap::GraphEdgeMap::IsEdgeIncluded()
{
	return m_bIncluded;
}

int GraphEdgeMap::NumPotentialCorrs()
{
	return (int)m_PotentialCorrCollections.size();
}

double GraphEdgeMap::EdgeValue(EdgeValueType val)
{
	if (val==EV_CONSISTENCY || (val==EV_AUTO && m_fConsistencyEdgeValue>=0))
		return m_fConsistencyEdgeValue;
	else if (val==EV_BEST_MAP || (val==EV_AUTO && m_fIntegratedEdgeValue>=0))
		return m_fIntegratedEdgeValue;
	else if (val==EV_SHAPE_DESCRIPTOR || (val==EV_AUTO && m_fConsistencyEdgeValue>=0))
		return m_fShapeDescriptorEdgeValue;

	assert(false);
	return 0;
}

int GraphEdgeMap::ExploreDeeper(const VKStringList & collName, 
								const VKStringList & corrValName)
{
	if ((m_iExploredCollections+1) >= collName.count())
		return -1;
	if (m_iExploredCollections>=-1)
		m_iExploredCollections++;
	
	return m_iExploredCollections;
}

void GraphEdgeMap::AddPotentialLink(const VKString & corrName, 
									const VKString & corrValue,
									int corrID)
{
	m_PotentialCorrCollections.push_back(corrID);
	m_sCorrCollName.push_back(corrName);
	m_sCorrCollValue.push_back(corrValue);
}

int GraphEdgeMap::CurrLinkID()
{
	if (IsEdgeIncluded())
		return m_iSelectedBestID;
	else
		return -1;
}

int GraphEdgeMap::NumLinks()
{
	return (int)m_PotentialCorrCollections.size();
}

VKString GraphEdgeMap::GetLinkCollName(int linkID)
{
	return m_sCorrCollName[linkID];
}

int GraphEdgeMap::GetLinkCorrID(int linkID)
{
	return m_PotentialCorrCollections[linkID];
}

VKString GraphEdgeMap::GetLinkCorrValName(int linkID)

{
	return m_sCorrCollValue[linkID];
}


void GraphEdgeMap::SetDescriptorValue(double value)
{
	m_fShapeDescriptorEdgeValue = value;
}

int GraphEdgeMap::NumConsistencyOptimizations()
{
	return m_iNumConsistencyOptimizations;
}

void GraphEdgeMap::AddDynamicLink(const VKStringList & collName,
							const VKStringList & corrValName,
							int mapID)
{
	assert(false);
	// TODO: optimizeforconsistency: update dynamic links (do not cache dynamic)
	// 
}

bool GraphEdgeMap::OptimizeForConsistency(double thresholdToRemove,
										  std::vector<double>*perLinkScores,
										  std::vector<double>*perLinkCorrVals)
{
	m_iNumConsistencyOptimizations++;
	
	std::map<int, double> * diffLink = GetDiffusedLink();
	assert (diffLink!=NULL);
	
	double currScore = 0;
	int currBestLink = m_iSelectedBestID;
	if (m_iSelectedBestID >= 0)
		currScore = CalculateConsistencyScoreForLink(m_iSelectedBestID, diffLink);
	
	if (perLinkCorrVals!=NULL)
		(*perLinkCorrVals) = m_fIntegratedCorrValue;
	
	if (perLinkScores!=NULL)
	{
		perLinkScores->clear();
		for (int i=0; i<NumLinks(); i++)
			(*perLinkScores).push_back(-1);
		if (m_iSelectedBestID>=0)
			(*perLinkScores)[m_iSelectedBestID] = currScore;
	}
	
	VKString name1 = Node(0)->NodeModel()->ModelName().c_str();
	VKString name2 = Node(1)->NodeModel()->ModelName().c_str();
	
//	if (name1=="chair0009" && name2=="chair0026")
//	{
//		std::cout<<"OPTIMIZING: "<<name1.c_str()<<" -> "<<name2.c_str()<<std::endl;
//		std::cout<<"\tStartValue["<<currBestLink<<"] = "<<currScore<<std::endl;
//	}
	
	for (int i=0; i<NumLinks(); i++)
	{
		if (i==m_iSelectedBestID)
			continue;
				
		UpgradeOrder(currBestLink);
		UpgradeOrder(m_iSelectedBestID);	// keep current to be in cache
		double anotherScore = CalculateConsistencyScoreForLink(i, diffLink, currScore);
//		if (name1=="chair0009" && name2=="chair0026")
//		{
//			std::cout<<"\tValue["<<i<<"] = "<<anotherScore<<std::endl;
//		}
		if (perLinkScores!=NULL)
			(*perLinkScores)[i] = anotherScore;
		
		if (anotherScore > currScore)
		{
			currScore = anotherScore;
			currBestLink = i;
		}
	}
//	if (name1=="chair0009" && name2=="chair0026")
//	{
//		std::cout<<"CHOSE: "<<currBestLink<<" w/ "<<currScore<<std::endl;
//	}
	bool diffVal = false;
	if (currScore < thresholdToRemove)
	{
		diffVal = IsEdgeIncluded();
		m_iSelectedBestID = -1;
		ExcludeEdge();
	}
	else
	{
		diffVal = (m_iSelectedBestID != currBestLink);
		m_iSelectedBestID = currBestLink;
		IncludeEdge();
	}
	m_fConsistencyEdgeValue = currScore;
	
	if (m_pGraph->m_bConservativeOnMemory)
	{
		ClearDiffLinkCache();
		ClearCache();
	}
	return diffVal;
}

double GraphEdgeMap::CurrentConsistencyScore()
{
	std::map<int, double> * diffLink = GetDiffusedLink();
	assert(diffLink!=NULL);
	m_fConsistencyEdgeValue = CalculateConsistencyScoreForLink(m_iSelectedBestID, diffLink);
	return m_fConsistencyEdgeValue;
}

double GraphEdgeMap::CurrentCorrValue()
{
	int linkID = m_iSelectedBestID;
	if (linkID < 0)
		return 0;
	assert(linkID < (int)m_fIntegratedCorrValue.size());
	return m_fIntegratedCorrValue[linkID];
}

double GraphEdgeMap::CalculateConsistencyScoreForLink(int linkID, 
													  std::map<int, double> * diffMap,
													  double )
{
	
	double consistencyScore = 0;
	ModelDatabase * db = m_pNode1->NodeModel()->GetDB();
	PointSet * inPnts = m_pNode1->NodeModel()->PointCloud();
	int nPoints = inPnts->NumPoints();
	FillCacheForLink(linkID);
	std::map<int, double> * link = GetCache(linkID);
	for (int i=0; i<nPoints; i++)
	{
		if (link[i].size() < diffMap[i].size())
			consistencyScore += InnerProduct(link[i], diffMap[i]);
		else
			consistencyScore += InnerProduct(diffMap[i], link[i]);
	}
	
	double globalScale = m_fIntegratedCorrValue[linkID];
	
	return globalScale * consistencyScore / nPoints;
}

double GraphEdgeMap::InnerProduct(std::map<int, double> & f1, 
								  std::map<int, double> & f2)
{
	double innerProduct = 0;
	for (std::map<int, double>::iterator iter1 = f1.begin(); 
		 iter1 != f1.end(); iter1++)
	{
		std::map<int, double>::iterator iter2 = f2.find(iter1->first);
		if (iter2 != f2.end())
			innerProduct += iter1->second * iter2->second;
	}
	return innerProduct;
}

std::map<int, double> * GraphEdgeMap::GetDiffusedLink()
{
	int clusterID = -1;
	CollectionAggregatePairwise * pairAgg = GetMyPairAgg(&clusterID);
	if (pairAgg==NULL)
		return NULL;

	assert(m_pNode1!=NULL);
	assert(m_pNode1->NodeModel()!=NULL);
	ModelDatabase * db = m_pNode1->NodeModel()->GetDB();
	assert(db!=NULL);
	PointSet * inPnts = m_pNode1->NodeModel()->PointCloud();
	int nPoints = inPnts->NumPoints();

	if (m_aStoreCurrentDiffusedMap==NULL)
		m_aStoreCurrentDiffusedMap = new std::map<int, double>[nPoints];
	
	assert(m_pGraph!=NULL);
	assert(m_pGraph->m_InClusterToGlobalID!=NULL);
	assert(m_pGraph->m_InClusterToGlobalIDSize!=NULL);
	assert(clusterID < (int)m_pGraph->m_mCCsToNodes.size());
	
	pairAgg->FillValsForPair(m_pNode1->NodeModel()->ModelID(), 
							 m_pNode2->NodeModel()->ModelID(),
							 m_aStoreCurrentDiffusedMap, nPoints);	
	
//	for (int i=0; i<nPoints; i++)
//	{
//		double norm = 0;
//		
//		std::vector<ModelPoint> corrs;
//		std::vector<double> vals;
//		
//		pairAgg->GetCorrs(inPnts->GetPoint(i), &corrs, &vals);
//		m_aStoreCurrentDiffusedMap[i].clear();
//		
//		for (int j=0; j<(int)corrs.size(); j++)
//		{
//			
//			if (corrs[j].GetModel() != m_pNode2->NodeModel())
//				continue;
//			
//			m_aStoreCurrentDiffusedMap[i][corrs[j].PntCloudID()] = vals[j];
//			norm += vals[j];			
//		}
//	
//		if (norm != 0)
//		{
//			std::map<int, double>::iterator iter;
//			for (iter = m_aStoreCurrentDiffusedMap[i].begin(); 
//				 iter != m_aStoreCurrentDiffusedMap[i].end(); iter++)
//				iter->second /= norm;
//		}
//		else
//			assert(m_aStoreCurrentDiffusedMap[i].size()==0);
//	}
	
	return m_aStoreCurrentDiffusedMap;
}

void GraphEdgeMap::OptimizeForLinkValue()
{
	IncludeEdge();
	if (m_fSortedIntegratedCorrValue.size() == 0)
	{
		ModelDatabase * db = Node(0)->NodeModel()->GetDB();
		for (int i=0; i<(int)m_PotentialCorrCollections.size(); i++)
		{
			int corrID = m_PotentialCorrCollections[i];
			VKString corrName = m_sCorrCollName[i];
			VKString corrValue = m_sCorrCollValue[i];
			
			CollectionSinglePair * singlePair = db->GetCorrSinglePair(corrName, corrID);
			CorrValue * corrVal = singlePair->GetCorrValueByName(corrValue);
			assert(singlePair->ModelFrom()==Node(0)->NodeModel());
			assert(singlePair->ModelTo()==Node(1)->NodeModel());
			double globInt = corrVal->IntegratedValue("auto", "auto");
			m_fSortedIntegratedCorrValue.push_back(Sortable(-globInt, NULL, i));
			m_fIntegratedCorrValue.push_back(globInt);
		}
		std::sort(m_fSortedIntegratedCorrValue.begin(), 
				  m_fSortedIntegratedCorrValue.end());
	}
	assert(m_fSortedIntegratedCorrValue.size()>0);
	m_iSelectedBestID = m_fSortedIntegratedCorrValue[0].id;
	//m_iSelectedBestID = 0;
}

void GraphEdgeMap::UpdateDiffusionMap()
{
	assert(false);	// note: doesn't handle multi-class
//	// remove previous
//	if (m_iPreviousSelectedBest >= 0)
//		AddValuesDueToLink(m_iPreviousSelectedBest, diffMap, -1.);
//	
//	m_iPreviousSelectedBest = -1;	// nothing added
//	
//	// add new
//	if (IsEdgeIncluded())
//	{
//		assert(m_iSelectedBestID>=0);
//		AddValuesDueToLink(m_iSelectedBestID, diffMap, 1.);;
//		// remember what was added
//		m_iPreviousSelectedBest = m_iSelectedBestID;
//	}
}

void GraphEdgeMap::AddValuesDueToCurrentLink()
{
	if (IsEdgeIncluded() && m_iSelectedBestID>=0)
		AddValuesDueToLink(m_iSelectedBestID, 1.);
}

CollectionAggregatePairwise * GraphEdgeMap::GetMyPairAgg(int * saveClusterID)
{
	int c1ID = Node(0)->m_iClusterID;
	int c2ID = Node(1)->m_iClusterID;
	assert(c1ID==c2ID && c1ID>=0);

	if (saveClusterID!=NULL)
		*saveClusterID = c1ID;
		
	if (c1ID < 0)
		return NULL;
	
	assert(c1ID < (int)m_pGraph->m_aPairAgg.size());	
	
	return m_pGraph->m_aPairAgg[c1ID];
}

void GraphEdgeMap::AddValuesDueToLink(int linkID, double weight)
{
	CollectionAggregatePairwise * aggMap = GetMyPairAgg();
	if (aggMap==NULL)
		return;
	
	FillCacheForLink(linkID);
	ModelDatabase * db = Node(0)->NodeModel()->GetDB();
	std::map<int, double> * currCache = GetCache(linkID);
	PointSet * pnts1 = Node(0)->NodeModel()->PointCloud();
	PointSet * pnts2 = Node(1)->NodeModel()->PointCloud();
	assert(linkID < (int)m_fIntegratedCorrValue.size());
	double globalScale = weight * m_fIntegratedCorrValue[linkID];
	for (int i=0; i<pnts1->NumPoints(); i++)
	{	
		int p1 = db->GlobalID(pnts1->GetPoint(i));
		for (std::map<int, double>::iterator iter = currCache[i].begin();
			 iter != currCache[i].end(); iter++)
		{
			int p2 = db->GlobalID(pnts2->GetPoint(iter->first));
			aggMap->AddEntry(p1, p2, globalScale * iter->second);
		}
	}
}

void GraphEdgeMap::FillCacheForLink(int linkID)
{
	if (m_CachedMaps == NULL)
		InitCache();
	std::map<int, int >::iterator iter = m_CorrIDToCachedMapID.find(linkID);
	if (iter != m_CorrIDToCachedMapID.end())	// cache filled
		return;
	
	LoadNewCachedCorr(linkID);
	std::map<int, double> * currCache = GetCache(linkID);
	
	ModelDatabase * db = Node(0)->NodeModel()->GetDB();
	if (linkID>=(int)m_PotentialCorrCollections.size() || linkID<0)
	{
		std::cout<<"[ERROR] invalid linkID: ";
		std::cout<<linkID<<" / "<<m_PotentialCorrCollections.size()<<std::endl;;
	}
	assert(linkID < (int)m_PotentialCorrCollections.size());
	int corrID = m_PotentialCorrCollections[linkID];
	VKString corrName = m_sCorrCollName[linkID];
	VKString corrValue = m_sCorrCollValue[linkID];

	CollectionSinglePair * singlePair = db->GetCorrSinglePair(corrName, corrID);
	assert(singlePair!=NULL);
	CorrValue * corrVal = singlePair->GetCorrValueByName(corrValue);
	assert(corrVal!=NULL);
	assert(singlePair->ModelFrom()==Node(0)->NodeModel());
	assert(singlePair->ModelTo()==Node(1)->NodeModel());
	
	PointSet * pnts1 = singlePair->ModelFrom()->PointCloud();
	
	for (int i=0; i<pnts1->NumPoints(); i++)
	{
		ModelPoint p1 = pnts1->GetPoint(i);
		std::vector<ModelPoint> corrs;
		singlePair->GetCorrs(p1, &corrs);
		for (int j=0; j<(int)corrs.size(); j++)
		{
			ModelPoint p2 = corrs[j];
			double cValue = corrVal->Value(p1, p2);
			int p2ID = p2.PntCloudID();
			assert(p2ID>=0);
			currCache[i][p2ID] = cValue;
		}
	}
}

void GraphEdgeMap::InitCache()
{
	m_CachedMaps = new std::map<int, double> *[m_iMaxCacheSize];
	int nPoints = Node(0)->NodeModel()->PointCloud()->NumPoints();
	for (int i=0; i<m_iMaxCacheSize; i++)
	{
		m_CachedMaps[i] = new std::map<int, double>[nPoints];
		m_iAvailableCacheIDs.push_back(i);
	}
}

void GraphEdgeMap::ClearDiffLinkCache()
{
	if (m_aStoreCurrentDiffusedMap!=NULL)
	{
		delete [] m_aStoreCurrentDiffusedMap;
		m_aStoreCurrentDiffusedMap = NULL;
	}	
}
void GraphEdgeMap::ClearCache()
{	
	if (m_CachedMaps!=NULL)
	{
		for (int i=0; i<m_iMaxCacheSize; i++)
			delete [] m_CachedMaps[i];
		delete [] m_CachedMaps;
		m_CachedMaps = NULL;
		m_iAvailableCacheIDs.clear();
		m_CorrIDToCachedMapID.clear();
		m_CachedMapsOrder.clear();
	}
}

void GraphEdgeMap::ClearCached(int linkID)
{
	std::map<int, int >::iterator iter = m_CorrIDToCachedMapID.find(linkID);
	assert (iter != m_CorrIDToCachedMapID.end());
	int nPoints = Node(0)->NodeModel()->PointCloud()->NumPoints();
	for (int i=0; i<nPoints; i++)
		m_CachedMaps[iter->second][i].clear();
	
	m_iAvailableCacheIDs.push_back(iter->second);
	m_CorrIDToCachedMapID.erase(iter);
}

void GraphEdgeMap::UpgradeOrder(int linkID)
{
	if (m_CachedMapsOrder.size()==0)
		return;
	if (m_CachedMapsOrder[0] == linkID)
	{
		m_CachedMapsOrder.erase(m_CachedMapsOrder.begin());
		m_CachedMapsOrder.push_back(linkID);
	}
}

void GraphEdgeMap::LoadNewCachedCorr(int linkID)
{
	while ((int)m_CachedMapsOrder.size() >= m_iMaxCacheSize)
	{
		int clearCorr = m_CachedMapsOrder[0];
		ClearCached(clearCorr);
		m_CachedMapsOrder.erase(m_CachedMapsOrder.begin());
	}
	
	assert(m_iAvailableCacheIDs.size() > 0);
	int availCacheID = m_iAvailableCacheIDs[0];
	m_iAvailableCacheIDs.erase(m_iAvailableCacheIDs.begin());
	
	m_CachedMapsOrder.push_back(linkID);
	m_CorrIDToCachedMapID[linkID] = availCacheID;
}

std::map<int, double> * GraphEdgeMap::GetCache(int linkID)
{
	std::map<int, int >::iterator iter = m_CorrIDToCachedMapID.find(linkID);
	assert(iter!=m_CorrIDToCachedMapID.end());
	return m_CachedMaps[iter->second];
}

void GraphEdgeMap::WriteEdge(FILE * binaryFile)
{
	if (m_pNode1 == NULL || m_pNode2==NULL)
	{
		m_iExploredCollections = -100;
		fwrite(&m_iExploredCollections, sizeof(int), 1, binaryFile);
		return;	// invalid
	}
	fwrite(&m_iExploredCollections, sizeof(int), 1, binaryFile);
	int ids[2];
	assert(m_pNode1!=NULL && m_pNode2!=NULL);
	ids[0] = m_pNode1->NodeID();
	ids[1] = m_pNode2->NodeID();

	fwrite(ids, sizeof(int), 2, binaryFile);
	fwrite(&m_bIncluded, sizeof(bool), 1, binaryFile);
	
	fwrite(&m_iPreviousSelectedBest, sizeof(int), 1, binaryFile);
	fwrite(&m_iSelectedBestID, sizeof(int), 1, binaryFile);

	fwrite(&m_fConsistencyEdgeValue, sizeof(double), 1, binaryFile);
	fwrite(&m_fIntegratedEdgeValue, sizeof(double), 1, binaryFile);
	fwrite(&m_fShapeDescriptorEdgeValue, sizeof(double), 1, binaryFile);

	int nCorrIDs = (int)m_PotentialCorrCollections.size();
	assert(m_sCorrCollName.count()==nCorrIDs);
	assert(m_sCorrCollValue.count()==nCorrIDs);
	fwrite(&nCorrIDs, sizeof(int), 1, binaryFile);
	if (nCorrIDs>0)
	{
		int * allCorrIDs = new int[nCorrIDs];
		for (int i=0; i<nCorrIDs; i++)
			allCorrIDs[i] = m_PotentialCorrCollections[i];
		fwrite(allCorrIDs, sizeof(int), nCorrIDs, binaryFile);
		int ok=fflush(binaryFile);
		assert(ok==0);
		delete [] allCorrIDs;

		VKString collNameStr = m_sCorrCollName.join(" ");
		int nChars1 = strlen(collNameStr.c_str());
		assert(nChars1>0);
		fwrite(&nChars1, sizeof(int), 1, binaryFile);
		fwrite(collNameStr.c_str(), sizeof(char), nChars1+1, binaryFile);
		
		VKString corrValNameStr = m_sCorrCollValue.join(" ");
		int nChars2 = strlen(corrValNameStr.c_str());
		assert(nChars2>0);
		fwrite(&nChars2, sizeof(int), 1, binaryFile);
		fwrite(corrValNameStr.c_str(), sizeof(char), nChars2+1, binaryFile);
	}
	
	int nIntegratedVals = (int)m_fIntegratedCorrValue.size();
	fwrite(&nIntegratedVals, sizeof(int), 1, binaryFile);
	if (nIntegratedVals>0)
	{
		double * allIntegratedVals = new double[nIntegratedVals];
		for (int i=0; i<nIntegratedVals; i++)
			allIntegratedVals[i] = m_fIntegratedCorrValue[i];
		fwrite(allIntegratedVals, sizeof(double), nIntegratedVals, binaryFile);
		int ok=fflush(binaryFile);
		assert(ok==0);
		delete [] allIntegratedVals;
	}

	// NOT SAVED: (ok?)
	//	std::vector<Sortable> m_fPreviousConsistencyValues;
	//	std::vector<Sortable> m_fSortedIntegratedCorrValue;	
}

void GraphEdgeMap::ReadEdge(FILE * binaryFile)
{
	fread(&m_iExploredCollections, sizeof(int), 1, binaryFile);	
	if (m_iExploredCollections == -100)
		return;	// invalid
	int ids[2];
	
	fread(ids, sizeof(int), 2, binaryFile);
	assert(m_pNode1!=NULL && m_pNode2!=NULL);
	assert(ids[0] == m_pNode1->NodeID());
	assert(ids[1] == m_pNode2->NodeID());
	
	fread(&m_bIncluded, sizeof(bool), 1, binaryFile);
	
	fread(&m_iPreviousSelectedBest, sizeof(int), 1, binaryFile);
	fread(&m_iSelectedBestID, sizeof(int), 1, binaryFile);
	
	fread(&m_fConsistencyEdgeValue, sizeof(double), 1, binaryFile);
	fread(&m_fIntegratedEdgeValue, sizeof(double), 1, binaryFile);
	fread(&m_fShapeDescriptorEdgeValue, sizeof(double), 1, binaryFile);
	
	int nCorrIDs;
	fread(&nCorrIDs, sizeof(int), 1, binaryFile);
//	std::cout<<"EDGE: e="<<m_iExploredCollections<<" nD=["<<ids[0]<<" "<<ids[1]<<"] inc="<<m_bIncluded;
//	std::cout<<" ps="<<m_iPreviousSelectedBest<<" s="<<m_iSelectedBestID<<" c= "<<m_fConsistencyEdgeValue;
//	std::cout<<" i="<<m_fIntegratedEdgeValue<<" sd="<<m_fShapeDescriptorEdgeValue<<std::endl;
//	std::cout<<" norrs="<<nCorrIDs<<std::endl;
	if (nCorrIDs>0)
	{
		int * allCorrIDs = new int[nCorrIDs];
		fread(allCorrIDs, sizeof(int), nCorrIDs, binaryFile);
		for (int i=0; i<nCorrIDs; i++)
			m_PotentialCorrCollections.push_back(allCorrIDs[i]);
		delete [] allCorrIDs;
		
		int nChars1;
		fread(&nChars1, sizeof(int), 1, binaryFile);
		char * collNameStr = new char[nChars1+1];
		fread(collNameStr, sizeof(char), nChars1+1, binaryFile);
		m_sCorrCollName = VKString(collNameStr).split(" ");
 		delete [] collNameStr;

		int nChars2;
		fread(&nChars2, sizeof(int), 1, binaryFile);
//		std::cout<<"nCahrs: "<<nChars2<<std::endl;
		char * corrValNameStr = new char[nChars2+1];
		fread(corrValNameStr, sizeof(char), nChars2+1, binaryFile);
		m_sCorrCollValue = VKString(corrValNameStr).split(" ");
		if (m_sCorrCollValue.count()!=nCorrIDs
			|| m_sCorrCollName.count()!=nCorrIDs)
		{
			std::cout<<"[ERROR] Parsing error for Dynamig Graph Edges."<<std::endl;
			std::cout<<m_sCorrCollValue.count()<<" != "<<nCorrIDs<<std::endl;
			for (int i=0; i<m_sCorrCollValue.count(); i++)
				std::cout<<"\t"<<i<<": "<<m_sCorrCollValue[i].c_str()<<std::endl;
			std::cout<<m_sCorrCollName.count()<<" != "<<nCorrIDs<<std::endl;
			for (int i=0; i<m_sCorrCollName.count(); i++)
				std::cout<<"\t"<<i<<": "<<m_sCorrCollName[i].c_str()<<std::endl;
		}
		assert(m_sCorrCollValue.count()==nCorrIDs);
		delete []corrValNameStr;
	}
	
	
	int nIntegratedVals;// = (int)m_fIntegratedCorrValue.size();
	fread(&nIntegratedVals, sizeof(int), 1, binaryFile);
	if (nIntegratedVals>0)
	{
		double * allIntegratedVals = new double[nIntegratedVals];
		fread(allIntegratedVals, sizeof(double), nIntegratedVals, binaryFile);
		for (int i=0; i<nIntegratedVals; i++)
			m_fIntegratedCorrValue.push_back(allIntegratedVals[i]);
		delete [] allIntegratedVals;
	}
	
	if (m_bIncluded)
	{
		ExcludeEdge();
		IncludeEdge();
	}
}

/////////// DynamicGraphLogParams ///////////
DynamicGraphLogParams::IterationStep::IterationStep()
{
	m_fConsistencyScore = 0;
	m_iStepType = -1;
	m_pEigenvalues = NULL;
	m_bDrewStep = false;
}

DynamicGraphLogParams::DynamicGraphLogParams(CollectionDiffuseDynamicGraph* graph)
{
	m_pGraph = graph;
//	m_DirPrefix = m_pGraph->m_pDB->GetCollectionDir(m_pGraph->m_sCorrCollName);
	m_DirPrefix = m_pGraph->m_pDB->GetMainDir()+"/Images/"+m_pGraph->m_sCollectionName+"/";
	VKString subdir("");
	
	int id = 0;
	while (m_pGraph->m_pDB->DirExists(m_DirPrefix + subdir))
		subdir = VKString::number(id++) + "/";
	m_DirPrefix = m_DirPrefix + subdir;
}

void DynamicGraphLogParams::WriteHTMLLog(DatabaseVisWindow * window, 
										 bool renderImages,
										 bool redraw)
{	
	if (m_Iterations.size() == 0)
		return;
	
	bool drewAll = true;
	for (int i=0; i<(int)m_Iterations.size() && drewAll; i++)
		drewAll = m_Iterations[i].m_bDrewStep;
	if (drewAll && !redraw)
		return;
		
	window->DisableApp();
	
	VKString htmlFile;
	
	m_pGraph->m_pDB->CreateDirsOnPath(m_DirPrefix);
	htmlFile = m_DirPrefix + "index.html";
	
	std::ofstream htmlStream(htmlFile.c_str()); 
	assert(htmlStream.is_open());
	WriteHTMLHeaderBegin(htmlStream, "Dynamic Graph Log");
						 
	for (int i=0; i<(int)m_Iterations.size(); i++)
	{
		std::map<int, std::vector<int> > perNodeEdges;
		
		IterationStep & step = m_Iterations[i];
		
		step.m_bDrewStep = true;
		htmlStream<<"<h2>ITER "<<i<<" ";
		if (i==0 || i==1)
			htmlStream<<"Graph Init ";
		if (step.m_iStepType==0)
			htmlStream<<" : Diffusion";
		else if (step.m_iStepType==1) 
			htmlStream<<" : Consistency";
		else
			htmlStream<<" : Unknown";
		htmlStream<<"</h2>\n";
		
		// write eigenvalues
		if (step.m_pEigenvalues!=NULL)
		{
			VKString fileOnly = VKString("Iter") + VKString::number(i) + "_evals.txt";
			step.m_pEigenvalues->WriteVectorASCII((m_DirPrefix + "/" + fileOnly).c_str());
			htmlStream<<"<a href=\""<<fileOnly.c_str()<<"\"> Eigenvalues</a>:<br>";
			htmlStream<<"&nbsp;&nbsp;&nbsp;&nbsp;Sum="<<step.m_pEigenvalues->Norm(1);
			int nVals = step.m_pEigenvalues->Rows();
			assert(nVals >= 3);
			htmlStream<<" E=["<<(*(step.m_pEigenvalues))(0);
			htmlStream<<" "<<(*(step.m_pEigenvalues))(1);
			htmlStream<<" "<<(*(step.m_pEigenvalues))(2);
			htmlStream<<" ... "<<(*(step.m_pEigenvalues))(nVals-1)<<"]<br>\n";
		}
		
		// write edge consistency
		if (step.m_EdgeIDs.size() > 0)
		{
			VKString fileOnly = VKString("Iter") + VKString::number(i) + "_Edges.html";
			VKString srtEdgeFile = m_DirPrefix + "/" + fileOnly;
			htmlStream<<"<a href=\""<<fileOnly.c_str()<<"\"> Edge Consistency</a><br>";
			std::ofstream htmlEdgeStream(srtEdgeFile.c_str());
			VKString title = VKString("Iter ") + VKString::number(i) + " edges";
			assert(htmlEdgeStream.is_open());
			WriteHTMLHeaderBegin(htmlEdgeStream, title);
			htmlEdgeStream<<"<table border=1>\n";
			htmlEdgeStream<<"\t<tr>\n";
			WriteHTMLTableRowItem(htmlEdgeStream, "<b>Edge</b>");
			WriteHTMLTableRowItem(htmlEdgeStream, "<b>Img</b>");
			WriteHTMLTableRowItem(htmlEdgeStream, "<b>Consistency</b>");
			WriteHTMLTableRowItem(htmlEdgeStream, "<b>Correspondence</b>");
			htmlEdgeStream<<"\t</tr>\n";
			assert(step.m_EdgeIDs.size()==step.m_aPerEdgeLinkIDs.size()
				   && step.m_EdgeIDs.size()==step.m_aPerEdgeConsistency.size()
				   && step.m_EdgeIDs.size()==step.m_aPerEdgeCorrValue.size());
			for (int e = 0; e < (int)step.m_EdgeIDs.size(); e++)
			{
				int edgeID = step.m_EdgeIDs[e];
				int linkID = step.m_aPerEdgeLinkIDs[e];
				double consistency = step.m_aPerEdgeConsistency[e];
				double corrValue = step.m_aPerEdgeCorrValue[e];
				GraphEdgeMap & edge = m_pGraph->m_aAllEdges[edgeID];
				
				htmlEdgeStream<<"\t<tr>\n";
				VKString m1 = edge.Node(0)->NodeModel()->ModelName();
				VKString m2 = edge.Node(1)->NodeModel()->ModelName();
				WriteHTMLTableRowItem(htmlEdgeStream, m1+" -> "+m2);
				
				perNodeEdges[edge.Node(0)->NodeID()].push_back(e);
				perNodeEdges[edge.Node(1)->NodeID()].push_back(e);
				
				if (renderImages)
				{
					if (linkID >= 0)
					{
						VKString imgStr = GetLinkImageName(edgeID, linkID)+".png";
						imgStr = VKString("<img width=200 src=\"") + imgStr + "\">";
						WriteLinkImage(edgeID, linkID, window);
						WriteHTMLTableRowItem(htmlEdgeStream, imgStr);
					}
					else
						WriteHTMLTableRowItem(htmlEdgeStream, "EXCLUDED");
				}
				WriteHTMLTableRowItem(htmlEdgeStream, VKString::number(consistency));
				WriteHTMLTableRowItem(htmlEdgeStream, VKString::number(corrValue));
				htmlEdgeStream<<"\t</tr>\n";
			}
			
			WriteHTMLHeaderEnd(htmlEdgeStream);
			htmlEdgeStream.close();
		}
		
		// write modified edges
		if (step.m_ModifiedEdges.size() > 0)
		{
			VKString fileOnly = VKString("Iter") + VKString::number(i) + "_Modified.html";
			VKString modifFile = m_DirPrefix + "/" + fileOnly;
			htmlStream<<"<a href=\""<<fileOnly.c_str()<<"\"> Modified Edges</a><br>";
			std::ofstream htmlModifStream(modifFile.c_str());
			assert(htmlModifStream.is_open());
			VKString title = VKString("Iter ") + VKString::number(i) + " modified edges";
			WriteHTMLHeaderBegin(htmlModifStream, title);
			
			htmlModifStream<<"<table border=1>\n";
			htmlModifStream<<"\t<tr>\n";
			WriteHTMLTableRowItem(htmlModifStream, "<b>Edge</b>");
			WriteHTMLTableRowItem(htmlModifStream, "<b>Old Link</b>");
			WriteHTMLTableRowItem(htmlModifStream, "<b>New Link</b>");
			htmlModifStream<<"\t</tr>\n";
			assert(step.m_ModifiedEdges.size()==step.m_ModifiedOldLinks.size()
				   && step.m_ModifiedEdges.size()==step.m_ModifiedNewLinks.size()
				   && step.m_ModifiedEdges.size()==step.m_AllScores.size()
				   && step.m_ModifiedEdges.size()==step.m_AllCorrValues.size());
			
			for (int e=0; e<(int)step.m_ModifiedEdges.size(); e++)
			{
				assert( e < (int)step.m_AllScores.size());
				assert( e < (int)step.m_AllCorrValues.size());
				int edgeID = step.m_ModifiedEdges[e];
				int oldLink = step.m_ModifiedOldLinks[e];
				int newLink = step.m_ModifiedNewLinks[e];
				double oldScore = -1;
				double newScore = -1;
				if (step.m_AllScores[e].size()>0)
				{
					assert(oldLink < (int)step.m_AllScores[e].size());
					assert(newLink < (int)step.m_AllScores[e].size());
					if (oldLink>=0)
						oldScore = step.m_AllScores[e][oldLink];
					if (newLink>=0)
						newScore = step.m_AllScores[e][newLink];
				}
				
				double oldCorrValue = -1;
				double newCorrValue = -1;
				if (step.m_AllCorrValues[e].size()>0)
				{
					assert(oldLink < (int)step.m_AllCorrValues[e].size());
					assert(newLink < (int)step.m_AllCorrValues[e].size());
					if (oldLink>=0)
						oldCorrValue = step.m_AllCorrValues[e][oldLink];
					if (newLink>=0)
						newCorrValue = step.m_AllCorrValues[e][newLink];
				}				
				if (newLink==oldLink)
					continue;	// skip if same (e.g. excluded->excluded)
				
				GraphEdgeMap & edge = m_pGraph->m_aAllEdges[edgeID];
				
				htmlModifStream<<"\t<tr>\n";
				VKString m1 = edge.Node(0)->NodeModel()->ModelName();
				VKString m2 = edge.Node(1)->NodeModel()->ModelName();
				if (oldLink!=-1 && newLink!=-1)
					WriteHTMLTableRowItem(htmlModifStream, m1+" -> "+m2+"<br>MODIFIED");
				else
					WriteHTMLTableRowItem(htmlModifStream, m1+" -> "+m2);
				
				if (oldLink >= 0)
				{
					VKString imgStr="";
					if (renderImages)
					{
						imgStr = GetLinkImageName(edgeID, oldLink)+".png";
						imgStr = VKString("<img src=\"") + imgStr + "\"><br>";
						WriteLinkImage(edgeID, oldLink, window);
					}
					imgStr += VKString("Consistency=")+VKString::number(oldScore)+"<br>";
					imgStr += VKString("Similarity=")+VKString::number(oldCorrValue)+"<br>";
					imgStr += VKString("LinkID=")+VKString::number(oldLink);
					WriteHTMLTableRowItem(htmlModifStream, imgStr);
				}
				else
					WriteHTMLTableRowItem(htmlModifStream, "EXCLUDED");

				if (newLink >= 0)
				{
					VKString imgStr="";
					if (renderImages)
					{
						imgStr = GetLinkImageName(edgeID, newLink)+".png";
						imgStr = VKString("<img src=\"") + imgStr + "\"><br>";
						WriteLinkImage(edgeID, newLink, window);
					}
					imgStr += VKString("Consistency=")+VKString::number(newScore)+"<br>";
					imgStr += VKString("Similarity=")+VKString::number(newCorrValue)+"<br>";
					imgStr += VKString("LinkID=")+VKString::number(newLink);
					WriteHTMLTableRowItem(htmlModifStream, imgStr);
				}
				else
					WriteHTMLTableRowItem(htmlModifStream, "EXCLUDED");
				
				htmlModifStream<<"\t</tr>\n";
			}
			
			WriteHTMLHeaderEnd(htmlModifStream);
			htmlModifStream.close();
		}
			
		// write nodes
		if (step.m_NodeIDs.size()>0)
		{
			VKString fileOnly = VKString("Iter") + VKString::number(i) + "_Nodes.html";
			VKString nodeFile = m_DirPrefix + "/" + fileOnly;
			htmlStream<<"<a href=\""<<fileOnly.c_str()<<"\"> Node Distances</a><br>";
			std::ofstream htmlNodesStream(nodeFile.c_str());
			assert(htmlNodesStream.is_open());
			VKString title = VKString("Iter ") + VKString::number(i) + " node distances";
			WriteHTMLHeaderBegin(htmlNodesStream, title);
			
			htmlNodesStream<<"<table border=1>\n";
			htmlNodesStream<<"\t<tr>\n";
			WriteHTMLTableRowItem(htmlNodesStream, "<b>Node</b>");
			WriteHTMLTableRowItem(htmlNodesStream, "<b>Distances</b>");
			WriteHTMLTableRowItem(htmlNodesStream, "<b>Links</b>");
			htmlNodesStream<<"\t</tr>\n";
			
			for (int j=0; j<(int)step.m_NodeIDs.size(); j++)
			{
				htmlNodesStream<<"\t<tr>\n";
				int nodeID = step.m_NodeIDs[j];
				Model3D * model = m_pGraph->m_aAllNodes[nodeID].NodeModel();
				
				std::vector<double> nodePerPntErr = step.m_pPerPointDistances[j];
				double nodeErr = step.m_PerNodeDistances[j];
				std::sort(nodePerPntErr.begin(), nodePerPntErr.end());
				
				VKString nodeEntry = model->ModelName();
				double maxErr = nodePerPntErr[nodePerPntErr.size()-1];
				nodeEntry+=VKString("<br>MeanErr=") + VKString::number(nodeErr);
				nodeEntry+=VKString("<br>MinErr=") + VKString::number(nodePerPntErr[0]);
				nodeEntry+=VKString("<br>MaxErr=") + VKString::number(maxErr);
				WriteHTMLTableRowItem(htmlNodesStream, nodeEntry);
				
				VKString img = WriteNodeImage(step.m_pPerPointDistances[j], 
											  nodeID, i, window);
				img = VKString("<img src=") + img+" width=200>";
				WriteHTMLTableRowItem(htmlNodesStream, img);
				
				VKString linksStr("");
				std::map<int, std::vector<int> >::iterator links;
				links = perNodeEdges.find(nodeID);
				if (links != perNodeEdges.end())
				{
					for (int l=0; l<(int)links->second.size(); l++)
					{
						int e = links->second[l];
						assert(e < (int)step.m_EdgeIDs.size());
						
						int edgeID = step.m_EdgeIDs[e];
						int linkID = step.m_aPerEdgeLinkIDs[e];
						double score = step.m_aPerEdgeConsistency[e];
						
						VKString linkImgName = GetLinkImageName(edgeID, linkID) + ".png";
						linksStr += VKString("<a href=") + linkImgName + ">";
						linksStr += VKString("E_") + VKString::number(edgeID);
						linksStr += VKString("_L=") + VKString::number(linkID);
						linksStr += VKString("_S=") + VKString::number(score)+"</a><br>";
					}
				}
				WriteHTMLTableRowItem(htmlNodesStream, linksStr);
				
				htmlNodesStream<<"\t</tr>\n";
			}
			htmlNodesStream<<"</table>\n";
			WriteHTMLHeaderEnd(htmlNodesStream);
			htmlNodesStream.close();
		}

	}
	WriteHTMLHeaderEnd(htmlStream);
	htmlStream.close();
	window->EnableApp();
	WriteLog();
}

void DynamicGraphLogParams::WriteHTMLTableRowItem(std::ofstream & textStream, 
												  const VKString & text)
{
	textStream<<"\t\t<td>\n"<<"\t\t\t"<<text.c_str()<<"\n"<<"\t\t</td>\n";
}

void DynamicGraphLogParams::WriteHTMLHeaderBegin(std::ofstream & textStream,
												 const VKString & title)
{
	textStream<<"<html>\n<head>\n<title>"<<title.c_str();
	textStream<<"</title>\n</head>\n<body>\n";
}

void DynamicGraphLogParams::WriteHTMLHeaderEnd(std::ofstream & textStream)
{
	textStream<<"</body>\n</html>\n";
	textStream.close();
}

void DynamicGraphLogParams::WriteLinkImage(int edgeID, int linkID,
										   DatabaseVisWindow * window)
{
	std::map<int, std::map< int, bool> >::iterator iter = m_WrittenLinks.find(edgeID);
	if (iter != m_WrittenLinks.end())
	{
		if (iter->second.find(linkID)!= iter->second.end())
			return;	// image exists
	}
	m_WrittenLinks[edgeID][linkID] = true;

	VKString filename = m_DirPrefix + GetLinkImageName(edgeID, linkID);	
	m_pGraph->m_pDB->CreateDirsOnPathToFile(filename);
	
	if (m_pGraph->m_pDB->FileExists(filename+".png"))
		return;

	GraphEdgeMap & edge = m_pGraph->m_aAllEdges[edgeID];
	int corrID = edge.GetLinkCorrID(linkID);
	VKString collName = edge.GetLinkCollName(linkID);
	VKString corrValueName = edge.GetLinkCorrValName(linkID);
	
	// render given model
	CorrCollection * coll = m_pGraph->m_pDB->GetCorrCollection(collName);

	CollectionPairwise * pairwiseColl =  coll->SafeCastToPairwise();
	VKString collType=coll->GetCollectionType();
	if (pairwiseColl!=NULL && pairwiseColl->NumModelCorrs()>0)
		collType=pairwiseColl->GetModelCorr(0)->GetCollectionType();
	
	if (coll!=NULL && collType=="SinglePairRigidAlign")
	{
		window->SetCorrFlag(DatabaseVisWindow::RENDER_CORRS_SPECIAL_1, true);
		window->SetCorrFlag(DatabaseVisWindow::RENDER_CORRS_SET, true);
		window->SetDesignRender(false);
		ModelDatabaseRenderingParams * dbRend = m_pGraph->m_pDB->GetRenderParams();
		
		int modelID = edge.Node(0)->NodeModel()->ModelID();
		dbRend->m_mWindowToModelID[0] = modelID;
		
		//corrValueName
		window->SetCorrToRender(collName, corrID, "none");
		
		window->StartDrawToBuffer();
		window->Draw();
		BufferID & bufferID = window->EndDrawToBuffer();	
		
		double cx, cy, scale, dx, dy;
		window->GetWindowParams(0, &cx, &cy, &scale, &dx, &dy); 
		int x0 = (int)((cx + window->Width())/2 - dx/4);
		int y0 = (int)((cy + window->Height())/2 - dy/4);
		int x1 = (int)((cx + window->Width())/2 + dx/4);
		int y1 = (int)((cy + window->Height())/2 + dy/4);
		
		window->WriteCurrentImage(bufferID, filename, true, x0, y0, x1, y1);
	}
	else if (coll!=NULL)
	{
		window->SetCorrFlag(DatabaseVisWindow::RENDER_CORRS_EDGES, true);
		window->SetCorrFlag(DatabaseVisWindow::RENDER_CORRS_SET, true);
		window->SetDesignRender(false);
		ModelDatabaseRenderingParams * dbRend = m_pGraph->m_pDB->GetRenderParams();
		
		assert(dbRend->m_mWindowToModelID.size() >= 3);
		for (int i=0; i<dbRend->m_mWindowToModelID.size(); i++)
			dbRend->m_mWindowToModelID[i] = -1;
			
		int modelID = edge.Node(0)->NodeModel()->ModelID();
		int model2ID = edge.Node(1)->NodeModel()->ModelID(); 
		dbRend->m_mWindowToModelID[0] = modelID;
		dbRend->m_mWindowToModelID[2] = model2ID;
		
		//corrValueName
		window->SetCorrToRender(collName, corrID, "none");
		
		window->StartDrawToBuffer();
		window->Draw();
		window->Draw();
		BufferID & bufferID = window->EndDrawToBuffer();	
		
//		double cx, cy, scale, dx, dy;
//		window->GetWindowParams(0, &cx, &cy, &scale, &dx, &dy); 
//		int x0 = (int)((cx + window->Width())/2 - dx/4);
//		int y0 = (int)((cy + window->Height())/2 - dy/4);
//		int x1 = (int)((cx + window->Width())/2 + dx/4);
//		int y1 = (int)((cy + window->Height())/2 + dy/4);
		
		int x0 = 0;
		int y0 = 0;
		int x1 = window->Width();
		int y1 = window->Height();
		
		window->WriteCurrentImage(bufferID, filename, true, x0, y0, x1, y1);
	}
	window->SetCorrFlag(DatabaseVisWindow::RENDER_CORRS_SPECIAL_1, false);		
	window->SetCorrFlag(DatabaseVisWindow::RENDER_CORRS_EDGES, false);
	window->SetCorrFlag(DatabaseVisWindow::RENDER_CORRS_SET, false);
	window->SetCorrToRender("none", -1, "none");
	window->SetDesignRender(true);			
}

VKString DynamicGraphLogParams::WriteNodeImage(std::vector<double> & nodeDists, 
											   int nodeID, int iterID, 
											   DatabaseVisWindow * window)
{
	VKString nodeStr = VKString::number(nodeID);
	VKString iterStr = VKString::number(iterID);
	VKString imgFile = m_DirPrefix + "/AllNodes/NodeImg_"+nodeStr + "_"+iterStr;
	
	// create model value
	VKString modelValueName("DynamicGraphNode");
	Model3D * model = m_pGraph->m_aAllNodes[nodeID].NodeModel();
	ModelValue * modelValue = model->GetModelValue(modelValueName);
	if (modelValue!=NULL)
		model->RemoveModelValue(modelValueName);
	modelValue = new ModelValue(model, modelValueName, false, false);
	
	// set per point values
	PointSet * pnts = model->PointCloud();
	assert(pnts->NumPoints()==(int)nodeDists.size());
	for (int i=0; i<(int)nodeDists.size(); i++)
		modelValue->SetValue(pnts->GetPoint(i), nodeDists[i]);
	
	// attach model value to a model
	model->AddModelValue(modelValueName, modelValue);
	
	// render model / write image
	model->RenderingParams()->SetFlags((int)Model3DRenderingParams::RENDER_VALUE);
	model->RenderingParams()->SetModelValueToRender(modelValueName);
	
	ModelDatabaseRenderingParams * dbRend = m_pGraph->m_pDB->GetRenderParams();
	dbRend->m_mWindowToModelID[0] = model->ModelID();
	window->SetDesignRender(false);
	window->StartDrawToBuffer();
	window->Draw();
	BufferID & bufferID = window->EndDrawToBuffer();	
	
	double cx, cy, scale, dx, dy;
	window->GetWindowParams(0, &cx, &cy, &scale, &dx, &dy); 
	int x0 = (int)((cx + window->Width())/2 - dx/4);
	int y0 = (int)((cy + window->Height())/2 - dy/4);
	int x1 = (int)((cx + window->Width())/2 + dx/4);
	int y1 = (int)((cy + window->Height())/2 + dy/4);
	
	window->WriteCurrentImage(bufferID, imgFile, true, x0, y0, x1, y1);
	window->SetDesignRender(true);
	
	model->RenderingParams()->SetFlags((int)Model3DRenderingParams::RENDER_FACES);
	
	return VKString("./AllNodes/NodeImg_")+nodeStr + "_"+iterStr + ".png";
}

VKString DynamicGraphLogParams::GetLinkImageName(int edgeID, int linkID)
{
	VKString retVal = VKString("AllLinks/Link_") + VKString::number(edgeID);
	return retVal + "_" + VKString::number(linkID);
}

void DynamicGraphLogParams::NewDiffusionStep()
{
	m_Iterations.push_back(IterationStep());
	Last().m_iStepType = 0;
}

void DynamicGraphLogParams::NewConsistencyStep()
{
	m_Iterations.push_back(IterationStep());
	Last().m_iStepType = 1;
}

DynamicGraphLogParams::IterationStep & DynamicGraphLogParams::Last()
{
	return m_Iterations[m_Iterations.size()-1];
}

void DynamicGraphLogParams::AddEigenvalues(LinAlgVectorReal * evals)
{
	assert(Last().m_pEigenvalues==NULL);
	Last().m_pEigenvalues = new LinAlgVectorReal(*evals);
}

void DynamicGraphLogParams::AddPerEdgeConsistencyScores(std::vector<double> & scores,
														std::vector<double> & corrVals,
														std::vector<int> & edgeIDs, 
														std::vector<int> & linkIDs)
{	
	assert(scores.size()==edgeIDs.size() && linkIDs.size()==edgeIDs.size());
	assert(Last().m_EdgeIDs.size()==0);
	Last().m_EdgeIDs = edgeIDs;
	Last().m_aPerEdgeConsistency = scores;
	Last().m_aPerEdgeLinkIDs = linkIDs;	
	Last().m_aPerEdgeCorrValue = corrVals;
}

void DynamicGraphLogParams::AddPerEdgeUpdates(std::vector<int> & edgeIDs, 
								std::vector<int> & oldLinks,
								std::vector<int> & newLinks,
								std::vector<std::vector<double> > & allScores,
								std::vector<std::vector<double> > & allCorrValues)
{
	assert(Last().m_ModifiedEdges.size()==0);
	Last().m_ModifiedEdges = edgeIDs;
	Last().m_ModifiedOldLinks = oldLinks;
	Last().m_ModifiedNewLinks = newLinks;
	Last().m_AllScores = allScores;
	Last().m_AllCorrValues = allCorrValues;
}

void DynamicGraphLogParams::AddNodeScores(std::vector<int> & nodeIDs, 
										  std::vector<double> & scores, 
										  std::vector<std::vector<double> > & ppScore)
{
	Last().m_NodeIDs = nodeIDs;
	Last().m_PerNodeDistances = scores;
	Last().m_pPerPointDistances = ppScore;
}

void DynamicGraphLogParams::WriteLog()
{
	VKString collDir = m_pGraph->m_pDB->GetCollectionDir(m_pGraph->m_sCollectionName);
	VKString logFile = collDir + "/Iterations.log.bin";
	FILE * binFile = fopen(logFile.c_str(), "wb");
	assert(binFile!=NULL);
	int nIters= (int)m_Iterations.size();
	fwrite(&nIters, sizeof(int), 1, binFile);
	
	for (int i=0; i<(int)m_Iterations.size(); i++)
	{
		IterationStep & step = m_Iterations[i];
		fwrite(&(step.m_iStepType), sizeof(int), 1, binFile);
		
		WriteVector(binFile, step.m_EdgeIDs);
		WriteVector(binFile, step.m_aPerEdgeConsistency);
		WriteVector(binFile, step.m_aPerEdgeCorrValue);
		WriteVector(binFile, step.m_aPerEdgeLinkIDs);
		
		WriteVector(binFile, step.m_ModifiedEdges);
		WriteVector(binFile, step.m_ModifiedOldLinks);
		WriteVector(binFile, step.m_ModifiedNewLinks);
		WriteVectors(binFile, step.m_AllScores);
		WriteVectors(binFile, step.m_AllCorrValues);
		
		WriteVector(binFile, step.m_NodeIDs);
		WriteVector(binFile, step.m_PerNodeDistances);
		WriteVectors(binFile, step.m_pPerPointDistances);
		
		fwrite(&(step.m_fConsistencyScore), sizeof(double), 1, binFile);
		fwrite(&(step.m_bDrewStep), sizeof(bool), 1, binFile);
	}
	fclose(binFile);
}

void DynamicGraphLogParams::ReadLog()
{
	VKString collDir = m_pGraph->m_pDB->GetCollectionDir(m_pGraph->m_sCollectionName);
	VKString logFile = collDir + "/Iterations.log.bin";
	FILE * binFile = fopen(logFile.c_str(), "rb");
	assert(binFile!=NULL);
	int nIters= (int)m_Iterations.size();
	fread(&nIters, sizeof(int), 1, binFile);
	
	for (int i=0; i<(int)nIters; i++)
	{
		m_Iterations.push_back(IterationStep());
		IterationStep & step = m_Iterations[i];
		fread(&(step.m_iStepType), sizeof(int), 1, binFile);
		
		ReadVector(binFile, step.m_EdgeIDs);
		ReadVector(binFile, step.m_aPerEdgeConsistency);
		ReadVector(binFile, step.m_aPerEdgeCorrValue);
		ReadVector(binFile, step.m_aPerEdgeLinkIDs);
		
		ReadVector(binFile, step.m_ModifiedEdges);
		ReadVector(binFile, step.m_ModifiedOldLinks);
		ReadVector(binFile, step.m_ModifiedNewLinks);
		ReadVectors(binFile, step.m_AllScores);
		ReadVectors(binFile, step.m_AllCorrValues);
		
		ReadVector(binFile, step.m_NodeIDs);
		ReadVector(binFile, step.m_PerNodeDistances);
		ReadVectors(binFile, step.m_pPerPointDistances);
		
		fread(&(step.m_fConsistencyScore), sizeof(double), 1, binFile);
		fread(&(step.m_bDrewStep), sizeof(bool), 1, binFile);
	}
	fclose(binFile);
}

void DynamicGraphLogParams::WriteVector(FILE* binFile, std::vector<int> & vec)
{
	int nVals = (int)vec.size();
	
	fwrite(&nVals, sizeof(int), 1, binFile);
	
	if (nVals>0)
	{
		int * vals = new int[nVals];
		for (int i=0; i<nVals; i++)
			vals[i] = vec[i];
		fwrite(vals, sizeof(int), nVals, binFile);
		delete [] vals;
	}
}

void DynamicGraphLogParams::WriteVector(FILE* binFile, std::vector<double> & vec)
{
	int nVals = (int)vec.size();
	
	fwrite(&nVals, sizeof(int), 1, binFile);
	
	if (nVals>0)
	{
		double * vals = new double[nVals];
		for (int i=0; i<nVals; i++)
			vals[i] = vec[i];
		fwrite(vals, sizeof(double), nVals, binFile);
		delete [] vals;
	}	
}

void DynamicGraphLogParams::ReadVector(FILE* binFile, std::vector<int> & vec)
{
	vec.clear();
	int nVals;
	
	fread(&nVals, sizeof(int), 1, binFile);
	
	if (nVals>0)
	{
		int * vals = new int[nVals];
		fread(vals, sizeof(int), nVals, binFile);
		for (int i=0; i<nVals; i++)
			vec.push_back(vals[i]);
		delete [] vals;
	}
}

void DynamicGraphLogParams::ReadVector(FILE* binFile, std::vector<double> & vec)
{
	vec.clear();
	int nVals;
	
	fread(&nVals, sizeof(int), 1, binFile);
	
	if (nVals>0)
	{
		double * vals = new double[nVals];
		fread(vals, sizeof(double), nVals, binFile);
		for (int i=0; i<nVals; i++)
			vec.push_back(vals[i]);
		delete [] vals;
	}
}

void DynamicGraphLogParams::WriteVectors(FILE* binFile, 
										 std::vector<std::vector<double> > & vecs)
{
	int nVecs = (int)vecs.size();
	fwrite(&nVecs, sizeof(int), 1, binFile);
	for (int i=0; i<nVecs; i++)
		WriteVector(binFile, vecs[i]);
}

void DynamicGraphLogParams::ReadVectors(FILE* binFile, 
										std::vector<std::vector<double> > & vecs)
{
	vecs.clear();
	int nVecs =0;
	fread(&nVecs, sizeof(int), 1, binFile);
	for (int i=0; i<nVecs; i++)
	{
		vecs.push_back(std::vector<double>());
		ReadVector(binFile, vecs[i]);
	}
}




