/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TimeProfiler.h"
#include "LSFuzzy.h"

////// LSFuzzyPerGroup ///////
LSFuzzyPerGroup::LSFuzzyPerGroup(std::vector<int> & grouping, 
								 int * globalPntToGroup,
								 int groupID, double threshold, 
								 double interpWeight)
: m_Grouping(grouping)
{
	m_aGlobalPntToGroup = globalPntToGroup;
	assert(groupID < (int)m_Grouping.size());
	int nFromPnts = m_Grouping[groupID];
	m_GroupedRowVals = new std::map<int, double> *[nFromPnts];
	int nGroups = (int)m_Grouping.size();
	m_aRowFilled = new bool[nFromPnts];
	for (int i=0; i<nFromPnts; i++)
	{
		m_GroupedRowVals[i] = new std::map<int, double>[nGroups];
		m_aRowFilled[i] = false;
	}
		
	m_fInterpWeight = interpWeight;
	m_fMaxValThreshold = threshold;	
	m_iGroupID = groupID;
	
	m_iGroupOffset = 0;
	m_iTotalNPoints = 0;
	for (int i=0; i<(int)m_Grouping.size(); i++)
	{
		if (i < m_iGroupID)
			m_iGroupOffset += m_Grouping[i];
		m_iTotalNPoints += m_Grouping[i];
	}
	
	m_iMatrixNumAlignEquations = 0;
	m_pLSEquations = NULL;
	m_pLSEquationsVector = NULL;
	m_pFuzzyValues = NULL;
	
}

LSFuzzyPerGroup::~LSFuzzyPerGroup()
{
	ClearMtxData();
	for (int i=0; i<m_Grouping[m_iGroupID]; i++)
		delete [] m_GroupedRowVals[i];
	delete [] m_GroupedRowVals;
	delete [] m_aRowFilled;
}

void LSFuzzyPerGroup::ClearMtxData()
{
	if (m_pLSEquations!=NULL)
		delete m_pLSEquations;
	if (m_pFuzzyValues!=NULL)
		delete m_pFuzzyValues;
	if (m_pLSEquationsVector!=NULL)
		delete m_pLSEquationsVector;
	m_pLSEquations = NULL;
	m_pFuzzyValues = NULL;
	m_pLSEquationsVector = NULL;
	
	for (int i=0; i<m_Grouping[m_iGroupID]; i++)
	{
		m_aRowFilled[i] = false;
		for (int j=0; j<(int)m_Grouping.size(); j++)
			m_GroupedRowVals[i][j].clear();
	}
}

std::map<int, double> * LSFuzzyPerGroup::GetCorrs(int pntFrom)
{
	int locPnt = FillRow(pntFrom);
	assert(m_aRowFilled[locPnt]);
	
	return m_GroupedRowVals[locPnt];
}

void LSFuzzyPerGroup::InitMtx(std::vector<int> & aij,
							  std::vector<double> & avals)
{
	if (m_pLSEquations!=NULL)	// matrix exists
		return;
	
	m_iMatrixNumAlignEquations = (int)avals.size();
	assert(aij.size() == 2 * avals.size());
	assert(m_pLSEquations==NULL);
	
	int mtxFormat = LinAlgMatrixSparseReal::SUPPORT_CX_SPARSE;
	int nRows = (int)avals.size() + m_Grouping[m_iGroupID];
	int nCols = m_iTotalNPoints;
	int nz = 2 * (int)avals.size() + m_Grouping[m_iGroupID];
	m_pLSEquations = new LinAlgMatrixSparseReal(nRows, nCols, mtxFormat, nz);
	
	for (int i=0; i<(int)avals.size(); i++)
	{
		m_pLSEquations->AddVal(i, aij[2*i+0], avals[i]);
		m_pLSEquations->AddVal(i, aij[2*i+1], -avals[i]);
	}
	
	int r0 = (int)avals.size();
	int c0 = m_iGroupOffset;
	for (int i=0; i<m_Grouping[m_iGroupID]; i++)
		m_pLSEquations->AddVal(r0+i, c0+i, m_fInterpWeight);
	
	m_pLSEquationsVector = new LinAlgVectorReal(nRows, 0);
	m_pFuzzyValues = new LinAlgVectorReal(nCols, 0);

	LinAlgMatrixSparseReal::MatrixFactorizationType factorize;
	factorize = LinAlgMatrixSparseReal::MAT_FACT_CHOLESKY_SYMM;

	m_pLSEquations->Factorize(factorize);
}

int LSFuzzyPerGroup::FillRow(int pntFrom)
{
	assert(pntFrom >= m_iGroupOffset);
	int locPnt = pntFrom - m_iGroupOffset;
	assert(locPnt < m_Grouping[m_iGroupID]);
	
	if (m_aRowFilled[locPnt])
		return locPnt;
	
	m_aRowFilled[locPnt] = true;

	assert(m_pLSEquations!=NULL 
		   && m_pFuzzyValues!=NULL 
		   && m_pLSEquationsVector!=NULL);
	
	m_pLSEquationsVector->SetValue(m_iMatrixNumAlignEquations + locPnt, m_fInterpWeight);
	m_pLSEquations->SolveFactorized(*m_pLSEquationsVector, *m_pFuzzyValues);
	m_pLSEquationsVector->SetValue(m_iMatrixNumAlignEquations + locPnt, 0);
	
	for (int i=0; i<m_pFuzzyValues->Rows(); i++)
	{
		if (m_pFuzzyValues->GetValue(i) > m_fMaxValThreshold)
		{
			int g2 = m_aGlobalPntToGroup[i];
			m_GroupedRowVals[locPnt][g2][i] = m_pFuzzyValues->GetValue(i);
		}
	}
	return locPnt;
}





////// LSFuzzy ///////
LSFuzzy::LSFuzzy(LSFuzzyType tp,
				 std::vector<int> & grouping)
{
	m_sMatlabScript = "none";
	m_Type = tp;
	m_pRadiosityFormFactors = NULL;
	m_pRadiosityEmitted = NULL;
	m_pRadiosityEnergy = NULL;

	//m_bComputeAllFuzzy = true;
	m_bComputeAllFuzzy = false;
	
	m_Grouping = grouping;
	m_iTotalNPoints = 0;
	for (int i=0; i<(int)m_Grouping.size(); i++)
		m_iTotalNPoints+=m_Grouping[i];

	m_FuzzyPerGroup = new LSFuzzyPerGroup*[(int)m_Grouping.size()];	
	m_aGlobalPntToGroup = new int [m_iTotalNPoints];
	
	int pc = 0;
	for (int i=0; i<(int)m_Grouping.size(); i++)
	{
		for (int j=0; j<m_Grouping[i]; j++)
		{
			m_aGlobalPntToGroup[pc] = i;
			pc++;
		}
		m_FuzzyPerGroup[i] = NULL;
	}
	
	m_iCorrSize = m_iTotalNPoints;
	m_aiTempClosestIDs = new int[m_iCorrSize];
	m_aiTempClosestVals = new double[m_iCorrSize];
	
	// precomputed correspondences
	m_aPrecomputedCorrs = new std::map<int, double> * [m_iTotalNPoints];
	m_aIsRowPrecomputed = new bool [m_iTotalNPoints];
	
	for (int i=0; i<m_iTotalNPoints; i++)
	{
		m_aPrecomputedCorrs[i] = NULL;
		m_aIsRowPrecomputed[i] = false;
	}

}

LSFuzzy::~LSFuzzy()
{
	ClearAll();
	delete [] m_aGlobalPntToGroup;
	delete [] m_FuzzyPerGroup;
	
	if (m_pRadiosityFormFactors!=NULL)
		m_pRadiosityFormFactors;
	
	if (m_pRadiosityEmitted!=NULL)
		delete m_pRadiosityEmitted;
	
	if (m_pRadiosityEmitted!=NULL)	
		delete m_pRadiosityEnergy;
	
	for (int i=0; i<m_iTotalNPoints; i++)
	{
		if (m_aPrecomputedCorrs[i]!=NULL)
			delete [] m_aPrecomputedCorrs[i];
	}
	delete [] m_aPrecomputedCorrs;
	delete [] m_aIsRowPrecomputed;
}

void LSFuzzy::SetMatlabScript(const VKString & scriptName)
{
	m_sMatlabScript = scriptName;
}

void LSFuzzy::ClearAll()
{
	m_Aij.clear();
	m_Aval.clear();
	for (int i=0; i<(int)m_Grouping.size(); i++)
		if (m_FuzzyPerGroup[i]!=NULL)
			m_FuzzyPerGroup[i]->ClearMtxData();
}

std::map<int, double> * LSFuzzy::GetAllCorrs(int pntFrom)
{
	if(m_Type==TP_LS_FUZZY)
	{
		FillRow(pntFrom);	
		int groupID1 = m_aGlobalPntToGroup[pntFrom];

		assert(m_FuzzyPerGroup[groupID1]!=NULL);
		std::map<int, double> * fuzzyCorrs = NULL;
		fuzzyCorrs = m_FuzzyPerGroup[groupID1]->GetCorrs(pntFrom);
		return fuzzyCorrs;
	}
	else
	{
		double threshold = 0.01;
		if (m_aIsRowPrecomputed[pntFrom])
			return m_aPrecomputedCorrs[pntFrom];
		
		m_aIsRowPrecomputed[pntFrom] = true;
		
		int nGroups = (int)m_Grouping.size();
		m_aPrecomputedCorrs[pntFrom] = new std::map<int, double>[nGroups];
		
		for (int i=0; i<m_iTotalNPoints; i++)
		{
			(*m_pRadiosityEmitted)(i) = 0;
			(*m_pRadiosityEnergy)(i) = 0;
		}
		
		(*m_pRadiosityEmitted)(pntFrom) = 1.;
		
		m_pRadiosityFormFactors->SolveFactorized(*m_pRadiosityEmitted, *m_pRadiosityEnergy);
		
		for (int i=0; i<m_iTotalNPoints; i++)
		{
			double val = (*m_pRadiosityEnergy)(i);
			if (val > threshold)
				m_aPrecomputedCorrs[pntFrom][GetGroup(i)][i] = val;
		}
		return m_aPrecomputedCorrs[pntFrom];
	}
}

int LSFuzzy::BestCorr(int pntFrom, int groupTo, double * value)
{
	std::map<int, double> * fuzzyCorrs = GetAllCorrs(pntFrom);
	int bestID = -1;
	double bestValue = -NUM_DBL_MAX;
	for (std::map<int, double>::iterator iter = fuzzyCorrs[groupTo].begin();
		 iter != fuzzyCorrs[groupTo].end(); iter++)
	{
		if (bestValue < iter->second)
		{
			bestValue = iter->second;
			bestID = iter->first;
		}
	}
	if (bestID==-1)
	{
		std::cout<<"[WARNING] Could not find best correspondence"<<std::endl;
		bestID=0;
		*value = 0;
	}
	else
		*value = bestValue;
	return bestID;
}

double LSFuzzy::FuzzyCorr(int pntFrom, int pntTo)
{
	std::map<int, double> * fuzzyCorrs = GetAllCorrs(pntFrom);
	
	int groupID2 = m_aGlobalPntToGroup[pntTo];

	std::map<int, double>::iterator iter = fuzzyCorrs[groupID2].find(pntTo);
	if (iter==fuzzyCorrs[groupID2].end())
		return 0;
	else
		return iter->second;
}

void LSFuzzy::FuzzyCorrs(int pntFrom, int ** closestIDs, 
						 double ** closestVals, int * nCorrs)
{
	std::map<int, double> * fuzzyCorrs = GetAllCorrs(pntFrom);
	
	*closestIDs = m_aiTempClosestIDs;
	*closestVals = m_aiTempClosestVals;
	
	*nCorrs=0;
	for (int i=0; i<(int)m_Grouping.size(); i++)
	{
		for (std::map<int, double>::iterator iter = fuzzyCorrs[i].begin();
			 iter != fuzzyCorrs[i].end(); iter++)
		{
			assert(*nCorrs < m_iCorrSize);
			
			(*closestIDs)[*nCorrs] = iter->first;
			(*closestVals)[*nCorrs] = iter->second;
			(*nCorrs)++;
		}
	}
}

void LSFuzzy::AddValue(int p1, int p2, double val)
{
	m_Aij.push_back(p1);
	m_Aij.push_back(p2);
	m_Aval.push_back(val);
}

void LSFuzzy::InitGroup(int groupID)
{
	assert(m_Type==TP_LS_FUZZY);
	if (m_FuzzyPerGroup[groupID]==NULL)
	{
		m_FuzzyPerGroup[groupID] = new LSFuzzyPerGroup(m_Grouping, 
													   m_aGlobalPntToGroup,
													   groupID, 0.01, 
													   100);
		m_FuzzyPerGroup[groupID]->InitMtx(m_Aij, m_Aval);
	}
}

void LSFuzzy::Embed(const VKString & prefix, 
					double * timeInSec)
{
	if (timeInSec!=NULL)
		*timeInSec = (double)time(NULL);
	LinAlgMatrixSparseReal::MatrixFactorizationType factorize;
	factorize = LinAlgMatrixSparseReal::MAT_FACT_CHOLESKY_SYMM;
	if (m_Type==TP_LS_FUZZY)
	{
		std::cout<<"QR Factorization ("<<m_Grouping.size()<<"):"<<std::endl;	
	
		if (m_bComputeAllFuzzy)
		{
			PrecomputeAllFuzzy();
		}
		else
		{
			TimeProfiler profiler;
			for (int i=0; i<(int)m_Grouping.size(); i++)
			{
				profiler.WriteProgress("QR Factorize", i, (int)m_Grouping.size());
				InitGroup(i);
			}
			std::cout<<std::endl;
		}
	}
	else if (m_Type==TP_RADIOSITY)
	{		
		double * rowNorm = new double[m_iTotalNPoints];
		for (int i=0; i<m_iTotalNPoints; i++)
			rowNorm[i] = 0;
		
		for (int i=0; i<(int)m_Aval.size(); i++)
		{
			int p1 = m_Aij[2*i+0];
			int p2 = m_Aij[2*i+1];
			double val = m_Aval[i];
			rowNorm[p1] += val;
			rowNorm[p2] += val;
		}
		LinAlgMatrixSparseReal::SupportedMatrixFormats format;
		format = LinAlgMatrixSparseReal::SUPPORT_CX_SPARSE;
		m_pRadiosityFormFactors = new LinAlgMatrixSparseReal(m_iTotalNPoints, 
															 m_iTotalNPoints, format);
		
		for (int i=0; i<(int)m_Aval.size(); i++)
		{
			int p1 = m_Aij[2*i+0];
			int p2 = m_Aij[2*i+1];
			double val = m_Aval[i];

			m_pRadiosityFormFactors->AddVal(p1, p2, -.99 * val / rowNorm[p1]);
			m_pRadiosityFormFactors->AddVal(p2, p1, -.99 * val / rowNorm[p2]);
		}
		for (int i=0; i<m_iTotalNPoints; i++)
			m_pRadiosityFormFactors->AddVal(i, i, 1.);
		m_pRadiosityEmitted = new LinAlgVectorReal(m_iTotalNPoints);
		m_pRadiosityEnergy = new LinAlgVectorReal(m_iTotalNPoints);
		delete [] rowNorm;
		std::cout<<"Factorizing matrix "<<m_iTotalNPoints<<"^2:..."<<std::flush;	
		m_pRadiosityFormFactors->Factorize(factorize);
		std::cout<<" - DONE!"<<std::endl;
	}
	else if (m_Type==TP_MATLAB_SCRIPT)
	{
		// write script input
		WriteAijIntoSparseMatrix(prefix + ".matrix.sim.bin");
		WriteGrouping(prefix + ".grouping.bin");
		
		// run script
		RunMatlab(m_sMatlabScript, prefix);
		
		// read script output
		ReadSparseMatrixIntoPrecompCorrs(prefix+".matrix.diff.bin");
	}
	if (timeInSec!=NULL)
		*timeInSec = (double)time(NULL)-(double)*timeInSec;
}

void LSFuzzy::RunMatlab(const VKString & scriptName, 
						const VKString & argsPrefix)
{
	VKString matlabParams = " -nodisplay -nosplash -r \"loadPaths(); ";
	matlabParams = matlabParams + scriptName+"('"+argsPrefix+"'); exit;\"";
	
	VKString matlabExe1 = "matlab";
	VKString matlabExe2 = "/usr/local/matlab/bin/matlab";
	
	VKString cmd = VKString("if [ \"`command -v ")+matlabExe1+"`\" != \"\" ]; then ";
	cmd = cmd + matlabExe1 + matlabParams + "; else ";
	cmd = cmd + matlabExe2 + matlabParams + "; fi ";
	assert(system(NULL));
	
	std::cout<<"MATLAB CMD: "<<cmd.c_str()<<std::endl;
	system(cmd.c_str());
}

void LSFuzzy::WriteAijIntoSparseMatrix(const VKString & sparseMtx)
{
	assert(2*m_Aval.size()==m_Aij.size());
	assert(m_Aval.size()>0);
	
	int nVals = (int)m_Aval.size();
	
	int * rows = new int[nVals];
	int * cols = new int[nVals];
	double * vals = new double[nVals];
	
	for (int i=0; i<nVals; i++)
	{
		rows[i] = m_Aij[2*i+0];
		cols[i] = m_Aij[2*i+1];
		vals[i] = m_Aval[i];
	}
	
	FILE * file = fopen(sparseMtx.c_str(), "wb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot write "<<sparseMtx.c_str()<<std::endl;
	assert(file!=NULL);
	fwrite(&m_iTotalNPoints, sizeof(int), 1, file);
	fwrite(&m_iTotalNPoints, sizeof(int), 1, file);
	fwrite(&nVals, sizeof(int), 1, file);
	
	fwrite(rows, sizeof(int), nVals, file);
	fwrite(cols, sizeof(int), nVals, file);
	fwrite(vals, sizeof(double), nVals, file);
	
	fclose(file);
	
	delete [] rows;
	delete [] cols;
	delete [] vals;
}

void LSFuzzy::WriteGrouping(const VKString & groupingStr)
{
	int nGroups = (int)m_Grouping.size();
	double * groupIDs = new double[nGroups];
	for (int i=0; i<nGroups; i++)
		groupIDs[i] = (double)m_Grouping[i];
	
	FILE * file = fopen(groupingStr.c_str(), "wb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot write "<<groupingStr.c_str()<<std::endl;	
	assert(file!=NULL);
	fwrite(&nGroups, sizeof(int), 1, file);
	fwrite(groupIDs, sizeof(double), nGroups, file);
	fclose(file);
	
	delete [] groupIDs;
}

void LSFuzzy::ReadSparseMatrixIntoPrecompCorrs(const VKString & sparseMtx)
{	
	// read sparse matrix
	int m, n, nz;
	int * rows;
	int * cols;
	double * vals;
	FILE * file = fopen(sparseMtx.c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot read "<<sparseMtx.c_str()<<std::endl;

	assert(file!=NULL);
	fread(&m, sizeof(int), 1, file);
	fread(&n, sizeof(int), 1, file);
	fread(&nz, sizeof(int), 1, file);
	assert(m==n && m==m_iTotalNPoints);
	
	rows = new int[nz];
	cols = new int[nz];
	vals = new double[nz];
	
	fread(rows, sizeof(int), nz, file);
	fread(cols, sizeof(int), nz, file);
	fread(vals, sizeof(double), nz, file);
	fclose(file);
	
	// FILL:
	//	std::map<int, double> ** m_aPrecomputedCorrs;
	//	bool * m_aIsRowPrecomputed;
	
	int nGroups = (int)m_Grouping.size();
	// clear data
	for (int i=0; i<m_iTotalNPoints; i++)
	{
		m_aIsRowPrecomputed[i] = true;

		if (m_aPrecomputedCorrs[i]==NULL)
			m_aPrecomputedCorrs[i] = new std::map<int, double>[nGroups];
		else
		{
			for (int j=0; j<nGroups; j++)
				m_aPrecomputedCorrs[i][j].clear();
		}
	}
	
	// fill data
	for (int i=0; i<nz; i++)
	{
		int r = rows[i];
		int c = cols[i];
		double v = vals[i];
		m_aPrecomputedCorrs[r][GetGroup(c)][c] = v;
	}
	
	delete [] rows;
	delete [] cols;
	delete [] vals;
}

void LSFuzzy::PrecomputeAllFuzzy()
{
	
	// write *bin file

	// run matlab scripts (in parallel w/ qsub or in sequence)
		// check matlab path
	
	// 
}

int LSFuzzy::GetGroup(int pnt)
{
	return m_aGlobalPntToGroup[pnt];
}

void LSFuzzy::FillRow(int pntFrom)
{
	assert(m_Type==TP_LS_FUZZY);
	int groupID = m_aGlobalPntToGroup[pntFrom];
	if (m_FuzzyPerGroup[groupID]==NULL)
		InitGroup(groupID);
	m_FuzzyPerGroup[groupID]->FillRow(pntFrom);
}

void LSFuzzy::Write(const char * dir, const char * prefix)
{
	std::string str(dir);
	str = str+ prefix+"_lsfuzzy.bin";
	
	FILE * file = fopen(str.c_str(), "wb");
	int sizeVals=m_Aval.size();
	fwrite(&sizeVals, sizeof(int), 1, file);
	
	assert(sizeVals>0);
	double * vals = new double[sizeVals];
	int * ids = new int[2*sizeVals];
	
	for (int i=0; i<sizeVals; i++)
	{
		vals[i] = m_Aval[i];
		ids[2*i+0] = m_Aij[2*i+0];
		ids[2*i+1] = m_Aij[2*i+1];
	}
	
	fwrite(vals, sizeof(double), sizeVals, file);
	fwrite(ids, sizeof(int), 2*sizeVals, file);
		
	delete [] vals;
	delete [] ids;
	fclose(file);
}

void LSFuzzy::Read(const char * dir, const char * prefix)
{
	std::string str(dir);
	str = str+ prefix+"_lsfuzzy.bin";
	
	FILE * file = fopen(str.c_str(), "rb");
	int sizeVals;
	fread(&sizeVals, sizeof(int), 1, file);
	assert(sizeVals>0);
	
	double * vals = new double[sizeVals];
	int * ids = new int[2*sizeVals];
	
	fread(vals, sizeof(double), sizeVals, file);
	fread(ids, sizeof(int), 2*sizeVals, file);
	
	for (int i=0; i<sizeVals; i++)
		AddValue(ids[2*i+0], ids[2*i+1], vals[i]);
	
	delete [] vals;
	delete [] ids;
	fclose(file);
}
