/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CollectionIdentityPointsets.h"

#include "Model3D.h"
#include "DatabaseVisWindow.h"
#include "CorrCollection.h"
#include "ModelDatabase.h"
#include "ModelValueLabel.h"
#include "DistanceEuclidean.h"
#include "DistanceGeodesic.h"
#include "Sortable.h"
#include "CollectionGroundTruth.h"

/******************      Model3D            *****************/
Model3D::Model3D(const VKString & filename, const VKString & modelName, 
				 int nApproxPnts, ModelDatabase * db)
{
	m_pIsotropciMesh = NULL;
	m_pMeshTreeForRigidMatch = NULL;
	m_pKdTreeForRigidMatch = NULL;
	m_bCentroid = false;
	m_bRadius = false;
	m_bPCAWithUpDefined = false;
	m_bPCAAxisDefined = false;
	m_aPntsForRigidMatch = NULL;
    m_bFastIntersectionBuffer = false;
	m_GeomType = INVALID_GEOM;
	m_pDB = db;
	m_pRenderingParams = NULL;
	m_iApproximationPnts = nApproxPnts;
	m_sFilename = filename;
	m_sGeometryFile = filename;
	m_sModelName = modelName;
	m_iID = -1;
	m_pMesh = NULL;
	
	m_iIDTempPointSet = 0;
	m_iIDTempModelValue = 0;
	m_pSearchTree = NULL;
	m_pCurrNhd = NULL;
	m_pFaceCentroids = NULL;
    m_pKdTreeFaceCentroids = NULL;        
}

Model3D::~Model3D()
{
	if (m_aPntsForRigidMatch!=NULL)
		delete [] m_aPntsForRigidMatch;
}

const R3Mesh * Model3D::GetMesh() const
{
	assert(m_pMesh!=NULL);
	return m_pMesh;
}

void Model3D::InitRenderingParams(ParamParser & visParams, 
								  DatabaseVisWindow * window)
{
	m_pRenderingParams = new Model3DRenderingParams(visParams, window);
	if (InMemory())
		m_pRenderingParams->InitializeRenderer(this);
}

void Model3D::WriteToDB()
{
	VKString filename = m_pDB->GetModelFile(m_sModelName);
	VKString dirPath = m_pDB->GetModelDir(m_sModelName);
	m_sGeometryFile = dirPath + "Geometry/" + m_sModelName + ".ply";
	m_pDB->CreateDirsOnPathToFile(filename);
	m_pDB->CreateDirsOnPathToFile(dirPath);
	std::ofstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot write: "<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	textStream<<"ModelData:"<<"\n";
	textStream<<"\tGeometry\t"<<m_sGeometryFile.c_str()<<"\n";
	textStream<<"\tGeomType\t"<<GeomTypeToStr(m_GeomType).c_str()<<"\n";
	m_pDB->CreateDirsOnPathToFile(m_sGeometryFile);
	if (m_pMesh!=NULL)
		m_pMesh->WriteFile(m_sGeometryFile.c_str());
	else
		assert(ModelDatabase::FileExists(m_sGeometryFile));
	
	// model values: texture, pre-calculated values, etc...
	if (m_ModelValuesCameWithExternalFile.count()>0)
	{
		textStream<<"\tModelValues\t[";
		for (int i=0; i<m_ModelValuesCameWithExternalFile.count(); i++)
			textStream<<m_ModelValuesCameWithExternalFile[i].c_str()<<" ";
		textStream<<"]\n";

		for (int i=0; i<m_ModelValuesCameWithExternalFile.count(); i++)
			GetModelValue(m_ModelValuesCameWithExternalFile[i])->WriteModelValue(m_pDB);
	}
	
	//model point sets (e.g. ground truth)
	if (m_PntSetsCameWithExternalFile.count()>0)
	{
		textStream<<"\tModelPntSets\t[";
		for (int i=0; i<m_PntSetsCameWithExternalFile.count(); i++)
			textStream<<m_PntSetsCameWithExternalFile[i].c_str()<<" ";
		textStream<<"]\n";
		
		for (int i=0; i<m_PntSetsCameWithExternalFile.count(); i++)
			GetPointSet(m_PntSetsCameWithExternalFile[i])->WritePointSet(m_pDB);
	}
	
	// model correspondences (e.g. ground truth maps)
	if (m_CorrCollectionsCameWithExternalFile.count()>0)
	{
		textStream<<"\tCorrCollections\t[";
		for (int i=0; i<m_CorrCollectionsCameWithExternalFile.count(); i++)
			textStream<<m_CorrCollectionsCameWithExternalFile[i].c_str()<<" ";
		textStream<<"]\n";
	}

	if (m_MyPreferableCamera.count() > 0)
	{
		VKString camStr = m_MyPreferableCamera.join(" ");
		textStream<<"\tCameraSettings\t["<<camStr.c_str()<<"]\n";	
	}
	
	if (m_sBIMQueryTriplet.count()==3)
	{
		textStream<<"\tQueryTriplet\t["<<m_sBIMQueryTriplet[0].c_str()<<" ";
		textStream<<m_sBIMQueryTriplet[1].c_str()<<" ";
		textStream<<m_sBIMQueryTriplet[2].c_str()<<"]\n";
	}
	textStream.close();
}

void Model3D::ReadFromDB()
{
	VKString filename = m_pDB->GetModelFile(m_sModelName);
	if (!m_pDB->FileExists(filename))
	{
		std::cout<<"\n[ERROR] Model file does not exist: "<<filename.c_str()<<std::endl;
		assert(m_pDB->FileExists(filename));
	}
	ParamParser modelParams(filename);
	bool valid;
	VKString geometry = modelParams.GetStrValue("ModelData", "Geometry", valid);
	if (!valid)
	{
		std::cout<<"\n[ERROR] Could not read model file: "<<m_sModelName.c_str()<<std::endl;
		assert(valid);
	}
	VKString geomType = modelParams.GetStrValue("ModelData", "GeomType", valid);
	if (valid)
		m_GeomType = GeomTypeStrToType(geomType);
	else 
		m_GeomType = UNIFORM_POLYGON_SOUP;
	    
	bool exists = m_pDB->FileExists(geometry);
	if (!exists)
	{
		VKString oldFailure = geometry;
		VKStringList tryOther = geometry.split("/");
		assert(tryOther.count()>0);
		tryOther[0] = m_pDB->GetMainDir();
		geometry = tryOther.join("/");
		exists = m_pDB->FileExists(geometry);
		if (!exists)
			std::cout<<"[ERROR] Could not find geometry file: "<<oldFailure.c_str()<<std::endl;		
		assert(exists);
	}
	m_pMesh = new R3Mesh();
	m_pMesh->ReadFile(geometry.c_str());
	
	//read model values
	m_ModelValuesCameWithExternalFile = modelParams.GetStrValues("ModelData", "ModelValues", valid);
	for (int i=0; i<m_ModelValuesCameWithExternalFile.count(); i++)
	{
		bool loaded = ModelValue::ReadModelValue(m_sModelName, 
												 m_ModelValuesCameWithExternalFile[i], 
												 m_pDB);
		assert(loaded);
	}	
	
	// read point sets
	m_PntSetsCameWithExternalFile = modelParams.GetStrValues("ModelData", "ModelPntSets", valid);
	for (int i=0; i<m_PntSetsCameWithExternalFile.count(); i++)
	{
		bool loaded = PointSet::ReadPointSet(m_sModelName, 
											 m_PntSetsCameWithExternalFile[i],
											 m_pDB);
		if (!loaded)
		{
			std::cout<<"[ERROR] Failed to read point set: m="<<m_sModelName.c_str();
			std::cout<<" pnts="<<m_PntSetsCameWithExternalFile[i].c_str()<<std::endl;
		}
		assert(loaded);
	}	
	
	// read correspondences (see ModelDatabase)
	VKStringList corrStr = modelParams.GetStrValues("ModelData", "CorrCollections", valid);
	m_CorrCollectionsCameWithExternalFile = corrStr;
	//std::cout<<"Found External collections; "<<corrStr.count()<<std::endl;
	//m_GeomType = IRREGULAR_POLYGON_SOUP;
	//std::cout<<"[WARNING] Temporarily disabled reading corres / point sets: Model3D"<<std::endl;
	
	m_sBIMQueryTriplet = modelParams.GetStrValues("ModelData", "QueryTriplet", valid);
	
	m_MyPreferableCamera = modelParams.GetStrValues("ModelData", "CameraSettings", valid);
}

bool Model3D::LoadToMemory()
{
	if (m_pMesh!=NULL)
		return true;		// already loaded
	
	bool remesh = m_iApproximationPnts>0;
	bool needToWrite = true;
	if (m_sFilename == m_pDB->GetModelFile(m_sModelName))	// in the DB
	{
		needToWrite = remesh;
		ReadFromDB();
	}
	else if (m_sFilename.endsWith(".dae"))	// COLLADA file
	{
#ifdef OPEN_COLLADA
		ReadFromCOLLADA();
		m_GeomType = IRREGULAR_POLYGON_SOUP;
#else
		std::cout<<"[ERROR] COLLADA files are not supported. ";
		std::cout<<"Define INCLUDE_GOOGLE_SKETCHUP=\"true\" in the makefile. "<<std::endl;	
#endif
	}
	else if (m_sFilename.endsWith(".psb"))	// file from the Princeton Shape Benchmark 
	{
		ReadFromPSB();
	}
	else if (m_sFilename.endsWith(".obj") || m_sFilename.endsWith(".off") || 
			 m_sFilename.endsWith(".ray") || m_sFilename.endsWith(".ply") || 
			 m_sFilename.endsWith(".cat") || m_sFilename.endsWith(".ifs") ||
			 m_sFilename.endsWith(".stl") || m_sFilename.endsWith(".wrl"))
	{
		assert(m_pMesh==NULL);
		m_pMesh = new R3Mesh();
		m_pMesh->ReadFile(m_sFilename.c_str());
		m_GeomType = IRREGULAR_POLYGON_SOUP;
	}
	else if (m_sFilename.endsWith(".skp"))
	{
#ifdef GOOGLE_SKETCHUP
		ReadFromSketchup();
		m_GeomType = IRREGULAR_POLYGON_SOUP;
#else
		m_GeomType = INVALID_GEOM;
		std::cout<<"[ERROR] Sketchup files are not supported. ";
		std::cout<<"Define INCLUDE_GOOGLE_SKETCHUP=\"true\" in the makefile. "<<std::endl;
#endif
	}
	else 
	{
		m_GeomType = INVALID_GEOM;
		std::cout<<"[ERROR] File is in unknown format: "<<m_sFilename.c_str()<<std::endl;
	}
	
	assert(m_pMesh!=NULL);
	
	for (int i=0; i<m_ModelValuesCameWithExternalFile.count(); i++)
		if (!m_pDB->m_ModelValueNames.contains(m_ModelValuesCameWithExternalFile[i]))
			m_pDB->m_ModelValueNames.push_back(m_ModelValuesCameWithExternalFile[i]);
	
	if (m_GeomType==IRREGULAR_POLYGON_SOUP && remesh)
		Remesh();
	
	//if (needToWrite)
		WriteToDB();
    
	if (m_pRenderingParams!=NULL)
		m_pRenderingParams->InitializeRenderer(this);
	return true;
}

void Model3D::NormalizeScale()
{
	R3Point p = m_pMesh->Centroid();
	double rad = m_pMesh->AverageRadius(&p);
	for (int i=0; i<m_pMesh->NVertices(); i++)
	{
		R3MeshVertex * v = m_pMesh->Vertex(i);
		R3Point pos = m_pMesh->VertexPosition(v) / rad;
		m_pMesh->SetVertexPosition(v, pos);
	}
}

void Model3D::CameWithPointSet(const VKString & pntSetName)
{
	if (!m_PntSetsCameWithExternalFile.contains(pntSetName))
		m_PntSetsCameWithExternalFile.push_back(pntSetName);
}

void Model3D::CameWithCollection(const VKString & collectionName)
{
	if (!m_CorrCollectionsCameWithExternalFile.contains(collectionName))
		m_CorrCollectionsCameWithExternalFile.push_back(collectionName);
}

void Model3D::Remesh()
{
	m_GeomType = UNIFORM_POLYGON_SOUP;
//	std::cout<<"Remeshing "<<m_sFilename.c_str()<<std::endl;
	// subdivide mesh to have m_iApproximationPnts
	R3Mesh * oldMesh = m_pMesh;
	//R3Mesh * newMesh = SubdivideMeshLongestEdge(m_pMesh, m_iApproximationPnts);
	//R3Mesh * newMesh = SubdivideMeshLargestAreaFace(m_pMesh, m_iApproximationPnts);
	double minDist = FLT_MAX;
	R3Mesh * newMesh = SubdivideMeshFarthestFaceCentroid(m_pMesh, m_iApproximationPnts, &minDist);
	//newMesh->SubdivideLongEdges(minDist/2);
	
	// todo: re-project
//	assert(m_CorrCollectionsCameWithExternalFile.count()==0);
//	assert(m_PntSetsCameWithExternalFile.count()==0);
	if (m_CorrCollectionsCameWithExternalFile.count()>0
		|| m_PntSetsCameWithExternalFile.count()>0)
	{
		std::cout<<"[WARNING] Input collections and point sets might have changed ";
		std::cout<<"due to remeshing. Potential serious errors"<<std::endl;
		for (int i=0; i<m_PntSetsCameWithExternalFile.count(); i++)
		{
			PointSet * pnts = GetPointSet(m_PntSetsCameWithExternalFile[i]);
			assert(pnts!=NULL);
			pnts->Remeshed(oldMesh, newMesh);
		}
	}
	
	if (newMesh==NULL)
	{
		std::cout<<"[ERROR] Could not remesh model: "<<m_sFilename.c_str()<<std::endl;
		assert(false);
	}
	VKString tempFile = m_pDB->GetMainDir() + "tempmesh.ply";
	m_pDB->CreateDirsOnPathToFile(tempFile);
	newMesh->WriteFile(tempFile.c_str());
	delete newMesh;	// weird bug NFaces saved != NFaces read
	newMesh = new R3Mesh();
	newMesh->ReadFile(tempFile.c_str());
	
	// ID nearest vertices
	std::vector<int> newMeshVtxToOldMeshVtx;
	//	PointSet * pntVertices = new PointSet("TempPointSetAllVertices");
	//	pntVertices->InitFromAllVertices(this);
	//	R3Kdtree<const PointInSet *> * kdTree = pntVertices->GetKdTree();
	//	for (int i=0; i<newMesh->NVertices(); i++)
	//	{
	//		const PointInSet * closest = kdTree->FindClosest(R3Point(newMesh->VertexPosition(newMesh->Vertex(i))), 0, FLT_MAX);
	//		newMeshVtxToOldMeshVtx.push_back(closest->id);
	//	}
	//	delete pntVertices;
	
	if (m_ModelValuesCameWithExternalFile.count()==0)
	{
		// set new mesh
		m_pMesh = newMesh;	
	}
	else	// re-project values
	{
		// Find nearest faces map from new mesh to old mesh (to transfer labels)
		std::vector<int> newMeshVtxToOldMeshFace;
		R3MeshSearchTree * searchtree = new R3MeshSearchTree(oldMesh);
		bool printError = false;
		for (int i=0; i<newMesh->NVertices(); i++)
		{
			R3MeshIntersection closest;
			R3Point vpos = newMesh->VertexPosition(newMesh->Vertex(i));
			searchtree->FindClosest(vpos, closest, 0, FLT_MAX);
			if (closest.face==NULL)
			{
				if (printError)
				{
					std::cout<<"[WARNING] Failed to re-project mesh : "<<m_sFilename.c_str()<<std::endl;
					std::cout<<"\tNFaces: "<<m_pMesh->NFaces();
					std::cout<<" Pnt="<<vpos.X()<<", "<<vpos.Y()<<", "<<vpos.Z()<<", "<<std::endl;
				}
				newMeshVtxToOldMeshFace.push_back(-1);
			}
			else
				newMeshVtxToOldMeshFace.push_back(oldMesh->FaceID(closest.face));
		}
		delete searchtree;
		
		// set new mesh
		m_pMesh = newMesh;
		
		// update model values (that depend on a mesh)
		for (int i=0; i<m_ModelValuesCameWithExternalFile.count(); i++)
		{
			ModelValue * modelValue = GetModelValue(m_ModelValuesCameWithExternalFile[i]);
			assert (modelValue!=NULL);
			ModelValue * updatedValue = modelValue->CopyModelValue(oldMesh, newMesh, 
																   newMeshVtxToOldMeshVtx,
																   newMeshVtxToOldMeshFace);
			assert(updatedValue!=NULL);
			
			RemoveModelValue(m_ModelValuesCameWithExternalFile[i]);
			AddModelValue(m_ModelValuesCameWithExternalFile[i], updatedValue);
			updatedValue->WriteModelValue(m_pDB);
		}
	}
	
	delete oldMesh;
}

bool Model3D::AreFacesSame(R3Mesh * origMesh, R3MeshFace * f1, R3MeshFace * f2)
{
	assert(f1!=NULL && f2!=NULL && origMesh!=NULL);
	R3MeshVertex * v1[3];
	R3MeshVertex * v2[3];
	for (int i=0; i<3; i++)
		v1[i] = origMesh->VertexOnFace(f1, i);
	for (int i=0; i<3; i++)
		v2[i] = origMesh->VertexOnFace(f2, i);
	
	if (f1==f2)
		return false;	// actually same pointer - it's ok
	
	if ((v1[0]==v2[0] && v1[1]==v2[1] && v1[2]==v2[2])
		|| (v1[0]==v2[0] && v1[1]==v2[2] && v1[2]==v2[1])
		|| (v1[0]==v2[1] && v1[1]==v2[0] && v1[2]==v2[2])
		|| (v1[0]==v2[1] && v1[1]==v2[2] && v1[2]==v2[0])
		|| (v1[0]==v2[2] && v1[1]==v2[0] && v1[2]==v2[1])
		|| (v1[0]==v2[2] && v1[1]==v2[1] && v1[2]==v2[0]))
		return true;
	return false;
}

void Model3D::RemoveRepeatedTriangles(R3Mesh * origMesh)
{
	for (int f=0; f<origMesh->NFaces(); f++)
	{
		
		R3MeshFace * face = origMesh->Face(f);
		bool deleteMe = false;
		
		for (int fof=0; fof<3 && !deleteMe; fof++)
		{
			R3MeshFace * cf = origMesh->FaceOnFace(face, fof);
			if (cf!=NULL)
				deleteMe = AreFacesSame(origMesh, face, cf);
		}
		
		if (!deleteMe)
		{
			for (int vof=0; vof<3 && !deleteMe; vof++)
			{
				R3MeshVertex * v = origMesh->VertexOnFace(face, vof);
				assert(v!=NULL);
				for (int fov = 0; fov < origMesh->VertexValence(v) && !deleteMe; fov++)
				{
					R3MeshEdge * e = origMesh->EdgeOnVertex(v, fov);
					
					R3MeshFace * cf = origMesh->FaceOnEdge(e, 0);
					if (cf!=NULL)
						deleteMe = AreFacesSame(origMesh, face, cf);
					
					if (!deleteMe)
					{
						cf = origMesh->FaceOnEdge(e, 1);
						if (cf!=NULL)
							deleteMe = AreFacesSame(origMesh, face, cf);
					}
				}
			}
		}

		if (deleteMe)
		{

			origMesh->DeleteFace(face);
			f--;
		}
	}
}

void Model3D::SubdivideFace(R3Mesh * mesh, R3MeshFace * f, 
							std::vector<R3MeshFace*> * newFaces, 
							std::vector<R3MeshFace*> * delFaces)
{
	// Find vertices/edges bounding face
	R3MeshVertex *v0 = mesh->VertexOnFace(f, 0);
	R3MeshEdge *e0 = mesh->EdgeOnFace(f, v0, RN_CCW);
	R3MeshVertex *v1 = mesh->VertexAcrossEdge(e0, v0);
	R3MeshEdge *e1 = mesh->EdgeOnFace(f, v1, RN_CCW);
	R3MeshVertex *v2 = mesh->VertexAcrossEdge(e1, v1);
	R3MeshEdge *e2 = mesh->EdgeOnFace(f, v2, RN_CCW);
	
//	assert(e0!=e1 && e0!=e2 && e1!=e2);
//	assert(v0!=v1 && v0!=v2 && v1!=v2);
	int m = mesh->FaceMaterial(f);
	int s = mesh->FaceSegment(f);
	
	if (delFaces!=NULL)
	{
		for (int i=0; i<3; i++)
		{
			R3MeshFace * fof = mesh->FaceOnFace(f, i);
			if (fof!=NULL)
				delFaces->push_back(fof);
		}
		delFaces->push_back(f);
	}

//	std::vector<R3MeshVertex*> tempDebugVertex;
//	R3MeshFace * debugF1 = mesh->FaceOnEdge(e0, 0);
//	R3MeshFace * debugF2 = mesh->FaceOnEdge(e0, 1);
//	for (int i=0; i<3 && debugF1; i++)
//		tempDebugVertex.push_back(mesh->VertexOnFace(debugF1, i));
//	for (int i=0; i<3 && debugF2; i++)
//		tempDebugVertex.push_back(mesh->VertexOnFace(debugF2, i));

	// Delete face		
	mesh->DeleteFace(f);
	 
	// Subdivide edges
	R3MeshVertex *ve0 = mesh->SubdivideEdge(e0);
	R3MeshVertex *ve1 = mesh->SubdivideEdge(e1);
	R3MeshVertex *ve2 = mesh->SubdivideEdge(e2);

	// Create new faces
//	// DEBUG BEGINS
//	std::cout<<"Adding face"<<std::endl;
//	R3MeshVertex * nv[3];
//	R3MeshEdge * ne[3];
//	nv[0] = v0;
//	nv[1] = ve0;
//	nv[2] = ve2;
//	for (int i=0; i<3; i++)
//	{
//		std::cout<<"\tvtx["<<i<<"] = "<<nv[i]<<" valence="<<mesh->VertexValence(nv[i])<<std::endl;
//		if (i>0 && mesh->VertexValence(nv[i])!=3)
//		{
//			std::cout<<"\tStrange Valence vertex edges:"<<std::endl;
//			std::cout<<"\tVertexID="<<mesh->VertexID(nv[i])<<" DELETED FACE="<<f<<std::endl;
//			if (tempDebugVertex.size()>=3)
//				std::cout<<"\t\tTRI1: "<<debugF1<<" ["<<tempDebugVertex[0]<<", "<<tempDebugVertex[1]<<", "<<tempDebugVertex[2]<<"] "<<std::endl;
//			if (tempDebugVertex.size()>=6)
//				std::cout<<"\t\tTRI2: "<<debugF2<<" ["<<tempDebugVertex[3]<<", "<<tempDebugVertex[4]<<", "<<tempDebugVertex[5]<<"] "<<std::endl;
//			for (int j=0; j<mesh->VertexValence(nv[i]); j++)
//			{
//				R3MeshEdge * edge = mesh->EdgeOnVertex(nv[i], j);
//				std::cout<<"\t\tE: "<<mesh->VertexOnEdge(edge, 0)<<" -> "<<mesh->VertexOnEdge(edge, 1)<<std::endl;
//			}
//		}
//		ne[i] = mesh->EdgeBetweenVertices(nv[i], nv[(i+1)%3]);
//	}
//	
//	for (int i=0; i<3; i++)
//	{
//		std::cout<<"\tedge["<<i<<"] = "<<ne[i]<<std::endl;
//		if (ne[i]!=0)
//		{
//			std::cout<<"\tEDGE["<<i<<"] = ("<<mesh->VertexOnEdge(ne[i], 0)<<" -> "<<std::flush;
//			std::cout<<mesh->VertexOnEdge(ne[i], 1)<<"). f0="<<mesh->FaceOnEdge(ne[i], 0)<<std::flush;
//			std::cout<<" f1="<<mesh->FaceOnEdge(ne[i], 1)<<std::endl;
//		}
//	}
//	// DEBUG ENDS
	
	R3MeshFace *f1 = mesh->CreateFace(v0, ve0, ve2);  
	R3MeshFace *f2 = mesh->CreateFace(v1, ve1, ve0);  
	R3MeshFace *f3 = mesh->CreateFace(v2, ve2, ve1);  
	R3MeshFace *f4 = mesh->CreateFace(ve0, ve1, ve2); 
//	std::cout<<"Subdivision: "<<f1<<" "<<f2<<" "<<f3<<" "<<f4<<std::endl;

	assert(f1!=NULL);
	assert(f2!=NULL);
	assert(f3!=NULL);
	assert(f4!=NULL);
	
	if (newFaces!=NULL)
	{
		if (f1!=NULL)
			newFaces->push_back(f1);
		if (f2!=NULL)
			newFaces->push_back(f2);
		if (f3!=NULL)
			newFaces->push_back(f3);
		if (f4!=NULL)
			newFaces->push_back(f4);
	}

	// Set face materials
	mesh->SetFaceMaterial(f1, m);
	mesh->SetFaceMaterial(f2, m);
	mesh->SetFaceMaterial(f3, m);
	mesh->SetFaceMaterial(f4, m);

	// Set face segments
	mesh->SetFaceSegment(f1, s);
	mesh->SetFaceSegment(f2, s);
	mesh->SetFaceSegment(f3, s);
	mesh->SetFaceSegment(f4, s);
		
}

// actually subdivides face adjacent to an edge with farthest centroid. 
R3Mesh * Model3D::SubdivideMeshFarthestFaceCentroid(R3Mesh * origMesh, int approxPnts, double * minDist)
{
	assert(minDist!=NULL);
	*minDist = FLT_MAX;
	R3Mesh * newMesh = CloneMesh(origMesh);
	RemoveRepeatedTriangles(newMesh);
	
	//vector of (minimal) distances from all edges (centroids)
	std::vector<Sortable> minDistances;
	
	// fill distances from edge centroids to vertices
		// prepare vertices kd-tree
	R3Point * vertices = new R3Point[newMesh->NVertices()];
	RNArray<R3Point *> pnts;
	for (int i=0; i<newMesh->NVertices(); i++)
	{
		vertices[i] = newMesh->VertexPosition(newMesh->Vertex(i));
		pnts.Insert(&(vertices[i]));
	}
	R3Kdtree<R3Point*> kdtree(pnts);
	
		// fill minimal distance from edge centroid to a vertex
	for (int e=0; e<newMesh->NEdges(); e++)
	{
		R3MeshEdge * edge = newMesh->Edge(e);
		if (newMesh->FaceOnEdge(edge, 0)!=NULL || newMesh->FaceOnEdge(edge, 1)!=NULL)
		{
			assert(newMesh->FaceOnEdge(edge, 0) != newMesh->FaceOnEdge(edge, 1));
			R3Point p = newMesh->EdgeMidpoint(edge);
			R3Point * closest = kdtree.FindClosest(p);
			assert(closest!=NULL);
			minDistances.push_back(Sortable(R3Distance(p, *closest), edge));
		}
	}
	std::sort(minDistances.begin(), minDistances.end());
	
	std::vector<R3Point> addedPoints;
//	std::vector<R3MeshVertex*> addedVertices;
	// add approxPnts - farthest centroids
	while ((int)addedPoints.size() < approxPnts)
	{	
		R3MeshEdge * farthestEdge = (R3MeshEdge*)minDistances[(int)minDistances.size()-1].ptr;
		*minDist = vkMin(minDistances[(int)minDistances.size()-1].value, *minDist);
		int edgeID = newMesh->EdgeID(farthestEdge);
		assert(edgeID>=0 && edgeID<newMesh->NEdges());
		R3Point newPnt = newMesh->EdgeMidpoint(farthestEdge);

		// subdivide face (adjacent to farthest edge)
		R3MeshFace * f1 = newMesh->FaceOnEdge(farthestEdge, 0);
		R3MeshFace * f2 = newMesh->FaceOnEdge(farthestEdge, 1);
		if ((f1==NULL)												// if only one face exist - pick it
			|| (f1!=NULL && f2!=NULL								// if both faces exist
				&& newMesh->FaceArea(f1) < newMesh->FaceArea(f2)))	//		subdivide face with larger area
			f1 = f2;
		assert(f1!=NULL);

		std::vector<R3MeshEdge *> delEdges;
		std::vector<R3MeshEdge *> newEdges;
		
		// add old edges (to be deleted by the subdivision)
		for (int eonf=0; eonf<3; eonf++)
			delEdges.push_back(newMesh->EdgeOnFace(f1, eonf));
		
		std::vector<R3MeshFace*> newFaces;
		std::vector<R3MeshFace*> delFaces;
		SubdivideFace(newMesh, f1, &newFaces, &delFaces);
		//if (newFaces.size()!=4)
		//	std::cout<<"[ERROR] N new Faces = "<<newFaces.size()<<std::endl;
		
		//assert(newFaces.size()>0);
		assert(newFaces.size()==4);
		R3MeshFace * centerFace = newFaces[newFaces.size()-1];
		
//		R3MeshFace * centerFace = newMesh->SubdivideFace(f1);
		assert(centerFace!=NULL);

		// add new edges to a list
		for (int vonf=0; vonf<3; vonf++)
		{
			R3MeshVertex * vtx = newMesh->VertexOnFace(centerFace, vonf);
			//addedVertices.push_back(vtx);

			for (int eonv=0; eonv < newMesh->VertexValence(vtx); eonv++)
			{
				R3MeshEdge * edge = newMesh->EdgeOnVertex(vtx, eonv);
				R3MeshVertex * vtx2 = newMesh->VertexAcrossEdge(edge, vtx);
				if (vonf==0
					|| (vonf==1 && vtx2 != newMesh->VertexOnFace(centerFace, 0))
					|| (vonf==2 && vtx2 != newMesh->VertexOnFace(centerFace, 0)
						&&  vtx2 != newMesh->VertexOnFace(centerFace, 1)))
					newEdges.push_back(edge);
			}
		}
		
		// for new edges - calculate distances to existing points
		for (int ne=0; ne<(int)newEdges.size(); ne++)
		{
			R3MeshEdge * edge = newEdges[ne];
			R3Point p = newMesh->EdgeMidpoint(edge);
			R3Point * closest = kdtree.FindClosest(p);
			assert(closest!=NULL);
			double dist = R3Distance(p, *closest);
			
			for (int nv = 0; nv < (int)addedPoints.size(); nv++)
				dist = vkMin(dist, R3Distance(p, addedPoints[nv]));
			
			minDistances.push_back(Sortable(dist, edge));
			
		}

		// update centroids (min distances)
		for (int e=0; e<(int)minDistances.size(); e++)
		{
			bool deleted = false;
			for (int de=0; de<(int)delEdges.size() && !deleted; de++)
				if (minDistances[e].ptr==delEdges[de])
					deleted = true;
			if (deleted)	// if deleted - remove from the list
			{
				minDistances.erase(minDistances.begin()+e);
				e--;
			}
			else			// update min with newpoint
			{
				R3MeshEdge * edge = (R3MeshEdge*)minDistances[e].ptr;
				R3Point p = newMesh->EdgeMidpoint(edge);
				minDistances[e].value = vkMin(minDistances[e].value, R3Distance(p, newPnt));
			}
		}

		std::sort(minDistances.begin(), minDistances.end());
		addedPoints.push_back(newPnt);
	}
	
	delete [] vertices;
	
	return newMesh;
}

R3Mesh * Model3D::SubdivideMeshLargestAreaFace(R3Mesh * origMesh, int approxPnts)
{
	R3Mesh * newMesh = CloneMesh(origMesh);
	
	bool slow = false;
	double nMinRatio = 10;		
	if (slow)
	{
		int nInserted = 0;
		double largestToSmallestRatio = 1000;
		TimeProfiler timer;
		while ((newMesh->NVertices() < approxPnts || largestToSmallestRatio > nMinRatio)
				&& nInserted < approxPnts)
		{
			nInserted++;
			double minArea = FLT_MAX;
			double maxArea = 0;
			R3MeshFace * maxAreaFace = NULL;
			for (int f=0; f<newMesh->NFaces(); f++)
			{
				R3MeshFace * face= newMesh->Face(f);
				double area = newMesh->FaceArea(face);
				if (area>0)
					minArea = vkMin(area, minArea);
				if (area > maxArea)
				{
					maxArea = area;
					maxAreaFace = face;
				}
			}
			assert(maxAreaFace!=NULL);
			assert(minArea>0);
			largestToSmallestRatio = maxArea / minArea;

			if (largestToSmallestRatio > nMinRatio)
				SubdivideFace(newMesh, maxAreaFace, NULL, NULL);
		}
	}
	else
	{	
		std::vector<Sortable> facesByArea;
		for (int f=0; f<newMesh->NFaces(); f++)
		{
			R3MeshFace * face= newMesh->Face(f);
			double area = newMesh->FaceArea(face);
			if (area>0)
				facesByArea.push_back(Sortable(area, face));
			
			for (int fof=0; fof<3; fof++)
				assert(newMesh->EdgeOnFace(face, fof)!=NULL);
		}
		assert(facesByArea.size()>0);
		std::sort(facesByArea.begin(), facesByArea.end());

		std::map<void *, bool> deletedFaces;
		
		int nInserted = 0;
		double largest = facesByArea[(int)facesByArea.size()-1].value;
		double smallest = facesByArea[0].value;
		double largestToSmallestRatio = largest / smallest;
		
		// Insert up to approxPnts until ensured that:
		//		1. Model has at least approxPnts
		//		2. LaregstArea  / Smallest Area < 2
		while ((newMesh->NVertices() < approxPnts || largestToSmallestRatio > nMinRatio) 
			   && nInserted < approxPnts)
		{
			Sortable largestFace = facesByArea[(int)facesByArea.size()-1];
		
			// check if face was deleted
			std::map<void *, bool>::iterator iter = deletedFaces.find(largestFace.ptr);
			assert(largestFace.ptr!=NULL);
			if (iter == deletedFaces.end())		// NOT DELETED: subdivide the face
			{
				nInserted+=3;
				int currNFaces = newMesh->NFaces();
				R3MeshFace * currFace = (R3MeshFace*)largestFace.ptr;
				assert(currFace!=NULL);
				std::vector<R3MeshFace*> newFaces;
				std::vector<R3MeshFace*> delFaces;
				
				SubdivideFace(newMesh, currFace, &newFaces, &delFaces);
				
				for (int i=0; i<(int)delFaces.size(); i++)
					if (delFaces[i]!=currFace)
						deletedFaces[delFaces[i]] = true;
				
				for (int i=0; i<(int)newFaces.size(); i++)
					facesByArea.push_back(Sortable(newMesh->FaceArea(newFaces[i]), newFaces[i]));
				std::sort(facesByArea.begin(), facesByArea.end());	// TODO: do quick insert!
			}
			else	// DELETED: never use it for lookup again
				deletedFaces.erase(iter);

			facesByArea.erase(facesByArea.begin() + ((int)facesByArea.size()-1));
		}
	}
	return newMesh;
}

R3Mesh * Model3D::SubdivideMeshLongestEdge(R3Mesh * origMesh, int approxPnts)
{
	R3Mesh * newMesh = CloneMesh(origMesh);
	if (newMesh->NVertices() < approxPnts 
		&& newMesh->NVertices()!=0 && newMesh->NEdges()!=0)
	{
		std::vector<Sortable> edgesWithLength;
		for (int i=0; i<newMesh->NEdges(); i++)
		{
			R3MeshEdge * edge = newMesh->Edge(i);
			edgesWithLength.push_back(Sortable(newMesh->EdgeLength(edge), edge, -1));
		}
		
		while(newMesh->NVertices() < approxPnts)
		{
			std::sort(edgesWithLength.begin(), edgesWithLength.end());
			
			R3MeshEdge * edge = (R3MeshEdge*)edgesWithLength[edgesWithLength.size()-1].ptr;
			R3MeshVertex * vertex = newMesh->SubdivideEdge(edge);
			edgesWithLength.erase(edgesWithLength.begin()+(edgesWithLength.size()-1));
			for (int i=0; i<newMesh->VertexValence(vertex); i++)
			{
				R3MeshEdge * edge = newMesh->EdgeOnVertex(vertex, i);
				edgesWithLength.push_back(Sortable(newMesh->EdgeLength(edge), edge, -1));
			}
		}
	}
	return newMesh;
}

void Model3D::FreeMemory()
{
	// TODO: remove point sets / point clouds
	
	if (m_pMesh!=NULL)
	{
		delete m_pMesh;
		m_pMesh = NULL;
	}
}

Model3D * Model3D::CreateMyCopy()
{
    // TODO: copy
    assert(m_PointSets.size()==0);
    assert(m_ModelValues.size()==0);
    assert(m_TempPointSets.size()==0);
    assert(m_TempModelValues.size()==0);
    assert(m_ModelDistances.size()==0);
    assert(m_ShapeDescriptors.size()==0);
    
    int cpID = 1;
    VKString copyModelName = m_sModelName + "_cp" + VKString::number(cpID);
    while (m_pDB->GetModelByName(copyModelName)!=NULL)
        copyModelName = m_sModelName + "_cp" + VKString::number(++cpID);
    Model3D * newModel = new Model3D(m_sFilename, copyModelName, 
                                     m_iApproximationPnts, m_pDB);
    if (m_pMesh!=NULL)
        newModel->m_pMesh = CloneMesh(m_pMesh);
    newModel->InitRenderingParams(m_pRenderingParams->m_VisParams,
                                  m_pRenderingParams->m_pWindow);
    
    newModel->m_GeomType = m_GeomType;
    return newModel;
}

Model3D * Model3D::SplitModel(std::vector<int> & facesInNewModel)
{
    std::vector<bool> isFaceInNew;
    int nFaces = m_pMesh->NFaces();
    for (int i=0; i<nFaces; i++)
        isFaceInNew.push_back(false);
    
    for (int i=0; i<(int)facesInNewModel.size(); i++)
        isFaceInNew[facesInNewModel[i]] = true;
    
    std::vector<int> facesInOldModel;
    for (int i=0; i<nFaces; i++)
        if (!isFaceInNew[i])
            facesInOldModel.push_back(i);

    Model3D * newModel = CreateMyCopy();

    std::sort(facesInOldModel.begin(), facesInOldModel.end());
    for (int i=(int)facesInOldModel.size()-1; i>=0; i--)
        newModel->RemoveFaceOnMesh(facesInOldModel[i]);
    
    std::sort(facesInNewModel.begin(), facesInNewModel.end());
    for (int i=(int)facesInNewModel.size()-1; i>=0; i--)
        RemoveFaceOnMesh(facesInNewModel[i]);

    return newModel;
}


void Model3D::RemoveFaceOnMesh(int faceID)
{
    assert(faceID < m_pMesh->NFaces());
    R3MeshFace * face = m_pMesh->Face(faceID);
    R3MeshEdge * e1 = m_pMesh->EdgeOnFace(face, 0);
    R3MeshEdge * e2 = m_pMesh->EdgeOnFace(face, 1);
    R3MeshEdge * e3 = m_pMesh->EdgeOnFace(face, 2);
    R3MeshVertex * v1 = m_pMesh->VertexOnFace(face, 0);
    R3MeshVertex * v2 = m_pMesh->VertexOnFace(face, 1);
    R3MeshVertex * v3 = m_pMesh->VertexOnFace(face, 2);
    
    m_pMesh->DeleteFace(face);
    
    if (m_pMesh->FaceOnEdge(e1) == NULL)
        m_pMesh->DeleteEdge(e1);
    if (m_pMesh->FaceOnEdge(e2) == NULL)
        m_pMesh->DeleteEdge(e2);
    if (m_pMesh->FaceOnEdge(e3) == NULL)
        m_pMesh->DeleteEdge(e3);
    
    if (m_pMesh->VertexValence(v1)==0)
        m_pMesh->DeleteVertex(v1);
    if (m_pMesh->VertexValence(v2)==0)
        m_pMesh->DeleteVertex(v2);
    if (m_pMesh->VertexValence(v3)==0)
        m_pMesh->DeleteVertex(v3);
    
    if (m_pFaceCentroids!=NULL)
        delete [] m_pFaceCentroids;
    if (m_pKdTreeFaceCentroids!=NULL)
        delete m_pKdTreeFaceCentroids;
    if (m_pSearchTree!=NULL)
        delete m_pSearchTree;
    
    for (std::map<VKString, FaceNhdStruct *>::iterator iter = m_mFaceNhds.begin(); 
         iter != m_mFaceNhds.end(); iter++)
        delete iter->second;
    
    m_mFaceNhds.clear();
    m_ConnectedComponents.clear();
    m_FaceCCID.clear();
    m_FaceNeighbors.clear();
    
    // TODO: update
    assert(m_PointSets.size()==0);
    assert(m_ModelValues.size()==0);
    assert(m_TempPointSets.size()==0);
    assert(m_TempModelValues.size()==0);
    assert(m_ModelDistances.size()==0);
    assert(m_ShapeDescriptors.size()==0);
}

void Model3D::FillCCFaces(int seedFace, std::vector<int> & saveFaceIDs)
{
    FillConnectedComponentsFaces(false);
    assert(seedFace >= 0 && seedFace < (int)m_FaceCCID.size());
    int cID = m_FaceCCID[seedFace];
    assert(cID>=0 && cID < m_ConnectedComponents.size());
    for (int i=0; i<(int)m_ConnectedComponents[cID].size(); i++)
        saveFaceIDs.push_back(m_ConnectedComponents[cID][i]);
}

// rendering
void Model3D::Draw(int subwindowID)
{
	//std::cout<<"Drawing - 1"<<std::endl;
	//bool erasePrecomputed = true;
	m_pRenderingParams->LoadMyCamera(subwindowID);
	
    if (m_bFastIntersectionBuffer)
        m_pRenderingParams->DrawIntersectInfoToFramebuffer(subwindowID);
    
	VisWindowRenderer * renderer = m_pRenderingParams->m_pWindow->GetRenderer();
	
	if (m_pMesh==NULL || m_pMesh->NVertices()==0)
		renderer->RenderText(R3Point(0,0,0), "Invalid Model", 
							 0, 0, 1, VisWindowRenderer::TEXT_SIZE_LARGE);

	// draw MODEL FACES
	if (RenderingParams()->GetPrecomputed()==NULL &&
		m_pRenderingParams->IsFlagOn(Model3DRenderingParams::RENDER_FACES))
	{
		R3Mesh * mesh = GetMeshForRigidMatch(false);
		if (mesh!=NULL)
		{
			glEnable(GL_LIGHTING);
			glDisable(GL_CULL_FACE);

			static GLfloat surfaceMaterial[4];
			RenderingParams()->m_pWindow->StandardSurfaceColor(surfaceMaterial);
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
			mesh->Draw();
		}
	}
	
	// draw MODEL EDGES
	if (m_pRenderingParams->IsFlagOn(Model3DRenderingParams::RENDER_EDGES))
	{
		if (m_pMesh!=NULL)
		{
			glDisable(GL_LIGHTING);
			glDisable(GL_CULL_FACE);

			glColor3d(1., .5, .5);
			m_pMesh->DrawEdges();
		}
	}
	
	// draw POINT SET
	if (m_pRenderingParams->IsFlagOn(Model3DRenderingParams::RENDER_POINT_SET))
	{
		PointSet * pntSet = GetPointSet(m_pRenderingParams->GetPointSetToRender());
		int descriptorID = m_pRenderingParams->m_pWindow->m_iPointDescriptorToRender;
		VKString descriptorName = "none";
		if (descriptorID>=0)
			descriptorName = m_pDB->m_PointDescriptorNames[descriptorID];
		
		//std::cout<<"Drawing pnts: "<<ModelName().c_str()<<std::endl;
		if (pntSet!=NULL)
			pntSet->Draw(RenderingParams()->m_SelectedPnt, 
						 m_pRenderingParams->m_pWindow, descriptorName, this);
		//std::cout<<"Drawing pnts: "<<ModelName().c_str()<<" - DONE!"<<std::endl;
	}
	
	// draw MODEL VALUE
	//std::cout<<"Draw Model Value ? "<<m_pRenderingParams->IsFlagOn(Model3DRenderingParams::RENDER_VALUE)<<std::endl;
	if (m_pRenderingParams->IsFlagOn(Model3DRenderingParams::RENDER_VALUE))
	{
		//std::cout<<"ModelName: "<<m_sModelName.c_str()<<std::endl;
		//std::cout<<"ModelValue: "<<m_pRenderingParams->GetModelValueToRender().c_str()<<std::endl;
		ModelValue * modelValue = GetModelValue(m_pRenderingParams->GetModelValueToRender());
		
		if (modelValue!=NULL)
			modelValue->DrawAsMesh(m_pRenderingParams->m_pWindow);
	}
	
//	// draw SELECTED point
//	if (!RenderingParams()->m_SelectedPnt.Invalid())
//	{
//		static GLfloat bluePnt[4];
//		bluePnt[0] = 0;
//		bluePnt[1] = 0;
//		bluePnt[2] = 1.;
//		bluePnt[3] = 1.;
//		
//		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, bluePnt); 
//		
//		R3Point p = RenderingParams()->m_SelectedPnt.GetPosition();
//		glEnable(GL_LIGHTING);
//		R3Sphere(p, RenderingParams()->GetStandardRadius()).Draw();
//	}
	
	// draw / erase PRECOMPUTED colors (due to ModelValue, CorrValue, etc...)
//	if (erasePrecomputed)
//		RenderingParams()->FreePrecomputedColors();
	
	if (RenderingParams()->GetPrecomputed()!=NULL
		&& (!m_pRenderingParams->IsFlagOn(Model3DRenderingParams::RENDER_POINT_SET)
			 || m_pRenderingParams->IsFlagOn(Model3DRenderingParams::RENDER_FACES)))
		RenderingParams()->GetPrecomputed()->Draw(RenderingParams()->m_pWindow);	
	
	m_pRenderingParams->RestoreCameraAfterLoad(subwindowID);
	
	//std::cout<<"Drawing - 2"<<std::endl;
}

Model3DRenderingParams * Model3D::RenderingParams()
{
	return m_pRenderingParams;
}

Model3D::GeometryType Model3D::GetGeometryType()
{
	return m_GeomType;
}

// 
const VKString & Model3D::ModelName() const
{
	return m_sModelName;
}

const VKString & Model3D::ModelFile() const
{
	return m_sFilename;
}

const VKString & Model3D::ModelGeometryFile() const
{
	return m_sGeometryFile;
}

int Model3D::ModelID() const
{
	return m_iID;
}

void Model3D::UpdateID(int newID)
{
    m_iID = newID;
}

bool Model3D::InMemory()
{
	return (m_pMesh!=NULL);
}

R3Box Model3D::BBox()
{
	if (m_pMesh!=NULL)
    {
		R3Point p0 = m_pMesh->VertexPosition(m_pMesh->Vertex(0));
        R3Box box(p0, p0);
        for (int i=0; i<m_pMesh->NVertices(); i++)
			box.Union(m_pMesh->VertexPosition(m_pMesh->Vertex(i)));
        return box;
    }
	else 
		return R3Box(R3Point(0,0,0), R3Point(1,1,1));
}

VKString Model3D::StandardTempStrPrefix()
{
	return "$^@(`TEMP_DATA_";
}

PointSet * Model3D::PointCloud()
{
	PointSet * retVal = GetPointSet("ConnectedPointCloud");
	if (retVal==NULL)
	{
		std::cout<<"[ERROR] Cannot find point cloud in model: "<<ModelName().c_str()<<std::endl;
		assert(retVal!=NULL);
	}
	return retVal;
}

ModelPoint Model3D::CheckIfPointIsFromTheCloud(const ModelPoint & point)
{
	PointSet * pntSet = PointCloud();
	assert(pntSet!=NULL);
	int setID = -1;
	ModelPoint retVal = point;
	if (pntSet->ContainsExact(point, &setID))
		retVal.SetPointCloudID(setID);

	return retVal;
}

PointSet * Model3D::GetPointSet(const VKString & pointSetName)
{
	if (pointSetName.startsWith(StandardTempStrPrefix().c_str()))
	{
		VKString copy = pointSetName;
		bool ok;
		int tempID = copy.replace(StandardTempStrPrefix(), "").toInt(&ok);
		assert(ok);
		std::map<int, PointSet * >::iterator iter = m_TempPointSets.find(tempID);
		if (iter!=m_TempPointSets.end())
			return iter->second;
	}
	else 
	{
		std::map<VKString, PointSet * >::iterator iter = m_PointSets.find(pointSetName);
		if (iter!=m_PointSets.end())
			return iter->second;
	}
	return NULL;
}

void Model3D::AddPointSet(const VKString & pntSetName, PointSet * pointSet)
{
	assert(m_PointSets.find(pntSetName)==m_PointSets.end());
	m_PointSets[pntSetName] = pointSet;
	
	for (int i=0; i<pointSet->NumPoints(); i++)
	{
		if (pointSet->GetPoint(i).GetModel() != this)
		{
			std::cout<<"\n[ERROR]Model "<<ModelName().c_str()<<" with id "<<ModelID()<<std::endl;;
			std::cout<<"\tPoint "<<i<<" thinks it belongs to ";
			std::cout<<((Model3D*)pointSet->GetPoint(i).GetModel())->ModelName().c_str()<<std::endl;
			std::cout<<"\tSet="<<pntSetName.c_str()<<std::endl;
			assert(pointSet->GetPoint(i).GetModel() == this);
		}
	}
}

VKString Model3D::AddTemporaryPointSet(PointSet * pointSet)
{
	assert(m_TempPointSets.find(m_iIDTempPointSet)==m_TempPointSets.end());
	m_TempPointSets[m_iIDTempPointSet] = pointSet;
	m_iIDTempPointSet++;
	return StandardTempStrPrefix() + VKString::number(m_iIDTempPointSet-1);
}

void Model3D::RemovePointSet(const VKString & pointSetName)
{
	if (pointSetName.startsWith(StandardTempStrPrefix().c_str()))
	{
		VKString copy = pointSetName;
		bool ok;
		int tempID = copy.replace(StandardTempStrPrefix(), "").toInt(&ok);
		assert(ok);
		std::map<int, PointSet * >::iterator iter = m_TempPointSets.find(tempID);
		assert (iter!=m_TempPointSets.end());
		delete iter->second;
		m_TempPointSets.erase(iter);
	}
	else 
	{
		std::map<VKString, PointSet * >::iterator iter = m_PointSets.find(pointSetName);
		assert (iter!=m_PointSets.end());
		delete iter->second;
		m_PointSets.erase(iter);
	}
}

// Model Values
ModelValue * Model3D::GetModelValue(const VKString & modelValueName)
{
	if (modelValueName.startsWith(StandardTempStrPrefix().c_str()))
	{
		VKString copy = modelValueName;
		bool ok;
		int tempID = copy.replace(StandardTempStrPrefix(), "").toInt(&ok);
		assert(ok);
		std::map<int, ModelValue * >::iterator iter = m_TempModelValues.find(tempID);
		if (iter!=m_TempModelValues.end())
			return iter->second;
	}
	else 
	{
		std::map<VKString, ModelValue * >::iterator iter = m_ModelValues.find(modelValueName);
		if (iter!=m_ModelValues.end())
			return iter->second;
	}
	return NULL;
}

void Model3D::AddModelValue(const VKString & modelValueName, ModelValue * modelValue)
{
	if (!m_pDB->m_ModelValueNames.contains(modelValueName))
		m_pDB->m_ModelValueNames.push_back(modelValueName);
	
	assert(m_ModelValues.find(modelValueName)==m_ModelValues.end());
	m_ModelValues[modelValueName] =  modelValue;
}

VKString Model3D::AddTemporaryModelValue(ModelValue * modelValue)
{
	assert(m_TempModelValues.find(m_iIDTempPointSet)==m_TempModelValues.end());
	m_TempModelValues[m_iIDTempPointSet] = modelValue;
	m_iIDTempModelValue++;
	return StandardTempStrPrefix() + VKString::number(m_iIDTempPointSet-1);
	
}

void Model3D::RemoveModelValue(const VKString & modelValueName)
{
	if (modelValueName.startsWith(StandardTempStrPrefix().c_str()))
	{
		VKString copy = modelValueName;
		bool ok;
		int tempID = copy.replace(StandardTempStrPrefix(), "").toInt(&ok);
		assert(ok);
		std::map<int, ModelValue * >::iterator iter = m_TempModelValues.find(tempID);
		assert (iter!=m_TempModelValues.end());
		delete iter->second;
		m_TempModelValues.erase(iter);
	}
	else 
	{
		std::map<VKString, ModelValue * >::iterator iter = m_ModelValues.find(modelValueName);
		assert (iter!=m_ModelValues.end());
		delete iter->second;
		m_ModelValues.erase(iter);
	}
}

//void Model3D::AddInterpolator(const VKString & interpName, CollectionIdentityPointsets * interp)
//{
//	m_Interpolators[interpName] = interp;
//}
//
//CollectionIdentityPointsets * Model3D::GetInterpolator(const VKString & interpName)
//{
//	std::map<VKString, CollectionIdentityPointsets*>::iterator iter = m_Interpolators.find(interpName);
//	if (iter!=m_Interpolators.end())
//		return iter->second;
//	return NULL;
//}

CollectionIdentityPointsets * Model3D::InterpolatorMap(const VKString & mapName)
{
	CorrCollection * coll = m_pDB->GetCorrCollection(mapName);
	if (coll==NULL)
	{
		std::cout<<"[ERROR] Interpolating map not found: "<<mapName.c_str()<<std::endl;
		assert(coll!=NULL);
	}
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	CollectionSinglePair * singlePair = collPairwise->GetModelCorr(ModelID());
	assert(singlePair!=NULL);
	assert(singlePair->ModelFrom()==this && singlePair->ModelTo()==this);
	assert(singlePair->GetCollectionType()=="CollectionIdentityPointsets");
	return (CollectionIdentityPointsets*)singlePair;
}

CollectionIdentityPointsets * Model3D::VerticesFromPointsInterpolator()
{
	VKString collName = CollectionIdentityPointsets::DefaultVerticesToPointCloudMap();
	return InterpolatorMap(collName);
}

CollectionIdentityPointsets * Model3D::FacesFromPointsInterpolator()
{
	VKString collName = CollectionIdentityPointsets::DefaultFacesToPointCloudMap();
	return InterpolatorMap(collName);
}

CollectionIdentityPointsets * Model3D::PointsToFacesMap()
{
	VKString collName = CollectionIdentityPointsets::DefaultFacesToPointCloudMap();
	return InterpolatorMap(collName+"Inverse");
}

CollectionIdentityPointsets * Model3D::PointsToVerticesMap()
{
	VKString collName = CollectionIdentityPointsets::DefaultVerticesToPointCloudMap();
	return InterpolatorMap(collName+"Inverse");
}


// Distances
DistanceEuclidean * Model3D::DefaultEuclideanDistance()
{
	for (std::map<VKString, ModelDistance * >::iterator iter = m_ModelDistances.begin();
		 iter != m_ModelDistances.end(); iter++)
	{
		if (iter->second->DistanceType() == "DistanceEuclidean")
			return (DistanceEuclidean*)iter->second;
	}
	return NULL;
}

ModelDistance * Model3D::DefaultGeodesicDistance()
{
	for (std::map<VKString, ModelDistance * >::iterator iter = m_ModelDistances.begin();
		 iter != m_ModelDistances.end(); iter++)
	{
		if (iter->second->DistanceType() == "DistanceGeodesic")
			return (DistanceGeodesic*)iter->second;
	}
	return NULL;
}

ModelDistance * Model3D::GetModelDistance(const VKString & modelDistanceMetric)
{
	std::map<VKString, ModelDistance * >::iterator iter;
	iter = m_ModelDistances.find(modelDistanceMetric);
	if (iter!=m_ModelDistances.end())
		return iter->second;
	return NULL;
}

void Model3D::AddModelDistance(const VKString & modelDistanceName, ModelDistance * modelDistance)
{
	assert(m_ModelDistances.find(modelDistanceName)==m_ModelDistances.end());
	m_ModelDistances[modelDistanceName] = modelDistance;
}

R3MeshSearchTree * Model3D::GetSearchTree()
{
	if (m_pSearchTree==NULL)
		m_pSearchTree = new R3MeshSearchTree((R3Mesh*)GetMesh());

	return m_pSearchTree;
}

// shape descriptors
ShapeDescriptor * Model3D::GetShapeDescriptor(const VKString & descriptorName)
{
	std::map<VKString, ShapeDescriptor * >::iterator iter = m_ShapeDescriptors.find(descriptorName);
	if (iter==m_ShapeDescriptors.end())
		return NULL;
	return iter->second;
}

void Model3D::AddShapeDescriptor(const VKString & descriptorName, ShapeDescriptor * descriptor)
{
	assert(m_ShapeDescriptors.find(descriptorName)==m_ShapeDescriptors.end());
	m_ShapeDescriptors[descriptorName] = descriptor;
}


/******************* READ COLLADA *****************/
#ifdef OPEN_COLLADA
void Model3D::ReadFromCOLLADA()
{
	//std::cout<<"Trying to read COLLADA"<<std::endl;
	// fill m_ModelValuesCameWithExternalFile
	
	Model3DColladaParser::LoadColladaFile(this);
	assert(m_pMesh!=NULL);
	//NormalizeScale();
}
#endif

/******************* READ SKETCHUP *****************/
#ifdef GOOGLE_SKETCHUP
void Model3D::ReadFromSketchup()
{
	ISkpApplication* ppApp = NULL;
	HRESULT hr = GetSketchUpSkpApplication(&ppApp);
	assert (hr==S_OK && ppApp!=NULL);

	ISkpFileReader * fileReader = NULL;	
	assert(ppApp!=NULL);
	hr = ppApp->QueryInterface(IID_ISkpFileReader, (void **)&fileReader);
	assert (hr==S_OK && fileReader!=NULL);
	
	ISkpDocument * skpDocument=NULL;
	//BSTR  filename = SysAllocString("../../../Data/GoogleWarehouse/chair/chair5.skp");
	BSTR  filename = SysAllocString(m_sFilename.c_str());
	hr = fileReader->OpenFile(filename, &skpDocument);
	if (hr!=S_OK || skpDocument==NULL)
	{
		switch(hr)
		{
			case E_SKP_SKETCHUP_MODEL_INVALID:
				std::cout<<"[ERROR] Invalid sketchup model."<<std::endl;
				break;
			case E_SKP_SKETCHUP_MODEL_MORE_RECENT:
				std::cout<<"[ERROR] Invalid sketchup model version."<<std::endl;
				break;
			case E_SKP_FILE_NOT_FOUND:
				std::cout<<"[ERROR] Sketchup file not found."<<std::endl;
				break;
			default:
				std::cout<<"[ERROR] Loading sketchup model. errno="<<hr<<std::endl;
		}
		assert(hr==S_OK && skpDocument!=NULL);
	}
}
#endif

void Model3D::ReadFromPSB()
{
	ParamParser parser(m_sFilename);
	bool valid;
	VKString geomType = parser.GetStrValue("ModelPSB", "GeometryType", valid);
	if (valid)
		m_GeomType = GeomTypeStrToType(geomType);
	else
		m_GeomType = IRREGULAR_POLYGON_SOUP;
		
	int lastSlash = m_sFilename.lastIndexOf("/");
	VKString dirPrefix="";
	if (lastSlash!=-1)
	{
		dirPrefix = m_sFilename;
		dirPrefix = dirPrefix.left(lastSlash) + "/";
	}
	
	// create geometry
	VKString geomFile = parser.GetStrValue("ModelPSB", "Geometry");
	geomFile = dirPrefix + geomFile;
	m_pMesh = new R3Mesh();
	m_pMesh->ReadFile(geomFile.c_str());
	//NormalizeScale();

	// create segmentations
	VKString segFile  = parser.GetStrValue("ModelPSB", "Segments", valid);
	if (valid && segFile!="none")
	{
		segFile = dirPrefix + segFile;
		VKStringList segIDToLabel;
		std::ifstream segFileStream(segFile.c_str());
		if (!segFileStream.is_open())
		{
			std::cout<<"[ERROR] Cannot open segfile: "<<segFile.c_str()<<std::endl;
			assert(segFileStream.is_open());
		}
		while (!segFileStream.eof())
		{
			VKString label;
			int id;
			label.readToken(segFileStream);
			segFileStream>>id;
			if (segFileStream.eof())
				break;
			assert(id < 10000);	/// sanity check
			while (id >= segIDToLabel.count())
				segIDToLabel.push_back(VKString("UnlabeledSeg") + VKString::number(id));
			
			label.replace("-", "");
			for (int i=0; i<=9; i++)
				label.replace(VKString::number(i), "");		
			
			segIDToLabel[id]=label;
		}
		segFileStream.close();
		
		VKString labelingName = "PSBLabels";
		ModelValueLabel * labels = new ModelValueLabel(this, labelingName);
		
		for (int i=0; i<segIDToLabel.count(); i++)
			labels->AssignLabel(i, segIDToLabel[i]);
		
		// iterate through vertices
		for (int i=0; i<(int)m_pMesh->NVertices(); i++)
		{
			R3MeshVertex * v = m_pMesh->Vertex(i);
			int segmentID = -1;
			if (m_pMesh->VertexValence(v)>0)
			{
				R3MeshFace * face = m_pMesh->FaceOnVertex(v);
			
				if (face!=NULL)
					segmentID = m_pMesh->FaceSegment(face);
			}
				
			labels->AssignSegmentID(ModelPoint(i, this), segmentID);
		}

		// iterate through faces
		for (int i=0; i<(int)m_pMesh->NFaces(); i++)
			labels->AssignSegmentIDToFace(i, m_pMesh->FaceSegment(m_pMesh->Face(i)));
		
		AddModelValue(labelingName, labels);
		m_ModelValuesCameWithExternalFile.push_back(labelingName);
	}
	
	// create ground truth: correspondences
	VKStringList corrSets  = parser.GetStrValues("ModelPSB", "CorrPntSets", valid);
	if (valid && corrSets.count()>0)
	{
		VKStringList corrGroups  = parser.GetStrValues("ModelPSB", "CorrGroups", valid);
		assert(valid && corrGroups.count()==corrSets.count());

		CollectionGroundTruth * gtMap = NULL;		
		VKString gtTrueName = CollectionGroundTruth::DefaultGroundTruthMapName();
		m_CorrCollectionsCameWithExternalFile.push_back(gtTrueName);
		CorrCollection * coll = m_pDB->GetCorrCollection(gtTrueName);

		if (coll==NULL)
		{
			gtMap = new CollectionGroundTruth(gtTrueName, m_pDB);
			m_pDB->AddCorrCollection(gtTrueName, gtMap);
		}
		else
		{
			assert(coll->GetCollectionType()=="CollectionGroundTruth");
			gtMap = (CollectionGroundTruth*)coll;
		}
		// create or find ground truth map
		assert(gtMap!=NULL);
		
		for (int i=0; i<corrSets.count(); i++)
		{
			
			VKString pntSetName = corrSets[i];
			VKString classSetName = corrGroups[i];
			
			if (pntSetName!="AllVertices"
				&& pntSetName!="AllFaces")	// if not exception: read point set
			{
				VKString pntSetFile = dirPrefix + pntSetName;
				pntSetName = classSetName+"_FeatPnts";
				m_PntSetsCameWithExternalFile.push_back(pntSetName);
				PointSet * newPnts = new PointSet(pntSetName);
				newPnts->InitFromExternal(this, pntSetFile);
				AddPointSet(pntSetName, newPnts);
			}
			// add point set / class name to collection
			gtMap->AddModelFeaturePntsGT(this, pntSetName, classSetName);
		}
	}

	
	// TODO: create ground truth: symmetries
//	VKString symFile  = parser.GetStrValue("ModelPSB", "Symm", valid);
//	if (valid && symFile!="none")
//		symFile = "none";
//	

	// load camera: Camera, CameraLine
	VKString camera  = parser.GetStrValue("ModelPSB", "Camera", valid);
	if (valid && camera!="none")
	{
		camera = dirPrefix + camera;
		VKString camLine  = parser.GetStrValue("ModelPSB", "CameraLine");
		bool ok;
		int cLine = camLine.toInt(&ok);
		assert(ok);
		
		std::ifstream textStream(camera.c_str());
		assert(textStream.is_open());
		VKString cam;
		cam.readLine(textStream);
		for (int i=0; i<cLine; i++)
			cam.readLine(textStream);
		m_MyPreferableCamera = cam.split(" ");
		textStream.close();
	}
	
	m_sBIMQueryTriplet = parser.GetStrValues("ModelPSB", "QueryTriplet", valid);
}

R3Mesh * Model3D::CloneMesh(const R3Mesh* mesh)
{
	R3Mesh * newMesh = new R3Mesh();
	for (int i=0; i<mesh->NVertices(); i++)
	{
		const R3MeshVertex * vtx = mesh->Vertex(i);
		R3Point p = mesh->VertexPosition(vtx);
		R3Vector n = mesh->VertexNormal(vtx);
		newMesh->CreateVertex(p, n);
	}
	
	for (int i=0; i<mesh->NFaces(); i++)
	{
		const R3MeshFace * face = mesh->Face(i);
		
		newMesh->CreateFace(newMesh->Vertex(mesh->VertexID(mesh->VertexOnFace(face, 0))),
							newMesh->Vertex(mesh->VertexID(mesh->VertexOnFace(face, 1))),
							newMesh->Vertex(mesh->VertexID(mesh->VertexOnFace(face, 2))));
	}
	
	return newMesh;
}

void Model3D::FastIntersectionBuffer(bool enable)
{
    m_bFastIntersectionBuffer = enable;
}

// CC processing
int Model3D::FillConnectedComponentsFaces(bool createLabeling)
{
	if (m_ConnectedComponents.size()>0)	// assume it's done
		return (int)m_ConnectedComponents.size();
	
	// fill connected components
	int currCC = 0;
	std::vector<bool> markedFace;
	for (int i=0; i<m_pMesh->NFaces(); i++)
	{
		markedFace.push_back(false);
		m_FaceCCID.push_back(-1);
	}
	int unmarked = m_pMesh->NFaces();
	while (unmarked>0)
	{
		m_ConnectedComponents.push_back(std::vector<int>());
		std::vector<int> facesToVisit;
		for (int i=0; i<m_pMesh->NFaces(); i++)
		{
			if (!markedFace[i])
			{
				facesToVisit.push_back(i);
				break;
			}
		}
		while (facesToVisit.size()>0)
		{
			if (!markedFace[facesToVisit[0]])
			{
				markedFace[facesToVisit[0]] = true;
				unmarked--;
				m_ConnectedComponents[currCC].push_back(facesToVisit[0]);
				R3MeshFace * f = m_pMesh->Face(facesToVisit[0]);
				for (int fnhdID=0; fnhdID<3; fnhdID++)
				{
					R3MeshFace * nf = m_pMesh->FaceOnFace(f, fnhdID);
					if (nf!=NULL)
					{
						int fofid = m_pMesh->FaceID(nf);
						if (!markedFace[fofid])
							facesToVisit.push_back(fofid);
					}
				}
			}
			facesToVisit.erase(facesToVisit.begin());
		}
		currCC++;
	}
	
	for (int i=0; i<(int)m_ConnectedComponents.size(); i++)
	for (int j=0; j<(int)m_ConnectedComponents[i].size(); j++)
		m_FaceCCID[m_ConnectedComponents[i][j]] = i;
	
	if (createLabeling)		// create labels ModelValue
	{
		VKString labelsName = "ConnectedComponents";
		ModelValueLabel * ccLabels = new ModelValueLabel(this, labelsName);
		
		for (int i=0; i<m_pMesh->NFaces(); i++)
			ccLabels->AssignSegmentIDToFace(i, m_FaceCCID[i]);
		for (int i=0; i<(int)m_ConnectedComponents.size(); i++)
			ccLabels->AssignLabel(i, VKString::number(i));
		AddModelValue(labelsName, ccLabels);
		if (!m_pDB->m_ModelValueNames.contains(labelsName))
			m_pDB->m_ModelValueNames.push_back(labelsName);
	}
	return currCC;
}




void Model3D::CreateSingleCCForFaces()
{
	assert(false);
	if (m_FaceNeighbors.size()>0)
		return;
	int nCC = FillConnectedComponentsFaces(true);
//	for (int i=0; i<m_pMesh->NFaces(); i++)
//		m_FaceNeighbors.push_back(std::vector<int>());
//	if (nCC>1)
//	{
//		std::vector<std::vector<int> > ccs = m_ConnectedComponents;
//		std::vector<std::vector<double> > ccdists;
//		std::vector<std::vector<int[2]> > nearestFaces;
//		std::vector<R3Kdtree<PointWithID*> > centroids;
//		
//		// find faces centroids
//		PointWithID * faceCentroids = new PointWithID[m_pMesh->NFaces()];
//		for (int i=0; i<m_pMesh->NFaces(); i++)
//		{
//			faceCentroids[i].pnt = m_pMesh->FaceCentroid(m_pMesh->Face(i));
//			faceCentroids[i].id = i;
//		}
//		
//		// fill kd trees to search for nearest facess
//		for (int i=0; i<(int)m_ConnectedComponents.size(); i++)
//		{
//			RNArray<PointWithID*> pnts;
//			for (int j=0; j<(int)m_ConnectedComponents[i].size(); j++)
//				pnts.push_back(&(faceCentroids[m_ConnectedComponents[i][j]]));
//			centroids.push_back(R3Kdtree<PointWithID*>(pnts));
//		}
//		
//		// fill minimal distances between connected components
//		for (int i=0; i<(int)m_ConnectedComponents.size(); i++)
//		{
//			ccdists.push_back(std::vector<double>());
//			nearestFaces.push_back(std::vector<int[2]>());
//			for (int j=i+1; j<(int)m_ConnectedComponents.size(); j++)
//			{
//				double minDist = MAX_FLT;
//				for (int f=0; f<(int)m_ConnectedComponents[i].size(); f++)
//				{
//					
//				}
//				ccdists.push_back();
//			}
//		}
//		
//		// repeat until 1 CC left
//		while (ccs.size()>1)
//		{
//			// merge closest CCs (at a face)
//			
//			// update CCs and distances
//		}
//	    delete [] faceCentroids;
//	}
	
	// fill random faces
//	int faceID1 = m_ConnectedComponents[0][0];
//	for (int i=1; i<(int)m_ConnectedComponents.size(); i++)
//		m_FaceNeighbors[faceID1].push_back(m_ConnectedComponents[i][0]);
	
	// fill default neighborhood
	for (int i=0; i<m_pMesh->NFaces(); i++)
	{
		R3MeshFace * f = m_pMesh->Face(i);
		for (int fnhdID=0; fnhdID<3; fnhdID++)
		{
			R3MeshFace * nf = m_pMesh->FaceOnFace(f, fnhdID);
			if (nf!=NULL)
			{
				int fofid = m_pMesh->FaceID(nf);
				m_FaceNeighbors[i].push_back(fofid);
			}
		}
	}
}

bool Model3D::SameCluster(int face1, int face2)
{
	return m_FaceCCID[face1] == m_FaceCCID[face2];
}

void Model3D::SingleCCFaceNeighbors(int faceID, std::vector<int> * neighbors)
{
	assert(faceID < (int)m_FaceNeighbors.size());
	*neighbors = m_FaceNeighbors[faceID];
}

void Model3D::CreateFaceNhd(const VKString & name, double threshold)
{
	if (m_mFaceNhds.find(name)!=m_mFaceNhds.end())
	{
		SelectFaceNhd(name);
		return;
	}
	
	Model3D::FaceNhdStruct * nhd = new Model3D::FaceNhdStruct();
	nhd->nhd = new std::vector<int>[m_pMesh->NFaces()];
	nhd->weights = new std::vector<double>[m_pMesh->NFaces()];
		
	if (m_pFaceCentroids==NULL)
	{
		// find faces centroids
		m_pFaceCentroids = new PointWithID[m_pMesh->NFaces()*4];
		for (int i=0; i<m_pMesh->NFaces(); i++)
		{
			R3MeshFace * f = m_pMesh->Face(i);
			m_pFaceCentroids[4*i + 0].pnt = m_pMesh->FaceCentroid(f);
			m_pFaceCentroids[4*i + 0].id = i;
			
			m_pFaceCentroids[4*i + 1].pnt = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 0));
			m_pFaceCentroids[4*i + 1].id = i;
			m_pFaceCentroids[4*i + 2].pnt = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 1));
			m_pFaceCentroids[4*i + 2].id = i;
			m_pFaceCentroids[4*i + 3].pnt = m_pMesh->VertexPosition(m_pMesh->VertexOnFace(f, 2));
			m_pFaceCentroids[4*i + 3].id = i;
			
		}

		// fill kd trees to search for nearest facess
		RNArray<PointWithID*> pnts;
		for (int i=0; i<m_pMesh->NFaces(); i++)
			pnts.Insert(&(m_pFaceCentroids[i]));
		
		m_pKdTreeFaceCentroids = new R3Kdtree<PointWithID*>(pnts);
	}
	
	DistanceEuclidean * dist = DefaultEuclideanDistance();
	assert(dist!=NULL);
	
	double norm = dist->GetStandardNorm();
	double sigma2 = pow(threshold*.5, 2);
	for (int i=0; i<m_pMesh->NFaces(); i++)
	{
		RNArray<PointWithID*> neighbors;
		m_pKdTreeFaceCentroids->FindAll(m_pFaceCentroids[i].pnt, 0, norm * threshold, neighbors);
		std::map<int, bool> added;
		for (int j=0; j<neighbors.NEntries(); j++)
		{
			if (added.find(neighbors[j]->id) == added.end())
			{
				added[neighbors[j]->id] = true;
				nhd->nhd[i].push_back(neighbors[j]->id);
				double dist = R3Distance(neighbors[j]->pnt, m_pFaceCentroids[i].pnt) / norm;
				nhd->weights[i].push_back(exp(-dist / sigma2));
			}
		}
	}
	
	m_mFaceNhds[name] = nhd;
	SelectFaceNhd(name);
}

void Model3D::SelectFaceNhd(const VKString & faceNhdName)
{
	std::map<VKString, FaceNhdStruct *>::iterator iter = m_mFaceNhds.find(faceNhdName);
	assert(iter!=m_mFaceNhds.end());
	m_pCurrNhd = iter->second;
}

void Model3D::FaceNhd(int faceID, std::vector<int> ** nhd, std::vector<double> ** weights)
{
	assert(m_pCurrNhd!=NULL);
	*nhd = &(m_pCurrNhd->nhd[faceID]);
	*weights = &(m_pCurrNhd->weights[faceID]);
}

Model3D::GeometryType Model3D::GeomTypeStrToType(const VKString & geomStr)
{
	if (geomStr=="IRREGULAR_POLYGON_SOUP")
		return IRREGULAR_POLYGON_SOUP;
	else if (geomStr=="UNIFORM_POLYGON_SOUP")
		return UNIFORM_POLYGON_SOUP;
	else if (geomStr=="UNIFORM_MANIFOLD")
		return UNIFORM_MANIFOLD;
	
	assert(false);
	return INVALID_GEOM;
}

VKString Model3D::GeomTypeToStr(GeometryType tp)
{
	
	if (tp==IRREGULAR_POLYGON_SOUP)
		return "IRREGULAR_POLYGON_SOUP";
	else if (tp == UNIFORM_POLYGON_SOUP)
		return "UNIFORM_POLYGON_SOUP";
	else if (tp == UNIFORM_MANIFOLD)
		return "UNIFORM_MANIFOLD";

	assert(false);
	return "INVALID_GEOM";
}

ModelDatabase * Model3D::GetDB()
{
	return m_pDB;
}

R3Point Model3D::GetCentroid()
{
	if (!m_bCentroid)
	{
		m_bCentroid = true;
		m_Centroid = m_pMesh->Centroid();
	}
	return m_Centroid;
}

double Model3D::GetMeshRadius()
{
	if (!m_bRadius)	
	{
		m_bRadius = true;
		GetCentroid();
		m_fRadius = m_pMesh->AverageRadius(&m_Centroid);
	}
	return m_fRadius;
}

const R3Triad & Model3D::ComputePCAAxis()
{
//	std::cout<<"PCAAXIS: "<<m_bPCAAxisDefined<<std::endl;
//	std::cout<<"\tX=["<<m_PCAAxis[0].X()<<" "<<m_PCAAxis[0].Y()<<" "<<m_PCAAxis[0].Z()<<"] "<<std::endl;
//	std::cout<<"\tY=["<<m_PCAAxis[1].X()<<" "<<m_PCAAxis[1].Y()<<" "<<m_PCAAxis[1].Z()<<"] "<<std::endl;
//	std::cout<<"\tZ=["<<m_PCAAxis[2].X()<<" "<<m_PCAAxis[2].Y()<<" "<<m_PCAAxis[2].Z()<<"] "<<std::endl;
	if (!m_bPCAAxisDefined)
	{
		m_bPCAAxisDefined = true;
		GetCentroid();
		m_PCAAxis = m_pMesh->PrincipleAxes(&m_Centroid, m_afPCAVariance);
		m_ChangeOfBasisToPCA = m_PCAAxis.Matrix();
		m_ChangeOfBasisFromPCA = m_PCAAxis.InverseMatrix();
	}
	return m_PCAAxis;
}

int Model3D::FindLeastSimilarToMe(const R3Triad & axis, const R3Vector & vec)
{
	int ls=-1;
	double wc = 1;
	for (int i=0; i<3; i++)	// find up
	{
		double c = axis[i].Dot(vec);
		if (vkAbs(c)<wc)
		{
			wc = vkAbs(c);
			ls = i;
		}
	}
	assert(ls!=-1);
	return ls;
}

const R3Triad & Model3D::ComputePCAWithUpVector(R3Vector * upVec)
{
	if (upVec==NULL)
		return ComputePCAAxis();

	if (!m_bPCAWithUpDefined)
	{
		m_bPCAWithUpDefined = true;
		upVec->Normalize();
//		{	// THIS IS SIMPLE (but incorrect) VERSION OF 2D PCA
//			R3Triad pca = ComputePCAAxis();
//			int notUp = FindLeastSimilarToMe(pca, *upVec);
//			R3Vector axis1a = *upVec;
//			R3Vector axis1b = pca[notUp];
//			R3Vector axis1c = axis1a % axis1b;
//			m_PCAAxisWithUp = R3Triad(axis1a, axis1b, axis1c);
//			return m_PCAAxisWithUp;
//		}	// TODO: ABove is not right - strange bug below :(
		
		// find (arbitrary) 2D plane orthogonal to the up vector
		int notUp = FindLeastSimilarToMe(R3xyz_triad, *upVec);
		R3Vector axis1a = *upVec;
		R3Vector axis1b = R3xyz_triad[notUp];
		R3Vector axis1c = axis1a % axis1b;
		axis1b = axis1a % axis1c;
		R3Triad xyzWithUp(axis1a, axis1b, axis1c);
		xyzWithUp.Normalize();
		const R4Matrix & xyzToUp = xyzWithUp.Matrix();
		
		// Get centroid
		R3Point centroid = GetCentroid();
		
		// Compute covariance matrix
		RNArea area = 0;
		RNScalar m[4] = { 0 };
		assert(m_pMesh);
		for (int i = 0; i < m_pMesh->NFaces(); i++) 
		{
			R3MeshFace *face = m_pMesh->Face(i);
			RNArea face_area = m_pMesh->FaceArea(face);
			R3Point face_centroid = m_pMesh->FaceCentroid(face);

			R3Point xyzVec(face_centroid[0] - centroid[0], 
						   face_centroid[1] - centroid[1],
						   face_centroid[2] - centroid[2]);
			
			R3Point xyzWithUpVec = xyzToUp * xyzVec;
			
			double x = xyzWithUpVec[1];
			double y = xyzWithUpVec[2];
			m[0] += face_area * x*x;
			m[1] += face_area * x*y;
			m[2] += face_area * x*y;	
			m[3] += face_area * y*y;
			area += face_area;
		}
		
		// Normalize covariance matrix
		if (area == 0) return R3xyz_triad;
		for (int i = 0; i < 4; i++) m[i] /= area;
		
		// Compute eigenvalues and eigenvectors
		RNScalar U[4];
		RNScalar W[2];
		RNScalar Vt[4];
		RNSvdDecompose(2, 2, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
		
		// Copy principle axes into more convenient form
		// W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
		R3Vector axes[3];
		axes[0] = *upVec;
//		for (int i = 0; i < 2; i++) {
//			for (int j = 0; j < 2; j++) {
//				axes[1+i][j] = Vt[2*i+j];
//			}
//		}
		
		axes[1] = Vt[0] * xyzWithUp[1]  + Vt[1] * xyzWithUp[2];
		axes[2] = Vt[2] * xyzWithUp[1]  + Vt[3] * xyzWithUp[2];
				
		// Flip axes so that "heavier" on positive side for first two dimensions
		int positive_count[2] = { 0, 0 };
		int negative_count[2] = { 0, 0 };
		for (int i = 0; i < m_pMesh->NVertices(); i++) {
			R3MeshVertex *vertex = m_pMesh->Vertex(i);
			R3Vector vertex_vector = m_pMesh->VertexPosition(vertex) - centroid;
			for (int j = 1; j < 2; j++) {
				RNScalar dot = axes[j].Dot(vertex_vector);
				if (dot > 0.0) positive_count[j]++;
				else negative_count[j]++;
			}
		}
		for (int j =1; j < 2; j++) {
			if (positive_count[j] < negative_count[j]) {
				axes[j].Flip();
			}
		}
		
		// Set third axis to form orthonormal triad with other two
		axes[2] = axes[0] % axes[1];
		axes[0].Normalize();
		axes[1].Normalize();
		axes[2].Normalize();

		
		// Just checking
		assert(RNIsEqual(axes[0].Length(), 1.0, RN_BIG_EPSILON));
		assert(RNIsEqual(axes[1].Length(), 1.0, RN_BIG_EPSILON));
		assert(RNIsEqual(axes[2].Length(), 1.0, RN_BIG_EPSILON));
		assert(RNIsZero(axes[0].Dot(axes[1]), RN_BIG_EPSILON));
		assert(RNIsZero(axes[1].Dot(axes[2]), RN_BIG_EPSILON));
		assert(RNIsZero(axes[0].Dot(axes[2]), RN_BIG_EPSILON));
		
//		// Return variances (eigenvalues)
//		if (variances) {
//			variances[0] = W[0];
//			variances[1] = W[1];
//		}
		
		// Return triad
		m_PCAAxisWithUp = R3Triad(axes[0], axes[1], axes[2]);		
	}
	return m_PCAAxisWithUp;
}

int Model3D::NPntsForRigidMatch()
{
	return PointCloud()->NumPoints();
}

PointInSet * Model3D::GetPntsForRigidMatch(bool anisotropic)
{
	assert(!anisotropic);
	if (m_aPntsForRigidMatch==NULL ||
		m_bAnisotropicScale!=anisotropic)
	{
		PointSet * pnts = PointCloud();

		int nPoints = pnts->NumPoints();
		if (m_aPntsForRigidMatch==NULL)
			m_aPntsForRigidMatch = new PointInSet[nPoints];
		
		m_bAnisotropicScale = anisotropic;
		if (m_bAnisotropicScale)
		{
			m_aPntsForRigidMatchArray = new R3Point[nPoints];
			for (int i=0; i<nPoints; i++)
			{
				m_aPntsForRigidMatchArray[i] = MakeIsotropic(pnts->GetPoint(i).GetPosition());
				
				m_aPntsForRigidMatch[i].m_pXFormedPos = &(m_aPntsForRigidMatchArray[i]);
				m_aPntsForRigidMatch[i].pnts = pnts;
				m_aPntsForRigidMatch[i].id = i;
			}
		}
		else
		{
			for (int i=0; i<nPoints; i++)
			{
				m_aPntsForRigidMatch[i].pnts = pnts;
				m_aPntsForRigidMatch[i].id = i;
			}
		}
	}
		
	return m_aPntsForRigidMatch;
}

R3Point Model3D::MakeIsotropic(const R3Point & p)
{
	assert(false);
	ComputePCAAxis();
	
	// project to PCA
//	std::cout<<"variance=["<<m_afPCAVariance[0]<<" "<<m_afPCAVariance[1]<<" "<<m_afPCAVariance[2];
//	std::cout<<"]"<<std::endl;
//	std::cout<<"Pnt: "<<p.X()<<" "<<p.Y()<<" "<<p.Z()<<std::endl;
	R3Vector v(p - GetCentroid());
	
//	std::cout<<"v: "<<v.X()<<" "<<v.Y()<<" "<<v.Z()<<std::endl;	
	R3Point pcaPnt = m_ChangeOfBasisToPCA * R3Point(v.X(), v.Y(), v.Z());
	
	// scale based on variance
	double scale = 1;
	pcaPnt.X(pcaPnt.X() / scale/pow(m_afPCAVariance[0], .5));
	pcaPnt.Y(pcaPnt.Y() / scale/pow(m_afPCAVariance[1], .5));
	pcaPnt.Z(pcaPnt.Z() / scale/pow(m_afPCAVariance[2], .5));
	
	// project to XYZ
	R3Point ptmp = m_ChangeOfBasisFromPCA * pcaPnt;
//	std::cout<<"ptmp: "<<ptmp.X()<<" "<<ptmp.Y()<<" "<<ptmp.Z()<<std::endl;	
	
	return m_ChangeOfBasisFromPCA * pcaPnt;
}

R3MeshSearchTree * Model3D::GetMeshSearchTreeForRigidMatch(bool anisotropic)
{
	if (anisotropic)
	{
		if (m_pMeshTreeForRigidMatch==NULL)
		{// create tree / create mesh
			m_pIsotropciMesh = CloneMesh(m_pMesh);
			for (int i=0; i<m_pIsotropciMesh->NVertices(); i++)
			{
				R3MeshVertex * v = m_pIsotropciMesh->Vertex(i);
				R3Point xformed = MakeIsotropic(m_pIsotropciMesh->VertexPosition(v));
				m_pIsotropciMesh->SetVertexPosition(v, xformed);
			}
			m_pMeshTreeForRigidMatch = new R3MeshSearchTree(m_pIsotropciMesh); 
		}
		assert(m_pMeshTreeForRigidMatch!=NULL);
		return m_pMeshTreeForRigidMatch;
	}
	return GetSearchTree();
}

R3Mesh * Model3D::GetMeshForRigidMatch(bool anisotropic)
{
	if (anisotropic)
	{
		if (m_pIsotropciMesh==NULL)
			GetMeshSearchTreeForRigidMatch(anisotropic);
		assert(m_pIsotropciMesh!=NULL);
		return m_pIsotropciMesh;
	}
	else
		return m_pMesh;
}

R3Kdtree<const PointInSet *> * Model3D::GetKdTreeForRigidMatch(bool anisotropic)
{
	if (anisotropic)
	{
		if (m_pKdTreeForRigidMatch==NULL)
		{	// create kd-tree
			GetPntsForRigidMatch(anisotropic);
			assert(m_aPntsForRigidMatch!=NULL);
			
			RNArray<const PointInSet *> array;
			
			for (int i = 0; i < NPntsForRigidMatch(); i++) 
				array.Insert(&(m_aPntsForRigidMatch[i]));
			
			m_pKdTreeForRigidMatch = new R3Kdtree<const PointInSet *>(array, 
																	  &global_PointPosition, 
																	  NULL);
		}
		assert(m_pKdTreeForRigidMatch!=NULL);
		return m_pKdTreeForRigidMatch;
	}
	return PointCloud()->GetKdTree();
}

