/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __Model3DRenderingParams_H
#define __Model3DRenderingParams_H

#include "VKString.h"
#include "ParamParser.h"

#include "PointSet.h"
#include "ModelValue.h"
#include "CorrValue.h"
#include "ModelDistance.h"
#include "VisWindow.h"

class Model3D;
class DatabaseVisWindow;
class ModelPerVertexValues;

/**
 * Rendering parameters for a model
 */
class Model3DRenderingParams
{
	public:
		friend class Model3D;
		
		enum ModelRenderFlags
		{
			RENDER_FACES		= 1,
			RENDER_EDGES		= 2,
			RENDER_POINT_SET	= 4,
			RENDER_CORRS		= 8, 
			RENDER_VALUE		= 16,		// includes textures, colors, etc...
			RENDER_DISTANCE		= 32,
			RENDER_PNT_NORMALS	= 64,
			RENDER_XXX2			= 128,
			RENDER_XXX3			= 256,
			RENDER_VAL_ISOCONT	= 512,		// isocontours for functions on surfaces
			RENDER_XXX5			= 1024,
			RENDER_XXX6			= 2048,
			RENDER_XXX7			= 4096,
			RENDER_XXX8			= 8192,
			//			RENDER_CORRS_EDGES	= 16384,	// edges between corresponding points
			//			RENDER_CORRS_PNTS	= 32768,	// points colors
			//			RENDER_CORRS_MVALUE	= 65536,	// model value
			//			RENDER_CORRS_CVALUE	= 131072,	// correspondence value
		};
		static const int RENDER_ALL_FLAGS	= 0xFFFFFFFF;			// mask: all flags
		static const int RENDER_CORRS_STYLE_MASK	= 0x3C000;		// mask: correspondences
		static const int RENDER_MAIN_FLAGS_MASK		= 0x3F;			// mask: main flags
		
		enum InteractiveSelectionFlags
		{
			SELECT_CHECK_NORMAL = 1,
			SELECT_CHECK_DISTANCE = 2
		};
	
		// some initialization
		Model3DRenderingParams(ParamParser & visParams, 
							   DatabaseVisWindow * window);
		void InitializeRenderer(Model3D * refModel);	// might required a point cloud
		
		// change rendering mode
		void FlipFlag(ModelRenderFlags flag);
		void SetFlag(ModelRenderFlags flag, bool on);
		void SetFlags(int flags, int mask = RENDER_ALL_FLAGS);
		bool IsFlagOn(ModelRenderFlags flag);
		void ClearFlags();
		
		// camera operations
		void WindowResize(int newW, int newH);
		void Rotate(int subwindow, int x, int y, int dx, int dy);
		void Translate(int subwindow, int x, int y, int dx, int dy);
		void Scale(int subwindow, int x, int y, int dx, int dy);
		void UpdateCamera(const R3Vector & upVec, const R3Vector & towardsVec, 
						  R3Point * camPos = NULL, double * distRad = NULL);
		R3Vector GetCamTowards(int subwindowID);
		R3Vector GetCamUp(int subwindowID);
		R3Point GetCamPosition(int subwindowID);
		double GetRadius();
		
		void LoadMyCamera(int subWindow);
//		R4Matrix MyCameraTransformMatrix(int subwindow);
//		R4Matrix MyCameraTransformMatrixInverse(int subwindow);
		void RestoreCameraAfterLoad(int subWindow);
		//void LoadMyLights();
		R4Matrix CameraMatrix(int subwindowID, bool inverse = false);
		R4Matrix CurrCameraMatrix();
		R4Matrix CurrCameraInverseMatrix();
		R3Point TransformMyPntToCurrCamera(const R3Point & pntIn, int subwindowID = -1);
        R3Point TransformMyPntToCurrCameraInverse(const R3Point & pntIn, int subwindowID = -1);
        
		
		// lists to render
		void SetDistanceToRender(const VKString & distName);
		void SetPointSetToRender(const VKString & pntSetName);
		void SetModelValueToRender(const VKString & modelValueName);
		//		void SetCorrValueToRender(const VKString & corrValueName);
		//		void SetCollectionToRender(const VKString & collectionName);
		//		void ChangeCollectionID(int changeID);
		
		VKString GetDistanceToRender();
		VKString GetPointSetToRender();
		VKString GetModelValueToRender();
		//		VKString GetCorrValueToRender();
		//		VKString GetCollectionToRender();	
		//		int GetCurrCollectionID();
		
		void ChangeRenderingRadius(double change);
		double GetStandardRadius();
		
		void SelectRegion(int x, int y, double brushSize, std::vector<int> & saveIDs, 
						  int selectionFlags, const VKString & pntSet = "auto");
		void SelectPoint(int x, int y, int selectionFlags);
		void SelectPoint(int selectID);
		int SelectVertex(int x, int y);
		ModelPoint GetSelectedPoint();
    
        virtual void DrawIntersectInfoToFramebuffer(int subwindowID, 
                                                    bool * trianglesToIgnore = NULL);
        virtual ModelPoint QuickIntersection(int x, int y, double brushSize, int selectionFlags);
        static R3Point IntersectFace(const R3Ray & ray, 
                                     const R3Point & fp1, const R3Point & fp2, const R3Point & fp3);
        static R3Point IntersectFace(const R3Ray & ray, R3Mesh * mesh, R3MeshFace * face);
    
		static double CheckIntersection(const R3Point & samplePos, 
                                        const R3Ray & ray, double radius);
		
		ModelPerVertexValues * GetPrecomputed();
		void SetPrecomputed(ModelPerVertexValues * colors);
		void FreePrecomputedColors();
		void SetCurrValueRange(const VKString & valueType, double minValue, double maxValue,
							   double aveValue, double meanValue);
		
	
		std::vector<R3Point> m_RenderAdditionalPnts;
	
		VKString m_sValueType;
		double m_fMinValue;
		double m_fMaxValue;
		double m_fAveValue;
		double m_fMeanValue;
		
		R3Ray CreateMyRay(int x, int y);
	
		virtual void WriteCameras(std::ofstream & textStream);
		virtual void ReadCameras(std::ifstream & textStream);
	
	//protected:
		ModelPerVertexValues * m_pPrecomputedColors;
		
		ModelPoint m_SelectedPnt;
		PointSet * GetCandidateSelectionPoints();
		
		
		// list
		//VKString m_sCollectionName;
		//int m_iCollectionID;
		
		VKString m_sDistanceName;
		VKString m_sPntSetName;
		VKString m_sModelValueName;
		//VKString m_sCorrValueName;
		
		std::map<VKString, ModelPerVertexValues *> m_CachedVals;
		
		// renderer
		DatabaseVisWindow * m_pWindow;
		Model3D * m_pModel;
		
		// camera
		// default camera
		R3Point m_CameraPos;
		R3Point m_Origin;
		R3Vector m_Towards;
		R3Vector m_Up;
		double m_Radius;
		double m_fRendFactor;
		
		// window-specific camera
		R3Viewer * GetViewer(int subwindow);
		std::map<int, R3Viewer *> m_SubwindowToViewer;
		std::map<int, R3Camera> m_SubwindowToCamera;
    
    // read more on openGL framebuffers: http://www.songho.ca/opengl/gl_fbo.html
		
		//std::map<int, VisWindow::BufferID> m_SubwindowToIntersectionFramebuffer; 
	
		BufferID m_IntersectionBuffer;
		bool m_bBufferExists;
		bool m_bBufferFilled;
	
		int m_iCameraID;
		int m_iSubwindowID;
		
		// rendering details
		int m_iRenderingFlags;
		RNRgb m_DefaultColor;
		double m_fDefaultAlpha;
		
		// params from the file
		ParamParser & m_VisParams;
		
		double m_fCameraTranslateScale;	// do not adjust, this is pretty arbitrary constant
};

#endif

