/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TimeProfiler.h"
#include "ModelDatabase.h"
#include <stdlib.h>
#include <sys/stat.h>


///////////// ModelDatabaseRenderingParams /////////////////
ModelDatabaseRenderingParams::ModelDatabaseRenderingParams(DatabaseVisWindow * dbWindow, 
														   ParamParser & renderParams)
: m_DrawParams(renderParams)
{
	m_pWindow = dbWindow;

	std::vector<int> nMainWnd = renderParams.GetIntValues("Visualization", "NMainWnd");
	std::vector<int> nAdditionalWnd = renderParams.GetIntValues("Visualization", "NAddWnd");
	assert(nMainWnd.size()==2 && nAdditionalWnd.size()==2);
	for (int i=0; i<nMainWnd[0]*nMainWnd[1] + nAdditionalWnd[0]*nAdditionalWnd[1]; i++)
		m_mWindowToModelID.push_back(i);
}

///////////// ModelDatabase /////////////////
ModelDatabase::ModelDatabase(const VKString & dbDir, 
							 const VKString & dbName, 
							 int nApproxPntsPerModel)
{
	m_sDBName = dbName;
	m_sDirName = dbDir;
	m_iNApproxPnts = nApproxPntsPerModel;
	m_pRenderingParams = NULL;
	
	m_NGlobalPoints = -1;
	m_aModelPntToGlobalID = NULL;
	m_aGlobalIDToModelPntPair = NULL;

}

ModelDatabase::~ModelDatabase()
{
	// TODO: delete models and maps
}

void ModelDatabase::InitRenderingParams(DatabaseVisWindow * dbWindow, ParamParser & renderParams)
{
	m_pRenderingParams = new ModelDatabaseRenderingParams(dbWindow, renderParams);
	for (int i=0; i<(int)m_pRenderingParams->m_mWindowToModelID.size(); i++)
		if (m_pRenderingParams->m_mWindowToModelID[i] >= (int)m_Models.size())
			m_pRenderingParams->m_mWindowToModelID[i] = 0;
	
	for (int i=0; i<(int)m_Models.size(); i++)
	{
		m_Models[i]->InitRenderingParams(m_pRenderingParams->m_DrawParams, 
										 m_pRenderingParams->m_pWindow);
		m_pRenderingParams->m_ViewableModelNames.push_back(m_Models[i]->ModelName());
	}
	m_pRenderingParams->m_iModelListStart = 0;
	m_pRenderingParams->m_iModelListEnd = renderParams.GetIntValue("Visualization", 
																   "ModelsInSubwindow");
}

int ModelDatabase::NGlobalPnts()
{
	if (m_NGlobalPoints==-1)
		InitPointToGlobalIDMaps();
	return m_NGlobalPoints;
}

int ModelDatabase::GlobalID(const ModelPoint & pnt)
{
	if (m_NGlobalPoints==-1)
		InitPointToGlobalIDMaps();
	return MatrixPosition(pnt, m_aModelPntToGlobalID);
}

ModelPoint ModelDatabase::GetPoint(int globalID)
{
	if (m_NGlobalPoints==-1)
		InitPointToGlobalIDMaps();
	return PositionToPoint(globalID, this, m_aGlobalIDToModelPntPair);
}

void ModelDatabase::InitPointToGlobalIDMaps()
{
	if (m_NGlobalPoints==-1)
	{
		assert(m_aModelPntToGlobalID==NULL && m_aGlobalIDToModelPntPair==NULL);
		m_NGlobalPoints = ConstructPointToGlobalIDMap(this, 
													  &m_aModelPntToGlobalID, 
													  &m_aGlobalIDToModelPntPair);
		assert(m_aModelPntToGlobalID!=NULL && m_aGlobalIDToModelPntPair!=NULL);
	}
}

int ModelDatabase::MatrixPosition(const ModelPoint & pnt, 
									  int ** pointToGlobalID)
{
	int modelID = pnt.GetModel()->ModelID();
	//assert(modelID>=0 && modelID < NModels);
	int pntID = pnt.PntCloudID();
	//assert(pntID>=0 && pntID < NPoints);
	//return modelID * NPoints + pntID;
	return pointToGlobalID[modelID][pntID];
}

ModelPoint ModelDatabase::PositionToPoint(int position, ModelDatabase * db, 
										  int * globalIDToPoint)
{
	//	int modelID = position / NPoints;
	//	int pntID = position % NPoints;
	int modelID = globalIDToPoint[2 * position + 0];
	int pntID = globalIDToPoint[2 * position + 1];
	assert(modelID < db->NumModels());
	PointSet * pnts = db->GetModelByID(modelID)->PointCloud();
	assert(pntID < pnts->NumPoints());
	return pnts->GetPoint(pntID);
}

int ModelDatabase::ConstructPointToGlobalIDMap(ModelDatabase * db, 
												   int *** pointToGlobalID, 
												   int ** globalIDToPoint)
{
	int NModels = db->NumModels();	
	int nPoints = 0;
	for (int i=0; i<NModels; i++)
		nPoints += db->GetModelByID(i)->PointCloud()->NumPoints();
	
	(*pointToGlobalID) = new int*[NModels];	
	(*globalIDToPoint) = new int[2 * nPoints];
	int globalID = 0;
	for (int i=0; i<NModels; i++)
	{
		int pointsOnModel = db->GetModelByID(i)->PointCloud()->NumPoints();
		(*pointToGlobalID)[i] = new int[pointsOnModel];
		
		for (int j=0; j<pointsOnModel; j++)
		{
			(*pointToGlobalID)[i][j] = globalID;		
			(*globalIDToPoint)[2*globalID + 0] = i;
			(*globalIDToPoint)[2*globalID + 1] = j;
			globalID++;
		}
	}
	
	assert(nPoints==globalID);
	return nPoints;
	
}

VKString ModelDatabase::DatabaseName()
{
	return m_sDBName;
}

ModelDatabaseRenderingParams * ModelDatabase::GetRenderParams()
{
	return m_pRenderingParams;
}

void ModelDatabase::InitializeDB(const VKString & dataPrefix, 
								 VKStringList & dataDirs, 
								 VKStringList & dataFiles,
                                 const VKString & modelPrefix,
                                 bool addInAnyCase, 
								 bool loadToMem,
								 bool keepInMem)
{
	int numModels = 0;
	assert(dataDirs.count()==dataFiles.count());
	
	for (int i=0; i<dataDirs.count(); i++)
	{
		TimeProfiler timer;
		
		VKString currDir = dataPrefix + dataDirs[i] + "/";

        std::cout<<"Attempting to read: "<<(currDir + dataFiles[i]).c_str()<<std::endl;
		std::ifstream dbStream((currDir + dataFiles[i]).c_str());
		if (dbStream.is_open())	// try to get a database description - use it for loading
		{
			dbStream.close();
			std::cout<<"Parsing: "<<(currDir + dataFiles[i]).c_str()<<std::endl;
			ParamParser databaseContent(currDir + dataFiles[i]);
			int n = (int)databaseContent.m_ParamModules.size();
			int currProg=0;
			std::cout<<"InitDB: "<<dataDirs[i].c_str()<<" - "<<n<<" models "<<std::endl;
			std::map<VKString, std::map<VKString, VKStringList> >::iterator iter;
			for (iter = databaseContent.m_ParamModules.begin(); 
				 iter != databaseContent.m_ParamModules.end();
				 iter++)
			{
				timer.WriteProgress("InitDB", currProg++, n);
				VKString entryTp = databaseContent.GetStrValue(iter->first, "EntryType");
				VKString modelPath = databaseContent.GetStrValue(iter->first, "Filepath");

				//std::cout<<"\tLoading: "<<modelPath.c_str()<<std::endl;
				bool nameExists;
				VKString proposedName = databaseContent.GetStrValue(iter->first, "ModelName", 
																	nameExists);
                VKString newModelName = modelPrefix + (nameExists ? proposedName : iter->first);
                int counter=1;
                while (GetModelByName(newModelName)!=NULL)
                    newModelName = modelPrefix + iter->first + "_v"+VKString::number(counter++);
				AddModel(currDir + "/"+ modelPath, newModelName, loadToMem, keepInMem);
				numModels++;
			}
			std::cout<<std::endl;
		}
		else	// if not available - name models by directory
		{
			std::cout<<"[WARNING] Could not load DB="<<dataDirs[i].c_str();
			std::cout<<" - no descriptor"<<std::endl;
		}
	}
	
	std::cout<<"Loaded "<<numModels<<" models"<<std::endl;
}

Model3D * ModelDatabase::GetModelInSubwindow(int subwindow)
{
	return GetModelByID(GetModelIDInSubwindow(subwindow));
}

int ModelDatabase::GetModelIDInSubwindow(int subwindow)
{
	assert(subwindow>=0 && subwindow < (int)GetRenderParams()->m_mWindowToModelID.size());
	return GetRenderParams()->m_mWindowToModelID[subwindow];
}

void ModelDatabase::AddModel(Model3D * model, bool loadToMem, bool keepInMem)
{
	static TimeProfiler timer;
	if (model->m_iID==-1)
		model->m_iID = (int)m_Models.size();
	m_Models.push_back(model);
	if (m_MapModelNameToModelID.find(model->ModelName()) != m_MapModelNameToModelID.end())
	{
		std::cout<<"[ERROR] Model with the name "<<model->ModelName().c_str();
		std::cout<<" aready exists"<<std::endl;
		assert(false);
	}
	else 
	{
		m_MapModelNameToModelID[model->ModelName()] = model->m_iID;
	}
	
	if (loadToMem)
	{
		if (!model->LoadToMemory())	// could not load
			std::cout<<"[WARNING] Could not load model: "<<model->ModelName().c_str()<<std::endl;	
		
		if (!keepInMem)
			model->FreeMemory();
	}
}

void ModelDatabase::RemoveModel(int modelID)
{
	// update collections
	for (std::map<VKString, CorrCollection * >::iterator iter = m_CorrCollections.begin();
		 iter != m_CorrCollections.end(); iter++)
		iter->second->RemoveModel(modelID);

    // update pointer to modelID map
	Model3D * model = GetModelByID(modelID);
    assert(model!=NULL);
    VKString eraseName = model->ModelName();
	m_RemovedModels.push_back(eraseName);
	
    assert(model->ModelID()==modelID);
    for (std::map<VKString, int>::iterator iter = m_MapModelNameToModelID.begin();
         iter != m_MapModelNameToModelID.end(); iter++)
    {
        if (iter->second > modelID)
            iter->second--;   
    }

	// remove name to model map
    std::map<VKString, int>::iterator iter = m_MapModelNameToModelID.find(eraseName);
    assert(iter!=m_MapModelNameToModelID.end());
    m_MapModelNameToModelID.erase(iter);

	// free memory and remove from model from the array
    delete model;
    m_Models.erase(m_Models.begin() + modelID);
	
	// update model ids
    for (int i=modelID; i < NumModels(); i++)
    {
        Model3D * updModel = GetModelByID(i);
        assert(updModel!=NULL);
        updModel->UpdateID(i);
    }
    
	// update models rendering 
    ModelDatabaseRenderingParams * renderParams = GetRenderParams();
	if (renderParams!=NULL)
	{
		int iEnd = renderParams->m_iModelListEnd;
		int iStart = renderParams->m_iModelListStart;
		renderParams->m_iModelListEnd =  iEnd - iStart;
		renderParams->m_iModelListStart = 0;
    
		renderParams->m_ViewableModelNames.remove(eraseName);
	}
}

void ModelDatabase::AddModel(const VKString & filename, const VKString & modelName, 
							 bool loadToMem, bool keepInMem)
{
	Model3D * newModel = new Model3D(filename, modelName, m_iNApproxPnts, this);
	AddModel(newModel, loadToMem, keepInMem);
}

Model3D * ModelDatabase::GetModelByName(const VKString & modelName)
{
	std::map<VKString, int>::iterator iter = m_MapModelNameToModelID.find(modelName);
	if (iter==m_MapModelNameToModelID.end())
	{
		//std::cout<<"[ERROR] Cannot find Model: "<<modelName.c_str()<<std::endl;
		//assert(iter!=m_MapModelNameToModelID.end());
        return NULL;    // let caller handle this
	}
	return GetModelByID(iter->second);
}

Model3D * ModelDatabase::GetModelByID(int modelID)
{
	if (modelID >= m_Models.size() || modelID<0)
	{
		//std::cout<<"[WARNING] Trying to get model ";
        //std::cout<<modelID<<" out of "<<m_Models.size()<<std::endl;
		//assert(modelID < (int)m_Models.size() && modelID>=0);
        return NULL;
	}
	return m_Models[modelID];
}

int ModelDatabase::NumModels() const
{
	return (int)m_Models.size();
}

void ModelDatabase::SwapModels(const VKString & modelName, int modelID)
{
	Model3D * takeMe = GetModelByName(modelName);
	Model3D * swapMe = GetModelByID(modelID);
	
	int takeMeOldID = takeMe->ModelID();
	int swapMeOldID = swapMe->ModelID();
	assert(swapMeOldID == modelID);
	
	VKString takeMeName = takeMe->ModelName();
	VKString swapMeName = swapMe->ModelName();
	assert(takeMeName==modelName);

	m_Models[takeMeOldID] = swapMe;
	m_Models[swapMeOldID] = takeMe;
	
	swapMe->UpdateID(takeMeOldID);
	takeMe->UpdateID(swapMeOldID);

	m_MapModelNameToModelID[swapMeName] = takeMeOldID;
	m_MapModelNameToModelID[takeMeName] = swapMeOldID;
	
	GetRenderParams()->m_ViewableModelNames[takeMeOldID] = swapMeName;
	GetRenderParams()->m_ViewableModelNames[swapMeOldID] = takeMeName;

}

void ModelDatabase::ToggleModelRenderFlag(Model3DRenderingParams::ModelRenderFlags flag, 
										  int windowID)
{
	if (windowID==-1)
	{
		for (int i=0; i<NumModels(); i++)
		{
			GetModelByID(i)->RenderingParams()->FlipFlag(flag);
			GetModelByID(i)->RenderingParams()->m_RenderAdditionalPnts.clear();
		}
	}
	else 
	{
		Model3D * model = GetModelByID(GetRenderParams()->m_mWindowToModelID[windowID]);
		model->RenderingParams()->FlipFlag(flag);
	}
}

void ModelDatabase::SetModelRenderFlag(Model3DRenderingParams::ModelRenderFlags flag, 
										bool on, int windowID)
{
	if (windowID==-1)
	{
		for (int i=0; i<NumModels(); i++)
			GetModelByID(i)->RenderingParams()->SetFlag(flag, on);
	}
	else 
	{
		GetModelByID(GetRenderParams()->m_mWindowToModelID[windowID])->RenderingParams()->SetFlag(flag, on);
	}
}

// DB file management
VKString ModelDatabase::GetMainDir()
{
	if (!m_sDirName.endsWith("/"))
		m_sDirName += "/";
	return m_sDirName;
}

VKString ModelDatabase::GetModelDir(const VKString & modelname)
{
	return GetMainDir() + "Models/" + modelname + "/";
}

VKString ModelDatabase::GetModelFile(const VKString & modelname)
{
	return GetModelDir(modelname) + modelname + ".model3d";
}

VKString ModelDatabase::GetDistanceFile(const VKString & modelname, 
										const VKString & distancename)
{
	return GetModelDir(modelname) + "Distances/" + distancename + ".dist";
}

VKString ModelDatabase::GetModelValueFile(const VKString & modelname, 
										  const VKString & valuename)
{
	return GetModelDir(modelname) + "ModelValues/" + valuename + ".vals";
}

VKString ModelDatabase::GetPointSetDir(const VKString & modelname, 
									   const VKString & pointsetname)
{
	return GetModelDir(modelname) + "PointSets/";
}

VKString ModelDatabase::GetPointSetFile(const VKString & modelname, 
										const VKString & pointsetname)
{
	return GetPointSetDir(modelname, pointsetname) + pointsetname + ".pnts";
}

bool ModelDatabase::DirExists(const VKString & dirname)
{
	struct stat st;
	if(stat(dirname.c_str(), &st) == 0)
		return true;
	return false;
}

bool ModelDatabase::FileExists(const VKString & filename)
{
	struct stat st;
	if(stat(filename.c_str(), &st) == 0)
		return true;
	return false;
}

bool ModelDatabase::CreateDirsOnPath(const VKString & dirname)
{
	VKString currDir("");
	VKStringList dirs = dirname.split("/");
	for (int i=0; i<dirs.count(); i++)
	{
		currDir += dirs[i]+"/";
		if (!DirExists(currDir))
		{
			bool ok = mkdir(currDir.c_str(), 07777);
			assert(ok==0);
		}
			//system((VKString("mkdir ") + currDir).c_str());
	}
}

bool ModelDatabase::CreateDirsOnPathToFile(const VKString & filename)
{
	VKString currDir("");
	VKStringList dirs = filename.split("/");
	//std::cout<<"Filename: "<<filename.c_str()<<std::endl;
	for (int i=0; i<dirs.count() - 1; i++)
	{
		currDir += dirs[i]+"/";
		if (!DirExists(currDir))
		{
			//system((VKString("mkdir ") + currDir).c_str());
			bool ok = mkdir(currDir.c_str(), 07777);
			assert(ok==0);
		}
	}
}

bool ModelDatabase::ReadFromDir()
{
	VKString dbFile = GetMainDir() + "database.txt";
	if (!FileExists(dbFile))
		return false;
	
	std::cout<<"Loading database from directory: "<<std::endl;
	TimeProfiler timer;
	
	VKString impossibleModelName("!$@#*&$(_IMPOSSIBLE_MODEL_NAME_!@#$*)(!@#");
	ParamParser dbParams(dbFile);
	
	VKStringList modelNames;
	std::map<VKString, std::map<VKString, VKStringList> >::iterator iter;
	for (iter = dbParams.m_ParamModules.begin(); iter != dbParams.m_ParamModules.end(); iter++)
	{
		VKString modelName = dbParams.GetStrValue(iter->first, "ModelName");
		int modelID = dbParams.GetIntValue(iter->first, "ModelID");
		
		while (modelID >= modelNames.count())
			modelNames.push_back(impossibleModelName);
		modelNames[modelID] = modelName;
	}
	
	for (int i=0; i<modelNames.count(); i++)
	{
		timer.WriteProgress("ModelLoadProg", i, modelNames.count());
		assert(modelNames[i]!=impossibleModelName);
		AddModel(GetModelFile(modelNames[i]), modelNames[i], true, true);
	}
	std::cout<<std::endl;
	VKStringList corrs;
	for (int i=0; i<NumModels(); i++)
	{
		VKStringList & loadCrrs = GetModelByID(i)->m_CorrCollectionsCameWithExternalFile;
		for (int j=0; j<loadCrrs.count(); j++)
			if (!corrs.contains(loadCrrs[j]))
				corrs.push_back(loadCrrs[j]);
		
		VKStringList & loadPntSets = GetModelByID(i)->m_PntSetsCameWithExternalFile;
		for (int j=0; j<loadPntSets.count(); j++)
			if (!m_PointSetNames.contains(loadPntSets[j]))
				m_PointSetNames.push_back(loadPntSets[j]);
	}
	
	for (int i=0; i<corrs.count(); i++)
	{
		timer.WriteProgress("Corrs", i, corrs.count());
		bool couldRead = CorrCollection::ReadAndCreateCollection(corrs[i], this, NULL);
		if (!couldRead)
		{
			std::cout<<"[ERROR] Mandatory (input) collection is missing: ";
			std::cout<<corrs[i].c_str()<<std::endl;	
		}
		assert(couldRead);
	}
	if (corrs.count()!=0)
		std::cout<<std::endl;
	return true;
}

void ModelDatabase::WriteToDir()
{
	std::cout<<"Writing to dir"<<std::endl;
	VKString dbFile = GetMainDir() + "database.txt";
	CreateDirsOnPath(GetMainDir());

	std::ofstream textStream(dbFile.c_str());
	assert(textStream.is_open());
	
    std::cout<<"Writing DB ("<<NumModels()<<")"<<std::endl;
    TimeProfiler timer;
	VKStringList corrs;
	for (int i=0; i<NumModels(); i++)
	{
        timer.WriteProgress("progWriteDB", i, NumModels());
		textStream<<"Model"<<i<<":\n";
		textStream<<"\tModelName\t"<<m_Models[i]->ModelName().c_str()<<"\n";
		textStream<<"\tModelID\t"<<m_Models[i]->ModelID()<<"\n";
		
		GetModelByID(i)->WriteToDB();
		
		VKStringList & loadCrrs = GetModelByID(i)->m_CorrCollectionsCameWithExternalFile;
		for (int j=0; j<loadCrrs.count(); j++)
			if (!corrs.contains(loadCrrs[j]))
				corrs.push_back(loadCrrs[j]);
	}	
	for (int i=0; i<corrs.count(); i++)
	{
		CorrCollection * coll = GetCorrCollection(corrs[i]);
		assert(coll!=NULL);
		coll->WriteCollection();
	}
	std::cout<<std::endl;
	
	for (int i=0; i<m_RemovedModels.count(); i++)
	{
		VKString md = GetModelDir(m_RemovedModels[i]);
		VKString cmd = VKString("if [ -d ")+md+" ]; then rm -r "+md+"; fi";
		system(cmd.c_str());
	}
	m_RemovedModels.clear();
}

VKString ModelDatabase::GetCollectionDir(const VKString & corrCollection)
{
	return GetMainDir()+"CorrCollections/"+corrCollection+"/";
}

VKString ModelDatabase::GetCollectionFile(const VKString & corrCollection)
{
	return GetCollectionDir(corrCollection)+corrCollection+".coll";
}

VKString ModelDatabase::GetCollectionFileWithID(const VKString & corrCollection, int corrID)
{
	if (corrID<0)
		return GetCollectionFile(corrCollection);
	VKString numStr = VKString::number(corrID);
	return GetCollectionDir(corrCollection)+"/"+numStr+"/"+numStr+".coll";
}

VKString ModelDatabase::GetCorrValueFile(const VKString & corrCollection,
										 const VKString & corrValueName)
{
	return GetCollectionDir(corrCollection)+corrValueName+".corval";
}

VKString ModelDatabase::GetCorrValueFileWithID(const VKString & corrCollection,
											   int corrID, const VKString & corrValueName)
{
	if (corrID<0)
		return GetCorrValueFile(corrCollection, corrValueName);
	VKString numStr = VKString::number(corrID);
	return GetCollectionDir(corrCollection)+numStr+"/"+corrValueName+".corval";	
}

VKString ModelDatabase::GetAppDir(const VKString & appName)
{
	return GetMainDir() + "Applications/" + appName + "/";
}

VKString ModelDatabase::GetAppFile(const VKString & appName)
{
	return GetAppDir(appName) + appName + ".app";
}


CorrCollection * ModelDatabase::GetCorrCollection(const VKString & collectionName)
{
	std::map<VKString, CorrCollection * >::iterator iter;
	iter = m_CorrCollections.find(collectionName);
	if (iter==m_CorrCollections.end())
		return NULL;
	return iter->second;
}

void ModelDatabase::AddCorrCollection(const VKString & collectionName, 
									  CorrCollection * collection)
{
	assert(!m_CorrCollectionNames.contains(collectionName));
	m_CorrCollectionNames.push_back(collectionName);
	m_CorrCollections[collectionName]  = collection;
}

CollectionSinglePair * ModelDatabase::GetCorrSinglePair(const VKString & collectionName, 
														int corrID)
{
	assert(corrID>=0);
	CorrCollection * coll = GetCorrCollection(collectionName);
	assert(coll!=NULL);
	CollectionPairwise * collPair = coll->SafeCastToPairwise();
	assert(collPair);
	return collPair->GetModelCorr(corrID);
}

