/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ModelDatabase_H
#define __ModelDatabase_H

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "VKStringList.h"
#include "ParamParser.h"

#include "Model3D.h"
#include "CollectionSinglePair.h"
#include "CollectionPairwise.h"
#include "CorrCollection.h"

class DatabaseVisWindow;

class ModelDatabaseRenderingParams
{
	public:
		ModelDatabaseRenderingParams(DatabaseVisWindow * dbWindow, 
									 ParamParser & renderParams);
	
		DatabaseVisWindow * m_pWindow;
		ParamParser & m_DrawParams;
	
		std::vector<int> m_mWindowToModelID;
	
		int m_iModelListStart;
		int m_iModelListEnd;
		VKStringList m_ViewableModelNames;
};

class ModelDatabase
{
	public: 
		ModelDatabase(const VKString & dbDir, const VKString & dbName, int nApproxPntsPerModel);
		virtual ~ModelDatabase();
	
		virtual void InitRenderingParams(DatabaseVisWindow * dbWindow, 
										 ParamParser & renderParams);
		virtual ModelDatabaseRenderingParams * GetRenderParams();
        // hmm.. i guess this is redundant to 'GetModelInSubwindow()'
		Model3D * GetModelInMainWindow(int windowID);
		Model3D * GetModelInAdditionalWindow(int windowID);
	
		virtual bool ReadFromDir();
		virtual void WriteToDir();
	
		void InitializeDB(const VKString & dataPrefix, 
						  VKStringList & dataDirs, 
						  VKStringList & dataFiles, 
                          const VKString & modelPrefix = "", 
                          bool addInAnyCase = true, 
						  bool loadModel = true,
						  bool keepInMem = true);
		void AddModel(Model3D * model, bool loadToMem, bool keepInMem);
		void AddModel(const VKString & filename, const VKString & modelName,
					  bool loadToMem, bool keepInMem);
        void RemoveModel(int modelID);
		VKString DatabaseName();
		int GetModelIDInSubwindow(int subwindow);
		Model3D * GetModelInSubwindow(int subwindow);
		
	// Correspondences and collections
		CorrCollection * GetCorrCollection(const VKString & collectionName);
		void AddCorrCollection(const VKString & collectionName, CorrCollection * collection);
		CollectionSinglePair * GetCorrSinglePair(const VKString & collectionName, int corrID);

		Model3D * GetModelByName(const VKString & modelName);
		Model3D * GetModelByID(int modelID);
		int NumModels() const;
		
		virtual void SwapModels(const VKString & modelName, int modelID);
	
	// visualization
		void ToggleModelRenderFlag(Model3DRenderingParams::ModelRenderFlags flag, int windowID);
		void SetModelRenderFlag(Model3DRenderingParams::ModelRenderFlags flag, bool on, int wndID);
	
	
		VKStringList m_ModelValueNames;
		VKStringList m_ShapeDescriptorNames;
		VKStringList m_DistanceNames;
		VKStringList m_PointSetNames;
		VKStringList m_CorrCollectionNames;
		int m_iCorrID;
		VKStringList m_CorrValueNames;
		VKStringList m_PointDescriptorNames;
	
	// Database management (save / load)
		VKString GetMainDir();
		VKString GetModelDir(const VKString & modelname);
	
		VKString GetModelFile(const VKString & modelname);

		VKString GetDistanceFile(const VKString & modelname, 
								 const VKString & distancename);
	
		VKString GetModelValueFile(const VKString & modelname, 
								   const VKString & valuename);
	
		VKString GetPointSetDir(const VKString & modelname, 
								const VKString & pointsetname);

		VKString GetPointSetFile(const VKString & modelname, 
								 const VKString & pointsetname);

		VKString GetCollectionDir(const VKString & corrCollection);
		VKString GetCollectionFile(const VKString & corrCollection);
		VKString GetCollectionFileWithID(const VKString & corrCollection, int corrID);
		VKString GetCorrValueFile(const VKString & corrCollection,
								  const VKString & corrValueName);
		VKString GetCorrValueFileWithID(const VKString & corrCollection,
										int corrID, const VKString & corrValueName);
	
		VKString GetAppDir(const VKString & appName);
		VKString GetAppFile(const VKString & appName);
	
		static bool DirExists(const VKString & dirname);
		static bool FileExists(const VKString & filename);
		static bool CreateDirsOnPath(const VKString & filename);
		static bool CreateDirsOnPathToFile(const VKString & dirname);
	
		void SetLoadedCollection(CorrCollection * collection, 
								 const VKString & modelname,
								 const VKString & collectionName);
		bool IsCollectionLoaded(const VKString & modelname,
								const VKString & collectionName);
		
		int NGlobalPnts();
		int GlobalID(const ModelPoint & pnt);
		ModelPoint GetPoint(int globalID);
	
	protected:	
		void InitPointToGlobalIDMaps();
		int m_NGlobalPoints;
		int ** m_aModelPntToGlobalID;
		int * m_aGlobalIDToModelPntPair;
		
		void InitGlobalIDs();
		static int ConstructPointToGlobalIDMap(ModelDatabase * db, 
										   int *** pointToGlobalID, 
										   int ** globalIDToPoint);
		static int MatrixPosition(const ModelPoint & pnt, int ** pointToGlobalID);
		static ModelPoint PositionToPoint(int id, ModelDatabase * db, 
										  int * globalIDToPoint);
	
		ModelDatabaseRenderingParams * m_pRenderingParams;
	
		VKString m_sDBName;
		VKString m_sDirName;
	
		// models
		std::map<VKString, int> m_MapModelNameToModelID;
		std::vector<Model3D *> m_Models;
			
		// maps (?)
		int m_iNApproxPnts;	
	
		std::map<VKString, CorrCollection * > m_CorrCollections;
	
		VKStringList m_RemovedModels;
};

#endif

