/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ModelDatabase.h"

#include "ModelPoint.h"
#include "Model3D.h"

void ModelPoint::InvalidateAllFields()
{
	m_iPntCloudID = -1;
	m_P = R3Point(-.01, 0.001, -.001);
	
	m_iVertID = -1;
	m_iTri = -1;
	m_afB[0] = 0;
	m_afB[1] = 0;
	m_afB[2] = 0;
	
	m_pModel = NULL;
	m_pMesh = NULL;
	m_PointTp = PNT_TP_INVALID;	
}

void ModelPoint::InitializeTriangle(int triangleID, double b1, double b2, double b3)
{
	m_iVertID = -1;
	m_iTri = triangleID;
	m_afB[0] = b1;
	m_afB[1] = b2;
	m_afB[2] = b3;
	assert(m_pMesh!=NULL);
	assert(m_iTri < m_pMesh->NFaces());
	CalculatePosition();
}

void ModelPoint::InitializeVertex(int vertID)
{
	m_iTri = -1;
	m_iVertID = vertID;
	assert(m_pMesh!=NULL);
	assert(m_pMesh->NVertices() > m_iVertID);
	CalculatePosition();
}

ModelPoint::ModelPoint()
{
	InvalidateAllFields();
}

ModelPoint::ModelPoint(const ModelPoint & other)
{
	*this = other;
}

// from R3Mesh
ModelPoint::ModelPoint(int triangleID, double b1, double b2, double b3, 
					   Model3D * model)
{
	InvalidateAllFields();

	m_pModel = model;
	assert(m_pModel!=NULL);
	m_pMesh = (R3Mesh*)GetModel()->GetMesh();
	assert(m_pMesh!=NULL);
	m_PointTp = PNT_TP_ON_MESH_SURFACE;
	
	assert(b1>=0 && b2>=0 && b3>=0);
	double norm  = b1+b2+b3;
	if(norm != 1.)
	{
		b1 /= norm;
		b2 /= norm;
		b3 /= norm;
	}
	
	if (b1==1. && b2==0. && b3==0.)
		InitializeVertex(m_pMesh->VertexID(m_pMesh->VertexOnFace(m_pMesh->Face(triangleID), 0)));
	else if (b1==0. && b2==1. && b3==0.)
		InitializeVertex(m_pMesh->VertexID(m_pMesh->VertexOnFace(m_pMesh->Face(triangleID), 1)));
	else if (b1==0. && b2==0. && b3==1.)
		InitializeVertex(m_pMesh->VertexID(m_pMesh->VertexOnFace(m_pMesh->Face(triangleID), 2)));
	else
		InitializeTriangle(triangleID, b1, b2, b3);
}

ModelPoint::ModelPoint(int vertexID, Model3D * model, int cloudID)
{
	InvalidateAllFields();

	m_pModel = model;
	assert(m_pModel!=NULL);
	m_pMesh = (R3Mesh*)GetModel()->GetMesh();
	assert(m_pMesh!=NULL);
	m_PointTp = PNT_TP_ON_MESH_SURFACE;	
	SetPointCloudID(cloudID);
	
	InitializeVertex(vertexID);
}

ModelPoint::ModelPoint(int triangleID, double b1, double b2, double b3, 
					   R3Mesh * flatMesh, Model3D * model)
{
	InvalidateAllFields();
	
	m_PointTp = PNT_TP_FLATTENED;
	m_pMesh = flatMesh;
	m_pModel = model;
	InitializeTriangle(triangleID, b1, b2, b3);
}

ModelPoint::~ModelPoint()
{
}

bool ModelPoint::operator == (const ModelPoint & other) const
{
	if (m_PointTp!=other.m_PointTp)
		return false;
		
	switch (m_PointTp) 
	{
		case PNT_TP_ON_MESH_SURFACE:
			if (m_iTri==other.m_iTri && m_iTri==-1)	// vertex id
				return m_iVertID==other.m_iVertID;
			else if (m_iVertID==other.m_iVertID && m_iVertID==-1)	// triangle id
				return m_afB[0]==other.m_afB[0] && m_afB[1]==other.m_afB[1]&& m_afB[2]==other.m_afB[2];
			return false;
		case PNT_TP_INVALID:
			return false;
		default:
			assert(false);
	}
	return false;
}

void ModelPoint::operator = (const ModelPoint & other)
{
	InvalidateAllFields();
	
	m_P = other.m_P;	
	m_iVertID = other.m_iVertID;
	m_iTri = other.m_iTri;
	for(int i=0; i<3; i++)
		m_afB[i] = other.m_afB[i];

	m_pModel = other.m_pModel;
	m_pMesh = other.m_pMesh;
	m_PointTp = other.m_PointTp;
	m_iPntCloudID = other.m_iPntCloudID;
}

void ModelPoint::SetPointCloudID(int id)
{
	m_iPntCloudID = id;
}

int ModelPoint::PntCloudID() const
{
	return m_iPntCloudID;
}

bool ModelPoint::operator < (const ModelPoint & other) const
{	
	assert(other.m_pModel!=NULL && m_pModel!=NULL);
	if (other.m_pModel != m_pModel)
		return m_pModel->ModelID() < other.m_pModel->ModelID();
	else if (m_iVertID<0 && other.m_iVertID>=0)	// vertex more important
		return true;
	else if (m_iVertID>=0 && other.m_iVertID<0)
		return false;
	else if (m_iVertID>=0 && other.m_iVertID>=0)
		return m_iVertID < other.m_iVertID;
	else 
	{
		assert(m_iTri>=0 && other.m_iTri>=0);
		if (m_afB[0] != other.m_afB[0])
			return m_afB[0] < other.m_afB[0];
		else if (m_afB[1] != other.m_afB[1])
			return m_afB[1] < other.m_afB[1];
		else if (m_afB[2] != other.m_afB[2])
			return m_afB[2] < other.m_afB[2];
	}

	assert(false);
}

bool ModelPoint::Invalid() const
{
	return m_PointTp==PNT_TP_INVALID;
}

int ModelPoint::TriID() const
{
	return m_iTri;
}

double ModelPoint::B(int coordID) const
{
	assert(coordID>=0 && coordID<3);
	return m_afB[coordID];
}

int ModelPoint::VertID(int coordID) const
{
	if (m_pMesh!=NULL && m_PointTp==PNT_TP_ON_MESH_SURFACE)
		return m_pMesh->VertexID(m_pMesh->VertexOnFace(m_pMesh->Face(m_iTri), coordID));
	else 
		return -1;
}

int ModelPoint::NearestVertex(bool * atVertex) const
{
	if (atVertex!=NULL)
		*atVertex = false;
	
	if (m_pMesh==NULL || m_PointTp!=PNT_TP_ON_MESH_SURFACE)
		return -1;

	if (atVertex!=NULL)
		*atVertex = (m_iVertID>=0);
	
	if (m_iVertID>=0)
		return m_iVertID;
	else if (B(0)>=B(1) && B(0)>=B(2))
		return VertID(0);
	else if (B(1)>=B(0) && B(1)>=B(2))
		return VertID(1);
	else
		return VertID(2);
}

double ModelPoint::Interpolate(double b1, double b2, double b3) const
{
	assert(m_iTri>=0);
	return b1 * m_afB[0] + b2 * m_afB[1] + b3 * m_afB[2];
}

const Model3D * ModelPoint::GetModel() const
{
	return m_pModel;
}

bool ModelPoint::SameModel(const ModelPoint & pnt) const
{
	return pnt.m_pModel == m_pModel;
}

R3Point ModelPoint::GetPosition() const
{
	return m_P;
}

R3Vector ModelPoint::Normal() const
{
	if (m_iVertID>=0)
		return m_pMesh->VertexNormal(m_pMesh->Vertex(m_iVertID));
	assert(false);
	return R3Vector(0,0,0);
}

ModelPoint::PointType ModelPoint::GetPointType() const
{
	return m_PointTp;
}


void ModelPoint::CalculatePosition()
{
	assert (m_PointTp==PNT_TP_ON_MESH_SURFACE || m_PointTp==PNT_TP_FLATTENED );
	assert(m_pMesh!=NULL);
	if (m_iTri >= 0)
	{
		assert(m_iTri < m_pMesh->NFaces());
		assert(m_iTri>=0);
		RNMagnitude mag[3];
		for (int i=0; i<3; i++)
			mag[i] = m_afB[i]; 
		m_P = m_pMesh->FacePoint(m_pMesh->Face(m_iTri), mag);
	}
	else 
	{
		assert(m_iVertID>=0);
		m_P = m_pMesh->VertexPosition(m_pMesh->Vertex(m_iVertID));
	}
}


void ModelPoint::WritePoint(std::ofstream & textStream) const
{
	textStream<<"ModelPoint "<<m_PointTp<<"\n";
	textStream<<m_P.X()<<" "<<m_P.Y()<<" "<<m_P.Z()<<"\n";
	textStream<<m_iPntCloudID<<" "<<m_iVertID<<" "<<m_iTri<<"\n";
	textStream<<m_afB[0]<<" "<<m_afB[1]<<" "<<m_afB[2]<<" "<<"\n";
	assert(m_pModel!=NULL);
	int id = m_pModel->ModelID();
	assert(id>=0);
	textStream<<id<<"\n";
}

void ModelPoint::ReadPoint(std::ifstream & textStream, ModelDatabase * db,
						   Model3D * model)
{
	std::string temp;
	textStream>>temp;
	assert(strcmp(temp.c_str(), "ModelPoint")==0);
	int pntTp=0;
	textStream>>pntTp;
	m_PointTp = (PointType)pntTp;
	double x, y, z;
	textStream>>x>>y>>z;
	m_P = R3Point(x,y,z);
	textStream>>m_iPntCloudID>>m_iVertID>>m_iTri;
	textStream>>m_afB[0]>>m_afB[1]>>m_afB[2];	
	int modelID;
	textStream>>modelID;
	
	if (model==NULL)
		m_pModel = db->GetModelByID(modelID);
	else
		m_pModel = model;
	assert(m_pModel!=NULL);
	m_pMesh = (R3Mesh*)m_pModel->GetMesh();
}


