/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Model3D.h"
#include "VkFunctions.h"

#include "PointDescriptor.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"

///////////// DescriptorsContainer //////////
PointDescriptorsContainer::PointDescriptorsContainer(int nDescriptors)
{
	m_iNDescriptors = nDescriptors;
	m_aPointDescriptors = new PointDescriptor*[m_iNDescriptors];
}

double PointDescriptorsContainer::PointSimilarity(int p1, int p2)
{
	assert(p1>=0 && p1 < m_iNDescriptors);
	assert(p2>=0 && p2 < m_iNDescriptors);
	return m_aPointDescriptors[p1]->MeasureSimilarity(*(m_aPointDescriptors[p2]));
}

void PointDescriptorsContainer::AddDescriptor(int id, PointDescriptor * descriptor)
{
	assert(id>=0 && id < m_iNDescriptors);
	m_aPointDescriptors[id] = descriptor;
}

PointDescriptor * PointDescriptorsContainer::GetDescriptor(int id)
{
	assert(id>=0 && id < m_iNDescriptors);
	return m_aPointDescriptors[id];
}

///////////// PointDescriptor //////////
PointDescriptor::PointDescriptor()
{
	m_iClassID = 0;
}

PointDescriptor::~PointDescriptor()
{
}

void PointDescriptor::Draw(const ModelPoint & pnt, DatabaseVisWindow * window, Model3D * model)
{
	static GLfloat greenPnt[4];
	greenPnt[0] = 0;
	greenPnt[1] = .6;
	greenPnt[2] = 0;
	greenPnt[3] = 1.;
	
	glEnable(GL_LIGHTING);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, greenPnt); 

	R3Sphere(pnt.GetPosition(), model->RenderingParams()->GetStandardRadius()).Draw();
}

double PointDescriptor::MeasureSimilarityGivenTransform(const PointDescriptor & p2CompareToMe, 
														const R3Affine & alignP2ToMe)
{
	// transform is not used by default
	MeasureSimilarity(p2CompareToMe);
}

void PointDescriptor::ConstructGaussSplat2D(double dx, double dy, double sigma1, double sigma2,
											double ** gaussSplat, int * splatNX, int * splatNY)
{
	double sigma1Sq = pow(sigma1, 2.);
	double sigma2Sq = pow(sigma2, 2.);
	double dist1 = sigma1 / 2.;
	double dist2 = sigma2 / 2.;
	
	int nx = ceil(dist1 / dx);
	int ny = ceil(dist2 / dy);
	*splatNX = 2 * nx + 1;
	*splatNY = 2 * ny + 1;
	
	*gaussSplat = new double[*splatNX * *splatNY];
	
	double norm = 0;
	for (int i=0; i<*splatNX; i++)
	for (int j=0; j<*splatNY; j++)
	{
		double x2 = pow((i-nx)*dx, 2.);
		double y2 = pow((j-ny)*dy, 2.);
		double val = exp(x2 / sigma1Sq) * exp(y2 / sigma2Sq);
		(*gaussSplat)[GetID(i, j, *splatNX, *splatNY)] = val;
		norm += val;
	}

//	for (int i=0; i<*splatNX; i++)
//	for (int j=0; j<*splatNY; j++)
//		(*gaussSplat)[GetID(i, j, *splatNX, *splatNY)] /= norm;
}


int PointDescriptor::GetID(int i, int j, int n1, int n2)
{
	assert(i<n1 && j < n2 && i>=0 && j>=0);
	return i * n2 + j;
}
				   
///////////// Point Descriptor SHD /////////////
PointDescriptorFeatureVector::PointDescriptorFeatureVector()
{
	m_pDescriptorVals = NULL;
	m_iClassID = 2;
}

PointDescriptorFeatureVector::~PointDescriptorFeatureVector()
{
	delete m_pDescriptorVals;
}

double PointDescriptorFeatureVector::MeasureSimilarity(const PointDescriptor & compareToMe)
{
	assert(m_pDescriptorVals!=NULL);
	assert(compareToMe.m_iClassID==m_iClassID);
	const PointDescriptorFeatureVector & compareFeatVec = 
										(const PointDescriptorFeatureVector &)compareToMe;
	
	assert(compareFeatVec.m_pDescriptorVals!=NULL);
	assert(m_pDescriptorVals->Rows()==compareFeatVec.m_pDescriptorVals->Rows());
	//double diff2 = 0;
	double corr = 0;
	for (int i=0; i<m_pDescriptorVals->Rows(); i++)
	{
		//std::cout<<"diff["<<i<<"] = "<<(*m_pDescriptorVals)(i);
		//std::cout<<"-"<<(*(compareFeatVec.m_pDescriptorVals))(i);
		//std::cout<<"="<<((*m_pDescriptorVals)(i) - (*(compareFeatVec.m_pDescriptorVals))(i));
		//std::cout<<std::endl;
		double v1 = (*m_pDescriptorVals)(i);
		double v2 = (*(compareFeatVec.m_pDescriptorVals))(i);
		//diff2 += pow(v1-v2, 2.);
		corr += v1*v2;
	}
	//std::cout<<"diff = "<<sqrt(diff2)<<" / "<<m_pDescriptorVals->Rows()<<std::endl;
	//return pow(vkMax(1. - sqrt(diff2) / m_pDescriptorVals->Rows(), 0.), 50);
	return pow(corr, 30);
}

void PointDescriptorFeatureVector::QuickBatchSphericalHarmonics(ModelDatabase * db, 
																const VKString & pntsName,
																bool forceRecompute,
								  std::map<Model3D*, std::vector<PointDescriptor*> > & output)
{
	int containerSize = 100;
	float * tempContainer = new float[containerSize];

	TimeProfiler timer;
	std::cout<<"Finding per point spherical harmonic descriptors"<<std::endl;
	for (int i=0; i<db->NumModels(); i++)
	{
		timer.WriteProgress("SHDPerPoint", i, db->NumModels());
		Model3D * model = db->GetModelByID(i);
		assert(model!=NULL);
		VKString geometryFile = model->ModelGeometryFile();
		VKString modelName = model->ModelName();
		PointSet * pnts = model->GetPointSet(pntsName);
		assert(pnts!=NULL);
		// calculate SHD
		VKString descrFile = db->GetPointSetDir(modelName, pntsName)+"SphericalHarmonics.shd";
		
		if (!db->FileExists(descrFile) || forceRecompute)
		{
			db->CreateDirsOnPathToFile(descrFile);

			VKString outPntsFile = db->GetPointSetDir(modelName, pntsName) + "PosAndNorm.pnts";
			
			FILE *fp = fopen(outPntsFile.c_str(), "wb");
			float coordinates[6];
			for (int pID = 0; pID < pnts->NumPoints(); pID++)
			{
				ModelPoint p = pnts->GetPoint(pID);
				R3Point pos = p.GetPosition();
				R3Vector norm = p.Normal();
				coordinates[0] = pos.X();
				coordinates[1] = pos.Y();
				coordinates[2] = pos.Z();
				coordinates[3] = norm.X();
				coordinates[4] = norm.Y();
				coordinates[5] = norm.Z();
				
				fwrite(coordinates, sizeof(float), 6, fp);
			}
			fclose(fp);
			
			VKString cmd = VKString("./ShapeDescriptors/SphericalHarmonics.sh ");
			cmd += geometryFile + " " + descrFile + " "+outPntsFile;
			//std::cout<<cmd.c_str()<<std::endl;
			//exit(0);
			system(cmd.c_str());
		}

		FILE *fp = fopen(descrFile.c_str(), "rb");
		std::vector<PointDescriptor*> & descriptors = output[model];
		for (int pID = 0; pID < pnts->NumPoints(); pID++)
		{
			int nSize;
			fread(&nSize, sizeof(int), 1, fp);
			if (nSize > containerSize)
			{
				delete [] tempContainer;
				containerSize = nSize*2;
				tempContainer = new float[containerSize];
			}
			
			fread(tempContainer, sizeof(float), nSize, fp);
			
			PointDescriptorFeatureVector * newDescr = new PointDescriptorFeatureVector();
			newDescr->m_pDescriptorVals = new LinAlgVectorReal(nSize);
			
//			std::cout<<"Model: "<<modelName.c_str()<<" point: "<<pID<<std::endl;
//			std::cout<<"\tDescriptor Size: "<<nSize<<std::endl;
//			std::cout<<"\t\t["<<std::flush;
			for (int j=0; j<nSize; j++)
			{
//				std::cout<<tempContainer[j]<<" "<<std::flush;
				(*(newDescr->m_pDescriptorVals))(j) = tempContainer[j];
			}
//			std::cout<<"]"<<std::endl;
			descriptors.push_back(newDescr);
		}
		fclose(fp);
	}
	delete [] tempContainer;
	std::cout<<std::endl;
}






