/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Sortable.h"

#include "ShapeDescriptor.h"
#include "Model3D.h"
#include "ModelDatabase.h"
#include "TimeProfiler.h"


ShapeDescriptor::ShapeDescriptor(Model3D * model)
{
	m_pDescriptorVals = NULL;
	m_pModel = model;
}

Model3D * ShapeDescriptor::GetModel()
{
	return m_pModel;
}

void ShapeDescriptor::FindMyNearestNeighbors(std::vector<ShapeDescriptor*> & otherDescriptors, 
											 int nNeighbors, 
											 std::vector<ShapeDescriptor*> & neighbors,
											 std::vector<double> * saveDistances)
{
	//std::cout<<"Neighbors="<<nNeighbors<<std::endl;
	if (m_pDescriptorVals==NULL)
		FillDescriptorValues();
	assert(m_pDescriptorVals!=NULL);
	int nDim = 	m_pDescriptorVals->Rows();
	std::vector<Sortable> distances;
	for (int i=0; i<(int)otherDescriptors.size(); i++)
	{
		ShapeDescriptor * other = otherDescriptors[i];
		if (other->m_pDescriptorVals==NULL)
			other->FillDescriptorValues();
		assert(other->m_pDescriptorVals!=NULL);
		
		assert(nDim == other->m_pDescriptorVals->Rows());
		double dist2 = 0;
		for (int j=0; j<nDim; j++)
			dist2 += pow((*m_pDescriptorVals)(j)-(*(other->m_pDescriptorVals))(j), 2);
		distances.push_back(Sortable(dist2, other, i));
	}
	
	std::sort(distances.begin(), distances.end());

	for (int i=0; i<vkMin(nNeighbors, (int)distances.size()); i++)
	{
		neighbors.push_back((ShapeDescriptor*)distances[i].ptr);
		if (saveDistances!=NULL)
			saveDistances->push_back(distances[i].value);
	}
}

double ShapeDescriptor::MeasureSimilarity(ShapeDescriptor * otherDescriptor)
{
	assert(m_pDescriptorVals!=NULL);
	assert(otherDescriptor!=NULL);
	assert(otherDescriptor->m_pDescriptorVals!=NULL);
	assert(otherDescriptor->m_pDescriptorVals->Rows()==m_pDescriptorVals->Rows());
	double innerProduct = 0;
	for (int i=0; i<m_pDescriptorVals->Rows(); i++)
		innerProduct += (*m_pDescriptorVals)(i) * (*(otherDescriptor->m_pDescriptorVals))(i);
	return innerProduct;
}

void ShapeDescriptor::FillDescriptorValues()
{
	// values need to be already defined
	assert(m_pDescriptorVals!=NULL);
}

void ShapeDescriptor::QuickBatchWorkSHD(ModelDatabase * db, bool forceRecompute,
										std::vector<ShapeDescriptor*> & descriptors)
{
	TimeProfiler timer;
	std::cout<<"Loading Spherical Harmonics descriptors:"<<std::endl;
	for (int i=0; i<db->NumModels(); i++)
	{
		timer.WriteProgress("currProg", i, db->NumModels());
		Model3D * model = db->GetModelByID(i);
		ShapeDescriptor * descriptor = new ShapeDescriptor(model);
		VKString filename = model->ModelGeometryFile();
		VKString modelName = model->ModelName();

		VKString descrFile = db->GetModelDir(modelName) + "SD/" + modelName + ".shd";
				
		if (!db->FileExists(descrFile) || forceRecompute)
		{
			db->CreateDirsOnPathToFile(descrFile);
			VKString cmd = VKString("./ShapeDescriptors/SphericalHarmonics.sh ");
			cmd += filename + " " + descrFile;
			system(cmd.c_str());
		}
		FILE * binFile = fopen(descrFile.c_str(), "rb");
		assert(binFile!=NULL);
		int nSize;
		fread(&nSize, sizeof(int), 1, binFile);
		float * values = new float[nSize];
		
		int nRead = fread(values, sizeof(float), nSize, binFile);
		assert(nRead == nSize);
		
		descriptor->m_pDescriptorVals = new LinAlgVectorReal(nSize);
		for (int i=0; i<nSize; i++)
			(*(descriptor->m_pDescriptorVals))(i) = values[i];
		descriptors.push_back(descriptor);
		
		delete [] values;
	}
}



