/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __VisWindow_H
#define __VisWindow_H

#include <vector>
#include <iostream>
#include <assert.h>

#include "gaps.h"

#include "ParamParser.h"
#include "VkFunctions.h"

class VisWindow;
class ModelPerVertexValues;
class ModelValuesInterpolatorDescriptor;

/**
 * Window renderer is a class that contains all implementation-specific details
 * for text rendering currently only Glut version is implemented, Qt is another option.
 * There is mutual dependence with VisWindow class - VisWindowRenderer makes calls 
 * for user input events, and VisWindow might make calls to issue special rendering commands.
 */
class VisWindowRenderer
{
	public:
		VisWindowRenderer(VisWindow * window);
	
		enum TextSize
		{
			TEXT_SIZE_SMALL,
			TEXT_SIZE_MEDIUM,
			TEXT_SIZE_LARGE,
		};

		virtual void RenderText(R3Point pnt, const VKString & text, 
								double r, double g, double b, TextSize size);
		virtual void RenderText(R2Point pnt, const VKString & text, 
								double r, double g, double b, TextSize size);

		virtual void TerminateWindow();
		virtual void UpdateScreen();
	protected:
		VisWindow * m_pWindow;
};

// the off-screen buffers
struct BufferID
{
	GLuint m_iFrameBufferObject;
	GLuint m_iDepthBuffer;
	GLuint m_iColorBuffer;
	GLuint m_iTextureBuffer;
	unsigned char * m_aColorBuffer;
	int m_iWidth;
	int m_iHeight;
};


/**
 * Vis window is a general window class. It has abstract functions for user input events
 * and handles some abstract rendering (subwindows / etc...)
 */
class VisWindow
{
	public:
		enum SpecialKeys
		{
			SPECIAL_NONE  = 0,
			SPECIAL_ALT   = 2,
			SPECIAL_CTRL  = 4,
			SPECIAL_SHIFT = 8,
		};
		enum MouseButton
		{
			MOUSE_LEFT,
			MOUSE_RIGHT,
			MOUSE_MIDDLE
		};
		enum MouseButtonState
		{
			MOUSE_BT_UP,
			MOUSE_BT_DOWN
		};
		VisWindow(ParamParser & visParams);
		
		void SetRenderer(VisWindowRenderer * renderer);
		VisWindowRenderer * GetRenderer();
		
		virtual void KeyPressed(const char key, int mouseX, int mouseY);
		virtual void KeyReleased(const char key, int mouseX, int mouseY);
		virtual void MouseButtonEvent(int x, int y, MouseButton mouseButtonID, 
									  MouseButtonState newState);
		virtual void MouseMotionEvent(int x, int y, int dx, int dy);
	
		virtual void UpdateMouseButtons(MouseButton mouseButtonID, MouseButtonState newState);
		virtual bool IsMouseButtonDown(MouseButton bt);
	
		virtual void SetSpecialKey(int specialKeyFlags);
		virtual bool Shift();
		virtual bool Ctrl();
		virtual bool Alt();
		
		virtual void Draw() = 0;
		virtual void InitializeGL();
		virtual void Resize(int w, int h);
		virtual int Width();
		virtual int Height();
	
		virtual void MapScalarToColor(double value, double minValue, double maxValue, 
									  double & r, double & g, double & b);
		virtual void MapIntToColor(int i, double & r, double & g, double & b);
		virtual void StandardSurfaceColor(double & r, double & g, double & b, double & a);
		
		virtual void MapIntToColor(int i, GLfloat material[4]);
		virtual void StandardSurfaceColor(GLfloat material[4]);
		
		virtual void MapIntToRadicalColor(int i, double & r, double & g, double & b);

		virtual void SubwindowPosition(int strID, int * top, int * left);
		virtual void StartRenderSubwindow(int top, int left);
		virtual VKString InSubWindow(int x, int y, bool * inWindow, 
									 double * xWindow, double * yWindow);
	
		// rendering global values
		virtual void ClearAllGlobalValues();
		virtual void AddToGlobalValuePool(const VKString & valueName, ModelPerVertexValues * values);
//		virtual void NormalizePoolAndCalculateColors(const VKString & valueName, 
//													 ModelValuesInterpolatorDescriptor * interpolate);
		virtual void NormalizePoolAndCalculateColors(const VKString & valueName, 
													 const VKString & interpolator="none");
			
		bool m_bGlobalColors;
		bool m_bDefaultDrawIsocontours;
		double m_fMinPercentile;
		double m_fMaxPercentile;
		
		bool InMainWndBuffer();
		void StartDrawToBuffer();
		BufferID & EndDrawToBuffer();
	
		void LoadCurrentBuffer();
		
		BufferID CreateBuffer();
	
	protected:
		int m_iCurrentFramebufferID;
		std::vector<BufferID> m_Buffers;

		// global values
		std::map<VKString, std::vector<ModelPerVertexValues*> > m_GlobalValuePool;
	
		// width / height
		int m_iWidth, m_iHeight;
		VisWindowRenderer * m_pRenderer;
		ParamParser & m_VisParams;
		MouseButtonState m_MouseStates[3];
		int m_iSpecialKeys;
		std::vector<RNRgb> m_MulticolorSequence;
		std::vector<VKString> m_AvailiableSubWindows;
};

#endif

